using MVR.FileManagementSecure;
using System;
using System.Collections.Generic;
using System.Net.Sockets;
using UnityEngine;


namespace JayJayWon
{

    public class BAInteropServer : BAInterop
    {
        public static void OnGetSupportedResourceTypes(object[] objects)
        {
            JSONStorableString exceptionCallback = objects[3] as JSONStorableString;
            try
            {
                HashSet<int> resourceTypesList = objects[0] as HashSet<int>;
                JSONStorableAction successCallback = objects[1] as JSONStorableAction;
                JSONStorableString failCallback = objects[2] as JSONStorableString;

                resourceTypesList.Clear();
                foreach (var resourceType in EnumManifest.GetEnumsList(ResourceType.enumManifestName))
                {
                    resourceTypesList.Add(resourceType);
                }

                successCallback.actionCallback.Invoke();
            }
            catch (Exception e)
            {
                exceptionCallback.val = e.ToString();
            }
        }

        public static void OnGetSupportedAndConcealedResourceTypes(object[] objects)
        {
            JSONStorableString exceptionCallback = objects[4] as JSONStorableString;
            try
            {
                HashSet<int> supportedResourceTypesList = objects[0] as HashSet<int>;
                HashSet<int> concealedResourceTypesList = objects[1] as HashSet<int>;
                JSONStorableAction successCallback = objects[2] as JSONStorableAction;
                JSONStorableString failCallback = objects[3] as JSONStorableString;

                supportedResourceTypesList.Clear();
                concealedResourceTypesList.Clear();
                foreach (var resourceType in EnumManifest.GetEnumsList(ResourceType.enumManifestName))
                {
                    supportedResourceTypesList.Add(resourceType);
                    if (resourceType!=ResourceType.presetAtom && Settings.IsResourceConcealed(resourceType,"")) concealedResourceTypesList.Add(resourceType);
                }

                successCallback.actionCallback.Invoke();
            }
            catch (Exception e)
            {
                exceptionCallback.val = e.ToString();
            }
        }


        public static void OnGetExistingAtomPresetResourceTypes(object[] objects)
        {
            JSONStorableString exceptionCallback = objects[3] as JSONStorableString;
            try
            {
                HashSet<string> resourceTypesList = objects[0] as HashSet<string>;
                JSONStorableAction successCallback = objects[1] as JSONStorableAction;
                JSONStorableString failCallback = objects[2] as JSONStorableString;

                resourceTypesList.Clear();
                foreach (var resourceType in ResourceManifest.existingAtomPresetResourceTypes)
                {
                    resourceTypesList.Add(resourceType);
                }

                successCallback.actionCallback.Invoke();
            }
            catch (Exception e)
            {
                exceptionCallback.val = e.ToString();
            }
        }

        public static void OnGetExistingAndConcealedAtomPresetResourceTypes(object[] objects)
        {
            JSONStorableString exceptionCallback = objects[4] as JSONStorableString;
            try
            {
                HashSet<string> existingResourceTypesList = objects[0] as HashSet<string>;
                HashSet<string> concealedResourceTypesList = objects[1] as HashSet<string>;
                JSONStorableAction successCallback = objects[2] as JSONStorableAction;
                JSONStorableString failCallback = objects[3] as JSONStorableString;

                existingResourceTypesList.Clear();
                concealedResourceTypesList.Clear();
                foreach (var resourceType in ResourceManifest.existingAtomPresetResourceTypes)
                {
                    existingResourceTypesList.Add(resourceType);
                    if (Settings.IsResourceConcealed(ResourceType.presetAtom, resourceType)) concealedResourceTypesList.Add(resourceType);
                }

                successCallback.actionCallback.Invoke();
            }
            catch (Exception e)
            {
                exceptionCallback.val = e.ToString();
            }
        }

        public static void OnGetTagNames(object[] objects)
        {
            JSONStorableString exceptionCallback = objects[4] as JSONStorableString;
            try
            {
                int resourceType = (int)objects[0];
                List<string> tagsList = objects[1] as List<string>;
                JSONStorableAction successCallback = objects[2] as JSONStorableAction;
                JSONStorableString failCallback = objects[3] as JSONStorableString;
                string atomPresetResourceName = (string) objects[5];

                if (!EnumManifest.GetEnumsList(ResourceType.enumManifestName).Contains(resourceType)) failCallback.val = "BAInteropServer.OnGetTagNames: Failure occured retrieving all Tag Names from BrowserAssist, due to an unrecognised resource type "+resourceType.ToString();
                else
                {
                    tagsList.Clear();
                    foreach (var tag in TagManifest.GetTags(resourceType, atomPresetResourceName))
                    {
                        tagsList.Add(tag.tagName);
                    }

                    successCallback.actionCallback.Invoke();
                }
            }
            catch (Exception e)
            {
                exceptionCallback.val = e.ToString();
            }
        }

        public static void OnGetCreatorNames(object[] objects)
        {
            JSONStorableString exceptionCallback = objects[4] as JSONStorableString;
            try
            {
                int resourceType = (int)objects[0];
                List<string> creatorsList = objects[1] as List<string>;
                JSONStorableAction successCallback = objects[2] as JSONStorableAction;
                JSONStorableString failCallback = objects[3] as JSONStorableString;
                string atomPresetResourceName = (string)objects[5];

                if (!EnumManifest.GetEnumsList(ResourceType.enumManifestName).Contains(resourceType)) failCallback.val = "BAInteropServer.OnGetCreatorNames: Failure occured retrieving all Creator Names from BrowserAssist, due to an unrecognised resource type " + resourceType.ToString();
                else
                {
                    creatorsList.Clear();

                    foreach (var creatorName in CreatorManifest.GetAllCreatorNames(resourceType, atomPresetResourceName)) creatorsList.Add(creatorName);
                    creatorsList.Sort();
                    successCallback.actionCallback.Invoke();
                }
            }
            catch (Exception e)
            {
                exceptionCallback.val = e.ToString();
            }
        }

        protected static string ValidateFilter(BAOrderedResourceFilter baInteropFilter)
        {
            if (baInteropFilter.resourceOrder!=BAResultsSortModeEnum.userDefault && !EnumManifest.GetEnumsList(BrowserSortMode.enumManifestName).Contains(baInteropFilter.resourceOrder)) return "BAInteropServer.ValidateFilter: Failure occured due to an unrecognised resource Sorting Order ("+ baInteropFilter.resourceOrder.ToString()+") in the filter specification.";
            if (baInteropFilter.creatorType != BACreatorTypeEnum.userDefault && baInteropFilter.creatorType != BACreatorTypeEnum.specificCreator && !EnumManifest.GetEnumsList(CreatorFilterChoices.enumManifestName).Contains(baInteropFilter.creatorType)) return "BAInteropServer.ValidateFilter: Failure occured due to an unrecognised Creator Type ("+ baInteropFilter.creatorType.ToString()+") in the filter specification.";
            if (baInteropFilter.creatorType==BACreatorTypeEnum.specificCreator && !CreatorManifest.GetAllCreatorNames().Contains(baInteropFilter.creatorName)) return "BAInteropServer.ValidateFilter: Failure occured due to an unrecognised Creator Name ("+ baInteropFilter.creatorName+") in the filter specification.";
            if (baInteropFilter.creatorType == BACreatorTypeEnum.localFolders && baInteropFilter.localFolder!="" && !FileManagerSecure.DirectoryExists(baInteropFilter.localFolder)) return "BAInteropServer.ValidateFilter: Failure occured due to a non-existent local folder ("+ baInteropFilter.localFolder+") in the filter specification.";
            if ((baInteropFilter.creatorType == BACreatorTypeEnum.specificCreator || baInteropFilter.creatorType == BACreatorTypeEnum.allPackageCreators) && baInteropFilter.varLicenseType != BALicenseTypeEnum.userDefault && !EnumManifest.GetEnumsList(LicenseType.enumManifestName).Contains(baInteropFilter.varLicenseType)) return "BAInteropServer.ValidateFilter: Failure occured due to an unrecognised VAR License Type ("+ baInteropFilter.varLicenseType.ToString()+") in the filter specification.";

            if (baInteropFilter.resourceNameFilterMode != BAResourceNameFilterModeEnum.userDefault && !EnumManifest.GetEnumsList(ResourceNameSearchMode.enumManifestName).Contains(baInteropFilter.resourceNameFilterMode)) return "BAInteropServer.ValidateFilter: Failure occured due to an unrecognised Resource Name Filter Mode Order (" + baInteropFilter.resourceNameFilterMode.ToString() + ") in the filter specification.";

            if (!EnumManifest.GetEnumsList(ResourceType.enumManifestName).Contains(baInteropFilter.resourceType)) return "BAInteropServer.ValidateFilter: Failure occured due to an unrecognised Resource Type (" + baInteropFilter.resourceType.ToString() + ") in the filter specification.";

            if (baInteropFilter.tagGroupType != BATagGroupTypeEnum.userDefault && baInteropFilter.tagGroupType != BATagGroupTypeEnum.userTag && !EnumManifest.GetEnumsList(TagGroups.enumManifestName).Contains(baInteropFilter.tagGroupType)) return "BAInteropServer.ValidateFilter: Failure occured due to an unrecognised Tag Group Type (" + baInteropFilter.tagGroupType.ToString() + ") in the filter specification.";

            if (baInteropFilter.tagGroupType==BATagGroupTypeEnum.userTag && !TagManifest.GetTagNames(baInteropFilter.resourceType, baInteropFilter.atomPresetResourceType).Contains(baInteropFilter.userTagName)) return "BAInteropServer.ValidateFilter: Failure occured due to an unrecognised Tag Name (" + baInteropFilter.userTagName+ ") in the filter specification.";

            if (!EnumManifest.GetEnumsList(ResourceType.enumManifestName).Contains(baInteropFilter.resourceType)) return "BAInteropServer.ValidateFilter: Failure occured due to an unrecognised Resource Type (" + baInteropFilter.resourceType.ToString() + ") in the filter specification.";

            HashSet<int> boolEnums = new HashSet<int>(new[] { -1, 1, 2 });

            if (!boolEnums.Contains(baInteropFilter.showHidden)) return "BAInteropServer.ValidateFilter: Failure occured due to an unrecognised enumeration (" + baInteropFilter.showHidden.ToString() + ") for the 'showHidden' data item in the filter specification.";

            if (!boolEnums.Contains(baInteropFilter.onlyFav)) return "BAInteropServer.ValidateFilter: Failure occured due to an unrecognised enumeration (" + baInteropFilter.onlyFav.ToString() + ") for the 'onlyFav' data item in the filter specification.";

            if (!boolEnums.Contains(baInteropFilter.onlyNew)) return "BAInteropServer.ValidateFilter: Failure occured due to an unrecognised enumeration (" + baInteropFilter.onlyNew.ToString() + ") for the 'onlyNew' data item in the filter specification.";

            return "";
        }
        public static void OnSelectResourceFilename(object[] objects)
        {
            JSONStorableString exceptionCallback = objects[4] as JSONStorableString;
            try
            {
                Dictionary<int, object> filterDict = (Dictionary<int, object>)objects[0];

                Vector3 vrBrowserPosition = (Vector3) objects[1];
                JSONStorableString selectedResourceFileName = objects[2] as JSONStorableString;
                JSONStorableString failCallback = objects[3] as JSONStorableString;
                string browserTitle = objects[5] as string;
                int desktopDisplayMode = (int) objects[6] ;
                int baKeepOpenOption = (int)objects[7];

                BAOrderedResourceFilter baInteropFilter = UnpackSortedResourceFilterDict(filterDict);
                string failureMessage = ValidateFilter(baInteropFilter);
                if (failureMessage != "") failCallback.val = failureMessage;
                else
                {
                    ResourceBrowserFilterSettings filter = new ResourceBrowserFilterSettings(baInteropFilter, BrowserAssist.callbackBrowserUI.browserFiltersUI as CallbackBrowserFiltersUI);

                    if (BrowserAssist.callbackBrowserUI != null)
                    {
                        BrowserAssist.callbackBrowserUI.InitCallback(filter, browserTitle, selectedResourceFileName, baKeepOpenOption, vrBrowserPosition);


                        BrowserAssist.callbackBrowserUI.SetActive();

                        if (!Globals.isVREnabled)
                        {
                            if (desktopDisplayMode == BADesktopDisplayModeEnum.userDefault) desktopDisplayMode = BADesktopDisplayModeEnum.rightDocked;
                            if (desktopDisplayMode == BADesktopDisplayModeEnum.fullWidth) BrowserAssist.callbackBrowserUI.FullWidthUI();
                            else if (desktopDisplayMode == BADesktopDisplayModeEnum.leftDocked) BrowserAssist.callbackBrowserUI.LeftDockUI();
                            else if (desktopDisplayMode == BADesktopDisplayModeEnum.rightDocked) BrowserAssist.callbackBrowserUI.RightDockUI();
                        }
                    }                   
                }
            }
            catch (Exception e)
            {
                exceptionCallback.val = e.ToString();
            }
        }
        public static void OnSetMainResourceFilters(object[] objects)
        {
            JSONStorableString exceptionCallback = objects[3] as JSONStorableString;
            try
            {
                Dictionary<int, object> filterDict = (Dictionary<int, object>)objects[0];

                int desktopDisplayMode = (int)objects[1];
                JSONStorableString failCallback = objects[2] as JSONStorableString;

                int selectMode = BASelectModeEnum.single;
                if (filterDict.ContainsKey(selectModeKey)) selectMode = (int)filterDict[selectModeKey];
                
                BAOrderedResourceFilter baInteropFilter = UnpackSortedResourceFilterDict(filterDict);
                string failureMessage = ValidateFilter(baInteropFilter);
                if (failureMessage != "") failCallback.val = failureMessage;
                else
                {
                    ResourceBrowserFilterSettings filter = new ResourceBrowserFilterSettings(baInteropFilter);

                    if (!Settings.IsResourceConcealed(filter.resourceType, filter.atomPresetResourceType))
                    {
                        if (BrowserAssist.mainBrowserUI != null)
                        {
                            var browserFiltersUI = BrowserAssist.mainBrowserUI.browserFiltersUI;
                           
                            if (filter.resourceType == ResourceType.presetAtom) browserFiltersUI.currentAtomPresetResourceType = filter.atomPresetResourceType;
                            else browserFiltersUI.currentResourceType = filter.resourceType;
                            browserFiltersUI.currentFilterSettings.CopyFrom(filter);

                            if (!Globals.isVREnabled)
                            {
                                if (desktopDisplayMode == BADesktopDisplayModeEnum.userDefault) desktopDisplayMode = BADesktopDisplayModeEnum.rightDocked;
                                if (desktopDisplayMode == BADesktopDisplayModeEnum.fullWidth) BrowserAssist.callbackBrowserUI.FullWidthUI();
                                else if (desktopDisplayMode == BADesktopDisplayModeEnum.leftDocked) BrowserAssist.callbackBrowserUI.LeftDockUI();
                                else if (desktopDisplayMode == BADesktopDisplayModeEnum.rightDocked) BrowserAssist.callbackBrowserUI.RightDockUI();
                            }

                            browserFiltersUI.currentFilterSettings.selectModeJSEnum.val = selectMode;
                        }
                    }
                    else
                    {
                        string resourceTypeName = EnumManifest.GetDisplayVal(ResourceType.enumManifestName, filter.resourceType);
                        if (filter.resourceType == ResourceType.presetAtom) resourceTypeName = "Preset: "+filter.atomPresetResourceType;
                        SuperController.LogMessage("BrowserAssist: Attempt to remotely select a concealed resource type '" + resourceTypeName + "' was blocked.");
                    }

                    
                }
            }
            catch (Exception e)
            {
                exceptionCallback.val = e.ToString();
            }
        }

        public static void OnGetFilteredResourceFilenames(object[] objects)
        {
            JSONStorableString exceptionCallback = objects[4] as JSONStorableString;
            try
            {
                Dictionary<int, object> filterDict = (Dictionary<int, object>)objects[0];

                HashSet<string> resourceFileNames = objects[1] as HashSet<string>;
                JSONStorableAction successCallback = objects[2] as JSONStorableAction;
                JSONStorableString failCallback = objects[3] as JSONStorableString;

                BAOrderedResourceFilter baInteropFilter = UnpackResourceFilterDict(filterDict);
                string failureMessage = ValidateFilter(baInteropFilter);
                if (failureMessage!="") failCallback.val = failureMessage;
                else
                {
                    ResourceBrowserFilterSettings filter = new ResourceBrowserFilterSettings(baInteropFilter);
                    int resourceNameSearchMode = Settings.resourceNameSearchModeJSESC.val;
                    if (baInteropFilter.resourceNameFilterMode == BAResourceNameFilterModeEnum.StartsWith) resourceNameSearchMode = ResourceNameSearchMode.StartsWith;
                    else if (baInteropFilter.resourceNameFilterMode == BAResourceNameFilterModeEnum.Contains) resourceNameSearchMode = ResourceNameSearchMode.Contains;
                    else if (baInteropFilter.resourceNameFilterMode == BAResourceNameFilterModeEnum.ContainsFullPath) resourceNameSearchMode = ResourceNameSearchMode.ContainsFullPath;

                    var rvges = ResourceManifest.GetResourceGroups(filter, resourceNameSearchMode);
                    resourceFileNames.Clear();
                    foreach (var resource in rvges)
                    {
                        resourceFileNames.Add(resource.resourceFullPathReference);
                    }

                    successCallback.actionCallback.Invoke();
                }
            }
            catch (Exception e)
            {
                exceptionCallback.val = e.ToString();
            }
        }

        public static void OnGetOrderedFilteredResourceFilenames(object[] objects)
        {
            JSONStorableString exceptionCallback = objects[4] as JSONStorableString;
            try
            {
                Dictionary<int, object> filterDict = (Dictionary<int, object>)objects[0];

                List<string> resourceFileNames = objects[1] as List<string>;
                JSONStorableAction successCallback = objects[2] as JSONStorableAction;
                JSONStorableString failCallback = objects[3] as JSONStorableString;

                BAOrderedResourceFilter baInteropFilter = UnpackSortedResourceFilterDict(filterDict);
                string failureMessage = ValidateFilter(baInteropFilter);
                if (failureMessage != "") failCallback.val = failureMessage;
                else
                {
                    ResourceBrowserFilterSettings filter = new ResourceBrowserFilterSettings(baInteropFilter);
                    int resourceNameSearchMode = Settings.resourceNameSearchModeJSESC.val;
                    if (baInteropFilter.resourceNameFilterMode == BAResourceNameFilterModeEnum.StartsWith) resourceNameSearchMode = ResourceNameSearchMode.StartsWith;
                    else if (baInteropFilter.resourceNameFilterMode == BAResourceNameFilterModeEnum.Contains) resourceNameSearchMode = ResourceNameSearchMode.Contains;
                    else if (baInteropFilter.resourceNameFilterMode == BAResourceNameFilterModeEnum.ContainsFullPath) resourceNameSearchMode = ResourceNameSearchMode.ContainsFullPath;

                    var rvges = ResourceManifest.GetResourceGroups(filter, resourceNameSearchMode);
                    resourceFileNames.Clear();
                    foreach (var resource in rvges)
                    {
                        resourceFileNames.Add(resource.resourceFullPathReference);
                    }

                    successCallback.actionCallback.Invoke();
                }
            }
            catch (Exception e)
            {
                exceptionCallback.val = e.ToString();
            }
        }

        private static BAOrderedResourceFilter UnpackCreatorFilterDict(Dictionary<int, object> filterDict)
        {
            string atomPresetResourceName = "";
            if (filterDict.ContainsKey(atomPresetResourceNameKey)) atomPresetResourceName = (string)filterDict[atomPresetResourceNameKey];
            var filter = new BAOrderedResourceFilter((int)filterDict[resourceTypeKey], atomPresetResourceName);
            filter.tagGroupType = (int)filterDict[tagGroupTypeKey];

            if (filterDict.ContainsKey(userTagNameNameKey)) filter.userTagName = (string)filterDict[userTagNameNameKey];

            filter.showHidden = (int)filterDict[showHiddenKey];
            filter.onlyFav = (int)filterDict[onlyFavKey];
            filter.onlyNew = (int)filterDict[onlyNewKey];

            return filter;
        }

        private static BAOrderedResourceFilter UnpackResourceFilterDict(Dictionary<int, object> filterDict)
        {
            var filter = UnpackCreatorFilterDict(filterDict);

            filter.creatorType = (int)filterDict[creatorTypeKey];
            if (filterDict.ContainsKey(creatorNameKey)) filter.creatorName = (string)filterDict[creatorNameKey];
            if (filterDict.ContainsKey(localFolderKey)) filter.localFolder = (string)filterDict[localFolderKey];
            if (filterDict.ContainsKey(localFolderIncSubFolderKey)) filter.includeSubFolders = (bool)filterDict[localFolderIncSubFolderKey];

            if (filterDict.ContainsKey(allClothingCreatorsKey)) filter.allClothingCreator = (bool)filterDict[allClothingCreatorsKey];
            if (filterDict.ContainsKey(clothingCreatorNameKey)) filter.clothingCreatorName = (string)filterDict[clothingCreatorNameKey];

            if (filterDict.ContainsKey(varLicenseKey)) filter.varLicenseType = (int)filterDict[varLicenseKey];
            if (filterDict.ContainsKey(orderByKey)) filter.resourceOrder = (int)filterDict[orderByKey];

            filter.resourceNameFilterMode = (int)filterDict[resourceNameFilterModeKey];
            filter.resourceName = (string)filterDict[resourceNameKey];
            

            return filter;
        }

        private static BAOrderedResourceFilter UnpackSortedResourceFilterDict(Dictionary<int, object> filterDict)
        {
            BAOrderedResourceFilter filter = UnpackResourceFilterDict(filterDict);

            filter.resourceOrder = (int)filterDict[orderByKey];

            return filter;
        }

    }
}
