﻿using MVR.FileManagementSecure;
using SimpleJSON;
using System;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace JayJayWon
{
    public class DataFileManifest
    {
        private static Dictionary<string, VARResourcesCoreDataFile> varResourcesCoreDataFiles = new Dictionary<string, VARResourcesCoreDataFile>();
        private static HashSet<VARResourcesCoreDataFile> saveVARResourcesCoreDataFiles = new HashSet<VARResourcesCoreDataFile>();

        private static Dictionary<string, VARGroupCoreDataFile> varGroupCoreDataFiles = new Dictionary<string, VARGroupCoreDataFile>();
        private static HashSet<VARGroupCoreDataFile> saveVARGroupCoreDataFiles = new HashSet<VARGroupCoreDataFile>();

        private static Dictionary<string, VARResourcesUserDataFile> varResourcesUserDataFiles = new Dictionary<string, VARResourcesUserDataFile>();
        private static HashSet<VARResourcesUserDataFile> saveVARResourcesUserDataFiles = new HashSet<VARResourcesUserDataFile>();

        private static Dictionary<string, VARGroupUserDataFile> varGroupUserDataFiles = new Dictionary<string, VARGroupUserDataFile>();
        private static HashSet<VARGroupUserDataFile> saveVARGroupUserDataFiles = new HashSet<VARGroupUserDataFile>();

        private static Dictionary<string, LocalResourceUserDataFile> localResourceUserDataFiles = new Dictionary<string, LocalResourceUserDataFile>();
        private static HashSet<LocalResourceUserDataFile> saveLocalResourceUserDataFiles = new HashSet<LocalResourceUserDataFile>();

        public static void ClearLocalResourceUserDataFile()
        {
            localResourceUserDataFiles.Clear();
            saveLocalResourceUserDataFiles.Clear();
            LocalResourceUserDataFile.ClearLocalDFDict();
        }

        public static void RegisterExisitingDFCoreVARResource(PackagedRVGE rvge, string coreDataFileName)
        {

            var existingDataFile = DataFile.GetVARResourcesCoreDataFile(rvge);
            if (existingDataFile != null)
            {
                if (existingDataFile.fileName == coreDataFileName)
                {
                    SuperController.LogError("BA.DataFileManifest.RegisterExisitingDFCoreVARResource: Attempt to re-register RVGE '"+rvge.resourceFullPathReference+"' in "+coreDataFileName);
                    return;
                }
                else throw new Exception("BA.DataFileManifest.RegisterExisitingDFCoreVARResource: Attempt to register RVGE '" + rvge.resourceFullPathReference + "' in " + coreDataFileName + "' when its already registered to '"+existingDataFile.fileName+"'");
            }
            else
            {
                VARResourcesCoreDataFile coreDataFile;
                if (varResourcesCoreDataFiles.ContainsKey(coreDataFileName)) coreDataFile = varResourcesCoreDataFiles[coreDataFileName];
                else
                {
                    coreDataFile = new VARResourcesCoreDataFile(coreDataFileName);
                    varResourcesCoreDataFiles.Add(coreDataFileName, coreDataFile);
                }

                coreDataFile.RegisterExistingVARResource(rvge);
            }

            
        }
        public static void RegisterExisitingDFUserVARResource(PackagedRVGE rvge, string userDataFileName)
        {
            var existingDataFile = DataFile.GetVARResourcesUserDataFile(rvge);
            if (existingDataFile != null)
            {
                if (existingDataFile.fileName == userDataFileName)
                {
                    SuperController.LogMessage("BA.DataFileManifest.RegisterExisitingDFUserVARResource: Attempt to re-register RVGE '" + rvge.resourceFullPathReference + "' in " + userDataFileName + ". This should not re-occur. Please report the issue to JayJayWon on Discord if it re-occurs regularly on BrowserAssist startup");
                    RegisterVARResourceUserSaveRequest(userDataFileName, true);
                    return;
                }
                else
                {
                    SuperController.LogMessage("BA.DataFileManifest.RegisterExisitingDFUserVARResource: Attempt to register RVGE '" + rvge.resourceFullPathReference + "' in " + userDataFileName + "' when its already registered to '" + existingDataFile.fileName + "'. This should not re-occur. Please report the issue to JayJayWon on Discord if it re-occurs regularly on BrowserAssist startup");
                    RegisterVARResourceUserSaveRequest(userDataFileName, true);
                    RegisterVARResourceUserSaveRequest(existingDataFile.fileName, true);
                }
            }
            else
            {
                VARResourcesUserDataFile userDataFile;

                if (varResourcesUserDataFiles.ContainsKey(userDataFileName)) userDataFile = varResourcesUserDataFiles[userDataFileName];
                else
                {
                    userDataFile = new VARResourcesUserDataFile(userDataFileName);
                    varResourcesUserDataFiles.Add(userDataFileName, userDataFile);
                }

                userDataFile.RegisterExistingVARResource(rvge);
            }          
        }


        public static void RegisterNewDFCoreVARResource(PackagedRVGE rvge)
        {
            var existingDataFile = DataFile.GetVARResourcesCoreDataFile(rvge);
            if (existingDataFile != null)
            {
                SuperController.LogError("BA.DataFileManifest.RegisterNewDFCoreVARResource: Attempt to re-register RVGE '" + rvge.resourceFullPathReference + "' in " + existingDataFile.fileName);
                return;
            }
            else
            {
                VARResourcesCoreDataFile coreDataFile = null;

                var freeCoreDataFiles = varResourcesCoreDataFiles.Values.Where(x => x.dataFileFreeCapacity).ToList();

                if (freeCoreDataFiles.Count > 0) coreDataFile = freeCoreDataFiles.First();
                else
                {
                    coreDataFile = new VARResourcesCoreDataFile(varResourcesCoreDataFiles.Keys.ToList());
                    varResourcesCoreDataFiles.Add(coreDataFile.fileName, coreDataFile);
                }
                coreDataFile.RegisterNewVARResource(rvge);
            }

                
        }

        public static void RegisterNewDFUserVARResource(PackagedRVGE rvge)
        {
            var existingDataFile = DataFile.GetVARResourcesUserDataFile(rvge);
            if (existingDataFile != null)
            {
                SuperController.LogError("BA.DataFileManifest.RegisterNewDFUserVARResource: Attempt to re-register RVGE '" + rvge.resourceFullPathReference + "' in " + existingDataFile.fileName);
                return;
            }
            else
            {
                VARResourcesUserDataFile userDataFile = null;
                var freeUserDataFiles = varResourcesUserDataFiles.Values.Where(x => x.dataFileFreeCapacity).ToList();
                if (freeUserDataFiles.Count > 0) userDataFile = freeUserDataFiles.First();
                else
                {
                    userDataFile = new VARResourcesUserDataFile(varResourcesUserDataFiles.Keys.ToList());
                    varResourcesUserDataFiles.Add(userDataFile.fileName, userDataFile);
                }
                userDataFile.RegisterNewVARResource(rvge);
            }

                
        }

        public static void RegisterExisitingDFCoreVARGroup(VARPackageVersionGroupEntry pvge, string coreDataFileName)
        {
            var existingDataFile = DataFile.GetVARGroupCoreDataFile(pvge);
            if (existingDataFile != null)
            {
                if (existingDataFile.fileName == coreDataFileName)
                {
                    SuperController.LogError("BA.DataFileManifest.RegisterExisitingDFCoreVARGroup: Attempt to re-register PVGE '" + pvge.varGroupName + "' in " + coreDataFileName);
                    return;
                }
                else throw new Exception("BA.DataFileManifest.RegisterExisitingDFCoreVARGroup: Attempt to register PVGE '" + pvge.varGroupName + "' in " + coreDataFileName + "' when its already registered to '" + existingDataFile.fileName + "'");
            }
            else
            {
                VARGroupCoreDataFile coreDataFile;

                if (varGroupCoreDataFiles.ContainsKey(coreDataFileName)) coreDataFile = varGroupCoreDataFiles[coreDataFileName];
                else
                {
                    coreDataFile = new VARGroupCoreDataFile(coreDataFileName);
                    varGroupCoreDataFiles.Add(coreDataFileName, coreDataFile);
                }

                coreDataFile.RegisterExistingVARGroup(pvge);
            }


            
        }

        public static void RegisterExisitingDFUserVARGroup(VARPackageVersionGroupEntry pvge, string userDataFileName)
        {
            var existingDataFile = DataFile.GetVARGroupUserDataFile(pvge);
            if (existingDataFile != null)
            {
                if (existingDataFile.fileName == userDataFileName)
                {
                    SuperController.LogError("BA.DataFileManifest.RegisterExisitingDFUserVARGroup: Attempt to re-register PVGE '" + pvge.varGroupName + "' in " + userDataFileName);
                }
                else
                {
                    SuperController.LogError("BA.DataFileManifest.RegisterExisitingDFUserVARGroup: Attempt to register PVGE '" + pvge.varGroupName + "' in " + userDataFileName + "' when its already registered to '" + existingDataFile.fileName + "'");
                }
                RegisterVARGroupUserSaveRequest(userDataFileName);
            }
            else
            {
                VARGroupUserDataFile userDataFile;

                if (varGroupUserDataFiles.ContainsKey(userDataFileName)) userDataFile = varGroupUserDataFiles[userDataFileName];
                else
                {
                    userDataFile = new VARGroupUserDataFile(userDataFileName);
                    varGroupUserDataFiles.Add(userDataFileName, userDataFile);
                }

                userDataFile.RegisterExistingVARGroup(pvge);
            }

                
        }


        public static void RegisterNewDFCoreVARGroup(VARPackageVersionGroupEntry pvge)
        {
            var existingDataFile = DataFile.GetVARGroupCoreDataFile(pvge);
            if (existingDataFile != null)
            {
                SuperController.LogError("BA.DataFileManifest.RegisterNewDFCoreVARGroup: Attempt to re-register PVGE '" + pvge.varGroupName + "' in " + existingDataFile.fileName);
                return;
            }
            else
            {
                VARGroupCoreDataFile coreDataFile = null;

                var freeCoreDataFiles = varGroupCoreDataFiles.Values.Where(x => x.dataFileFreeCapacity).ToList();

                if (freeCoreDataFiles.Count > 0) coreDataFile = freeCoreDataFiles.First();
                else
                {
                    coreDataFile = new VARGroupCoreDataFile(varGroupCoreDataFiles.Keys.ToList());
                    varGroupCoreDataFiles.Add(coreDataFile.fileName, coreDataFile);
                }
                coreDataFile.RegisterNewVARGroup(pvge);
            }
        }

        public static void RegisterNewDFUserVARGroup(VARPackageVersionGroupEntry pvge)
        {
            var existingDataFile = DataFile.GetVARGroupUserDataFile(pvge);
            if (existingDataFile != null)
            {
                SuperController.LogError("BA.DataFileManifest.RegisterNewDFUserVARGroup: Attempt to re-register PVGE '" + pvge.varGroupName + "' in " + existingDataFile.fileName);
                return;
            }
            else
            {
                VARGroupUserDataFile userDataFile = null;
                var freeUserDataFiles = varGroupUserDataFiles.Values.Where(x => x.dataFileFreeCapacity).ToList();
                if (freeUserDataFiles.Count > 0) userDataFile = freeUserDataFiles.First();
                else
                {
                    userDataFile = new VARGroupUserDataFile(varGroupUserDataFiles.Keys.ToList());
                    varGroupUserDataFiles.Add(userDataFile.fileName, userDataFile);
                }
                userDataFile.RegisterNewVARGroup(pvge);
            }            
        }

        public static void RegisterExisitingDFUserLocalResource(LocalRVGE rvge, string userDataFileName)
        {
            var existingDataFile = DataFile.GetLocalDataFile(rvge);
            if (existingDataFile != null)
            {
                if (existingDataFile.fileName == userDataFileName)
                {
                    SuperController.LogError("BA.DataFileManifest.RegisterExisitingDFUserLocalResource: Attempt to re-register RVGE '" + rvge.resourceFullPathReference + "' in " + userDataFileName);
                    return;
                }
                else throw new Exception("BA.DataFileManifest.RegisterExisitingDFUserLocalResource: Attempt to register RVGE '" + rvge.resourceFullPathReference + "' in " + userDataFileName + "' when its already registered to '" + existingDataFile.fileName + "'");
            }
            else
            {
                LocalResourceUserDataFile userDataFile;

                if (localResourceUserDataFiles.ContainsKey(userDataFileName)) userDataFile = localResourceUserDataFiles[userDataFileName];
                else
                {
                    userDataFile = new LocalResourceUserDataFile(userDataFileName);
                    localResourceUserDataFiles.Add(userDataFileName, userDataFile);
                }

                userDataFile.RegisterExistingLocalResource(rvge);
            }

                
        }


        public static void RegisterNewDFUserLocalResource(LocalRVGE rvge)
        {

            var existingDataFile = DataFile.GetLocalDataFile(rvge);
            if (existingDataFile != null)
            {
                SuperController.LogError("BA.DataFileManifest.RegisterNewDFUserLocalResource: Attempt to re-register RVGE '" + rvge.resourceFullPathReference + "' in " + existingDataFile.fileName);
                return;
            }
            else
            {
                LocalResourceUserDataFile userDataFile = null;
                var freeUserDataFiles = localResourceUserDataFiles.Values.Where(x => x.dataFileFreeCapacity).ToList();
                if (freeUserDataFiles.Count > 0) userDataFile = freeUserDataFiles.First();
                else
                {
                    userDataFile = new LocalResourceUserDataFile(localResourceUserDataFiles.Keys.ToList());
                    localResourceUserDataFiles.Add(userDataFile.fileName, userDataFile);
                }
                userDataFile.RegisterNewLocalResource(rvge);
            }               
        }

        public static void RegisterVARGroupCoreSaveRequest(VARGroupCoreDataFile dataFile)
        {
            saveVARGroupCoreDataFiles.Add(dataFile);
        }

        public static void RegisterVARGroupCoreSaveRequest(string varGroupCoreFileName)
        {
             if (varGroupCoreDataFiles.ContainsKey(varGroupCoreFileName)) RegisterVARGroupCoreSaveRequest(varGroupCoreDataFiles[varGroupCoreFileName]);
        }

        public static void RegisterVARGroupUserSaveRequest(VARGroupUserDataFile dataFile)
        {
            saveVARGroupUserDataFiles.Add(dataFile);
        }
        public static void RegisterVARGroupUserSaveRequest(string dataFileName)
        {
            if (varGroupUserDataFiles.ContainsKey(dataFileName)) RegisterVARGroupUserSaveRequest(varGroupUserDataFiles[dataFileName]);
        }

        public static void RegisterVARResourceCoreSaveRequest(string varResourceCoreFileName)
        {
            if (varResourcesCoreDataFiles.ContainsKey(varResourceCoreFileName)) RegisterVARResourceCoreSaveRequest(varResourcesCoreDataFiles[varResourceCoreFileName]);
        }
        public static void RegisterVARResourceCoreSaveRequest(VARResourcesCoreDataFile dataFile)
        {
            saveVARResourcesCoreDataFiles.Add(dataFile);
        }


        public static void RegisterVARResourceUserSaveRequest(string dataFileName, bool forceCreateMissingDataFile=false)
        {
            if (varResourcesUserDataFiles.ContainsKey(dataFileName)) RegisterVARResourceUserSaveRequest(varResourcesUserDataFiles[dataFileName]);
            else if (forceCreateMissingDataFile)
            {
                var userDataFile = new VARResourcesUserDataFile(dataFileName);
                varResourcesUserDataFiles.Add(dataFileName, userDataFile);
                RegisterVARResourceUserSaveRequest(userDataFile);
            }
        }
        public static void RegisterVARResourceUserSaveRequest(VARResourcesUserDataFile dataFile)
        {
            saveVARResourcesUserDataFiles.Add(dataFile);
        }

        public static void CancelAllVARSaveRequests()
        {
            saveVARResourcesUserDataFiles.Clear();
            saveVARResourcesCoreDataFiles.Clear();
            saveVARGroupUserDataFiles.Clear();
            saveVARGroupCoreDataFiles.Clear();

        }


        public static void RegisterLocalResourceUserSaveRequest(LocalResourceUserDataFile dataFile)
        {
            saveLocalResourceUserDataFiles.Add(dataFile);
        }

        public static void RegisterLocalResourceUserSaveRequest(string dataFileName)
        {
            if (!localResourceUserDataFiles.ContainsKey(dataFileName))
            {
                var userDataFile = new LocalResourceUserDataFile(dataFileName);
                localResourceUserDataFiles.Add(dataFileName, userDataFile);
            }

            RegisterLocalResourceUserSaveRequest(localResourceUserDataFiles[dataFileName]);
        }

        public static void Update()
        {
            
            if (saveVARResourcesCoreDataFiles.Count > 0)
            {
                foreach (var varResourcesData in saveVARResourcesCoreDataFiles)
                {
                    JSONClass jc = varResourcesData.GetJSON();
                    try
                    {
                        SuperController.singleton.SaveJSON(jc, varResourcesData.fileName);
                    }
                    catch (Exception e) { SuperController.LogError("Exception in BA.DataFileManifest.Update for saveVARResourcesCoreDataFiles " + varResourcesData.fileName + ": " + e); }
                }
                saveVARResourcesCoreDataFiles.Clear();
            }

            if (saveVARGroupCoreDataFiles.Count > 0)
            {

                foreach (var varGroupData in saveVARGroupCoreDataFiles)
                {
                    JSONClass jc = varGroupData.GetJSON();
                    try
                    {
                        SuperController.singleton.SaveJSON(jc, varGroupData.fileName);
                    }
                    catch (Exception e) { SuperController.LogError("Exception in BA.DataFileManifest.Update for saveVARGroupCoreDataFiles " + varGroupData.fileName + ": " + e); }                    
                }
                saveVARGroupCoreDataFiles.Clear();
            }

            if (saveVARResourcesUserDataFiles.Count > 0)
            {
                foreach (var varResourcesUserData in saveVARResourcesUserDataFiles)
                {
                    JSONClass jc = varResourcesUserData.GetJSON();
                    try
                    {
                        SuperController.singleton.SaveJSON(jc, varResourcesUserData.fileName);
                    }
                    catch (Exception e) { SuperController.LogError("Exception in BA.DataFileManifest.Update for saveVARResourcesUserDataFiles " + varResourcesUserData.fileName + ": " + e); }                    
                }
                saveVARResourcesUserDataFiles.Clear();
            }

            if (saveVARGroupUserDataFiles.Count > 0)
            {

                foreach (var varGroupUserData in saveVARGroupUserDataFiles)
                {
                    JSONClass jc = varGroupUserData.GetJSON();
                    try
                    {
                        SuperController.singleton.SaveJSON(jc, varGroupUserData.fileName);
                    }
                    catch (Exception e) { SuperController.LogError("Exception in BA.DataFileManifest.Update for saveVARGroupUserDataFiles " + varGroupUserData.fileName + ": " + e); }
                }
                saveVARGroupUserDataFiles.Clear();
            }
            if (saveLocalResourceUserDataFiles.Count > 0)
            {
                foreach (var localResourceUserData in saveLocalResourceUserDataFiles)
                {
                    JSONClass jc = localResourceUserData.GetJSON();
                    try {
                        if (jc["resources"].AsArray.Count == 0) FileManagerSecure.DeleteFile(localResourceUserData.fileName);
                        else  SuperController.singleton.SaveJSON(jc, localResourceUserData.fileName);
                    }
                    catch (Exception e) { SuperController.LogError("Exception in BA.DataFileManifest.Update for saveLocalResourceUserDataFiles " + localResourceUserData.fileName + ": " + e); }
                }
                saveLocalResourceUserDataFiles.Clear();
            }
        }
    }

    public abstract class DataFile
    {
        public const int MaxCapacity = 1000;
        public string fileName { get; protected set; }
        public DataFile(string fileName)
        {
            this.fileName = fileName;
        }

        protected static Dictionary<LocalRVGE, LocalResourceUserDataFile> rvgeLocalDataFileDict = new Dictionary<LocalRVGE, LocalResourceUserDataFile>();

        protected static Dictionary<PackagedRVGE, VARResourcesCoreDataFile> rvgeVARResourcesCoreDataFileDict = new Dictionary<PackagedRVGE, VARResourcesCoreDataFile>();

        protected static Dictionary<PackagedRVGE, VARResourcesUserDataFile> rvgeVARResourcesUserDataFileDict = new Dictionary<PackagedRVGE, VARResourcesUserDataFile>();

        protected static Dictionary<VARPackageVersionGroupEntry, VARGroupCoreDataFile> pvgeVARGroupCoreDataFileDict = new Dictionary<VARPackageVersionGroupEntry, VARGroupCoreDataFile>();

        protected static Dictionary<VARPackageVersionGroupEntry, VARGroupUserDataFile> pvgeVARGroupUserDataFileDict = new Dictionary<VARPackageVersionGroupEntry, VARGroupUserDataFile>();


        public static LocalResourceUserDataFile GetLocalDataFile(LocalRVGE rvge)
        {
            if (!rvgeLocalDataFileDict.ContainsKey(rvge)) return null;
            return rvgeLocalDataFileDict[rvge];
        }
        public static VARResourcesCoreDataFile GetVARResourcesCoreDataFile(PackagedRVGE rvge)
        {
            if (!rvgeVARResourcesCoreDataFileDict.ContainsKey(rvge)) return null;
            return rvgeVARResourcesCoreDataFileDict[rvge];
        }
        public static VARResourcesUserDataFile GetVARResourcesUserDataFile(PackagedRVGE rvge)
        {
            if (!rvgeVARResourcesUserDataFileDict.ContainsKey(rvge)) return null;
            return rvgeVARResourcesUserDataFileDict[rvge];
        }

        public static VARGroupCoreDataFile GetVARGroupCoreDataFile(VARPackageVersionGroupEntry pvge)
        {
            if (!pvgeVARGroupCoreDataFileDict.ContainsKey(pvge)) return null;
            return pvgeVARGroupCoreDataFileDict[pvge];
        }
        public static VARGroupUserDataFile GetVARGroupUserDataFile(VARPackageVersionGroupEntry pvge)
        {
            if (!pvgeVARGroupUserDataFileDict.ContainsKey(pvge)) return null;
            return pvgeVARGroupUserDataFileDict[pvge];
        }
        public abstract bool dataFileFreeCapacity { get; }
        public abstract int contentItemCount { get; }

        private static string GetDataFileName(int index, string folderName, string fileNameWithoutIndex, string fileExtension)
        {
            return folderName + "\\" + fileNameWithoutIndex + index.ToString() + fileExtension;
        }

        protected static string GetDataFileName(string folderName, string fileNameWithIndex, string fileExtension)
        {
            return folderName + "\\" + fileNameWithIndex + fileExtension;
        }

        protected static string GetNextDataFileName(string folderName, string fileName, string fileExtension, List<string> existingDataFileNames)
        {
            int dataFileIndex = 1;
            string nextVARGroupDataFileName = GetDataFileName(dataFileIndex, folderName, fileName, fileExtension);
            while (FileManagerSecure.FileExists(nextVARGroupDataFileName) || existingDataFileNames.Contains(nextVARGroupDataFileName))
            {
                dataFileIndex++;
                nextVARGroupDataFileName = GetDataFileName(dataFileIndex, folderName, fileName, fileExtension);
            }

            return nextVARGroupDataFileName;
        }

    }

    public class VARResourcesCoreDataFile : DataFile
    {
        private HashSet<PackagedRVGE> packagedRVGEs = new HashSet<PackagedRVGE>();

        public VARResourcesCoreDataFile(string coreFileName) : base(coreFileName)
        {
        }
        public VARResourcesCoreDataFile(List<string> existingCoreDataFileNames) : base("")
        {
            fileName = GetNextDataFileName(Const.VARResourcesCoreDataFilesFolder, Const.VARResourceDataFileName, Const.CoreDataFileExt, existingCoreDataFileNames);
        }

        public override bool dataFileFreeCapacity
        {
            get
            {
                return packagedRVGEs.Count < MaxCapacity;
            }
        }

        public override int contentItemCount
        {
            get
            {
                return packagedRVGEs.Count;
            }
        }

        public void RegisterExistingVARResource(PackagedRVGE rvge)
        {
            if (rvgeVARResourcesCoreDataFileDict.ContainsKey(rvge))
            {
                SuperController.LogError("BA.VARResourcesCoreDataFile.RegisterExistingVARResource: Invalid attempt to register RVGE '" + rvge.resourceFullPathReference + "' in DataFile '"+fileName+"' when its already registered to '" + rvgeVARResourcesCoreDataFileDict[rvge].fileName+"'");
                rvge.RegisterCoreDataFile(rvgeVARResourcesCoreDataFileDict[rvge]);
                return;
            }
            rvgeVARResourcesCoreDataFileDict.Add(rvge, this);
            packagedRVGEs.Add(rvge);
            rvge.RegisterCoreDataFile(this);

        }

        public void RegisterNewVARResource(PackagedRVGE rvge)
        {
            RegisterExistingVARResource(rvge);
            DataFileManifest.RegisterVARResourceCoreSaveRequest(this);
        }

        public void UnRegisterVARResource(PackagedRVGE rvge)
        {
            packagedRVGEs.Remove(rvge);
            DataFileManifest.RegisterVARResourceCoreSaveRequest(this);
        }

        public JSONClass GetJSON()
        {
            JSONClass jc = new JSONClass();
            jc["VARManifestStoreFormat"].AsInt = Const.VARManifestStoreFormat;

            jc["BAMajorVersion"].AsInt = Const.BAMajorVersion;
            jc["BAMinorVersion"].AsInt = Const.BAMinorVersion;
            jc["BAFixVersion"].AsInt = Const.BAFixVersion;

            JSONArray ja = new JSONArray();
            foreach (var rvge in packagedRVGEs)
            {
                ja[-1] = rvge.GetCoreJSON();
            }
            jc["resources"] = ja;

            return jc;
        }
    }

    public class VARResourcesUserDataFile : DataFile
    {
        private HashSet<PackagedRVGE> packagedRVGEs = new HashSet<PackagedRVGE>();

        public VARResourcesUserDataFile(string fileName) : base(fileName)
        {
        }

        public VARResourcesUserDataFile(List<string> existingUserDataFileNames) : base("")
        {
            fileName = GetNextDataFileName(Const.VARResourcesUserDataFilesFolder, Const.VARResourceDataFileName, Const.UserDataFileExt, existingUserDataFileNames);
        }

        public override bool dataFileFreeCapacity
        {
            get
            {
                return packagedRVGEs.Count < MaxCapacity;
            }
        }

        public override int contentItemCount
        {
            get
            {
                return packagedRVGEs.Count;
            }
        }

        public void RegisterExistingVARResource(PackagedRVGE rvge)
        {
            if (rvgeVARResourcesUserDataFileDict.ContainsKey(rvge))
            {
                SuperController.LogError("BA.VARResourcesUserDataFile.RegisterExistingVARResource: Invalid attempt to register RVGE '" + rvge.resourceFullPathReference + "' in DataFile '" + fileName + "' when its already registered to '" + rvgeVARResourcesUserDataFileDict[rvge].fileName + "'");
                rvge.RegisterUserDataFile(rvgeVARResourcesUserDataFileDict[rvge]);
                return;
            }

            packagedRVGEs.Add(rvge);
            rvgeVARResourcesUserDataFileDict.Add(rvge, this);
            rvge.RegisterUserDataFile(this);

        }

        public void RegisterNewVARResource(PackagedRVGE rvge)
        {
            RegisterExistingVARResource(rvge);
            DataFileManifest.RegisterVARResourceUserSaveRequest(this);
        }

        public void UnRegisterVARResource(PackagedRVGE rvge)
        {
            packagedRVGEs.Remove(rvge);
            DataFileManifest.RegisterVARResourceUserSaveRequest(this);
        }

        public JSONClass GetJSON()
        {
            JSONClass jc = new JSONClass();
            jc["VARUserDataStoreFormat"].AsInt = Const.VARUserDataStoreFormat;

            jc["BAMajorVersion"].AsInt = Const.BAMajorVersion;
            jc["BAMinorVersion"].AsInt = Const.BAMinorVersion;
            jc["BAFixVersion"].AsInt = Const.BAFixVersion;

            JSONArray ja = new JSONArray();
            foreach (var rvge in packagedRVGEs)
            {
                JSONClass rvgeJC = rvge.GetVARUserDataJSON();
                if (rvgeJC != null) ja[-1] = rvgeJC;
            }
            jc["resources"] = ja;

            return jc;
        }
    }
    public class VARGroupCoreDataFile : DataFile
    {
        private HashSet<VARPackageVersionGroupEntry> varPVGEs = new HashSet<VARPackageVersionGroupEntry>();
        public VARGroupCoreDataFile(string coreFileName) : base(coreFileName)
        {
        }

        public VARGroupCoreDataFile(List<string> existingCoreDataFileNames) : base("")
        {
            fileName = GetNextDataFileName(Const.VARGroupCoreDataFilesFolder, Const.VARGroupDataFileName, Const.CoreDataFileExt, existingCoreDataFileNames);
        }

        public override bool dataFileFreeCapacity
        {
            get
            {
                return varPVGEs.Count < MaxCapacity;
            }
        }

        public override int contentItemCount
        {
            get
            {
                return varPVGEs.Count;
            }
        }
        public void RegisterExistingVARGroup(VARPackageVersionGroupEntry pvge)
        {
            if (pvgeVARGroupCoreDataFileDict.ContainsKey(pvge))
            {
                SuperController.LogError("BA.VARGroupCoreDataFile.RegisterExistingVARGroup: Invalid attempt to register PVGE '" + pvge.varGroupName + "' in DataFile '" + fileName + "' when its already registered to '" + pvgeVARGroupCoreDataFileDict[pvge].fileName + "'");
                pvge.RegisterCoreDataFile(pvgeVARGroupCoreDataFileDict[pvge]);
                return;
            }

            varPVGEs.Add(pvge);
            pvgeVARGroupCoreDataFileDict.Add(pvge, this);
            pvge.RegisterCoreDataFile(this);
        }

        public void RegisterNewVARGroup(VARPackageVersionGroupEntry pvge)
        {
            RegisterExistingVARGroup(pvge);
            DataFileManifest.RegisterVARGroupCoreSaveRequest(this);
        }

        public void UnRegisterVARGroup(VARPackageVersionGroupEntry pvge)
        {
            varPVGEs.Remove(pvge);
            DataFileManifest.RegisterVARGroupCoreSaveRequest(this);
        }

        public JSONClass GetJSON()
        {
            JSONClass jc = new JSONClass();
            jc["VARManifestStoreFormat"].AsInt = Const.VARManifestStoreFormat;

            jc["BAMajorVersion"].AsInt = Const.BAMajorVersion;
            jc["BAMinorVersion"].AsInt = Const.BAMinorVersion;
            jc["BAFixVersion"].AsInt = Const.BAFixVersion;

            JSONArray ja = new JSONArray();
            foreach (var pvge in varPVGEs)
            {
                ja[-1] = pvge.GetVARGroupMetaDataJSON();
            }
            jc["varGroups"] = ja;

            return jc;
        }
    }

    public class VARGroupUserDataFile : DataFile
    {
        private HashSet<VARPackageVersionGroupEntry> varPVGEs = new HashSet<VARPackageVersionGroupEntry>();
        public VARGroupUserDataFile(string fileName) : base(fileName)
        {

        }

        public VARGroupUserDataFile(List<string> existingUserDataFileNames) : base("")
        {
            fileName = GetNextDataFileName(Const.VARGroupUserDataFilesFolder, Const.VARGroupDataFileName, Const.UserDataFileExt, existingUserDataFileNames);
        }

        public override bool dataFileFreeCapacity
        {
            get
            {
                return varPVGEs.Count < MaxCapacity;
            }
        }

        public override int contentItemCount
        {
            get
            {
                return varPVGEs.Count;
            }
        }
        public void RegisterExistingVARGroup(VARPackageVersionGroupEntry pvge)
        {
            if (pvgeVARGroupUserDataFileDict.ContainsKey(pvge))
            {
                SuperController.LogError("BA.VARGroupUserDataFile.RegisterExistingVARGroup: Invalid attempt to register PVGE '" + pvge.varGroupName + "' in DataFile '" + fileName + "' when its already registered to '" + pvgeVARGroupUserDataFileDict[pvge].fileName + "'");
                pvge.RegisterUserDataFile(pvgeVARGroupUserDataFileDict[pvge]);
                return;
            }


            varPVGEs.Add(pvge);
            pvgeVARGroupUserDataFileDict.Add(pvge, this);
            pvge.RegisterUserDataFile(this);
        }

        public void RegisterNewVARGroup(VARPackageVersionGroupEntry pvge)
        {
            RegisterExistingVARGroup(pvge);
            DataFileManifest.RegisterVARGroupUserSaveRequest(this);
        }

        public void UnRegisterVARGroup(VARPackageVersionGroupEntry pvge)
        {
            varPVGEs.Remove(pvge);
            DataFileManifest.RegisterVARGroupUserSaveRequest(this);
        }
        public JSONClass GetJSON()
        {
            JSONClass jc = new JSONClass();
            jc["VARUserDataStoreFormat"].AsInt = Const.VARUserDataStoreFormat;

            jc["BAMajorVersion"].AsInt = Const.BAMajorVersion;
            jc["BAMinorVersion"].AsInt = Const.BAMinorVersion;
            jc["BAFixVersion"].AsInt = Const.BAFixVersion;

            JSONArray ja = new JSONArray();
            foreach (var pvge in varPVGEs)
            {
                ja[-1] = pvge.GetVARGroupUserDataJSON();
            }
            jc["varGroups"] = ja;

            return jc;
        }
    }

    public class LocalResourceUserDataFile : DataFile
    {
        private HashSet<LocalRVGE> localRVGEs = new HashSet<LocalRVGE>();
        public LocalResourceUserDataFile(string fileName) : base(fileName)
        {

        }

        public LocalResourceUserDataFile(List<string> existingUserDataFileNames) : base("")
        {
            fileName = GetNextDataFileName(Const.LocalResourcesUserDataFolder, Const.LocalResourceDataFileName, Const.UserDataFileExt, existingUserDataFileNames);
        }

        public static void ClearLocalDFDict()
        {
            rvgeLocalDataFileDict.Clear();
        }

        public override bool dataFileFreeCapacity
        {
            get
            {
                return localRVGEs.Count < MaxCapacity;
            }
        }

        public override int contentItemCount
        {
            get
            {
                return localRVGEs.Count;
            }
        }
        public void RegisterExistingLocalResource(LocalRVGE rvge)
        {
            localRVGEs.Add(rvge);
            rvge.RegisterUserDataFile(this);
        }

        public void RegisterNewLocalResource(LocalRVGE rvge)
        {
            RegisterExistingLocalResource(rvge);
            DataFileManifest.RegisterLocalResourceUserSaveRequest(this);
        }

        public void PurgeLocalRVGE(LocalRVGE rvge)
        {
            localRVGEs.Remove(rvge);
            rvgeLocalDataFileDict.Remove(rvge);
            rvge.DeRegisterUserDataFile();
            DataFileManifest.RegisterLocalResourceUserSaveRequest(this);
        }

        public JSONClass GetJSON()
        {
            JSONClass jc = new JSONClass();
            jc["LocalUserDataStoreFormat"].AsInt = Const.LocalUserDataStoreFormat;

            jc["BAMajorVersion"].AsInt = Const.BAMajorVersion;
            jc["BAMinorVersion"].AsInt = Const.BAMinorVersion;
            jc["BAFixVersion"].AsInt = Const.BAFixVersion;

            JSONArray ja = new JSONArray();
            foreach (var rvge in localRVGEs)
            {
                ja[-1] = (rvge as LocalRVGE).GetCoreJSON();
            }
            jc["resources"] = ja;

            return jc;
        }
    }


}
