using System.Collections.Generic;
using UnityEngine;
using UnityEngine.SceneManagement;

namespace JayJayWon
{
    public static class EnumManifest
    {
        private static Dictionary<string, List<int>> enumManifestDict;
        private static Dictionary<string, string> enumValStoreValDict;
        private static Dictionary<string, string> enumValAltDisplayValDict;
        private static Dictionary<string, int> storeValEnumValDict;
        private static Dictionary<string, int> altDisplayValEnumValDict;

        static EnumManifest()
        {
            enumManifestDict = new Dictionary<string, List<int>>();
            enumValStoreValDict = new Dictionary<string, string>();
            enumValAltDisplayValDict = new Dictionary<string, string>();
            storeValEnumValDict = new Dictionary<string, int>();
            altDisplayValEnumValDict = new Dictionary<string, int>();
        }

        static public void AddEnum(string enumName, int enumVal)
        {
            string enumValKey = EnumValKey(enumName, enumVal);
            if (!enumManifestDict.ContainsKey(enumName)) enumManifestDict.Add(enumName, new List<int>());
            if (!enumManifestDict[enumName].Contains(enumVal)) enumManifestDict[enumName].Add(enumVal);
            else SuperController.LogError("BA.EnumManifest.AddEnum 1: Attempt to add existing enumeration value - " + enumValKey);
        }

        static private string EnumValKey(string enumName, int enumVal)
        {
            return enumName + "." + enumVal.ToString();
        }
        static private string EnumLabelKey(string enumName, string enumLabel)
        {
            return enumName + "." + enumLabel;
        }

        static public void AddEnum(string enumName, int enumVal, string storeValLabel)
        {
            AddEnum(enumName, enumVal);
            string enumValKey = EnumValKey(enumName, enumVal);
            string enumStoreLabelKey = EnumLabelKey(enumName, storeValLabel);
            if (!enumValStoreValDict.ContainsKey(enumName))
            {
                enumValStoreValDict.Add(enumValKey, storeValLabel);
                storeValEnumValDict.Add(enumStoreLabelKey, enumVal);
            }
            else SuperController.LogError("BA.EnumManifest.AddEnum 2: Attempt to add existing enumeration value - " + enumValKey);
        }

        static public void AddEnum(string enumName, int enumVal, string storeValLabel, string altLabel)
        {
            AddEnum(enumName, enumVal, storeValLabel);
            string enumValKey = EnumValKey(enumName, enumVal);
            string enumAltLabelKey = EnumLabelKey(enumName, altLabel);
            if (altLabel != null)
            {
                if (!enumValAltDisplayValDict.ContainsKey(enumName))
                {
                    enumValAltDisplayValDict.Add(enumValKey, altLabel);
                    altDisplayValEnumValDict.Add(enumAltLabelKey, enumVal);
                }
                else SuperController.LogError("BA.EnumManifest.AddEnum 3: Attempt to add existing enumeration value - " + enumValKey);
            }

        }

        static public bool Exists(string enumName, int enumVal)
        {
            if (!enumManifestDict.ContainsKey(enumName)) return false;
            if (!enumManifestDict[enumName].Contains(enumVal)) return false;
            return (true);
        }

        static public string GetDisplayVal(string enumName, int enumVal)
        {
            string enumValKey = EnumValKey(enumName, enumVal);
            if (Exists(enumName, enumVal))
            {
                if (enumValAltDisplayValDict.ContainsKey(enumValKey)) return enumValAltDisplayValDict[enumValKey];
                if (enumValStoreValDict.ContainsKey(enumValKey)) return enumValStoreValDict[enumValKey];
            }
            else SuperController.LogError("BA.EnumManifest.GetLabel: Attempt to retrieve display val for non-existing enumeration value - " + enumValKey);
            return ("");
        }
        static public string GetStoreVal(string enumName, int enumVal)
        {
            string enumValKey = EnumValKey(enumName, enumVal);
            if (Exists(enumName, enumVal))
            {
                if (enumValStoreValDict.ContainsKey(enumValKey)) return enumValStoreValDict[enumValKey];
            }
            else SuperController.LogError("BA.EnumManifest.GetStoreVal: Attempt to retrieve Store Value for non-existing enumeration value - " + enumValKey);
            return ("");
        }


        static public int GetEnumValFromStoreVal(string enumName, string storeValLabel)
        {
            string enumStoreLabelKey = EnumLabelKey(enumName, storeValLabel);
            if (storeValEnumValDict.ContainsKey(enumStoreLabelKey)) return storeValEnumValDict[enumStoreLabelKey];
            else SuperController.LogError("BA.EnumManifest.GetEnumValFromStoreVal: Attempt to retrieve enumeration Value for non-existing enumeration Label - " + enumStoreLabelKey);
            return (-1);
        }

        static public int GetEnumValFromDisplayVal(string enumName, string displayVal)
        {
            string enumStoreLabelKey = EnumLabelKey(enumName, displayVal);
            if (altDisplayValEnumValDict.ContainsKey(enumStoreLabelKey)) return altDisplayValEnumValDict[enumStoreLabelKey];
            else if (storeValEnumValDict.ContainsKey(enumStoreLabelKey)) return storeValEnumValDict[enumStoreLabelKey];
            else SuperController.LogError("BA.EnumManifest.GetEnumValFromDisplayVal: Attempt to retrieve enumeration Value for non-existing enumeration Display Val - " + enumStoreLabelKey);
            return (-1);
        }

        public static List<string> GetDisplayValList(string enumName)
        {
            List<string> displayVals = new List<string>();
            if (enumManifestDict.ContainsKey(enumName))
            {
                foreach (KeyValuePair<string, string> kvp in enumValStoreValDict)
                {
                    if (kvp.Key.StartsWith(enumName + "."))
                    {
                        if (enumValAltDisplayValDict.ContainsKey(kvp.Key)) displayVals.Add(enumValAltDisplayValDict[kvp.Key]);
                        else displayVals.Add(enumValStoreValDict[kvp.Key]);
                    }
                }
            }

            return (displayVals);
        }

        public static List<string> GetStoreValList(string enumName)
        {
            List<string> storeVals = new List<string>();
            if (enumManifestDict.ContainsKey(enumName))
            {
                foreach (KeyValuePair<string, string> kvp in enumValStoreValDict)
                {
                    if (kvp.Key.StartsWith(enumName + "."))
                    {
                        storeVals.Add(enumValStoreValDict[kvp.Key]);
                    }
                }
            }

            return (storeVals);
        }


        public static List<int> GetEnumsList(string enumName)
        {
            if (enumName == null) return null;
            List<int> enums = new List<int>();
            if (enumManifestDict.ContainsKey(enumName))
            {
                foreach (int enumVal in enumManifestDict[enumName])
                {

                    enums.Add(enumVal);
                }
            }

            return (enums);
        }
        public static SortedDictionary<int, string> GetEnumsDisplayValDict(string enumName)
        {
            if (enumName == null) return null;
            SortedDictionary<int, string> enumsDict = new SortedDictionary<int, string>();
            if (enumManifestDict.ContainsKey(enumName))
            {
                foreach (int enumVal in enumManifestDict[enumName])
                {
                    enumsDict.Add(enumVal, GetDisplayVal(enumName, enumVal));
                }
            }

            return (enumsDict);
        }

        public static Dictionary<int, string> GetEnumsStoreValDict(string enumName)
        {
            if (enumName == null) return null;
            Dictionary<int, string> enumsDict = new Dictionary<int, string>();
            if (enumManifestDict.ContainsKey(enumName))
            {
                foreach (int enumVal in enumManifestDict[enumName])
                {
                    enumsDict.Add(enumVal, GetStoreVal(enumName, enumVal));
                }
            }

            return (enumsDict);
        }

        public static Dictionary<string, int> GetStoreValEnumsDict(string enumName)
        {
            if (enumName == null) return null;
            Dictionary<string, int> enumsDict = new Dictionary<string, int>();
            if (enumManifestDict.ContainsKey(enumName))
            {
                foreach (int enumVal in enumManifestDict[enumName])
                {
                    enumsDict.Add(GetStoreVal(enumName, enumVal), enumVal);
                }
            }

            return (enumsDict);
        }
        public static Dictionary<string, int> GetDisplayValEnumsDict(string enumName)
        {
            if (enumName == null) return null;
            Dictionary<string, int> enumsDict = new Dictionary<string, int>();
            if (enumManifestDict.ContainsKey(enumName))
            {
                foreach (int enumVal in enumManifestDict[enumName])
                {
                    enumsDict.Add(GetDisplayVal(enumName, enumVal), enumVal);
                }
            }

            return (enumsDict);
        }

    }

    public static class DesktopDisplayMode
    {
        public const int rightDocked = BADesktopDisplayModeEnum.rightDocked;
        public const int fullWidth = BADesktopDisplayModeEnum.fullWidth;
        public const int leftDocked = BADesktopDisplayModeEnum.leftDocked;
        public const int leftMinimized = 4;
        public const int rightMinimized = 5;
    }

    public static class ActionUITabType
    {
        public const int primaryActionTab = 1;
        public const int tagsTab = 2;
        public const int positionTab = 3;
        public const int saveTab = 4;
        public const int clothingPresetTab = 5;       
        public const int sceneImportTab = 6;
        public const int quickFilterTab = 7;
        public const int infoTab = 20;


        private static Dictionary<int, string> actionUITabTypeLabels = new Dictionary<int, string>();

        public static string GetTabLabel(int tabType)
        {
            return actionUITabTypeLabels[tabType];
        }
        static ActionUITabType()
        {
            actionUITabTypeLabels.Add(primaryActionTab, "Action");
            actionUITabTypeLabels.Add(positionTab, "Position");
            actionUITabTypeLabels.Add(tagsTab, "Tags");
            actionUITabTypeLabels.Add(saveTab, "Save");
            actionUITabTypeLabels.Add(clothingPresetTab, "Presets");
            actionUITabTypeLabels.Add(sceneImportTab, "Import");
            actionUITabTypeLabels.Add(quickFilterTab, "Quick Filter");
            actionUITabTypeLabels.Add(infoTab, "Info");


        }
    }
    public static class FavHiddenFilterOptions
    {
        public const int standard = 1;
        public const int favOnly = 2;
        public const int showHidden = 3;
        public static string enumManifestName
        {
            get { return "FavHiddenFilterOptions"; }
        }
        static FavHiddenFilterOptions()
        {
            EnumManifest.AddEnum(enumManifestName, standard, "Standard");
            EnumManifest.AddEnum(enumManifestName, favOnly, "Only Favorites");
            EnumManifest.AddEnum(enumManifestName, showHidden, "Show Hidden");
        }
    }

    public static class ResourceNameSearchMode
    {
        public const int StartsWith = BAResourceNameFilterModeEnum.StartsWith;
        public const int Contains = BAResourceNameFilterModeEnum.Contains;
        public const int ContainsFullPath = BAResourceNameFilterModeEnum.ContainsFullPath;
        public static string enumManifestName
        {
            get { return "ResourceNameSearchMode"; }
        }
        static ResourceNameSearchMode()
        {
            EnumManifest.AddEnum(enumManifestName, StartsWith, "Starts with");
            EnumManifest.AddEnum(enumManifestName, Contains, "Contains");
            EnumManifest.AddEnum(enumManifestName, ContainsFullPath, "ContainsFullPath","Contains (Full Path)");
        }
    }

    public static class LicenseType
    {
        public const int Unknown = BALicenseTypeEnum.Unknown;
        public const int Any = BALicenseTypeEnum.Any;
        public const int PC = BALicenseTypeEnum.PC;
        public const int PCEA = BALicenseTypeEnum.PCEA;
        public const int Questionable = BALicenseTypeEnum.Questionable;
        public const int FC = BALicenseTypeEnum.FC;
        public const int CCBY = BALicenseTypeEnum.CCBY;
        public const int CCBYSA = BALicenseTypeEnum.CCBYSA;
        public const int CCBYND = BALicenseTypeEnum.CCBYND;
        public const int CCBYNC = BALicenseTypeEnum.CCBYNC;
        public const int CCBYNCSA = BALicenseTypeEnum.CCBYNCSA;
        public const int CCBYNCND = BALicenseTypeEnum.CCBYNCND;

        public static string enumManifestName
        {
            get { return "LicenseType"; }
        }
        static LicenseType()
        {
            EnumManifest.AddEnum(enumManifestName, Unknown, "Unknown", "Unknown");
            EnumManifest.AddEnum(enumManifestName, Any, "Any", "<i><b>Any</b></i>");
            EnumManifest.AddEnum(enumManifestName, PC, "PC", "PC");
            EnumManifest.AddEnum(enumManifestName, PCEA, "PC EA", "PC EA");
            EnumManifest.AddEnum(enumManifestName, Questionable, "Questionable", "Questionable");
            EnumManifest.AddEnum(enumManifestName, FC, "FC", "FC");
            EnumManifest.AddEnum(enumManifestName, CCBY, "CC BY", "CC BY");

            EnumManifest.AddEnum(enumManifestName, CCBYSA, "CC BY-SA", "CC BY-SA");
            EnumManifest.AddEnum(enumManifestName, CCBYND, "CC BY-ND", "CC BY-ND");
            EnumManifest.AddEnum(enumManifestName, CCBYNC, "CC BY-NC", "CC BY-NC");
            EnumManifest.AddEnum(enumManifestName, CCBYNCSA, "CC BY-NC-SA", "CC BY-NC-SA");
            EnumManifest.AddEnum(enumManifestName, CCBYNCND, "CC BY-NC-ND", "CC BY-NC-ND");
        }
    }
    public static class LeftRight
    {
        public const int left = 1;
        public const int right = 2;
        public const int both = 3;
        public const int neither = 4;

        public static string enumManifestName
        {
            get { return "LeftRight"; }
        }
        static LeftRight()
        {
            EnumManifest.AddEnum(enumManifestName, left);
            EnumManifest.AddEnum(enumManifestName, right);
            EnumManifest.AddEnum(enumManifestName, both);
            EnumManifest.AddEnum(enumManifestName, neither);
        }
    }

    public static class ResourceTargetCategory
    {
        public const int nonAtomTargetable = 1;
        public const int personAtomTargetable = 2;
        public const int otherAtomTargetable = 3;
        public const int anyAtomTargetable = 4;
        public const int textureAtomTargetable = 5;
        public const int anyTarget = 6;
        public const int subSceneTargetable = 7;
        public const int femaleAtomTargetable = 8;
        public const int maleAtomTargetable = 9;
        public const int cuaTargetable = 10;
        public const int coreControlTargetable = 11;
    }
    public static class ResourceCategory
    {
        public const int preset = 1;
        public const int scene = 2;
        public const int plugins = 3;
        public const int subScenes = 4;
        public const int clothing = 5;
        public const int cua = 6;
        public const int clothingItemPresets = 7;
        public const int sceneOrSessionPluginPresets = 8;
    }

    public static class ImportCategory
    {
        public const int personPresets = 1;
        public const int atomPresets = 2;
        public const int subScenes = 3;
        public const int animPatterns = 4;
        public const int mocap = 5;


        public static string enumManifestName
        {
            get { return "ImportCategory"; }
        }
        static ImportCategory()
        {
            EnumManifest.AddEnum(enumManifestName, personPresets, "Person Presets");
            EnumManifest.AddEnum(enumManifestName, atomPresets, "Atom Presets");
            EnumManifest.AddEnum(enumManifestName, animPatterns, "Animation Patterns");
            EnumManifest.AddEnum(enumManifestName, subScenes, "SubScenes");
            EnumManifest.AddEnum(enumManifestName, mocap, "Mocap");
        }
    }

    public static class ResourceType
    {
        public const int unknown = 0;
        public const int scene = BAResourceTypeEnum.scene;
        public const int presetAppearance = BAResourceTypeEnum.presetAppearance;
        public const int presetAnimation = BAResourceTypeEnum.presetAnimation;
        public const int presetBreastPhys = BAResourceTypeEnum.presetBreastPhys;
        public const int presetClothing = BAResourceTypeEnum.presetClothing;
        public const int presetGeneral = BAResourceTypeEnum.presetGeneral;
        public const int presetGlutePhysics = BAResourceTypeEnum.presetGlutePhysics;
        public const int presetHair = BAResourceTypeEnum.presetHair;
        public const int presetMorphs = BAResourceTypeEnum.presetMorphs;
        public const int presetPlugins = BAResourceTypeEnum.presetPlugins;
        public const int presetPose = BAResourceTypeEnum.presetPose;
        public const int presetSkin = BAResourceTypeEnum.presetSkin;
        public const int personTexture = BAResourceTypeEnum.personTexture;
        public const int atomTexture = BAResourceTypeEnum.atomTexture;
        public const int presetAtom = BAResourceTypeEnum.presetAtom;
        public const int plugins = BAResourceTypeEnum.plugins;
        public const int subScenes = BAResourceTypeEnum.subScenes;
        public const int femaleClothingItems = BAResourceTypeEnum.femaleClothingItems;
        public const int maleClothingItems = BAResourceTypeEnum.maleClothingItems;
        public const int cua = BAResourceTypeEnum.cua;
        public const int clothingItemPresets = BAResourceTypeEnum.clothingItemPresets;
        public const int presetSceneSessionPlugins = BAResourceTypeEnum.presetSceneSessionPlugins;

        private static Dictionary<int, int> resourceTypeToTargetCategory = new Dictionary<int, int>();
        private static Dictionary<int, int> resourceTypeToCategory = new Dictionary<int, int>();
        private static Dictionary<int, HashSet<int>> resourceCategoryToType = new Dictionary<int, HashSet<int>>();
        private static Dictionary<int, HashSet<int>> resourceTargetCategoryToType = new Dictionary<int, HashSet<int>>();
        private static Dictionary<int, HashSet<int>> resourceTypeToTagCategory = new Dictionary<int, HashSet<int>>();

        public static string enumManifestName
        {
            get { return "ResourceType"; }
        }
        static ResourceType()
        {
            EnumManifest.AddEnum(enumManifestName, scene, "Scene");
            AssignCategories(scene, ResourceCategory.scene, ResourceTargetCategory.nonAtomTargetable);
            AssignTags(scene, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, presetAppearance, "Preset Appearance");
            AssignCategories(presetAppearance, ResourceCategory.preset, ResourceTargetCategory.personAtomTargetable);
            AssignTags(presetAppearance, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, presetAnimation, "Preset Animation");
            AssignCategories(presetAnimation, ResourceCategory.preset, ResourceTargetCategory.personAtomTargetable);
            AssignTags(presetAnimation, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, presetBreastPhys, "Preset Breast Physics");
            AssignCategories(presetBreastPhys, ResourceCategory.preset, ResourceTargetCategory.personAtomTargetable);
            AssignTags(presetBreastPhys, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, presetClothing, "Preset Clothing");
            AssignCategories(presetClothing, ResourceCategory.preset, ResourceTargetCategory.personAtomTargetable);
            AssignTags(presetClothing, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, presetGlutePhysics, "Preset Glute Physics");
            AssignCategories(presetGlutePhysics, ResourceCategory.preset, ResourceTargetCategory.personAtomTargetable);
            AssignTags(presetGlutePhysics, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, presetGeneral, "Preset General");
            AssignCategories(presetGeneral, ResourceCategory.preset, ResourceTargetCategory.personAtomTargetable);
            AssignTags(presetGeneral, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, presetHair, "Preset Hair");
            AssignCategories(presetHair, ResourceCategory.preset, ResourceTargetCategory.personAtomTargetable);
            AssignTags(presetHair, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, presetMorphs, "Preset Morphs");
            AssignCategories(presetMorphs, ResourceCategory.preset, ResourceTargetCategory.personAtomTargetable);
            AssignTags(presetMorphs, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, presetPlugins, "Preset Plugins");
            AssignCategories(presetPlugins, ResourceCategory.preset, ResourceTargetCategory.personAtomTargetable);
            AssignTags(presetPlugins, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, presetPose, "Preset Pose");
            AssignCategories(presetPose, ResourceCategory.preset, ResourceTargetCategory.personAtomTargetable);
            AssignTags(presetPose, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, presetSkin, "Preset Skin");
            AssignCategories(presetSkin, ResourceCategory.preset, ResourceTargetCategory.personAtomTargetable);
            AssignTags(presetSkin, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, presetAtom, "Preset Atom");
            AssignCategories(presetAtom, ResourceCategory.preset, ResourceTargetCategory.otherAtomTargetable);
            AssignTags(presetAtom, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, plugins, "Plugins");
            AssignCategories(plugins, ResourceCategory.plugins, ResourceTargetCategory.anyTarget);
            AssignTags(plugins, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, subScenes, "SubScenes");
            AssignCategories(subScenes, ResourceCategory.subScenes, ResourceTargetCategory.subSceneTargetable);
            AssignTags(subScenes, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, femaleClothingItems, "Clothing (Female)");
            AssignCategories(femaleClothingItems, ResourceCategory.clothing, ResourceTargetCategory.femaleAtomTargetable);
            AssignTags(femaleClothingItems, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, maleClothingItems, "Clothing (Male)");
            AssignCategories(maleClothingItems, ResourceCategory.clothing, ResourceTargetCategory.maleAtomTargetable);
            AssignTags(maleClothingItems, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, cua, "Custom Unity Assets");
            AssignCategories(cua, ResourceCategory.cua, ResourceTargetCategory.cuaTargetable);
            AssignTags(cua, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, clothingItemPresets, "Clothing Item Presets");
            AssignCategories(clothingItemPresets, ResourceCategory.cua, ResourceTargetCategory.personAtomTargetable);
            AssignTags(clothingItemPresets, TagCategory.userDefined, TagCategory.gender);

            EnumManifest.AddEnum(enumManifestName, presetSceneSessionPlugins, "Scene & Session Plugin Presets");
            AssignCategories(presetSceneSessionPlugins, ResourceCategory.sceneOrSessionPluginPresets, ResourceTargetCategory.coreControlTargetable);
            AssignTags(presetSceneSessionPlugins, TagCategory.userDefined);


            //            EnumManifest.AddEnum(enumManifestName, personTexture, "Person Texture");

            //            EnumManifest.AddEnum(enumManifestName, atomTexture, "Atom Texture");


        }

        public static HashSet<int> originalSupportedResourceTypes
        {
            get
            {
                HashSet<int> resourceTypes = new HashSet<int>() { scene, presetAppearance, presetAnimation, presetBreastPhys , presetClothing , presetGeneral, presetGlutePhysics, presetHair , presetMorphs, presetPlugins, presetPose, presetSkin, plugins};

                return resourceTypes;
            }
        }

        public static HashSet<int> currentSupportedResourceTypes
        {
            get
            {
                HashSet<int> resourceTypes = new HashSet<int>(EnumManifest.GetEnumsList(enumManifestName));

                return resourceTypes;
            }
        }

        public static HashSet<string> currentSupportedResourceTypeNames
        {
            get
            {
                HashSet<string> resourceTypeNames = new HashSet<string>(EnumManifest.GetStoreValList(enumManifestName));

                return resourceTypeNames;
            }
        }

        private static void AssignTags (int resourceType, params int[] tagCategories)
        {
            if (!resourceTypeToTagCategory.ContainsKey(resourceType)) resourceTypeToTagCategory.Add(resourceType, new HashSet<int>());

            foreach (int tagCat in tagCategories)
            {
                resourceTypeToTagCategory[resourceType].Add(tagCat);
            }
        }
        private static void AssignCategories(int resourceType, int resourceCategory, int targetCategory)
        {
            resourceTypeToCategory.Add(resourceType, resourceCategory);
            resourceTypeToTargetCategory.Add(resourceType, targetCategory);

            if (!resourceCategoryToType.ContainsKey(resourceCategory)) resourceCategoryToType.Add(resourceCategory, new HashSet<int>());
            resourceCategoryToType[resourceCategory].Add(resourceType);

            if (!resourceTargetCategoryToType.ContainsKey(targetCategory)) resourceTargetCategoryToType.Add(targetCategory, new HashSet<int>());
            resourceTargetCategoryToType[targetCategory].Add(resourceType);
        }

        public static int GetResourceTargetCategory(int resourceType)
        {
            return resourceTypeToTargetCategory[resourceType];
        }
        public static int GetResourceCategory(int resourceType)
        {
            if (resourceTypeToCategory.ContainsKey(resourceType)) return resourceTypeToCategory[resourceType];
            return -1;
        }
        public static HashSet<int> GetTagCategories(int resourceType)
        {
            if (!resourceTypeToTagCategory.ContainsKey(resourceType)) resourceTypeToTagCategory.Add(resourceType, new HashSet<int>());
            return resourceTypeToTagCategory[resourceType];
        }

        public static HashSet<int> GetResourceTypesByCategory(int resourceCategory)
        {
            if (!resourceCategoryToType.ContainsKey(resourceCategory)) resourceCategoryToType.Add(resourceCategory, new HashSet<int>());
            return resourceCategoryToType[resourceCategory];
        }

        public static HashSet<int> GetResourceTypesByTargetCategory(int resourceCategory)
        {
            if (!resourceTargetCategoryToType.ContainsKey(resourceCategory)) resourceTargetCategoryToType.Add(resourceCategory, new HashSet<int>());
            return resourceTargetCategoryToType[resourceCategory];
        }

        public static string GetFileExtensionFilter(int resourceType)
        {
            int resourceCat = GetResourceCategory(resourceType);
            if (resourceCat == ResourceCategory.preset || resourceCat == ResourceCategory.sceneOrSessionPluginPresets) return "vap";
            if (resourceType == subScenes || resourceType == scene) return "json";
            if (resourceType == plugins) return "cs|cslist|dll";
            return "";
        }
        public static string GetPrimaryFileExtension(int resourceType)
        {
            int resourceCat = GetResourceCategory(resourceType);
            if (resourceCat == ResourceCategory.preset || resourceCat == ResourceCategory.sceneOrSessionPluginPresets) return "vap";
            if (resourceType == subScenes || resourceType == scene) return "json";
            if (resourceType == plugins) return "cs";
            return "";
        }
    }

    public static class ClothingCreatorFilterChoices
    {
        public const int all = BAClothingCreatorTypeEnum.allCreators;

        public static string enumManifestName
        {
            get { return "ClothingCreatorFilterChoices"; }
        }
        static ClothingCreatorFilterChoices()
        {
            EnumManifest.AddEnum(enumManifestName, all, "All Creators", "<i><b>All Creators</b></i>");

        }
    }

    public static class CreatorFilterChoices
    {
        public const int all = BACreatorTypeEnum.allLocalAndCreators;
        public const int allVARs = BACreatorTypeEnum.allPackageCreators;
        public const int localOnly = BACreatorTypeEnum.localFolders;

        public static string enumManifestName
        {
            get { return "CreatorFilterChoices"; }
        }
        static CreatorFilterChoices()
        {
            EnumManifest.AddEnum(enumManifestName, allVARs,"Package Creators", "<i><b>Package Creators</b></i>");
            EnumManifest.AddEnum(enumManifestName, localOnly, "Local Folders", "<i><b>Local Folders</b></i>");
            EnumManifest.AddEnum(enumManifestName, all,"Local & Packaged Content", "<i><b>Local & Packaged</b></i>");

        }
    }

    public static class BrowserSelectMode
    {
        public const int instant = BASelectModeEnum.instant;
        public const int single = BASelectModeEnum.single;
        public const int multi = BASelectModeEnum.multi;

        public static string enumManifestName
        {
            get { return "BrowserSelectMode"; }
        }
        static BrowserSelectMode()
        {
            EnumManifest.AddEnum(enumManifestName, instant, "Instant");
            EnumManifest.AddEnum(enumManifestName, single, "Single");
            EnumManifest.AddEnum(enumManifestName, multi, "Multi");
        }
    }

    public static class BrowserSortMode
    {
        public const int AtoZ = BAResultsSortModeEnum.AtoZ;
        public const int ZtoA = BAResultsSortModeEnum.ZtoA;
        public const int packageFolder = BAResultsSortModeEnum.packageFolder;
        public const int newDateTime = BAResultsSortModeEnum.newDateTime;
        public const int reverseNewDateTime = BAResultsSortModeEnum.reverseNewDateTime;
        public const int updatedDateTime = BAResultsSortModeEnum.updatedDateTime;
        public const int reverseUpdatedDateTime = BAResultsSortModeEnum.reverseUpdatedDateTime;

        public static string enumManifestName
        {
            get { return "BrowserSortMode"; }
        }
        static BrowserSortMode()
        {
            EnumManifest.AddEnum(enumManifestName, AtoZ, "A to Z");
            EnumManifest.AddEnum(enumManifestName, ZtoA, "Z to A");
            EnumManifest.AddEnum(enumManifestName, packageFolder, "Package & Folder");
            EnumManifest.AddEnum(enumManifestName, newDateTime, "Catalogued Date","New Date");
            EnumManifest.AddEnum(enumManifestName, reverseNewDateTime, "New Date\n(Reverse)");
            EnumManifest.AddEnum(enumManifestName, updatedDateTime, "Updated Date");
            EnumManifest.AddEnum(enumManifestName, reverseUpdatedDateTime, "Updated Date\n(Reverse)");

        }
    }
    public static class PluginTargetTypes
    {
        public const int sessionPlugins = 1;
        public const int scenePlugins = 2;

        public static string enumManifestName
        {
            get { return "PluginTargetTypes"; }
        }
        static PluginTargetTypes()
        {
            EnumManifest.AddEnum(enumManifestName, sessionPlugins, "Session Plugins");
            EnumManifest.AddEnum(enumManifestName, scenePlugins, "Scene Plugins");
        }
    }

    public static class CreateNewAtom
    {
        public const int newAtom = 1;
        public static string enumManifestName
        {
            get { return "CreateNewAtom"; }
        }
        static CreateNewAtom()
        {
            EnumManifest.AddEnum(enumManifestName, newAtom, "<New Atom>");
        }
    }

    public static class TriState
    {
        public const int active = 1;
        public const int inactive = 2;
        public const int indeterminate = 3;
        public static string enumManifestName
        {
            get { return "TriState"; }
        }
        static TriState()
        {
            EnumManifest.AddEnum(enumManifestName, active, "Active");
            EnumManifest.AddEnum(enumManifestName, inactive, "Inactive");
            EnumManifest.AddEnum(enumManifestName, indeterminate, "Indeterminate");
        }
    }

    public static class TagCategory
    {
        public const int all = 0;
        public const int bodyRegion = 1;
        public const int clothingType = 2;
        public const int gender = 3;
        public const int userDefined = 4;
        public const int pluginType = 5;
        public const int combi = 6;

        private static Dictionary<int, Color> tagCategoryColor = new Dictionary<int, Color>();
        public static string enumManifestName
        {
            get { return "TagCategory"; }
        }
        static TagCategory()
        {
            EnumManifest.AddEnum(enumManifestName, all, "All");
            EnumManifest.AddEnum(enumManifestName, bodyRegion, "Body Region");
            tagCategoryColor.Add(bodyRegion, new Color(0.75f, 0.75f, 1f));
            EnumManifest.AddEnum(enumManifestName, clothingType, "Clothing Type");
            tagCategoryColor.Add(clothingType, new Color(0.75f, 1f, 0.75f));
            EnumManifest.AddEnum(enumManifestName, gender, "Gender");
            tagCategoryColor.Add(gender, new Color(1f,0.75f,0.75f));
            EnumManifest.AddEnum(enumManifestName, pluginType, "Plugin Type");
            tagCategoryColor.Add(pluginType, new Color(0.75f, 1f, 0.75f));
            EnumManifest.AddEnum(enumManifestName, userDefined, "User");
            tagCategoryColor.Add(userDefined, new Color(0.75f, 1f, 1f));
            EnumManifest.AddEnum(enumManifestName, combi, "Combi-Tag");
            tagCategoryColor.Add(combi, new Color(0.75f, 0.5f, 1f));
        }

        public static Color GetColor(int tagCategory)
        {
            return tagCategoryColor[tagCategory];
        }
    }

    public static class TagGroups
    {
        public const int anyTags = BATagGroupTypeEnum.anyTags;
        public const int hiddenOnly = BATagGroupTypeEnum.hiddenOnly;
        public const int noTags = BATagGroupTypeEnum.noTags;
        public const int noUserTags = BATagGroupTypeEnum.noUserTags;
        public const int wornClothingOnly = BATagGroupTypeEnum.wornClothingOnly;


        public static string enumManifestName
        {
            get { return "TagGroups"; }
        }
        static TagGroups()
        {
            EnumManifest.AddEnum(enumManifestName, anyTags, "Any Tags", "<i><b>Any Tags</b></i>");
            EnumManifest.AddEnum(enumManifestName, hiddenOnly, "Only Hidden", "<i><b>Only Hidden</b></i>");
            EnumManifest.AddEnum(enumManifestName, wornClothingOnly, "Only Worn", "<i><b>Only Worn</b></i>");

            EnumManifest.AddEnum(enumManifestName, noTags, "No Tags", "<i><b>No Tags</b></i>");
            EnumManifest.AddEnum(enumManifestName, noUserTags, "No User Tags", "<i><b>No User Tags</b></i>");
        }
    }

    public static class VAMUILaunchButtonPosition
    {
        public const int disabled = 0;
        public const int toolBarGap = 1;
        public const int sceneBrowserSub = 2;
        public const int toolBarLeftInnerUpper = 3;
        public const int toolBarRightInnerUpper = 4;
        public static string enumManifestName
        {
            get { return "VAMUILaunchButtonPosition"; }
        }
        static VAMUILaunchButtonPosition()
        {
            EnumManifest.AddEnum(enumManifestName, disabled, "Disable");

            EnumManifest.AddEnum(enumManifestName, toolBarGap, "Toolbar Gap");
            EnumManifest.AddEnum(enumManifestName, sceneBrowserSub, "Scene Browser","Toolbar: Scene Browser");
            EnumManifest.AddEnum(enumManifestName, toolBarLeftInnerUpper, "toolBarLeftInnerUpper", "Toolbar Left");
            EnumManifest.AddEnum(enumManifestName, toolBarRightInnerUpper, "toolBarRightInnerUpper", "Toolbar Right");

        }
    }

    public static class TopUITabs
    {
        public const int generalSettings = 1;
        public const int vrSettings = 2;
        public const int desktopSettings = 3;
        public const int resourceSettings = 4;

    }

    public static class ResourceSettingsUITabs
    {
        public const int defaultFilters = 1;
        public const int tags = 2;
        public const int combiTags = 3;
        public const int creators = 4;

    }
    public static class GlobalSettingsUITabs
    {
        public const int general = 1;
        public const int resources = 2;
        public const int backups = 3;
        public const int import = 4;

    }

    public static class DesktopSettingsUITabs
    {
        public const int general = 1;
        public const int newAtomPosition = 2;

    }

    public static class VRSettingsUITabs
    {
        public const int general = 1;
        public const int newAtomPosition = 2;

    }


    public static class CustomRowWidth
    {
        public const int left = 1;
        public const int right = 2;
        public const int full = 3;
    }

    public static class AbsoluteRelative
    {
        public const int abs = 1;
        public const int rel = 2;
        public static string enumManifestName
        {
            get { return "AbsoluteRelative"; }
        }
        static AbsoluteRelative()
        {
            EnumManifest.AddEnum(enumManifestName, abs, "Absolute", "Abs");
            EnumManifest.AddEnum(enumManifestName, rel, "Relative", "Rel");

        }
    }

    public static class VRRelativePositionMode
    {
        public const int vrHead = 1;
        public const int vrLeftHand = 2;
        public const int vrRightHand = 3;

        public static string enumManifestName
        {
            get { return "VRRelativePositionMode"; }
        }
        static VRRelativePositionMode()
        {
            InitEnum(enumManifestName, vrHead, "VR Headset");
            InitEnum(enumManifestName, vrLeftHand, "VR Left Hand");
            InitEnum(enumManifestName, vrRightHand, "VR Right Hand");
        }

        static void InitEnum(string enumName, int enumVal, string storeValLabel)
        {
            EnumManifest.AddEnum(enumName, enumVal, storeValLabel);
        }
    }

    public static class DesktopRelativePositionMode
    {
        public const int lookCamera = 1;

        public static string enumManifestName
        {
            get { return "DesktopRelativePositionMode"; }
        }
        static DesktopRelativePositionMode()
        {
            InitEnum(enumManifestName, lookCamera, "Main Desktop View");
        }

        static void InitEnum(string enumName, int enumVal, string storeValLabel)
        {
            EnumManifest.AddEnum(enumName, enumVal, storeValLabel);
        }
    }
    public static class PersonGenderTypes
    {
        public const int firstPerson = 1;
        public const int firstFemale = 2;
        public const int firstMale = 3;

        public static string enumManifestName
        {
            get { return "PersonGenderTypes"; }
        }
        static PersonGenderTypes()
        {
            InitEnum(enumManifestName, firstPerson, "First Person");
            InitEnum(enumManifestName, firstFemale, "First Female");
            InitEnum(enumManifestName, firstMale, "First Male");
        }

        static void InitEnum(string enumName, int enumVal, string storeValLabel)
        {
            EnumManifest.AddEnum(enumName, enumVal, storeValLabel);
        }
    }

    public static class LogicOperands
    {
        public const int and = 1;
        public const int or = 2;

        public static string enumManifestName
        {
            get { return "LogicOperands"; }
        }
        static LogicOperands()
        {
            InitEnum(enumManifestName, and, "AND","and are");
            InitEnum(enumManifestName, or, "OR","or are");
        }

        static void InitEnum(string enumName, int enumVal, string storeValLabel,string altDisplayLabel)
        {
            EnumManifest.AddEnum(enumName, enumVal, storeValLabel, altDisplayLabel);
        }
    }

    public static class CombiLogicConditionOptions
    {
        public const int taggedWith = 1;
        public const int taggedWithAny = 2;
        public const int taggedWithAll = 3;
        public const int notTagged = 5;
        public const int notTaggedWithAny = 6;
        public const int taggedWithAnything = 7;
        public const int taggedWithAnyBodyRegion = 8;
        public const int taggedWithAnyClothingType = 9;
        public const int taggedWithAnyUserTag = 10;
        public const int everything = 19;
        public const int hidden = 20;
        public const int notHidden = 21;
        public const int favorite = 22;
        public const int notFavorite = 23;
        public const int New = 24;
        public const int notNew = 25;
        public const int varLicense = 26;
        public const int notVarLicense = 27;
        public const int packagedByCreator = 28;
        public const int notPackagedByCreator = 29;
        public const int inLocalFolder = 30;
        public const int notInLocalFolder = 31;
        public const int underLocalFolder = 32;
        public const int notUnderLocalFolder = 33;
        public const int varInFolder = 34;
        public const int varNotInFolder = 35;
        public const int varUnderFolder = 36;
        public const int varNotUnderFolder = 37;
        public const int createdBefore = 38;
        public const int createdAfter = 39;
        public const int updatedBefore = 40;
        public const int updatedAfter = 41;

        public static string enumManifestName
        {
            get { return "CombiLogicConditions"; }
        }
        static CombiLogicConditionOptions()
        {
            InitEnum(enumManifestName, everything, "everything", "everything");
            InitEnum(enumManifestName, hidden, "hidden", "hidden");
            InitEnum(enumManifestName, taggedWith, "taggedWith", "tagged with");
            InitEnum(enumManifestName, taggedWithAny, "taggedWithAny", "tagged with any of");
            InitEnum(enumManifestName, taggedWithAll, "taggedWithAll", "tagged with all of");
            InitEnum(enumManifestName, notTagged, "notTagged", "not tagged");
            InitEnum(enumManifestName, notTaggedWithAny, "notTaggedWithAny", "not tagged with any of");
            InitEnum(enumManifestName, taggedWithAnyBodyRegion, "taggedWithAnyClothingRegion", "on body regions");
            InitEnum(enumManifestName, taggedWithAnyClothingType, "taggedWithAnyClothingType", "of clothing types");
            InitEnum(enumManifestName, taggedWithAnyUserTag, "taggedWithAnyUserTag", "with user tags");
            InitEnum(enumManifestName, taggedWithAnything, "taggedWithAnything", "tagged with anything");
            InitEnum(enumManifestName, notHidden, "notHidden", "not hidden");
            InitEnum(enumManifestName, favorite, "favorite", "favorite");
            InitEnum(enumManifestName, notFavorite, "notFavorite", "not favorite");
            InitEnum(enumManifestName, New, "new", "new");
            InitEnum(enumManifestName, notNew, "notNew", "not new");
            InitEnum(enumManifestName, varLicense, "varLicensed", "in VAR licensed with");
            InitEnum(enumManifestName, notVarLicense, "notVarLicensed", "in VAR not licensed with");

            InitEnum(enumManifestName, packagedByCreator, "packagedByCreator", "packaged by Creator");
            InitEnum(enumManifestName, notPackagedByCreator, "notPackagedByCreator", "not packaged by Creator");

            InitEnum(enumManifestName, inLocalFolder, "inLocalFolder", "in local folder");
            InitEnum(enumManifestName, notInLocalFolder, "notInLocalFolder", "not in local folder");
            InitEnum(enumManifestName, underLocalFolder, "underLocalFolder", "under local folder");
            InitEnum(enumManifestName, notUnderLocalFolder, "notUnderLocalFolder", "not under local folder");

            InitEnum(enumManifestName, varInFolder, "varInFolder", "within VARs in folder");
            InitEnum(enumManifestName, varNotInFolder, "varNotInFolder", "within VARs not in folder");
            InitEnum(enumManifestName, varUnderFolder, "varUnderFolder", "within VARs under folder");
            InitEnum(enumManifestName, varNotUnderFolder, "varNotUnderFolder", "within VARs not under folder");

            InitEnum(enumManifestName, createdBefore, "createdBefore", "created before");
            InitEnum(enumManifestName, createdAfter, "createdAfter", "created after");
            InitEnum(enumManifestName, updatedBefore, "updatedBefore", "updated before");
            InitEnum(enumManifestName, updatedAfter, "updatedAfter", "updated after");
        }

        static void InitEnum(string enumName, int enumVal, string storeValLabel, string displayLabel)
        {
            EnumManifest.AddEnum(enumName, enumVal, storeValLabel, displayLabel);
        }
    }

    public static class CreatorFilterSortOrder
    {
        public const int defaultVal = 0;
        public const int alphabetically = 1;
        public const int resourceCount = 2;

        public static string enumManifestName
        {
            get { return "CreatorSortOrder"; }
        }
        static CreatorFilterSortOrder()
        {
            InitEnum(enumManifestName, defaultVal, "default", "default");
            InitEnum(enumManifestName, alphabetically, "alphabetically", "alphabetically");
            InitEnum(enumManifestName, resourceCount, "resourceCount", "by resource count");
        }

        static void InitEnum(string enumName, int enumVal, string storeValLabel, string altDisplayLabel)
        {
            EnumManifest.AddEnum(enumName, enumVal, storeValLabel, altDisplayLabel);
        }
    }


    public static class PositionRotationAxis
    {
        public const int xPosition = 1;
        public const int yPosition = 2;
        public const int zPosition = 3;
        public const int xRotation = 4;
        public const int yRotation = 5;
        public const int zRotation = 6;
    }

    public static class FolderBrowserType
    {
        public const int filterSelect = 1;
        public const int resourceMove = 2;
    }

    public static class FolderBrowserUIMode
    {
        public const int normal = 1;
        public const int moveFolder = 2;
        public const int deleteFolder = 3;
    }

    public static class CombTagUIMode
    {
        public const int normal = 1;
        public const int rename = 2;
        public const int deleteConfirm = 3;
    }
}

