﻿
using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.UI;



namespace JayJayWon
{
    public class CustomUIElement
    {
        protected UIDynamic _spacerLeft { get; set; }
        protected UIDynamic _spacerRight { get; set; }
        public UIDynamic _spacerActive { get; protected set; }

        protected float spacerHeight;
        public float height { get { return spacerHeight; } }
        protected HashSet<UIDynamicPopup> uidPopuplist;


        public CustomUIElement(float height)
        {
            spacerHeight = height;
            uidPopuplist = new HashSet<UIDynamicPopup>();
        }

        public void CloseAllPopups(UIDynamicPopup exception)
        {
            foreach (UIDynamicPopup popup in uidPopuplist)
            {
                if (popup != exception) popup.popup.visible = false;
            }
        }

        public virtual void ClearUI()
        {
            if (_spacerLeft != null) Globals.mvrScript.RemoveSpacer(_spacerLeft);
            if (_spacerRight != null) Globals.mvrScript.RemoveSpacer(_spacerRight);
            _spacerActive = null;
            uidPopuplist.Clear();
        }

        public GameObject CreateVScrollableContent(int anchorPreset, float horizontalSize, float verticalSize, Vector2 anchoredPositionOffset, float scrollBarWidth = 10f, float spacing = 0f)
        {
            var gameObject = UI.CreateVScrollableContent(_spacerActive.gameObject, new Color(1f,1f,1f,0.2f), anchorPreset, horizontalSize, verticalSize, anchoredPositionOffset,scrollBarWidth,spacing);
            return gameObject;
        }
        
        protected UIDynamicPopup InstantiatePopup(JSONStorableStringChooser jssc, bool scrollable)
        {
            Transform originalTransform = Globals.mvrScript.manager.configurableScrollablePopupPrefab.transform;
            if (!scrollable) originalTransform = Globals.mvrScript.manager.configurablePopupPrefab.transform;

            Transform popup = UnityEngine.Object.Instantiate(originalTransform, _spacerActive.gameObject.transform, false);
            UIDynamicPopup selectTargetPopup = popup.GetComponent<UIDynamicPopup>();

            selectTargetPopup.popup.showSlider = false;
            popup.SetParent(_spacerActive.gameObject.transform, false);
            jssc.popup = selectTargetPopup.popup;
            return (selectTargetPopup);
        }
        protected UIDynamicSlider InstantiateSlider(JSONStorableFloat jsf)
        {
            Transform originalTransform = Globals.mvrScript.manager.configurableSliderPrefab.transform;

            Transform slider = UnityEngine.Object.Instantiate(originalTransform, _spacerActive.gameObject.transform, false);
            UIDynamicSlider newSlider = slider.GetComponent<UIDynamicSlider>();

            slider.SetParent(_spacerActive.gameObject.transform, false);
            jsf.slider = newSlider.slider;
            return (newSlider);
        }

        private UIDynamicSlider CreateSlider(JSONStorableFloat jsf, float width, float height, float xOffsetPosition, float yOffsetPostion = 0f)
        {
            RectTransform spacerRT = _spacerActive.GetComponent<RectTransform>();

            spacerRT.anchorMin = new Vector2(0f, 1f);
            spacerRT.anchorMax = new Vector2(0f, 1f);
            spacerRT.pivot = new Vector2(0f, 1f);

            UIDynamicSlider newSlider = InstantiateSlider(jsf);
            
            RectTransform childRect = newSlider.transform.GetComponent<RectTransform>();
            childRect.anchorMin = new Vector2(0f, 1f);
            childRect.anchorMax = new Vector2(0f, 1f);
            childRect.pivot = new Vector2(0f, 1f);

            childRect.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);
            childRect.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
            childRect.transform.localPosition = new Vector3(xOffsetPosition, yOffsetPostion, 0f);

            return (newSlider);
        }

        // This prevents the popup panel from being rendered behind other items. Needs to be set at the end of the Frame to override the VAM default behaviour.
        protected IEnumerator ReparentPopup(UIDynamicPopup popupUID, bool forceRightParentPopup)
        {
            yield return new WaitForEndOfFrame();

            try
            {
                if (popupUID.popup.visible)
                {
                    if (forceRightParentPopup) popupUID.popup.popupPanel.transform.SetParent(_spacerRight.gameObject.transform.parent);
                    else popupUID.popup.popupPanel.transform.SetParent(_spacerActive.gameObject.transform.parent);
                }
            }
            catch (Exception e) { SuperController.LogError("BA.CustomUIElement.ReparentPopup: Exception caught:\n" + e); }
        }

        public UIDynamicPopup AddPopup(JSONStorableStringChooser jssc, UIPopup.OnOpenPopup callbackMethod, float width, float height, float xOffsetPosition, float yOffsetPosition = 0, float panelHeight = 0f, float popupRowHeight=50f,string labelOverride = null, float labelProportion = 0.4f)
        {
            UIDynamicPopup newPopup;
            if (panelHeight == 0f) newPopup = UI.CreatePopup(_spacerActive.gameObject, jssc, width, height, labelOverride, 30, xOffsetPosition, yOffsetPosition, anchorPreset: AnchorPresets.topLeft, popupRowHeight: popupRowHeight, labelProportion: labelProportion,rescaleText:false);
            else newPopup = UI.CreateScrollablePopup(_spacerActive.gameObject, jssc, width, height, labelOverride, 30, xOffsetPosition, yOffsetPosition, panelHeight, false, true, AnchorPresets.topLeft, labelProportion, popupRowHeight, 30, false,20);

            if (callbackMethod!=null) newPopup.popup.onOpenPopupHandlers += callbackMethod;

            newPopup.popup.onOpenPopupHandlers += () =>
            {
                //                UIAPluginUI.CloseAllPopups(newPopup);
                Globals.mvrScript.StartCoroutine(ReparentPopup(newPopup, false));
            };

            newPopup.popup.selectColor = Const.UIhighlightColor;

            uidPopuplist.Add(newPopup);

            return newPopup;
        }

        public UIDynamicToggle AddToggle (JSONStorableBool jsb, float width, float height, float xOffsetPosition, float yOffsetPosition, string labelOverride=null, float checkboxSize = 50f)
        {
            UIDynamicToggle newToggle= UI.CreateToggle(_spacerActive.gameObject,jsb,width,height, labelOverride, 30, xOffsetPosition,yOffsetPosition,AnchorPresets.topLeft, checkboxSize,false);
            return newToggle;
        }
        public UIDynamicTextField AddTextBox(JSONStorableString jss, float width, float height, float xOffsetPosition, float yOffsetPosition, TextAnchor textAlignment = TextAnchor.MiddleLeft)
        {

            UIDynamicTextField newTextField = UI.CreateTextBox(_spacerActive.gameObject, jss, width, height, 30, xOffsetPosition, yOffsetPosition, AnchorPresets.topLeft,textAlignment: textAlignment,rescaleText:false);
            return newTextField;
        }
        public UIDynamicTextField AddTextEntryBox(JSONStorableString jss, float width, float height, int fontSize, float xOffsetPosition, float yOffsetPosition, TextAnchor textAlignment = TextAnchor.MiddleLeft, InputField.LineType inputFieldType = InputField.LineType.MultiLineNewline, string placeHolderText = "", float hScrollBarHeight = 0f, float vScrollBarWidth = 10f, Action<string> textEntryChangeCallback = null, Action<string> textEntryEndCallback = null)
        {

            UIDynamicTextField newTextField = UI.CreateTextBox(_spacerActive.gameObject, jss, width, height, fontSize, xOffsetPosition, yOffsetPosition, AnchorPresets.topLeft,true,inputFieldType,placeHolderText,hScrollBarHeight,vScrollBarWidth, textEntryChangeCallback, textAlignment, textEntryEndCallback, false);
            return newTextField;
        }

        public UIDynamicTextField AddTextBox(string text, float width, float height, float xOffsetPosition, float yOffsetPosition)
        {
            JSONStorableString jss = new JSONStorableString("textField", text);
            UIDynamicTextField newTextField = UI.CreateTextBox(_spacerActive.gameObject, jss, width, height, 30, xOffsetPosition, yOffsetPosition, AnchorPresets.topLeft,false, rescaleText: false);
            return newTextField;
        }



        public UIDynamicButton AddButton(float width, float height, float xOffsetPosition, float yOffsetPosition, string label, UnityEngine.Events.UnityAction callback)
        {
            UIDynamicButton newButton = UI.CreateUIButton(_spacerActive.gameObject,width,height,label, 30, xOffsetPosition,yOffsetPosition,AnchorPresets.topLeft,false, rescaleText:false);
            if (callback!=null) newButton.button.onClick.AddListener(callback);
            return newButton;
        }

        public UIDynamicSlider AddSlider(JSONStorableFloat jsf, float width, float height, float xOffsetPostion, float yOffsetPostion)
        {
            UIDynamicSlider newSlider = CreateSlider(jsf, width, height, xOffsetPostion, yOffsetPostion);
            newSlider.label = jsf.name;

            return newSlider;
        }

        protected UIDynamicButton InstantiateUIButton()
        {
            Transform buttonTransform = UnityEngine.Object.Instantiate(Globals.mvrScript.manager.configurableButtonPrefab.transform, _spacerActive.gameObject.transform, false);
            UIDynamicButton selectTargetButton = buttonTransform.GetComponent<UIDynamicButton>();
            buttonTransform.SetParent(_spacerActive.gameObject.transform, false);
            return (selectTargetButton);
        }

        protected UIDynamicButton CreateButton(float width, float height, float xOffsetPosition, float yOffsetPostion = 0f)
        {
            RectTransform spacerRT = _spacerActive.GetComponent<RectTransform>();

            spacerRT.anchorMin = new Vector2(0f, 1f);
            spacerRT.anchorMax = new Vector2(0f, 1f);
            spacerRT.pivot = new Vector2(0f, 1f);

            UIDynamicButton targetButton = InstantiateUIButton();

            RectTransform childRect = targetButton.button.GetComponent<RectTransform>();
            childRect.anchorMin = new Vector2(0f, 1f);
            childRect.anchorMax = new Vector2(0f, 1f);
            childRect.pivot = new Vector2(0f,1f);

            childRect.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);
            childRect.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
            childRect.transform.localPosition = new Vector3(xOffsetPosition, yOffsetPostion, 0f);

            return (targetButton);
        }

        public void InitUI(int customRowWidth)
        {
            if (customRowWidth != CustomRowWidth.right)
            {
                _spacerLeft = Globals.mvrScript.CreateSpacer(false);
                _spacerActive = _spacerLeft;
                _spacerLeft.height = spacerHeight;


            }
            if (customRowWidth != CustomRowWidth.left)
            {
                _spacerRight = Globals.mvrScript.CreateSpacer(true);
                _spacerRight.height = spacerHeight;
            }
            if (customRowWidth == CustomRowWidth.right)
            {
                _spacerActive = _spacerRight;
            }
        }

    }


    public abstract class CustomRowUIElement : CustomUIElement
    {

        private Dictionary<int, UIDynamicButton> buttonDict = new Dictionary<int, UIDynamicButton>();

        private bool activeButtonHighlight;

        public CustomRowUIElement(float height = 50f, bool _activeButtonHighlight = true) : base(height)
        {
            activeButtonHighlight = _activeButtonHighlight;
        }


        protected void ClearAllButtonColors()
        {
            foreach (KeyValuePair<int, UIDynamicButton> kvp in buttonDict)
            {
                kvp.Value.buttonColor = new Color(0.85f, 0.85f, 0.85f, 1f);
            }
        }

        protected void ClearAllPopupColors(UIDynamicPopup popupException=null)
        {
            foreach (UIDynamicPopup popup in uidPopuplist)
            {
                GameObject popupBackground = popup.popup.transform.Find("Background").gameObject;
                Image backgroundImage = popupBackground.GetComponent<Image>();

                backgroundImage.color = new Color(0.85f, 0.85f, 0.85f, 1f);
                popup.labelTextColor = new Color(0.25f, 0.25f, 0.25f);
                if (popup!= popupException) popup.popup.currentValue = "Select...";
            }
        }
        protected void RegisterButtonAction(UIDynamicButton newButton, string label, int buttonRef, Action<int> callbackMethod)
        {
            newButton.label = label;

            newButton.button.onClick.AddListener(delegate
            {
                try
                {
                    if (activeButtonHighlight) SetButtonColorActive(buttonRef);
                    callbackMethod.Invoke(buttonRef);
                }
                catch (Exception e) { SuperController.LogError("Exception caught in BA.CutomRowUIElement.RegisterButtonCallback: " + e); }
            });

            buttonDict.Add(buttonRef, newButton);
        }

        public void SetButtonColorActive(int buttonRef)
        {
            ClearAllButtonColors();
            ClearAllPopupColors();
            buttonDict[buttonRef].buttonColor = Const.UIhighlightColor;
        }

        public override void ClearUI()
        {
            base.ClearUI();
            buttonDict.Clear();
        }


    }

    public class CustomRowGridlessUIElement : CustomRowUIElement
    {
        public CustomRowGridlessUIElement(float height = 50f, bool _activeButtonHighlight = true) : base(height, _activeButtonHighlight)
        {

        }


        public UIDynamicPopup AddGridlessPopupToRow(JSONStorableStringChooser jssc, UIPopup.OnOpenPopup callbackMethod, float width, float xOffsetPostion, float yOffsetPostion = 0, float panelHeight = 400f, bool scrollable = true)
        {
            UIDynamicPopup newPopup = AddPopup(jssc, callbackMethod, width,spacerHeight, xOffsetPostion, yOffsetPostion, panelHeight);           

            newPopup.popup.onValueChangeHandlers += (string val) =>
            {
                if (val != "Select...")
                {
                    ClearAllButtonColors();
                    ClearAllPopupColors(newPopup);
                    GameObject popupBackground = newPopup.popup.transform.Find("Background").gameObject;
                    Image backgroundImage = popupBackground.GetComponent<Image>();

                    backgroundImage.color = Const.UIhighlightColor;
                    newPopup.labelTextColor = Color.black;
                }
            };       
           
            return newPopup;
        }



        public UIDynamicButton AddGridlessButtonToRow(string label, int buttonRef, Action<int> callbackMethod, float width, float xOffsetPostion, float yOffsetPostion = 0)
        {
            UIDynamicButton newButton = CreateButton(width, spacerHeight, xOffsetPostion, yOffsetPostion);

            RegisterButtonAction(newButton, label, buttonRef, callbackMethod);

            return newButton;
        }

    }

    public class CustomRowGridUIElement : CustomRowUIElement
    {
        private int buttonPadding = 0;


        public CustomRowGridUIElement(float height = 50, bool _activeButtonHighlight = true) : base(height, _activeButtonHighlight)
        {

        }
        public UIDynamicPopup AddPopupToRow(JSONStorableStringChooser jssc, UIPopup.OnOpenPopup callbackMethod = null, float panelHeight = 400f, bool forceRightParentPopup = false, bool scrollable = true)
        {
            UIDynamicPopup newPopup = InstantiatePopup(jssc, scrollable);
            newPopup.label = jssc.label;
            newPopup.popup.onOpenPopupHandlers += callbackMethod;
            newPopup.popup.onOpenPopupHandlers += () =>
            {
                //                UIAPluginUI.CloseAllPopups(newPopup);
                Globals.mvrScript.StartCoroutine(ReparentPopup(newPopup, forceRightParentPopup));
            };
            newPopup.popup.popupPanelHeight = panelHeight;
            newPopup.popup.selectColor = Const.UIhighlightColor;

            uidPopuplist.Add(newPopup);

            return newPopup;
        }

        public UIDynamicButton AddButtonToRow(string label, int buttonRef, Action<int> callbackMethod)
        {
            UIDynamicButton newButton = InstantiateUIButton();
            RegisterButtonAction(newButton, label, buttonRef, callbackMethod);

            return newButton;
        }

        public void InitUI(int customRowWidth, int fixedElementsCount = 0)
        {
            base.InitUI(customRowWidth);

            int width = 525;
            if (customRowWidth == CustomRowWidth.full) width = 1050;

            if (fixedElementsCount > 0)
            {
                GridLayoutGroup glg = _spacerActive.gameObject.AddComponent<GridLayoutGroup>();
                glg.cellSize = new Vector2((width - ((fixedElementsCount - 1) * buttonPadding)) / fixedElementsCount, spacerHeight);

                glg.constraint = GridLayoutGroup.Constraint.FixedRowCount;
                glg.constraintCount = 1;
                glg.padding = new RectOffset(buttonPadding, buttonPadding, buttonPadding, buttonPadding);
                glg.spacing = new Vector2(buttonPadding, 0);
            }


        }

    }
    
    public abstract class BaseUITab
    {
        protected JSONStorableString helpTextJSS = new JSONStorableString("helpText", "");
        protected string defaultHelpText = "";

        protected class UIGroup
        {
            public int groupID;
            public BaseUITab uiTab;

            public void RefreshGroupUIStringChange(string val)
            {
                uiTab.RefreshGroupUI(groupID);
            }
        }


        public void ApplyHelp(UIDynamicSlider sliderUID, string helpText)
        {
            ApplyHelp(sliderUID.slider.gameObject, helpText);
            ApplyHelp(sliderUID.sliderControl.gameObject, helpText);
            ApplyHelp(sliderUID.rangeAdjustGroup.gameObject, helpText);
            ApplyHelp(sliderUID.quickButtonsGroup.gameObject, helpText);
            ApplyHelp(sliderUID.labelText.gameObject, helpText);
            ApplyHelp(sliderUID.defaultButton.gameObject, helpText);

            var inputTrans = sliderUID.transform.Find("ValueInputField");
            if (inputTrans != null) ApplyHelp(inputTrans.gameObject, helpText);

            var panelTrans = sliderUID.transform.Find("Panel");
            if (panelTrans != null) ApplyHelp(panelTrans.gameObject, helpText);
        }

        public void ApplyHelp(UIDynamicPopup popupUID, string helpText)
        {
            ApplyHelp(popupUID.popup.gameObject, helpText);
            ApplyHelp(popupUID.popup.popupPanel.gameObject, helpText);
        }
        public void ApplyHelp(UIDynamicToggle toggleUID, string helpText)
        {
            ApplyHelp(toggleUID.toggle.gameObject, helpText);
        }
        public void ApplyHelp(UIDynamicTextField textFieldUID, string helpText)
        {
            GameObject gameObject = textFieldUID.gameObject;
            if (textFieldUID.transform.GetComponent<InputField>() != null)
            {
                gameObject = textFieldUID.transform.GetComponent<InputField>().gameObject;
            }
            ApplyHelp(gameObject, helpText);
        }
        public void ApplyHelp(UIDynamicButton buttonUID, string helpText, bool warning = false)
        {
            ApplyHelp(buttonUID.button.gameObject, helpText, warning);
        }

        public void ClearHelp()
        {
            helpTextJSS.val = "";
            Image image = helpTextJSS.text.transform.parent.gameObject.GetComponent<Image>();
            image.color = Color.gray;
        }

        public void ApplyHelp(GameObject gameObject, string helpText, bool warning = false)
        {
            GeneralPointerEnterExitHelp contextHelp = gameObject.AddComponent<GeneralPointerEnterExitHelp>();
            contextHelp.Init(helpText, helpTextJSS, warning, defaultHelpText);
        }

        public class UIDynamicElementLists
        {
            public HashSet<UIDynamic> UIDSpacersList;
            public HashSet<UIDynamicButton> UIDButtonsList;
            public HashSet<UIDynamicSlider> UIDSlidersList;
            public HashSet<UIDynamicPopup> UIDPopupList;
            public HashSet<UIDynamicColorPicker> UIDColorPickerList;
            public HashSet<UIDynamicTextField> UIDTextFieldList;
            public HashSet<UIDynamicToggle> UIDToggleList;
            public HashSet<CustomUIElement> UIDCustomElementList;
            public HashSet<BaseUITab> childUITabs;

            public UIDynamicElementLists()
            {
                UIDSpacersList = new HashSet<UIDynamic>();
                UIDButtonsList = new HashSet<UIDynamicButton>();
                UIDSlidersList = new HashSet<UIDynamicSlider>();
                UIDPopupList = new HashSet<UIDynamicPopup>();
                UIDColorPickerList = new HashSet<UIDynamicColorPicker>();
                UIDTextFieldList = new HashSet<UIDynamicTextField>();
                UIDToggleList = new HashSet<UIDynamicToggle>();
                UIDCustomElementList = new HashSet<CustomUIElement>();
                childUITabs = new HashSet<BaseUITab>();
            }

            public void ClearUIDs()
            {
                MVRScript mvrs = Globals.mvrScript;
                foreach (UIDynamic spacer in UIDSpacersList) if (spacer != null) mvrs.RemoveSpacer(spacer);
                foreach (UIDynamicButton button in UIDButtonsList) if (button != null) mvrs.RemoveButton(button);
                foreach (UIDynamicSlider slider in UIDSlidersList) if (slider != null) mvrs.RemoveSlider(slider);
                foreach (UIDynamicPopup popup in UIDPopupList) if (popup != null) mvrs.RemovePopup(popup);
                foreach (UIDynamicColorPicker colorpicker in UIDColorPickerList) if (colorpicker != null) mvrs.RemoveColorPicker(colorpicker);
                foreach (UIDynamicTextField textfield in UIDTextFieldList) if (textfield != null) mvrs.RemoveTextField(textfield);
                foreach (UIDynamicToggle toggle in UIDToggleList) if (toggle != null) mvrs.RemoveToggle(toggle);
                foreach (CustomUIElement Element in UIDCustomElementList) Element.ClearUI();

                foreach (BaseUITab childUITab in childUITabs) childUITab.ClearAllGroupsUI();
                UIDSpacersList.Clear();
                UIDButtonsList.Clear();
                UIDSlidersList.Clear();
                UIDPopupList.Clear();
                UIDColorPickerList.Clear();
                UIDTextFieldList.Clear();
                UIDToggleList.Clear();
                UIDCustomElementList.Clear();
            }

            public int GetUIElementsCount()
            {
                int count = 0;
                foreach (UIDynamic spacer in UIDSpacersList) count += 1;
                foreach (UIDynamicButton button in UIDButtonsList) count += 1;
                foreach (UIDynamicSlider slider in UIDSlidersList) count += 1;
                foreach (UIDynamicPopup popup in UIDPopupList) count += 1;
                foreach (UIDynamicColorPicker colorpicker in UIDColorPickerList) count += 1;
                foreach (UIDynamicTextField textfield in UIDTextFieldList) count += 1;
                foreach (UIDynamicToggle toggle in UIDToggleList) count += 1;
                foreach (CustomUIElement customElement in UIDCustomElementList) count += 1;
                foreach (BaseUITab childUITab in childUITabs) count += 1;

                return count;
            }
            public void CloseAllPopup(UIDynamicPopup exception)
            {
                foreach (UIDynamicPopup UIDpopup in UIDPopupList)
                {
                    if (UIDpopup != exception) UIDpopup.popup.visible = false;
                }
                foreach (CustomUIElement customElement in UIDCustomElementList)
                {
                    customElement.CloseAllPopups(exception);
                }
                foreach (BaseUITab childUITab in childUITabs)
                {
                    childUITab.CloseAllPopups(exception);
                }
            }
        }

        private Dictionary<int, UIDynamicElementLists> uiElementGroups;
        private Dictionary<int, List<JSONStorableParam>> JSPsGroups;

        protected int uiGroupCount = 1;

        public BaseUITab()
        {
            uiElementGroups = new Dictionary<int, UIDynamicElementLists>();
            JSPsGroups = new Dictionary<int, List<JSONStorableParam>>();
        }
        protected void ForceUIRefreshOnValChange(JSONStorableParam jsp, int uiRefreshGroup, BaseUITab uiTab = null)
        {
            UIGroup groupLR = new UIGroup();
            groupLR.groupID = uiRefreshGroup;

            if (uiTab == null) groupLR.uiTab = this;
            else groupLR.uiTab = uiTab;

            switch (jsp.type)
            {
                case JSONStorable.Type.StringChooser:
                    RegisterJSSCRefreshUIonValChange(jsp as JSONStorableStringChooser, groupLR);
                    break;
                case JSONStorable.Type.String:
                    RegisterJSSRefreshUIonValChange(jsp as JSONStorableString, groupLR);
                    break;
                case JSONStorable.Type.Float:
                    RegisterJSFRefreshUIonValChange(jsp as JSONStorableFloat, groupLR);
                    break;
                case JSONStorable.Type.Bool:
                    RegisterJSBRefreshUIonValChange(jsp as JSONStorableBool, groupLR);
                    break;
                default:
                    SuperController.LogError("BA.BaseUITab.RegisterJSRefreshUIonValChange: unsupported JSONStorableParam type " + jsp.type.ToString());
                    break;
            }
        }
        private void RegisterJSSCRefreshUIonValChange(JSONStorableStringChooser jssc, UIGroup groupLR)
        {

            if (jssc.popup == null)
            {
                SuperController.LogError("BA.BaseUITab.RegisterJSSCRefreshUIonValChange - attempt to register JSSC with no associated UI Popup");
                return;
            }
            jssc.popup.onValueChangeHandlers += groupLR.RefreshGroupUIStringChange;
        }

        private void RegisterJSBRefreshUIonValChange(JSONStorableBool jsb, UIGroup groupLR)
        {
            if (jsb.toggle == null)
            {
                SuperController.LogError("BA.BaseUITab.RegisterJSBRefreshUIonValChange - attempt to register JSB with no associated UI Toggle");
                return;
            }
            jsb.toggle.onValueChanged.AddListener((value) => groupLR.RefreshGroupUIStringChange(""));
        }

        private void RegisterJSSRefreshUIonValChange(JSONStorableString jss, UIGroup groupLR)
        {
            if (jss.inputField == null)
            {
                SuperController.LogError("BA.BaseUITab.RegisterJSSRefreshUIonValChange - attempt to register JSS with no associated inputfield");
                return;
            }
            jss.inputField.onValueChanged.AddListener((text) => groupLR.RefreshGroupUIStringChange(text));

        }

        private void RegisterJSFRefreshUIonValChange(JSONStorableFloat jsf, UIGroup groupLR)
        {

            if (jsf.slider == null)
            {
                SuperController.LogError("BA.BaseUITab.RegisterJSFRefreshUIonValChange - attempt to register JSF with no associated Slider");
                return;
            }
            jsf.slider.onValueChanged.AddListener((floatVal) => groupLR.RefreshGroupUIStringChange(floatVal.ToString()));

        }

        protected abstract void CreateGroupUI(int groupID);

        public void RefreshGroupUI(int startGroupID)
        {

            CloseAllPopups(null);

            for (int groupID = startGroupID; groupID <= uiGroupCount; groupID++)
            {
                ClearGroupUI(groupID);
            }

            for (int groupID = startGroupID; groupID <= uiGroupCount; groupID++)
            {
                CreateGroupUI(groupID);
            }
        }

        private UIDynamicElementLists GetUIElementsFromGroup(int groupID, Dictionary<int, UIDynamicElementLists> uiElementsGroups)
        {
            UIDynamicElementLists uiElementGroup;

            if (uiElementsGroups.ContainsKey(groupID)) uiElementGroup = uiElementsGroups[groupID];
            else
            {
                uiElementGroup = new UIDynamicElementLists();
                uiElementsGroups.Add(groupID, uiElementGroup);
            }
            return (uiElementGroup);
        }

        protected void RegisterCustomUIElement(int groupID, CustomUIElement customElement)
        {
            UIDynamicElementLists uiElements = GetUIElementsFromGroup(groupID, uiElementGroups);     

            uiElements.UIDCustomElementList.Add(customElement);
        }

        protected void RegisterChildUITab(int groupID, BaseUITab childUITab)
        {
            UIDynamicElementLists uiElements = GetUIElementsFromGroup(groupID, uiElementGroups);
            uiElements.childUITabs.Add(childUITab);
        }

        protected UIDynamic CreateSpacer(int groupID, float height = 10f, bool rightSide = false)
        {
            UIDynamicElementLists uiElements = GetUIElementsFromGroup(groupID, uiElementGroups);

            UIDynamic spacer = Globals.mvrScript.CreateSpacer(rightSide);
            uiElements.UIDSpacersList.Add(spacer);
            spacer.height = height;
            return (spacer);
        }

        public void ClearAllGroupsUI()
        {
            CloseAllPopups(null);

            foreach (KeyValuePair<int, UIDynamicElementLists> kvp in uiElementGroups) ClearGroupUI(kvp.Key);

        }
        public virtual void ClearGroupUI(int groupID)
        {
            UIDynamicElementLists uiElements = GetUIElementsFromGroup(groupID, uiElementGroups);

            List<JSONStorableParam> jspListForGroup = null;
            uiElements.ClearUIDs();
            if (JSPsGroups.ContainsKey(groupID)) jspListForGroup = JSPsGroups[groupID];

            if (jspListForGroup != null)
            {
                jspListForGroup.Clear();
            }

        }
        public void CloseAllPopups(UIDynamicPopup exception)
        {
            foreach (var uiElements in uiElementGroups.Values.ToList()) uiElements.CloseAllPopup(exception);
        }

    }


    }
