﻿
using MVR.FileManagementSecure;
using System;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using UnityEngine;
using UnityEngine.UI;



namespace JayJayWon
{
    public static class BAPluginUI
    {
        private static CustomRowGridlessUIElement topTabMenuOptions;

        private static int currentTopTabSelection = TopUITabs.generalSettings;

        private static BaseUITab currentTabUI;

        private static GlobalUITab generalSettingsUITab;
        private static DesktopUITab desktopSettingsUITab;
        private static VRSettingsUITab vrSettingsUITab;
        private static ResourceSettingsUITab resourceSettingsUITab;

        private static UIDynamicButton generalSettingsUID;
        private static UIDynamicButton VRSettingsUID;
        private static UIDynamicButton desktopSettingsUID;
        private static UIDynamicPopup resourceSettingsUID;

        private static JSONStorableMultiEnumStringChooser resourceSettingsJSESC;

        public static void InitUI()
        {
            topTabMenuOptions = new CustomRowGridlessUIElement(80);
            topTabMenuOptions.InitUI(CustomRowWidth.full);

            generalSettingsUITab = new GlobalUITab();
            generalSettingsUID = topTabMenuOptions.AddGridlessButtonToRow("Global\nSettings", TopUITabs.generalSettings, TopTabSelected, 200, 0, 0);

            if (Globals.validStartup)
            {
                desktopSettingsUITab = new DesktopUITab();
                vrSettingsUITab = new VRSettingsUITab();
                resourceSettingsUITab = new ResourceSettingsUITab();

                resourceSettingsJSESC = new JSONStorableMultiEnumStringChooser("resourceType", null, null, topEnumManifestName: ResourceType.enumManifestName, topEnumCallbackFunction: ResourceSettingsTabSelected, mainValCallbackFunction: ResourceSettingsTabSelected);
                resourceSettingsJSESC.label = "Resource\nSettings";

                RePopulateResourceTypeSelector();


                VRSettingsUID = topTabMenuOptions.AddGridlessButtonToRow("VR\nSettings", TopUITabs.vrSettings, TopTabSelected, 200, 200, 0);
                desktopSettingsUID = topTabMenuOptions.AddGridlessButtonToRow("Desktop\nSettings", TopUITabs.desktopSettings, TopTabSelected, 200, 400, 0);

                resourceSettingsUID = topTabMenuOptions.AddGridlessPopupToRow(resourceSettingsJSESC, null, 450, 600, 0, 800, false);
                resourceSettingsUID.popup.currentValue = "Select...";
            }           

            generalSettingsUITab.RefreshGroupUI(1);

            topTabMenuOptions.SetButtonColorActive(currentTopTabSelection);

            currentTabUI = generalSettingsUITab;
        }

        public static void RePopulateResourceTypeSelector()
        {
            ResourceManifest.PopulateResourceTypeSelector(resourceSettingsJSESC);
        }

        public static void ResourceSettingsTabSelected(string value)
        {
            ResourceSettingsTabSelected(ResourceType.presetAtom);
        }
        private static void ResourceSettingsTabSelected(int val)
        {
            if (currentTopTabSelection!= TopUITabs.resourceSettings)
            {
                currentTopTabSelection = TopUITabs.resourceSettings;
                ClearCurrentTopTabsUI();
                
                currentTabUI = resourceSettingsUITab;             
            }
            resourceSettingsUITab.resourceType = val;
            if (val == ResourceType.presetAtom) resourceSettingsUITab.atomPresetResourceType = resourceSettingsJSESC.mainVal;

            if (PC.patreonContentEnabled || val == ResourceType.scene || val == ResourceType.presetAppearance)
            {
                if (val != ResourceType.presetAtom) BrowserAssist.mainBrowserUI.browserFiltersUI.currentResourceType = val;
                else BrowserAssist.mainBrowserUI.browserFiltersUI.currentAtomPresetResourceType = resourceSettingsJSESC.mainVal;
            }
            resourceSettingsUITab.RefreshGroupUI(1);
        }
        private static void TopTabSelected(int tabRef)
        {
            if (tabRef != currentTopTabSelection)
            {
                ClearCurrentTopTabsUI();
                switch (tabRef)
                {
                    case TopUITabs.generalSettings:
                        generalSettingsUITab.RefreshGroupUI(1);
                        currentTabUI = generalSettingsUITab;
                        break;
                    case TopUITabs.vrSettings:
                        vrSettingsUITab.RefreshGroupUI(1);
                        currentTabUI = vrSettingsUITab;
                        break;
                    case TopUITabs.desktopSettings:
                        desktopSettingsUITab.RefreshGroupUI(1);
                        currentTabUI = desktopSettingsUITab;
                        break;

                }
                currentTopTabSelection = tabRef;
            }
        }

        private static void ClearCurrentTopTabsUI()
        {
            currentTabUI.ClearAllGroupsUI();

        }

        public static void CloseAllPopups(UIDynamicPopup exception)
        {
            currentTabUI.CloseAllPopups(exception);
        }
    }

    public class GlobalUITab : BaseUITab
    {
        private CustomRowGridUIElement globalSettingsTabMenuOptions;
        private CustomUIElement globalSettingsCustomUI;
        private CustomUIElement helpTextCustomUI;

        private HashSet<ResourceTypeSettingsRowUI> resourceTypeSettingRows = new HashSet<ResourceTypeSettingsRowUI>();

        private HashSet<BackupSettingsRowUI> backupSettingRows = new HashSet<BackupSettingsRowUI>();

        private int currentGlobalSettingsTab = GlobalSettingsUITabs.general;

        public GlobalUITab()
        {
            if (!Globals.validStartup) currentGlobalSettingsTab = GlobalSettingsUITabs.backups;

            globalSettingsTabMenuOptions = new CustomRowGridUIElement();
            uiGroupCount = 2;

            globalSettingsCustomUI = new CustomUIElement(825);
            helpTextCustomUI = new CustomUIElement(200);
            defaultHelpText = "The Global Settings tab contain settings that impact general behavior of the Browser Assist UI - which are not specific to VR, Desktop or Resource types.";
            helpTextJSS.val = defaultHelpText;
            
        }

        private void GlobalSettingsTabSelected(int tabRef)
        {
            if (tabRef != currentGlobalSettingsTab)
            {
                currentGlobalSettingsTab = tabRef;
                RefreshGroupUI(2);
            }
        }

        protected override void CreateGroupUI(int groupID)
        {
            if (groupID == 1)
            {
                globalSettingsTabMenuOptions.InitUI(CustomRowWidth.full, 4);
                

                if (Globals.validStartup)
                {
                    globalSettingsTabMenuOptions.AddButtonToRow("General", GlobalSettingsUITabs.general, GlobalSettingsTabSelected);

                    globalSettingsTabMenuOptions.AddButtonToRow("Resource Types", GlobalSettingsUITabs.resources, GlobalSettingsTabSelected);                  
                }
                
                globalSettingsTabMenuOptions.AddButtonToRow("Backups", GlobalSettingsUITabs.backups, GlobalSettingsTabSelected);

                if (PC.patreonContentEnabled && Globals.validStartup) globalSettingsTabMenuOptions.AddButtonToRow("Import", GlobalSettingsUITabs.import, GlobalSettingsTabSelected);

                RegisterCustomUIElement(groupID, globalSettingsTabMenuOptions);

                globalSettingsTabMenuOptions.SetButtonColorActive(currentGlobalSettingsTab);
            }
            if (groupID == 2)
            {
                globalSettingsCustomUI.InitUI(CustomRowWidth.full);
                RegisterCustomUIElement(2, globalSettingsCustomUI);

                if (currentGlobalSettingsTab == GlobalSettingsUITabs.general) CreateUIGeneralSettings();

                if (currentGlobalSettingsTab == GlobalSettingsUITabs.resources) CreateUIResourceTypeSettings();

                if (currentGlobalSettingsTab == GlobalSettingsUITabs.backups) CreateUIBackupSettings();
                if (currentGlobalSettingsTab == GlobalSettingsUITabs.import) CreateUIImportSettings();
            }

                
        }

        private void CreateUIImportSettings()
        {
            defaultHelpText = "This Import settings tab allows settings that relate to the direct Importing of resources from Scenes.";


            helpTextJSS.val = defaultHelpText;

            helpTextCustomUI.InitUI(CustomRowWidth.full);

            var helpTextBox = helpTextCustomUI.AddTextBox(helpTextJSS, 1050f, 200f, 0f, 0f);
            helpTextBox.textColor = Color.white;
            helpTextBox.backgroundColor = new Color(0.4f, 0.5f, 0.2f, 0.4f);

            RegisterCustomUIElement(2, helpTextCustomUI);

            var disableNewAtomToggle = globalSettingsCustomUI.AddToggle(Settings.sceneImportDisableNewAtomJSB, 1050f, 50f, 0f, 0f, "Disable <New Atom> target option");
            ApplyHelp(disableNewAtomToggle, "Disable the <New Atom> option when selecting a target Atom for an Import action. If no valid target atoms are availbale in the currrent scene, then the <New Atom> option will remain available irrespective of this setting.");

            var enableSceneCachineToggle = globalSettingsCustomUI.AddToggle(Settings.sceneImportPersistCacheJSB, 1050f, 50f, 0, -120f, "Enable disk Caching of Scene contents for Import");
            ApplyHelp(enableSceneCachineToggle, "Enable Caching of Scene Contents. When active, the contents of any scene files are cached in a separate BrowserAssist cache disk store. This means their contents can be browsed more rapidly without needing to re-load scenes every time BrowserAssist is restarted. However, this does require some additional disk space - effectively doubling the space requirements of each scene file.");

            var clearCacheButton = globalSettingsCustomUI.AddButton(1050f, 50f, 0, -180f, "Clear Disk Cache", ClearCacheCallback);
            clearCacheButton.textColor = Color.white;
            clearCacheButton.buttonColor = Color.red;
            ApplyHelp(clearCacheButton, "This button will clear all cached scene files that are used by the BrowserAssist Import feature. This will save disk space but means that browsing of Scene import options is slower after each BrowserAssist restart.");

        }

        private void ClearCacheCallback()
        {
            SceneImportCache.ClearCache();
        }

        private JSONStorableString backupNamesFilterJSS = new JSONStorableString("backupNamesFilter", "");
        private GameObject backupsListContentGO;
        private GameObject backupsSettingsBoxGO;
        private UIDynamicButton resetButton;
        private UIDynamicButton cancelResetButton;
        private UIDynamicButton confirmResetButton;



        private void CreateUIBackupSettings()
        {
            if (Globals.validStartup)
            {
                defaultHelpText = "The Backups tab allows backups to be taken of the BrowserAssist meta data (e.g. Tags). These backups can be used to restore from in the event of an unexpected data corruption.\nNote that Clothing Tags are not included in backups and are not impacted by the BrowserAssist Restore or Reset processes.";
            }
            else
            {
                defaultHelpText = "An unexpected error occured when scanning the BrowserAssist data on startup. To preserve the integrity of the meta data, BrowserAssist will not operate until this issue is resolved.\n\nPrevious backups can be restored from this screen. Its advised that a backup of the corrupt data is taken before restoring a previous backup. If the restore process does not work, then please contact JayJayWon on Discord. Alternatively you can use the Reset function - but any meta data, such as Tags, will be lost from the current data set. Previous backups are maintained on a Reset so its advisable to perform a backup before doing a Reset.";
            }              
            
            helpTextJSS.val = defaultHelpText;

            helpTextCustomUI.InitUI(CustomRowWidth.full);

            var helpTextBox = helpTextCustomUI.AddTextBox(helpTextJSS, 1050f, 200f, 0f, 0f);
            helpTextBox.textColor = Color.white;
            helpTextBox.backgroundColor = new Color(0.4f, 0.5f, 0.2f, 0.4f);

            var createBackupButton = globalSettingsCustomUI.AddButton(300f, 43f, 0f, 0f, "Create Backup", CreateBackupButtonCallback);
            createBackupButton.buttonColor = Color.green;

            if (Globals.validStartup)
            {
                ApplyHelp(createBackupButton, "Create a new backup of the current BrowserAssist meta data. This will include data relating to resource Tags and other BrowserAssist persisted data. Once created the backup will be listed below with the current Date and Time as a backup name.");
            }
            else
            {
                ApplyHelp(createBackupButton, "Create a backup of the current BrowserAssist meta data. This will include data relating to resource Tags and other BrowserAssist persisted data. This backup will be marked as CORRUPT and wont be available for Restoring from - but may be useful for diagnosis or data retrieval by JayJayWon.");
            }
                

            backupNamesFilterJSS.val = "";
            var backupSettingsFilterTextBox = globalSettingsCustomUI.AddTextEntryBox(backupNamesFilterJSS, 540f, 50f, 30, 0f, -45f, TextAnchor.MiddleLeft, InputField.LineType.SingleLine, "Enter filter text for Backups list...", 0f, 0f, BackupsFilterNameChangeCallback);
            ApplyHelp(backupSettingsFilterTextBox, "Enter a search string to filter the list of Backups and allow quicker selection from a large number of backups.");

            backupsSettingsBoxGO = globalSettingsCustomUI.CreateVScrollableContent(AnchorPresets.topLeft, 1050f, 670f, new Vector2(0f, -100f), 20f, 3f);
            backupsListContentGO = backupsSettingsBoxGO.transform.Find("scrollableContentViewportGO/vScrollableContentGO").gameObject;

            RefreshBackupsSettingsList();

            resetButton = globalSettingsCustomUI.AddButton(300f, 45f, 0f, -780f, "Reset...", ResetCallback);
            resetButton.buttonColor = Color.red;
            resetButton.textColor = Color.white;
            ApplyHelp(resetButton, "Reset the BrowserAssist meta data. This effectively restores the plugin to its initial state and will result in a loss of any user meta data such as Tags etc. Backups are preserved when performing a Reset. The Reset process will also restart BrowserAssist which may take a noticeable period of time, during which VAM will become unresponsive.");


            cancelResetButton = globalSettingsCustomUI.AddButton(300f, 45f, 0f, -780f, "Cancel", CancelResetCallback);
            cancelResetButton.buttonColor = Color.green;
            ApplyHelp(cancelResetButton, "Cancel the reset of BrowserAssist meta data.");

            confirmResetButton = globalSettingsCustomUI.AddButton(300f, 45f, 310f, -780f, "Confirm", ConfirmResetCallback);
            confirmResetButton.buttonColor = Color.red;
            confirmResetButton.textColor = Color.white;
            ApplyHelp(confirmResetButton, "Confirm the reset the BrowserAssist meta data. This effectively restores the plugin to its initial state and will result in a loss of any user meta data such as Tags etc. Backups are preserved when performing a Reset. The Reset process will also restart BrowserAssist which may take a noticeable period of time, during which VAM will become unresponsive.");

            cancelResetButton.gameObject.SetActive(false);
            confirmResetButton.gameObject.SetActive(false);



            RegisterCustomUIElement(2, helpTextCustomUI);
        }

        private void ResetCallback()
        {
            resetButton.gameObject.SetActive(false);
            cancelResetButton.gameObject.SetActive(true);
            confirmResetButton.gameObject.SetActive(true);

            
        }

        private void CancelResetCallback()
        {
            resetButton.gameObject.SetActive(true);
            cancelResetButton.gameObject.SetActive(false);
            confirmResetButton.gameObject.SetActive(false);
        }

        private void ConfirmResetCallback()
        {


            resetButton.gameObject.SetActive(true);
            cancelResetButton.gameObject.SetActive(false);
            confirmResetButton.gameObject.SetActive(false);

#if VAM_GT_1_20_77_13
            BackupSettingsRowUI.DeleteAllMetaDataFiles();

            var pluginUID = Globals.mvrScript.storeId.Substring(0, Globals.mvrScript.storeId.IndexOf("_"));

            Globals.mvrScript.manager.ReloadPluginWithUID(pluginUID);
#else
            SuperController.LogMessage("Resetting of BrowserAssist is not supported in versions of VAM before 1.20.77.13.");
#endif
        }


        private void CreateBackupButtonCallback()
        {
            string newBackupFolder;
            string baseBackupFolderName;
            int nextUniqueID = 1;

            try
            {
                baseBackupFolderName = string.Format("{0:yyyy-MM-dd HH-mm}", DateTime.Now);

                string corruptedPostFix = "";
                if (!Globals.validStartup) corruptedPostFix = " CORRUPTED";

                string unusedBackupFolderName = baseBackupFolderName+ corruptedPostFix;
                
                while (FileManagerSecure.DirectoryExists(Const.MetaDataBackupFolder + "\\" + unusedBackupFolderName))
                {
                    unusedBackupFolderName = baseBackupFolderName + "(" + nextUniqueID + ")"+ corruptedPostFix;
                    nextUniqueID++;
                }

                newBackupFolder = Const.MetaDataBackupFolder + "\\" + unusedBackupFolderName;

                FileManagerSecure.CreateDirectory(newBackupFolder);
            }
            catch(Exception ex)
            {
                SuperController.LogError("BA.GeneralUITab.CreateBackupButtonCallback: Exception caught whilst creating backup folder. " + ex);
                return;
            }

            try
            {
                FileManagerSecure.CreateDirectory(newBackupFolder + "\\LocalResourcesUserData");
                FileManagerSecure.CreateDirectory(newBackupFolder + "\\VARGroupCoreData");
                FileManagerSecure.CreateDirectory(newBackupFolder + "\\VARGroupUserData");
                FileManagerSecure.CreateDirectory(newBackupFolder + "\\VARResourcesCoreData");
                FileManagerSecure.CreateDirectory(newBackupFolder + "\\VARResourcesUserData");

                Utils.CopyFolders(Const.LocalResourcesUserDataFolder, newBackupFolder + "\\LocalResourcesUserData", Const.LocalResourceDataFileName);

                Utils.CopyFolders(Const.VARGroupCoreDataFilesFolder, newBackupFolder + "\\VARGroupCoreData", Const.VARGroupDataFileName);

                Utils.CopyFolders(Const.VARGroupUserDataFilesFolder, newBackupFolder + "\\VARGroupUserData", Const.VARGroupDataFileName);

                Utils.CopyFolders(Const.VARResourcesCoreDataFilesFolder, newBackupFolder + "\\VARResourcesCoreData", Const.VARResourceDataFileName);

                Utils.CopyFolders(Const.VARResourcesUserDataFilesFolder, newBackupFolder + "\\VARResourcesUserData", Const.VARResourceDataFileName);

            }
            catch (Exception ex)
            {
                SuperController.LogError("BA.GeneralUITab.CreateBackupButtonCallback: Exception caught whilst copying files to backup folder. " + ex);

                var unusedFailedBackupFolderName = baseBackupFolderName;

                while (FileManagerSecure.DirectoryExists(Const.MetaDataBackupFolder + "\\" + unusedFailedBackupFolderName + " FAILED"))
                {
                    unusedFailedBackupFolderName = baseBackupFolderName + "(" + nextUniqueID + ")";
                    nextUniqueID++;
                }

                string failedBackupFolder = Const.MetaDataBackupFolder + "\\" + unusedFailedBackupFolderName + " FAILED";

                try
                {
                    FileManagerSecure.MoveFile(newBackupFolder, failedBackupFolder);
                }
                catch (Exception e)
                {
                    SuperController.LogError("BA.GeneralUITab.CreateBackupButtonCallback: Exception occured whilst trying to mark a failed backup folder as FAILED. " + e);
                }
            }           

            RefreshBackupsSettingsList();
        }



        private void BackupsFilterNameChangeCallback(string backupNameFilter)
        {
            backupNamesFilterJSS.val = backupNameFilter;
            RefreshBackupsSettingsList();

        }


        private JSONStorableString resourceTypeFilterJSS = new JSONStorableString("resourceTypeFilter", "");
        private GameObject resourceTypesListContentGO;
        private GameObject resourceTypeSettingsBoxGO;

        private void CreateUIResourceTypeSettings()
        {
            if (PC.patreonContentEnabled) defaultHelpText = "The Resource Type tab allows each Resource Type to be concealed. Once a Resource Type has been marked as 'Concealed' it will not longer be selectable in the Browser Filter or Resource Settings dropdowns. This enables less useful resouce types to be ignored and presents a simpler set of choices for the user.";
            else defaultHelpText = "The Resource Type tab allows each Resource Type to be concealed. Once a Resource Type has been marked as 'Concealed' it will not longer be selectable in the Browser Filter or Resource Settings dropdowns. This is useful in Patreons version of the plugin where many different types of resources are supported.";
            helpTextJSS.val = defaultHelpText;

            helpTextCustomUI.InitUI(CustomRowWidth.full);

            var helpTextBox = helpTextCustomUI.AddTextBox(helpTextJSS, 1050f, 200f, 0f, 0f);
            helpTextBox.textColor = Color.white;
            helpTextBox.backgroundColor = new Color(0.4f, 0.5f, 0.2f, 0.4f);

            resourceTypeFilterJSS.val = "";
            var resourceTypeSettingsFilterTextBox = globalSettingsCustomUI.AddTextEntryBox(resourceTypeFilterJSS, 540f, 50f, 30, 0f, 5f, TextAnchor.MiddleLeft, InputField.LineType.SingleLine, "Enter filter text for Resource list...", 0f, 0f, ResourceTypeFilterNameChangeCallback);
            ApplyHelp(resourceTypeSettingsFilterTextBox, "Enter a search string to filter the list of Resource Types and allow quicker selection from a large number of resource types.");
            resourceTypeSettingsBoxGO = globalSettingsCustomUI.CreateVScrollableContent(AnchorPresets.topLeft, 1050f, 770f, new Vector2(0f, -50f), 20f, 3f);
            resourceTypesListContentGO = resourceTypeSettingsBoxGO.transform.Find("scrollableContentViewportGO/vScrollableContentGO").gameObject;
            
            RefreshResourceTypeSettingsList();

            RegisterCustomUIElement(2, helpTextCustomUI);
        }

        private void ResourceTypeFilterNameChangeCallback(string resourceFilter)
        {
            resourceTypeFilterJSS.val = resourceFilter;
            RefreshResourceTypeSettingsList();

        }

        private void ClearResourceTypeSettingsList()
        {
            foreach (var resourceTypeRow in resourceTypeSettingRows)
            {
                resourceTypeRow.ClearUI();
            }
            resourceTypeSettingRows.Clear();
        }

        private void ClearBackupsSettingsList()
        {
            foreach (var row in backupSettingRows)
            {
                row.ClearUI();
            }
            backupSettingRows.Clear();
        }

        private string GetResourceTypeDisplayName(int resourceType, string atomPresetResourceType)
        {
            string displayName = EnumManifest.GetDisplayVal(ResourceType.enumManifestName, resourceType);

            if (resourceType == ResourceType.presetAtom)
            {
                displayName = "Preset: "+atomPresetResourceType;
            }

            return displayName;
        }

        public void RefreshBackupsSettingsList()
        {
            ClearBackupsSettingsList();

            string filter = backupNamesFilterJSS.val.ToLower();

            List<string> backupNames = new List<string>();
            string[] folders = new string[] { };

            try
            {
                folders = FileManagerSecure.GetDirectories(Const.MetaDataBackupFolder);
            }
            catch(Exception ex)
            {
                SuperController.LogError("BA.GeneralUITab.RefreshBackupsSettingsList: Exception occured attempting to retrieve the list of available backups.");
            }
            

            foreach (var folder in folders)
            {
                backupNames.Add(FileManagerSecure.GetFileName(folder));
            }

            backupNames.Reverse();

            foreach (string backupName in backupNames)
            {
                if (backupName.ToLower().Contains(filter))
                {
                    var backupSettingRow = new BackupSettingsRowUI(backupsListContentGO, backupName, this);
                    backupSettingRows.Add(backupSettingRow);
                }
            }
        }

        public void RefreshResourceTypeSettingsList()
        {           
            ClearResourceTypeSettingsList();

            string filter = resourceTypeFilterJSS.val.ToLower();

            List<int> resourceTypes = EnumManifest.GetEnumsList(ResourceType.enumManifestName);

            bool disableConcealResourceTypes= false;
            if (Settings.concealedResourceTypeCount >= resourceTypes.Count - 2) disableConcealResourceTypes = true;

            if (!PC.patreonContentEnabled)
            {
                if (Settings.IsResourceConcealed(ResourceType.scene,"") || Settings.IsResourceConcealed(ResourceType.presetAppearance, "")) disableConcealResourceTypes = true;
            }

            foreach (int resourceType in resourceTypes)
            {
                if (PC.patreonContentEnabled || resourceType == ResourceType.scene || resourceType == ResourceType.presetAppearance)
                {
                    if (resourceType != ResourceType.presetAtom)
                    {
                        string atomPresetResourceType = "";
                        if (ResourceType.GetResourceCategory(resourceType) == ResourceCategory.preset) atomPresetResourceType = "Person";

                        if (resourceType!=ResourceType.clothingItemPresets &&( filter == "" || GetResourceTypeDisplayName(resourceType, atomPresetResourceType).ToLower().Contains(filter)))
                        {
                            var tagSettingRow = new ResourceTypeSettingsRowUI(resourceTypesListContentGO, resourceType, atomPresetResourceType, this, disableConcealResourceTypes);
                            resourceTypeSettingRows.Add(tagSettingRow);
                        }

                    }
                    else
                    {
                        foreach (string atomPresetResourceType in ResourceManifest.existingAtomPresetResourceTypes)
                        {
                            if (filter == "" || GetResourceTypeDisplayName(resourceType, atomPresetResourceType).ToLower().Contains(filter))
                            {
                                var tagSettingRow = new ResourceTypeSettingsRowUI(resourceTypesListContentGO, resourceType, atomPresetResourceType, this, false);
                                resourceTypeSettingRows.Add(tagSettingRow);
                            }

                        }
                    }
                }
                

            }
        }

        private void CreateUIGeneralSettings()
        {
            var launchBAToggle = globalSettingsCustomUI.AddToggle(Settings.enableTopScreenLaunchButtonJSB, 1020, 50, 0, 0, "Enable BrowserAssist Launch button on VAM Top Menu");
            var launchButtonPlacementPopup = globalSettingsCustomUI.AddPopup(Settings.launchButtonPlacementJSESC, null, 1020, 80, 0, -60, labelProportion: 0.7f);
            var defaultResourceTypePopup = globalSettingsCustomUI.AddPopup(Settings.defaultResourceTypeJMESC, null, 1020, 80, 0, -150, panelHeight: 750f, labelProportion: 0.7f);
            var resourceNameSearchModePopup = globalSettingsCustomUI.AddPopup(Settings.resourceNameSearchModeJSESC, null, 1020, 80, 0, -240, labelProportion: 0.7f);
            var defaultTargetPersonType = globalSettingsCustomUI.AddPopup(Settings.defaultTargetPersonTypeJSE, null, 1020, 80, 0, -330, labelProportion: 0.7f);
            var defaultCreatorFilterSortOrder = globalSettingsCustomUI.AddPopup(Settings.defaultCreatorFilterOrderJSEC, null, 1020, 80, 0, -420, labelProportion: 0.7f);

            
            var enableBulkDeleteToggle = globalSettingsCustomUI.AddToggle(Settings.enableBulkDeleteJSB, 1020, 50, 0, -510, "Enable bulk delete of multi selected resources");
            var enableCSLReferencedPluginsToggle = globalSettingsCustomUI.AddToggle(Settings.enableCSLReferencedPluginsJSB, 1020, 50, 0, -570, "Enable browsing of csList referenced plugins");
            
            

            var disableMetaPurgeToggle = globalSettingsCustomUI.AddToggle(Settings.disableMetaPurgeOnMissingVARJSB, 1020, 50, 0, -630, "Disable purge of meta data on missing VAR");

            var enableQuickFilterToggle = globalSettingsCustomUI.AddToggle(Settings.enableQuickFiltersJSB, 1020, 50, 0, -690, "Enable Quick Filter buttons panel (Pateons only)");

            var rescanVARsButton = globalSettingsCustomUI.AddButton(1020, 50, 0, -760, "Rescan all VAR packages and update contents", delegate
            {
                VARPackageManifest.RescanAllVARResource();
            });

            helpTextCustomUI.InitUI(CustomRowWidth.full);

            var helpTextBox = helpTextCustomUI.AddTextBox(helpTextJSS, 1050f, 200f, 0f, 0f);
            helpTextBox.textColor = Color.white;
            ApplyHelp(launchBAToggle, "When this setting is enabled, a Browser Assist launch button will be displayed embeded into the VAM 'Top Menu' UI. The 'Top Menu' is the UI which is displayed when VAM first starts. It can also be accessed by pressing the star icon on the tool bar in the main VAM UI.");

            ApplyHelp(launchButtonPlacementPopup, "The launch button placement setting determines where the main BrowserAssist launch button is placed on the VAM UI toolbar. Options allow it to be placed over the VAM Scene Browser button (effectively replacing it) or to the left or right of the toolbar.");

            ApplyHelp(defaultResourceTypePopup, "This setting determines which Resource Type (e.g. Scene or Appearance preset) is selected by default when Browser Assist first starts.");

            ApplyHelp(resourceNameSearchModePopup, "The 'Resource Filter Name Search Mode' setting determines the behaviour of the resource name search filter in Browser Assist. The 'Starts With' option will filter to resources where the resource name starts with the search text. The two 'Contains' option will filter to resource names that contain the search text. 'Contains (Full Path)' will search for any matching text in the entire path name including the package name and folders.");

            ApplyHelp(defaultTargetPersonType, "This setting determines the default person gender when selecting a resource that has a person as it's target (e.g. Appearance Presets will select either a male or female Atom where available in the current scene). The setting can be overridden by setting the 'male' or 'female' tag on individual resources.");

            ApplyHelp(defaultCreatorFilterSortOrder, "This setting determines the default order which VAR Creators are listed in the Creator filter dropdown. This can either be alphabetical or by the quantity  of resources within VARs packaged by that Creator. The setting can be overriden per Resource in the Creators tab of each Resource Setting.");

            ApplyHelp(enableBulkDeleteToggle, "This setting determines whether a button is available on the BrowserAssist Action tab to delete multiple resources (and potentially their containing VARs) when many resources are selected using the multi-select feature. It can potentially delete a large number of files in a few clicks with no easy method to recover the files. As such, the feature should be used with caution.");

            ApplyHelp(enableBulkDeleteToggle, "This setting determines whether a button is available on the BrowserAssist Action tab to delete multiple resources (and potentially their containing VARs) when many resources are selected using the multi-select feature. It can potentially delete a large number of files in a few clicks with no easy method to recover the files. As such, the feature should be used with caution.");

            ApplyHelp(enableCSLReferencedPluginsToggle, "By default, when browsing plugins (only available to JayJayWon Patreons) any .cs file plugins that are referenced from a .cslist are not displayed. When this setting is enabled, then all .cs files will be displayed as an available plugin to load.");

            ApplyHelp(rescanVARsButton, "Performs a rescan of all VARs in the AddonPackages folder and identifies any new or removed resources inside the VAR pacakages. This may take some time, but only needs to be performed if any VAR packages have been unpacked, modified (i.e. resources added or removed) and re-packed.");

            ApplyHelp(disableMetaPurgeToggle, "When activated, this setting will disable the automated purging of meta data (e.g. Tags) that are associated with resources in VARs that are no longer present under the AddonPackages folder.");

            ApplyHelp(enableQuickFilterToggle, "When active, enables a panel in the main Browser UI that allows Quick Filter buttons to be set. This is allows the rapid switching between favorite filter settings. This feature is only avilable to JayJayWon Patreons.");



            RegisterCustomUIElement(2, helpTextCustomUI);

            helpTextBox.backgroundColor = new Color(0.4f, 0.5f, 0.2f, 0.4f);


        }
    }
    

    public class BackupSettingsRowUI
    {
        private GameObject rowGO;
        private GlobalUITab parentUI;

        private GameObject backupNameTextBoxGO;

        private UIDynamicButton restoreBackupButton;
        private UIDynamicButton deleteBackupButton;
        private UIDynamicButton cancelButton;
        private UIDynamicButton confirmRestoreBackupButton;
        private UIDynamicButton confirmDeleteBackupButton;

        private string backupName;

        private const int restoreConfirmMode = 1;
        private const int deleteConfirmMode = 2;
        private const int normalMode = 3;

        private int currentUIMode = normalMode;

        private string backupFolder { get
            {
                return Const.MetaDataBackupFolder + "\\" + backupName;
            } }

        public BackupSettingsRowUI(GameObject parentGO, string backupName, GlobalUITab parentUI)
        {
            rowGO = UI.AddChildGOImage(parentGO, new Color(0f, 0f, 0f, 0f), AnchorPresets.hStretchTop, 1030f, 60f, Vector2.zero);

            this.backupName = backupName;
            this.parentUI = parentUI;

            CreateUI();
        }

        public void ClearUI()
        {
            ClearUIContents();
            GameObject.Destroy(rowGO);
        }

        private void ClearUIContents()
        {
            if (backupNameTextBoxGO != null) GameObject.Destroy(backupNameTextBoxGO);
            if (restoreBackupButton != null) GameObject.Destroy(restoreBackupButton.gameObject);

            if (deleteBackupButton != null) GameObject.Destroy(deleteBackupButton.gameObject);
            if (cancelButton != null) GameObject.Destroy(cancelButton.gameObject);
            if (confirmRestoreBackupButton != null) GameObject.Destroy(confirmRestoreBackupButton.gameObject);
            if (confirmDeleteBackupButton != null) GameObject.Destroy(confirmDeleteBackupButton.gameObject);
        }

        private void RefreshUI()
        {
            ClearUIContents();
            CreateUI();
        }

        private void CreateUI()
        {
            bool invalidBackup = backupName.EndsWith("FAILED") || backupName.EndsWith("CORRUPTED");

            if (!FileManagerSecure.DirectoryExists(backupFolder + "\\LocalResourcesUserData")) invalidBackup = true;
            if (!FileManagerSecure.DirectoryExists(backupFolder + "\\VARGroupCoreData")) invalidBackup = true;
            if (!FileManagerSecure.DirectoryExists(backupFolder + "\\VARGroupUserData")) invalidBackup = true;
            if (!FileManagerSecure.DirectoryExists(backupFolder + "\\VARResourcesCoreData")) invalidBackup = true;
            if (!FileManagerSecure.DirectoryExists(backupFolder + "\\VARResourcesUserData")) invalidBackup = true;

            Color? backgroundColor = null;
            if (invalidBackup) backgroundColor = Color.red;

            backupNameTextBoxGO = UI.CreateBasicTextBox(rowGO, backupName, 700f, 60f, 30, 0f, 0f, AnchorPresets.middleLeft, true, 18, HorizontalWrapMode.Overflow, rescaleText: false, backgoundColor:backgroundColor);
            parentUI.ApplyHelp(backupNameTextBoxGO, "This is the name of the Backup that was previously taken. The format of the backup name gives the date and time that the backup was created. Backup data sets are stored in the VAM folder 'VAM\\Saves\\PluginData\\JayJayWon\\BrowserAssist\\MetaDataBackups'");

            if (currentUIMode == normalMode)
            {
                if (!invalidBackup)
                {
                    restoreBackupButton = UI.CreateUIButton(rowGO, 150f, 60f, "Restore...", 30, 705f, 0f, AnchorPresets.middleLeft, rescaleText: false);
                    restoreBackupButton.button.onClick.AddListener(delegate
                    {
                        currentUIMode = restoreConfirmMode;
                        RefreshUI();
                    });
                    parentUI.ApplyHelp(restoreBackupButton, "Restore the backup '"+backupName+"'. This will replace any current meta data (such as Tags) with the backed up data. The Restore process will also restart BrowserAssist which may take a noticeable period of time, during which VAM will become unresponsive. Its advisable to create a Backup of the curent meta data before performing a Restore.");


                }
                deleteBackupButton = UI.CreateUIButton(rowGO, 150f, 60f, "Delete...", 30, 860f, 0f, AnchorPresets.middleLeft, rescaleText: false);
                deleteBackupButton.button.onClick.AddListener(delegate
                {
                    currentUIMode = deleteConfirmMode;
                    RefreshUI();
                });

                parentUI.ApplyHelp(deleteBackupButton, "Delete the Backup data set '"+backupName+"'. This will permanently delete the backed up data set.");
            }
            else if (currentUIMode == restoreConfirmMode)
            {
                cancelButton = UI.CreateUIButton(rowGO, 150f, 60f, "Cancel", 30, 705f, 0f, AnchorPresets.middleLeft, rescaleText: false);
                cancelButton.button.onClick.AddListener(delegate
                {
                    currentUIMode = normalMode;
                    RefreshUI();
                });
                cancelButton.buttonColor = Color.green;

                parentUI.ApplyHelp(cancelButton, "Cancel the Restore");

                confirmRestoreBackupButton = UI.CreateUIButton(rowGO, 150f, 60f, "Confirm", 30, 860f, 0f, AnchorPresets.middleLeft, rescaleText: false);
                confirmRestoreBackupButton.button.onClick.AddListener(delegate
                {
                    RestoreBackup();
                    currentUIMode = normalMode;
                    RefreshUI();
                });
                confirmRestoreBackupButton.buttonColor = Color.red;
                confirmRestoreBackupButton.textColor = Color.white;
                parentUI.ApplyHelp(confirmRestoreBackupButton, "Confirm the Restore of the backup '"+backupName+"'. This will replace any current meta data (such as Tags) with the backed up data. The Restore process will also restart BrowserAssist which may take a noticeable period of time, during which VAM will become unresponsive. Its advisable to create a Backup before performing a Restore.");
            }
            else if (currentUIMode == deleteConfirmMode)
            {
                confirmDeleteBackupButton = UI.CreateUIButton(rowGO, 150f, 60f, "Confirm", 30, 705f, 0f, AnchorPresets.middleLeft, rescaleText: false);
                confirmDeleteBackupButton.button.onClick.AddListener(delegate
                {
                    DeleteBackup();
                    currentUIMode = normalMode;
                    parentUI.RefreshGroupUI(2);
                });
                confirmDeleteBackupButton.buttonColor = Color.red;
                confirmDeleteBackupButton.textColor = Color.white;
                parentUI.ApplyHelp(confirmDeleteBackupButton, "Confirm the Backup data set '"+backupName+"' deletion. This will permanently delete the backed up data set.");

                cancelButton = UI.CreateUIButton(rowGO, 150f, 60f, "Cancel", 30, 860f, 0f, AnchorPresets.middleLeft, rescaleText: false);
                cancelButton.button.onClick.AddListener(delegate
                {
                    currentUIMode = normalMode;
                    RefreshUI();
                });
                cancelButton.buttonColor = Color.green;
                parentUI.ApplyHelp(cancelButton, "Cancel the backup deletion.");
            }


        }

        private void DeleteBackup()
        {
            Utils.DeleteFolder(backupFolder);
        }

        private void RestoreBackup()
        {
            try
            {
#if VAM_GT_1_20_77_13
                DeleteAllMetaDataFiles();

                Utils.CopyFolders(backupFolder + "\\LocalResourcesUserData", Const.LocalResourcesUserDataFolder, Const.LocalResourceDataFileName);

                Utils.CopyFolders(backupFolder + "\\VARGroupCoreData", Const.VARGroupCoreDataFilesFolder, Const.VARGroupDataFileName);

                Utils.CopyFolders(backupFolder + "\\VARGroupUserData", Const.VARGroupUserDataFilesFolder, Const.VARGroupDataFileName);

                Utils.CopyFolders(backupFolder + "\\VARResourcesCoreData", Const.VARResourcesCoreDataFilesFolder, Const.VARResourceDataFileName);

                Utils.CopyFolders(backupFolder + "\\VARResourcesUserData", Const.VARResourcesUserDataFilesFolder, Const.VARResourceDataFileName);

                Settings.restoreNameJSS.val = backupName;
                Settings.SaveSettings();

                var pluginUID = Globals.mvrScript.storeId.Substring(0, Globals.mvrScript.storeId.IndexOf("_"));

                Globals.mvrScript.manager.ReloadPluginWithUID(pluginUID);
#else
                SuperController.LogMessage("Restoring BrowserAssist backups is not supported in versions of VAM before 1.20.77.13.");
#endif

                
            }
            catch(Exception ex)
            {
                SuperController.LogError("BA.BackupSettingsRowUI.RestoreBackup: Exception occured attempting to restore a backup. BrowserAssist is likely to be in a non stable state. Recommendation is to contact JayJayWon to advise on how to recover. "+ex);
                SuperController.LogMessage("WARNING: BrowserAssist backup restore has failed. The restore may be incomplete and data may not have been fully restored, even if no future errors occur. Its advisable that you report the issue to JayJayWon on Discord and provide any messages in the Error Log window. Please capture the Error Log now as the issue may not be re-producable and the exact full Error Log may be critical to assist diagnosis of the issue.");
            }
            

        }

        public static void DeleteAllMetaDataFiles()
        {          
            foreach (var file in FileManagerSecure.GetFiles(Const.LocalResourcesUserDataFolder))
            {
                FileManagerSecure.DeleteFile(file);
            }

            foreach (var file in FileManagerSecure.GetFiles(Const.VARGroupCoreDataFilesFolder))
            {
                FileManagerSecure.DeleteFile(file);
            }

            foreach (var file in FileManagerSecure.GetFiles(Const.VARGroupUserDataFilesFolder))
            {
                FileManagerSecure.DeleteFile(file);
            }

            foreach (var file in FileManagerSecure.GetFiles(Const.VARResourcesCoreDataFilesFolder))
            {
                FileManagerSecure.DeleteFile(file);
            }

            foreach (var file in FileManagerSecure.GetFiles(Const.VARResourcesUserDataFilesFolder))
            {
                FileManagerSecure.DeleteFile(file);
            }
        }

    }

    public class ResourceTypeSettingsRowUI
    {
        private GameObject rowGO;
        private GlobalUITab parentUI;

        private int resourceType;
        private string resourceTypeName;
        private string atomPresetResourceType;

        private GameObject resourceTypeTextBoxGO;

        private UIDynamicToggle resourceConcealToggle;

        private JSONStorableBool concealedJSB;

        private bool disableConcealResourceTypes;

        public ResourceTypeSettingsRowUI(GameObject parentGO, int resourceType, string atomPresetResourceType, GlobalUITab parentUI, bool disableConcealResourceTypes)
        {
            rowGO = UI.AddChildGOImage(parentGO, new Color(0f,0f,0f,0f), AnchorPresets.hStretchTop, 1030f, 60f, Vector2.zero);

            this.resourceType = resourceType;
            this.atomPresetResourceType = atomPresetResourceType;
            this.parentUI = parentUI;
            this.disableConcealResourceTypes = disableConcealResourceTypes;

            concealedJSB = new JSONStorableBool("concealed", Settings.IsResourceConcealed(resourceType,atomPresetResourceType), ConcealCallback);

            if (resourceType != ResourceType.presetAtom) resourceTypeName = EnumManifest.GetDisplayVal(ResourceType.enumManifestName, resourceType);
            else resourceTypeName = "Preset: " + atomPresetResourceType;

            CreateUI();
        }

        private void ConcealCallback(bool value)
        {
            List<int> resourceTypes = EnumManifest.GetEnumsList(ResourceType.enumManifestName);
            if (value)
            {
                Settings.SetConcealed(resourceType, atomPresetResourceType);  

                if (Settings.concealedResourceTypeCount >= resourceTypes.Count - 2 || !PC.patreonContentEnabled) parentUI.RefreshResourceTypeSettingsList();
            }
            else
            {
                Settings.UnsetConcealed(resourceType, atomPresetResourceType);
                if (Settings.concealedResourceTypeCount >= resourceTypes.Count - 3 || !PC.patreonContentEnabled) parentUI.RefreshResourceTypeSettingsList();
            }
            Settings.SaveSettings();

            BrowserAssist.mainBrowserUI.browserFiltersUI.RePopulateResourceTypeSelector();
            BAPluginUI.RePopulateResourceTypeSelector();
            Settings.RePopulateResourceTypeSelector();
        }

        public void ClearUI()
        {
            if (resourceTypeTextBoxGO != null) GameObject.Destroy(resourceTypeTextBoxGO);
            if (resourceConcealToggle != null) GameObject.Destroy(resourceConcealToggle.gameObject);
            GameObject.Destroy(rowGO);
        }

        private void RefreshUI()
        {
            if (resourceTypeTextBoxGO != null) GameObject.Destroy(resourceTypeTextBoxGO);
            if (resourceConcealToggle != null) GameObject.Destroy(resourceConcealToggle.gameObject);

            CreateUI();
        }

        private void CreateUI()
        {

            resourceTypeTextBoxGO = UI.CreateBasicTextBox(rowGO, resourceTypeName, 540f, 60f, 30, 0f, 0f, AnchorPresets.middleLeft, true, 18, HorizontalWrapMode.Overflow, rescaleText: false);
            parentUI.ApplyHelp(resourceTypeTextBoxGO, "This is the name of the Resource Type being configured.");

            resourceConcealToggle = UI.CreateToggle(rowGO, concealedJSB, 460f, 60f, "Conceal Resource Type", 26, 550f, 0f, AnchorPresets.middleLeft, 40f, rescaleText: false);

            if (disableConcealResourceTypes && !concealedJSB.val) resourceConcealToggle.toggle.interactable= false;

            if (resourceConcealToggle.toggle.interactable) 
            parentUI.ApplyHelp(resourceConcealToggle, "Conceal this Resource Type from the list of available filters. Once activated, the Browser Filter will not offer resources of the specified type as an option within BrowserAssist.");
            else parentUI.ApplyHelp(resourceConcealToggle, "Concealing all core Resource Types is not supported. Reveal a different core resource type if you want to conceal this one. A core resource type is defined as anything other than the Non-person Atom Presets.");


        }


    }



    public class DesktopUITab : BaseUITab
    {
        private CustomRowGridUIElement desktopSettingsTabMenuOptions;
        private CustomUIElement desktopSettingsCustomUI;
        private CustomUIElement helpTextCustomUI;

        private int currentDesktopSettingsTab = DesktopSettingsUITabs.general;
        public DesktopUITab()
        {
            desktopSettingsTabMenuOptions = new CustomRowGridUIElement();
            uiGroupCount = 2;

            desktopSettingsCustomUI = new CustomUIElement(825);
            helpTextCustomUI = new CustomUIElement(200);
            defaultHelpText = "The Desktop Settings tab contain settings that impact behavior of the Browser Assist UI when in Desktop mode (as opposed to VR mode)";
            helpTextJSS.val = defaultHelpText;


        }

        private void DesktopSettingsTabSelected(int tabRef)
        {
            if (tabRef != currentDesktopSettingsTab)
            {
                currentDesktopSettingsTab = tabRef;
                RefreshGroupUI(2);
            }
        }

        private void CreateUIGeneralSettings()
        {
            var uiScaleSlider = desktopSettingsCustomUI.AddSlider(Settings.uiScaleJSF, 1020, 120, 0, 0);
            uiScaleSlider.valueFormat = "F2";
            uiScaleSlider.ConfigureQuickButtons(-0.01f, -0.1f, -0.25f, -0.5f, 0.01f, 0.1f, 0.25f, 0.5f);


            var maxButtonsSlider = desktopSettingsCustomUI.AddSlider(Settings.maxButtonsPerPageDesktopJSF, 1020, 120, 0, -130);
            maxButtonsSlider.labelText.text = "Maximum number of Resources per Browser page";
            maxButtonsSlider.valueFormat = "F0";
            maxButtonsSlider.ConfigureQuickButtons(-10f, -50f, -100f, -1000f, 10f, 50f, 100f, 1000f);

            helpTextCustomUI.InitUI(CustomRowWidth.full);

            var helpTextBox = helpTextCustomUI.AddTextBox(helpTextJSS, 1050f, 200f, 0f, 0f);
            helpTextBox.textColor = Color.white;
            helpTextBox.backgroundColor = new Color(0.4f, 0.5f, 0.2f, 0.4f);


            ApplyHelp(uiScaleSlider, "The 'Desktop UI Scale' setting determines the scale of Browser Assist UI when in Desktop mode. Higher scales numbers make the UI larger and lower scale makes the UI smaller. ");

            ApplyHelp(maxButtonsSlider, "This setting determines the maximum number of resources icons that will be displayed on a single page of browser results, whilst in Desktop mode.");


            RegisterCustomUIElement(2, helpTextCustomUI);
        }

        private void CreateUINewAtomPositionSettings()
        {

            var xPosSlider = desktopSettingsCustomUI.AddSlider(Settings.desktopDefaultNewAtomPosRot.positionXJSF, 500, 120, 0, 0);
            xPosSlider.valueFormat = "F3";
            xPosSlider.label = "New Atom: Default Position X";

            var yPosSlider = desktopSettingsCustomUI.AddSlider(Settings.desktopDefaultNewAtomPosRot.positionYJSF, 500, 120, 0, -120);
            yPosSlider.valueFormat = "F3";
            yPosSlider.label = "New Atom: Default Position Y";

            var zPosSlider = desktopSettingsCustomUI.AddSlider(Settings.desktopDefaultNewAtomPosRot.positionZJSF, 500, 120, 0, -240);
            zPosSlider.valueFormat = "F3";
            zPosSlider.label = "New Atom: Default Position Z";

            var xRotSlider = desktopSettingsCustomUI.AddSlider(Settings.desktopDefaultNewAtomPosRot.rotationXJSF, 500, 120, 510, 0);
            xRotSlider.valueFormat = "F3";
            xRotSlider.label = "New Atom: Default Rotation X";

            var yRotSlider = desktopSettingsCustomUI.AddSlider(Settings.desktopDefaultNewAtomPosRot.rotationYJSF, 500, 120, 510, -120);
            yRotSlider.valueFormat = "F3";
            yRotSlider.label = "New Atom: Default Rotation Y";

            var zRotSlider = desktopSettingsCustomUI.AddSlider(Settings.desktopDefaultNewAtomPosRot.rotationZJSF, 500, 120, 510, -240);
            zRotSlider.valueFormat = "F3";
            zRotSlider.label = "New Atom: Default Rotation Z";

            var xPosAbsRelPopup = desktopSettingsCustomUI.AddPopup(Settings.desktopDefaultNewAtomPosRot.posXRelAbsJSE, null, 500, 60, 0, -360, 140, 60, "X Pos: Relative or Absolute", 0.8f);

            var yPosAbsRelPopup = desktopSettingsCustomUI.AddPopup(Settings.desktopDefaultNewAtomPosRot.posYRelAbsJSE, null, 500, 60, 0, -420, 140, 60, "Y Pos: Relative or Absolute", 0.8f);

            var zPosAbsRelPopup = desktopSettingsCustomUI.AddPopup(Settings.desktopDefaultNewAtomPosRot.posZRelAbsJSE, null, 500, 60, 0, -480, 140, 60, "Z Pos: Relative or Absolute", 0.8f);

            var xRotAbsRelPopup = desktopSettingsCustomUI.AddPopup(Settings.desktopDefaultNewAtomPosRot.rotXRelAbsJSE, null, 500, 60, 510, -360, 140, 60, "X Rot: Relative or Absolute", 0.8f);

            var yRotAbsRelPopup = desktopSettingsCustomUI.AddPopup(Settings.desktopDefaultNewAtomPosRot.rotYRelAbsJSE, null, 500, 60, 510, -420, 140, 60, "Y Rot: Relative or Absolute", 0.8f);

            var zRotAbsRelPopup = desktopSettingsCustomUI.AddPopup(Settings.desktopDefaultNewAtomPosRot.rotZRelAbsJSE, null, 500, 60, 510, -480, 140, 60, "Z Rot: Relative or Absolute", 0.8f);


            helpTextCustomUI.InitUI(CustomRowWidth.full);

            var helpTextBox = helpTextCustomUI.AddTextBox(helpTextJSS, 1050f, 200f, 0f, 0f);
            helpTextBox.textColor = Color.white;
            helpTextBox.backgroundColor = new Color(0.4f, 0.5f, 0.2f, 0.4f);


            ApplyHelp(xPosSlider, "The 'New Atom Default Position X' setting determines the default X coordinate for new atoms created by Browser Assist. The position can be absolute or relative to the desktop view point. If the position is relative, the X coordinate will offset the atom position to the left or right relative to the direction the user is facing.");

            ApplyHelp(yPosSlider, "The 'New Atom Default Position Y' setting determines the default Y coordinate for new atoms created by Browser Assist. The position can be absolute or relative to the desktop view point. If the position is relative, the Y coordinate will offset the atom position to the up or down relative to the direction the user is facing.");

            ApplyHelp(zPosSlider, "The 'New Atom Default Position Z' setting determines the default Z coordinate for new atoms created by Browser Assist. The position can be absolute or relative to the desktop view point. If the position is relative, the Z coordinate will offset the atom position to the further away or closer relative to the users view point.");

            ApplyHelp(xRotSlider, "The 'New Atom Default Rotation X' setting determines the default X rotation (Pitch) for new atoms created by Browser Assist. The rotation can be absolute or relative to the desktop view point. If the rotation is relative, the X rotation will offset the atom's pitch relative to the direction the user is facing.");

            ApplyHelp(yRotSlider, "The 'New Atom Default Rotation Y' setting determines the default Y rotation (Yaw) for new atoms created by Browser Assist. The rotation can be absolute or relative to the desktop view point. If the rotation is relative, the Y rotation will offset the atom's yaw relative to the direction the user is facing.");

            ApplyHelp(zRotSlider, "The 'New Atom Default Rotation Z' setting determines the default Z rotation (Roll) for new atoms created by Browser Assist. The rotation can be absolute or relative to the desktop view point. If the rotation is relative, the Z rotation will offset the atom's roll relative to the direction the user is facing.");

            ApplyHelp(xPosAbsRelPopup, "The 'X Pos: Relative or Absolute' setting determines whether default X position for new atoms (created by Browser Assist) is absolute or relative to the desktop view point.");

            ApplyHelp(yPosAbsRelPopup, "The 'Y Pos: Relative or Absolute' setting determines whether default Y position for new atoms (created by Browser Assist) is absolute or relative to the desktop view point.");

            ApplyHelp(zPosAbsRelPopup, "The 'Z Pos: Relative or Absolute' setting determines whether default Z position for new atoms (created by Browser Assist) is absolute or relative to the desktop view point.");

            ApplyHelp(xRotAbsRelPopup, "The 'X Rot: Relative or Absolute' setting determines whether default X rotation (Pitch) for new atoms (created by Browser Assist) is absolute or relative to the desktop view point.");

            ApplyHelp(yRotAbsRelPopup, "The 'Y Rot: Relative or Absolute' setting determines whether default Y rotation (Yaw) for new atoms (created by Browser Assist) is absolute or relative to the desktop view point.");

            ApplyHelp(zRotAbsRelPopup, "The 'Z Rot: Relative or Absolute' setting determines whether default Z rotation (Roll) for new atoms (created by Browser Assist) is absolute or relative to the desktop view point.");

            RegisterCustomUIElement(2, helpTextCustomUI);
        }

        protected override void CreateGroupUI(int groupID)
        {
            if (groupID == 1)
            {
                desktopSettingsTabMenuOptions.InitUI(CustomRowWidth.full, 4);
                desktopSettingsTabMenuOptions.AddButtonToRow("General", GlobalSettingsUITabs.general, DesktopSettingsTabSelected);
                desktopSettingsTabMenuOptions.AddButtonToRow("New Atom Position", GlobalSettingsUITabs.resources, DesktopSettingsTabSelected);

                RegisterCustomUIElement(groupID, desktopSettingsTabMenuOptions);

                desktopSettingsTabMenuOptions.SetButtonColorActive(currentDesktopSettingsTab);
            }
            if (groupID == 2)
            {
                desktopSettingsCustomUI.InitUI(CustomRowWidth.full);
                RegisterCustomUIElement(2, desktopSettingsCustomUI);

                if (currentDesktopSettingsTab == DesktopSettingsUITabs.general) CreateUIGeneralSettings();

                if (currentDesktopSettingsTab == DesktopSettingsUITabs.newAtomPosition) CreateUINewAtomPositionSettings();
            }

        }
    }

    public class VRSettingsUITab : BaseUITab
    {
        private CustomRowGridUIElement desktopSettingsTabMenuOptions;
        private CustomUIElement vrSettingsCustomUI;
        private CustomUIElement helpTextCustomUI;

        private int currentDesktopSettingsTab = DesktopSettingsUITabs.general;
        public VRSettingsUITab()
        {
            desktopSettingsTabMenuOptions = new CustomRowGridUIElement();
            uiGroupCount = 2;

            vrSettingsCustomUI = new CustomUIElement(825);
            helpTextCustomUI = new CustomUIElement(200);
            defaultHelpText = "The VR Settings tab contain settings that impact behavior of the Browser Assist UI when in VR mode (as opposed to Desktop mode)";
            helpTextJSS.val = defaultHelpText;
        }

        private void DesktopSettingsTabSelected(int tabRef)
        {
            if (tabRef != currentDesktopSettingsTab)
            {
                currentDesktopSettingsTab = tabRef;
                RefreshGroupUI(2);
            }
        }

        private void CreateUIGeneralSettings()
        {
            var dockVAMUIToggle = vrSettingsCustomUI.AddToggle(Settings.dockToVAMUIJSB, 500, 60, 0, 0);
            dockVAMUIToggle.label = "Dock BrowserAssist to VAM UI";

            var maxButtonsSlider = vrSettingsCustomUI.AddSlider(Settings.maxButtonsPerPageVRJSF, 1020, 120, 0, -70);
            maxButtonsSlider.labelText.text = "Maximum number of Resources per Browser page";
            maxButtonsSlider.valueFormat = "F0";
            maxButtonsSlider.ConfigureQuickButtons(-10f, -50f, -100f, -1000f, 10f, 50f, 100f, 1000f);

            helpTextCustomUI.InitUI(CustomRowWidth.full);

            var helpTextBox = helpTextCustomUI.AddTextBox(helpTextJSS, 1050f, 200f, 0f, 0f);
            helpTextBox.textColor = Color.white;
            helpTextBox.backgroundColor = new Color(0.4f, 0.5f, 0.2f, 0.4f);

            ApplyHelp(dockVAMUIToggle, "This setting determines if Browser Assist window is docked to VAM Main UI when in VR. If enabled, the Browser Assist window will behave simlarly to other settings available in VAM UI and can be repostioned using the VR controller B button.");

            ApplyHelp(maxButtonsSlider, "This setting determines the maximum number of resources icons that will be displayed on a single page of browser results, whilst in VR mode.");

            RegisterCustomUIElement(2, helpTextCustomUI);
        }

        private void CreateUINewAtomPositionSettings()
        {
            var xPosSlider = vrSettingsCustomUI.AddSlider(Settings.vrDefaultNewAtomPosRot.positionXJSF, 500, 120, 0, 0);
            xPosSlider.valueFormat = "F3";
            xPosSlider.label = "New Atom: Default Position X";

            var yPosSlider = vrSettingsCustomUI.AddSlider(Settings.vrDefaultNewAtomPosRot.positionYJSF, 500, 120, 0, -120);
            yPosSlider.valueFormat = "F3";
            yPosSlider.label = "New Atom: Default Position Y";

            var zPosSlider = vrSettingsCustomUI.AddSlider(Settings.vrDefaultNewAtomPosRot.positionZJSF, 500, 120, 0, -240);
            zPosSlider.valueFormat = "F3";
            zPosSlider.label = "New Atom: Default Position Z";

            var xRotSlider = vrSettingsCustomUI.AddSlider(Settings.vrDefaultNewAtomPosRot.rotationXJSF, 500, 120, 510, 0);
            xRotSlider.valueFormat = "F3";
            xRotSlider.label = "New Atom: Default Rotation X";

            var yRotSlider = vrSettingsCustomUI.AddSlider(Settings.vrDefaultNewAtomPosRot.rotationYJSF, 500, 120, 510, -120);
            yRotSlider.valueFormat = "F3";
            yRotSlider.label = "New Atom: Default Rotation Y";

            var zRotSlider = vrSettingsCustomUI.AddSlider(Settings.vrDefaultNewAtomPosRot.rotationZJSF, 500, 120, 510, -240);
            zRotSlider.valueFormat = "F3";
            zRotSlider.label = "New Atom: Default Rotation Z";

            var xPosAbsRelPopup = vrSettingsCustomUI.AddPopup(Settings.vrDefaultNewAtomPosRot.posXRelAbsJSE, null, 500, 60, 0, -360, 140, 60, "X Pos: Relative or Absolute", 0.8f);

            var yPosAbsRelPopup = vrSettingsCustomUI.AddPopup(Settings.vrDefaultNewAtomPosRot.posYRelAbsJSE, null, 500, 60, 0, -420, 140, 60, "Y Pos: Relative or Absolute", 0.8f);

            var zPosAbsRelPopup = vrSettingsCustomUI.AddPopup(Settings.vrDefaultNewAtomPosRot.posZRelAbsJSE, null, 500, 60, 0, -480, 140, 60, "Z Pos: Relative or Absolute", 0.8f);

            var xRotAbsRelPopup = vrSettingsCustomUI.AddPopup(Settings.vrDefaultNewAtomPosRot.rotXRelAbsJSE, null, 500, 60, 510, -360, 140, 60, "X Rot: Relative or Absolute", 0.8f);

            var yRotAbsRelPopup = vrSettingsCustomUI.AddPopup(Settings.vrDefaultNewAtomPosRot.rotYRelAbsJSE, null, 500, 60, 510, -420, 140, 60, "Y Rot: Relative or Absolute", 0.8f);

            var zRotAbsRelPopup = vrSettingsCustomUI.AddPopup(Settings.vrDefaultNewAtomPosRot.rotZRelAbsJSE, null, 500, 60, 510, -480, 140, 60, "Z Rot: Relative or Absolute", 0.8f);

            var relAtomPopup = vrSettingsCustomUI.AddPopup(Settings.vrDefaultNewAtomPosRot.relativeAtomNameJSME, null, 500, 60, 510, -550, 200, 60, "Relative to:", 0.6f);

            helpTextCustomUI.InitUI(CustomRowWidth.full);

            var helpTextBox = helpTextCustomUI.AddTextBox(helpTextJSS, 1050f, 200f, 0f, 0f);
            helpTextBox.textColor = Color.white;
            helpTextBox.backgroundColor = new Color(0.4f, 0.5f, 0.2f, 0.4f);

            ApplyHelp(xPosSlider, "The 'New Atom Default Position X' setting determines the default X coordinate for new atoms created by Browser Assist. The position can be absolute or relative to the VR headset or VR controllers. If the position is relative, the X coordinate will offset the atom position to the left or right relative to the direction the user is facing.");

            ApplyHelp(yPosSlider, "The 'New Atom Default Position Y' setting determines the default Y coordinate for new atoms created by Browser Assist. The position can be absolute or relative to the VR headset or VR controllers. If the position is relative, the Y coordinate will offset the atom position to the up or down relative to the direction the user is facing.");

            ApplyHelp(zPosSlider, "The 'New Atom Default Position Z' setting determines the default Z coordinate for new atoms created by Browser Assist. The position can be absolute or relative to the VR headset or VR controllers. If the position is relative, the Z coordinate will offset the atom position to the further away or closer relative to the users view point.");

            ApplyHelp(xRotSlider, "The 'New Atom Default Rotation X' setting determines the default X rotation (Pitch) for new atoms created by Browser Assist. The rotation can be absolute or relative to the VR headset or VR controllers. If the rotation is relative, the X rotation will offset the atom's pitch relative to the direction the user is facing.");

            ApplyHelp(yRotSlider, "The 'New Atom Default Rotation Y' setting determines the default Y rotation (Yaw) for new atoms created by Browser Assist. The rotation can be absolute or relative to the VR headset or VR controllers. If the rotation is relative, the Y rotation will offset the atom's yaw relative to the direction the user is facing.");

            ApplyHelp(zRotSlider, "The 'New Atom Default Rotation Z' setting determines the default Z rotation (Roll) for new atoms created by Browser Assist. The rotation can be absolute or relative to the VR headset or VR controllers. If the rotation is relative, the Z rotation will offset the atom's roll relative to the direction the user is facing.");

            ApplyHelp(xPosAbsRelPopup, "The 'X Pos: Relative or Absolute' setting determines whether default X position for new atoms (created by Browser Assist) is absolute or relative to the VR headset or VR controllers.");

            ApplyHelp(yPosAbsRelPopup, "The 'Y Pos: Relative or Absolute' setting determines whether default Y position for new atoms (created by Browser Assist) is absolute or relative to the VR headset or VR controllers.");

            ApplyHelp(zPosAbsRelPopup, "The 'Z Pos: Relative or Absolute' setting determines whether default Z position for new atoms (created by Browser Assist) is absolute or relative to the VR headset or VR controllers.");

            ApplyHelp(xRotAbsRelPopup, "The 'X Rot: Relative or Absolute' setting determines whether default X rotation (Pitch) for new atoms (created by Browser Assist) is absolute or relative to the VR headset or VR controllers.");

            ApplyHelp(yRotAbsRelPopup, "The 'Y Rot: Relative or Absolute' setting determines whether default Y rotation (Yaw) for new atoms (created by Browser Assist) is absolute or relative to the VR headset or VR controllers.");

            ApplyHelp(zRotAbsRelPopup, "The 'Z Rot: Relative or Absolute' setting determines whether default Z rotation (Roll) for new atoms (created by Browser Assist) is absolute or relative to the VR headset or VR controllers.");

            ApplyHelp(relAtomPopup, "If any of the atoms positions or rotations are set as 'Relative', then this setting determines whether that position or rotation is relative to the VR Headset, VR Left Controller or the VR right Controller.");

            RegisterCustomUIElement(2, helpTextCustomUI);

        }
        protected override void CreateGroupUI(int groupID)
        {
            if (groupID == 1)
            {
                desktopSettingsTabMenuOptions.InitUI(CustomRowWidth.full, 4);
                desktopSettingsTabMenuOptions.AddButtonToRow("General", GlobalSettingsUITabs.general, DesktopSettingsTabSelected);
                desktopSettingsTabMenuOptions.AddButtonToRow("New Atom Position", GlobalSettingsUITabs.resources, DesktopSettingsTabSelected);

                RegisterCustomUIElement(groupID, desktopSettingsTabMenuOptions);

                desktopSettingsTabMenuOptions.SetButtonColorActive(currentDesktopSettingsTab);
            }
            if (groupID == 2)
            {
                vrSettingsCustomUI.InitUI(CustomRowWidth.full);
                RegisterCustomUIElement(2, vrSettingsCustomUI);

                if (currentDesktopSettingsTab == DesktopSettingsUITabs.general) CreateUIGeneralSettings();

                if (currentDesktopSettingsTab == DesktopSettingsUITabs.newAtomPosition) CreateUINewAtomPositionSettings();
            }

            

            
        }
    }


}
