﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using UnityEngine.UI;
using static TabbedUIBuilder;



namespace JayJayWon
{
    public class ResourceSettingsUITab : BaseUITab
    {
        private CustomRowGridUIElement resourceSettingsTabMenuOptions;
        private CustomUIElement helpTextCustomUI;

        public int resourceType = ResourceType.scene;
        public string atomPresetResourceType = "";

        private CustomUIElement resourceSettingsCustomUI;

        private GameObject combiLogicListContentGO;
        private GameObject tagSettingsListContentGO;
        private GameObject creatorSettingsListContentGO;
        private GameObject combiLogicListTagSelectionContentGO;
        private GameObject combiLogicTagSelectionBoxGO;
        private UIDynamicTextField combiLogicTagFilterTextBox;

        private HashSet<TagSettingsRowUI> tagSettingRows = new HashSet<TagSettingsRowUI>();
        private HashSet<CreatorSettingsRowUI> creatorSettingRows = new HashSet<CreatorSettingsRowUI>();

        private HashSet<CombiTagConditionRowUI> combiTagConditionRows = new HashSet<CombiTagConditionRowUI>();
        private HashSet<GameObject> combiTagSelectionRows = new HashSet<GameObject>();
        private GameObject addCombiRowGO;
        private CombiTag draftCombiTag;
        private UIDynamicButton draftCombiTagApplyButton;
        private UIDynamicButton draftCombiTagCancelButton;
        private UIDynamicPopup combiTagSelectPopup;
        private UIDynamicButton combiTagNewButton;
        private UIDynamicButton combiTagRenameButton;
        private UIDynamicButton combiTagDeleteButton;

        private UIDynamicButton combiTagFoldersLeftScrollButton;
        private UIDynamicButton combiTagFoldersRightScrollButton;
        private UIDynamicTextField combiTagFoldersCountTextBox;

        private UIDynamicButton combiTagDeleteConfirmButton;
        private UIDynamicButton combiTagDeleteCancelButton;
        private UIDynamicTextField tagNameEditTextBox;
        private UIDynamicTextField tagNameWarningTextBox;
        private JSONStorableString newTagNameJSS;
        private JSONStorableString newTagNameWarningJSS;

        private int combiTagMode = CombTagUIMode.normal;

        private JSONStorableString tagSettingsFilterJSS = new JSONStorableString("tagSettingsFilter", "");
        private JSONStorableString creatorSettingsFilterJSS = new JSONStorableString("creatorSettingsFilter", "");

        private JSONStorableString combiTagFilterJSS = new JSONStorableString("combiTagFilter", "");

        private ResourceSettings resourceSettings { get { return Settings.GetResourceSettings(resourceType,atomPresetResourceType); } }

        private int currentResourceSelectionSettingsTab = ResourceSettingsUITabs.defaultFilters;

        private JSONStorableStringChooser combiTagsJSSC;

        private JSONStorableStringChooser yearJSSC;
        private JSONStorableStringChooser monthJSSC ;
        private JSONStorableStringChooser dayJSSC;

        public CombiTagConditionRowUI selectedCombiTagsRow { get; private set; } = null;

        public ResourceSettingsUITab()
        {
            resourceSettingsTabMenuOptions = new CustomRowGridUIElement();
            uiGroupCount = 2;

            combiTagsJSSC = new JSONStorableStringChooser("combiTags", null, "", "Select Combi-Tag Logic",CombiTagSelectCallback);
            newTagNameJSS = new JSONStorableString("newCombiTagName", "");
            newTagNameWarningJSS = new JSONStorableString("newCombiTagNameWarning", "");

            yearJSSC = new JSONStorableStringChooser("year", null, DateTime.Now.Year.ToString(), "Year:", delegate (string value)
            {
                var clc = selectedCombiTagsRow.clc;
                int year = int.Parse(value);
                clc.dateYear = year;
                dayJSSC.valNoCallback = clc.dateDay.ToString();
                UpdateDayChoices();
                RefreshCombiTagDraftStatusButtons();
                selectedCombiTagsRow.UpdateTagListText();

            });
            monthJSSC = new JSONStorableStringChooser("month", null, DateTime.Now.Month.ToString(), "Month:", delegate (string value)
            {
                var clc = selectedCombiTagsRow.clc;
                int month = int.Parse(value);
                clc.dateMonth = month;
                dayJSSC.valNoCallback = clc.dateDay.ToString();
                UpdateDayChoices();
                RefreshCombiTagDraftStatusButtons();
                selectedCombiTagsRow.UpdateTagListText();
            });
            dayJSSC = new JSONStorableStringChooser("day", null, DateTime.Now.Day.ToString(), "Day:", delegate (string value)
            {
                var clc = selectedCombiTagsRow.clc;
                int day = int.Parse(value);
                clc.dateDay = day;
                RefreshCombiTagDraftStatusButtons();
                selectedCombiTagsRow.UpdateTagListText();
            });

            resourceSettingsCustomUI = new CustomUIElement(820);
            helpTextCustomUI = new CustomUIElement(205);
        }

        protected override void CreateGroupUI(int groupID)
        {
            if (groupID == 1)
            {
                resourceSettingsTabMenuOptions.InitUI(CustomRowWidth.full, 4);
                resourceSettingsTabMenuOptions.AddButtonToRow("Default Filters", ResourceSettingsUITabs.defaultFilters, ResourceSettingsTabSelected);
                resourceSettingsTabMenuOptions.AddButtonToRow("Tags", ResourceSettingsUITabs.tags, ResourceSettingsTabSelected);
                resourceSettingsTabMenuOptions.AddButtonToRow("Combi-Tags", ResourceSettingsUITabs.combiTags, ResourceSettingsTabSelected);
                resourceSettingsTabMenuOptions.AddButtonToRow("Creators", ResourceSettingsUITabs.creators, ResourceSettingsTabSelected);
                RegisterCustomUIElement(groupID, resourceSettingsTabMenuOptions);

                resourceSettingsTabMenuOptions.SetButtonColorActive(currentResourceSelectionSettingsTab);
            }
            if (groupID == 2)
            {
                var resourceTypeDisplayName = EnumManifest.GetDisplayVal(ResourceType.enumManifestName, resourceSettings.defaultFilterSettings.resourceType);
                resourceSettingsCustomUI.InitUI(CustomRowWidth.full);
                RegisterCustomUIElement(groupID, resourceSettingsCustomUI);

                if (PC.patreonContentEnabled || resourceType == ResourceType.scene || resourceType == ResourceType.presetAppearance || resourceType== ResourceType.maleClothingItems || resourceType == ResourceType.femaleClothingItems)
                {
                    if (currentResourceSelectionSettingsTab == ResourceSettingsUITabs.defaultFilters) CreateUIDefaultFilterSettings(resourceTypeDisplayName);
                    else if (currentResourceSelectionSettingsTab == ResourceSettingsUITabs.tags) CreateUITagSettings(resourceTypeDisplayName);
                    else if (currentResourceSelectionSettingsTab == ResourceSettingsUITabs.combiTags) CreateUICombiTagSettings(resourceTypeDisplayName);
                    else if (currentResourceSelectionSettingsTab == ResourceSettingsUITabs.creators) CreateUICreatorSettings(resourceTypeDisplayName);


                }
                else
                {
                    resourceSettingsCustomUI.AddTextBox("JayJayWon Patreon subscribers have access to the full BrowserAssist plugin which allows for browsing of additional resource types, including " + resourceTypeDisplayName, 1000, 200, 0, 10);
                }
            }
        }

        private void CreateUICombiTagSettings(string resourceTypeDisplayName)
        {
            var combiTags = TagManifest.GetTags(resourceType, atomPresetResourceType, TagCategory.combi);
            defaultHelpText = "This tab allows 'Combi-Tags' to be created (and modified). Combi-Tags apply themselves to resources based on a combination of other tags, allowing automatic tagging of resources based on user specified criteria. Combi-tags can reference other combi-tags allowing nested logic criteria to be specified. Note that setting up looping criteria is blocked (i.e. you can't have CombiTag1 referencing ComiTag2 referencing CombiTag3 referencing CombiTag1).";
            helpTextJSS.val = defaultHelpText;

            helpTextCustomUI.InitUI(CustomRowWidth.full);
            var helpTextBox = helpTextCustomUI.AddTextBox(helpTextJSS, 1050f, 200f, 0f, 0f);
            helpTextBox.textColor = Color.white;
            helpTextBox.backgroundColor = new Color(0.4f, 0.5f, 0.2f, 0.4f);

            RegisterCustomUIElement(2, helpTextCustomUI);

            if (combiTags.Count > 0)
            {
                List<string> combiTagNames = new List<string>(combiTags.Select(x => x.tagName).ToList());

                combiTagsJSSC.choices = combiTagNames;
                if (!combiTagNames.Contains(combiTagsJSSC.val)) combiTagsJSSC.val = combiTagNames[0];

                if (combiTagMode != CombTagUIMode.rename)
                {
                    combiTagSelectPopup = resourceSettingsCustomUI.AddPopup(combiTagsJSSC, null, 600f, 50f, 0f, 0f, 400f, 50f, null, 0.6f);
                    ApplyHelp(combiTagSelectPopup, "Select an existing Combi-tag to view or modify it's configuration.");
                }
                if (combiTagMode == CombTagUIMode.deleteConfirm)
                {
                    combiTagDeleteConfirmButton = resourceSettingsCustomUI.AddButton(220f, 50f,600f, 0f, "Confirm Delete", ConfirmDeleteCombiTagCallback);
                    ApplyHelp(combiTagDeleteConfirmButton, "Confirm the permanent deletion of this Combi-tag");
                    combiTagDeleteConfirmButton.buttonColor = Color.red;
                    combiTagDeleteConfirmButton.textColor = Color.white;

                    combiTagDeleteCancelButton = resourceSettingsCustomUI.AddButton(210f, 50f, 850f, 0f, "Cancel Delete", CancelDeleteCombiTagCallback);
                    ApplyHelp(combiTagDeleteCancelButton, "Cancel the deletion of this combi-tag.");
                    combiTagDeleteCancelButton.buttonColor = Color.green;

                }
                else if (combiTagMode == CombTagUIMode.rename)
                {
                    newTagNameJSS.val = combiTagsJSSC.val;
                    tagNameEditTextBox = resourceSettingsCustomUI.AddTextEntryBox(newTagNameJSS, 400f, 60f, 30, 0f, 0f, TextAnchor.MiddleLeft, InputField.LineType.SingleLine, "", 0f, 0f, NewCombiTagNameChangeCallback, NewCombiTagNameCallback);
                    var inputComponent = tagNameEditTextBox.gameObject.GetComponent<InputField>();
                    inputComponent.ActivateInputField();
                    newTagNameWarningJSS.val = "";
                    tagNameWarningTextBox = resourceSettingsCustomUI.AddTextBox(newTagNameWarningJSS, 620f, 60f, 400f, 0f, TextAnchor.MiddleLeft);
                    ApplyHelp(tagNameEditTextBox, "Enter a new name for this combi-tag. Note that existing tag names can not be used and tag names can not end in a space character.");
                }
                else if (combiTagMode == CombTagUIMode.normal)
                {
                    combiTagRenameButton = resourceSettingsCustomUI.AddButton(150f, 50f, 610f, 0f, "Rename...", RenameCombiTagCallback);
                    ApplyHelp(combiTagRenameButton, "Rename the currently selected Combi-Tag.");
                    combiTagDeleteButton = resourceSettingsCustomUI.AddButton(150f, 50f, 760f, 0f, "Delete...", DeleteCombiTagCallback);
                    combiTagDeleteButton.buttonColor = Color.red;
                    combiTagDeleteButton.textColor = Color.white;
                    ApplyHelp(combiTagDeleteButton, "Delete the currently selected Combi-Tag.");
                    var combiLogicBoxGO = resourceSettingsCustomUI.CreateVScrollableContent(AnchorPresets.topLeft, 715f, 650f, new Vector2(60f, -105f), 20f, 3f);

                    combiLogicTagFilterTextBox = resourceSettingsCustomUI.AddTextEntryBox(combiTagFilterJSS, 275f, 50f, 28, 780f, -55f, TextAnchor.MiddleLeft, InputField.LineType.SingleLine, " Filter list...", 0f, 0f, CombiTagFilterNameChangeCallback);
                    combiLogicListContentGO = combiLogicBoxGO.transform.Find("scrollableContentViewportGO/vScrollableContentGO").gameObject;

                    combiLogicTagSelectionBoxGO = resourceSettingsCustomUI.CreateVScrollableContent(AnchorPresets.topLeft, 275f, 650f, new Vector2(780f, -105f), 20f, 3f);
                    combiLogicListTagSelectionContentGO = combiLogicTagSelectionBoxGO.transform.Find("scrollableContentViewportGO/vScrollableContentGO").gameObject;
                    

                    var textBox = resourceSettingsCustomUI.AddTextBox("    Tag resouces with this Combi-Tag....", 600f, 45f, 0f, -55f);
                    textBox.backgroundColor = new Color(0f, 0f, 0f, 0f);
                    textBox.textColor = Color.white;

                    if (draftCombiTag == null) CreateDraftCombiTag(combiTagsJSSC.val);


                    draftCombiTagApplyButton = resourceSettingsCustomUI.AddButton(150f, 50f, 610f, -770f, "Apply", ApplyCombiCallback);
                    ApplyHelp(draftCombiTagApplyButton, "Apply the modfications made to the currently selected Combi-Tag. Once applied, all resources (of the '"+resourceTypeDisplayName+"' type) will be updated to reflect the new combi tag configuration.");
                    draftCombiTagApplyButton.buttonColor = Color.green;
                    draftCombiTagCancelButton = resourceSettingsCustomUI.AddButton(150f, 50f, 450f, -770f, "Cancel", CancelCombiCallback);
                    draftCombiTagCancelButton.buttonColor = Color.red;
                    draftCombiTagCancelButton.textColor = Color.white;
                    ApplyHelp(draftCombiTagCancelButton, "Cancel the modfications made to the currently selected Combi-Tag.");
                    RefreshCombiTagConditionsList();

                    combiTagFoldersLeftScrollButton = resourceSettingsCustomUI.AddButton(50f, 50f, 775f, -760f, "<", delegate
                    {
                        if (folderPage > 1)
                        {
                            folderPage--;
                            RefreshCombiTagSelectionUI();
                        }
                    });
                    ApplyHelp(combiTagFoldersLeftScrollButton, "Display the previous page of folders for the current Combi Tag Condition");

                    combiTagFoldersRightScrollButton = resourceSettingsCustomUI.AddButton(50f, 50f, 1000f, -760f, ">", delegate
                    {
                        folderPage++;
                        RefreshCombiTagSelectionUI();
                    });
                    ApplyHelp(combiTagFoldersRightScrollButton, "Display the next page of folders for the current Combi Tag Condition");

                    combiTagFoldersCountTextBox = resourceSettingsCustomUI.AddTextBox(folderCountJSS, 175f, 50f, 825f, -760f, TextAnchor.MiddleLeft);
                    ApplyHelp(combiTagFoldersCountTextBox, "Displays the currently selected page of folders. If a large number of folders are available then you can use the filter box to reduce the number of folders available for selection.");
                    RefreshCombiTagSelectionUI();
                }

            }
            else
            {
                if (combiTagFoldersLeftScrollButton != null) combiTagFoldersLeftScrollButton.gameObject.SetActive(false);
                if (combiTagFoldersRightScrollButton != null) combiTagFoldersRightScrollButton.gameObject.SetActive(false);
                if (combiTagFoldersCountTextBox != null) combiTagFoldersCountTextBox.gameObject.SetActive(false);
            }

            if (combiTagMode == CombTagUIMode.normal)
            {
                combiTagNewButton = resourceSettingsCustomUI.AddButton(150f, 50f, 910f, 0f, "New", AddCombiTagCallback);
                ApplyHelp(combiTagNewButton, "Create a new Combi-Tag.");
                combiTagNewButton.buttonColor = Color.green;
            }
            
            RefreshCombiTagDraftStatusButtons();       
        }

        private void NewCombiTagNameCallback(string newTagName)
        {

            Tag tag = TagManifest.GetTag(resourceType, atomPresetResourceType, combiTagsJSSC.val);
            bool success = TagManifest.RenameTag(tag, newTagName);

            combiTagMode = CombTagUIMode.normal;
            if (success)
            {
                Settings.SaveSettings();
                combiTagsJSSC.valNoCallback = newTagName;
                CombiTagSelectCallback(newTagName);
                
            }
            else RefreshGroupUI(2);
            
        }

        private void NewCombiTagNameChangeCallback(string tagName)
        {
            Tag tag = TagManifest.GetTag(resourceType, atomPresetResourceType, combiTagsJSSC.val);
            string tagNameMod = TagManifest.NormalizeTagName(tagName);
            newTagNameJSS.inputField.text = tagNameMod;
            if (tagNameMod == "") newTagNameWarningJSS.val = "Blank tag names are not permitted";
            else if (tagNameMod != tag.tagName && TagManifest.GetTagNameCount(resourceType, atomPresetResourceType, tagNameMod) > 0) newTagNameWarningJSS.val = "Tag name already exists";
            else if (tagNameMod == "hide") newTagNameWarningJSS.val = "Tag name 'hide' is not permitted";
            else if (tagNameMod == "favorite") newTagNameWarningJSS.val = "Tag name 'favorite' is not permitted";
            else if (tagNameMod.EndsWith(" ")) newTagNameWarningJSS.val = "Tag names can not end with a space character";
            else newTagNameWarningJSS.val = "";
        }

        public void RefreshCombiTagSelectionUI()
        {
            if (selectedCombiTagsRow != null && selectedCombiTagsRow.clc.isMultiConfigurableCondition )
            {
                combiLogicTagFilterTextBox.gameObject.SetActive(selectedCombiTagsRow.clc.isMultiConfigurationFilterable);
                combiLogicTagSelectionBoxGO.SetActive(true);
                combiLogicListTagSelectionContentGO.SetActive(true);

                foreach (GameObject go in combiTagSelectionRows)
                {
                    GameObject.Destroy(go);
                }
                combiTagSelectionRows.Clear();

                if (selectedCombiTagsRow.clc.isTaggableCondition) RefreshCombiTagSelectionList();
                else if (selectedCombiTagsRow.clc.isLicenseCondition) RefreshLicenseSelectionList();
                else if (selectedCombiTagsRow.clc.isCreatorCondition) RefreshCreatorSelectionList();
                else if (selectedCombiTagsRow.clc.isVARFolderCondition) RefreshVARFolderSelectionList();
                else if (selectedCombiTagsRow.clc.isLocalFolderCondition) RefreshLocalFolderSelectionList();
                else if (selectedCombiTagsRow.clc.isDateCondition) RefreshDateSelection();

                if (selectedCombiTagsRow.clc.isLocalFolderCondition || selectedCombiTagsRow.clc.isVARFolderCondition)
                {
                    combiTagFoldersLeftScrollButton.gameObject.SetActive(true);
                    combiTagFoldersLeftScrollButton.button.interactable = folderPage > 1;
                    combiTagFoldersRightScrollButton.gameObject.SetActive(true);
                    combiTagFoldersRightScrollButton.button.interactable = folderPage < maxFolderPage;
                    combiTagFoldersCountTextBox.gameObject.SetActive(true);
                }
                else
                {
                    if (combiTagFoldersLeftScrollButton != null) combiTagFoldersLeftScrollButton.gameObject.SetActive(false);
                    if (combiTagFoldersRightScrollButton != null) combiTagFoldersRightScrollButton.gameObject.SetActive(false);
                    if (combiTagFoldersCountTextBox != null) combiTagFoldersCountTextBox.gameObject.SetActive(false);
                }
            }
            else
            {
                combiLogicTagFilterTextBox.gameObject.SetActive(false);
                combiLogicTagSelectionBoxGO.SetActive(false);
                combiLogicListTagSelectionContentGO.SetActive(false);
                if (combiTagFoldersLeftScrollButton!=null) combiTagFoldersLeftScrollButton.gameObject.SetActive(false);
                if (combiTagFoldersRightScrollButton != null) combiTagFoldersRightScrollButton.gameObject.SetActive(false);
                if (combiTagFoldersCountTextBox != null) combiTagFoldersCountTextBox.gameObject.SetActive(false);
            }
        }

        private UIDynamicButton lastTagsButtonActivate = null;

        private void RefreshCreatorSelectionList()
        {
            var creatorNames = CreatorManifest.GetAllCreatorNames(resourceType, atomPresetResourceType);
            var missingCreatorNames = selectedCombiTagsRow.clc.varCreators.Except(creatorNames);
            creatorNames.AddRange(missingCreatorNames);
            creatorNames.Sort();

            foreach (string creatorName in creatorNames)
            {
                if (combiTagFilterJSS.inputField.text != "" && !creatorName.ToLower().Contains(combiTagFilterJSS.inputField.text)) continue;

                 var rowGO = UI.AddChildGOImage(combiLogicListTagSelectionContentGO, new Color(0, 0, 0, 0), AnchorPresets.hStretchTop, 255f, 60f, Vector2.zero);
                combiTagSelectionRows.Add(rowGO);

                bool selected = selectedCombiTagsRow.clc.varCreators.Contains(creatorName);

                var toggle = UI.CreateToggle(rowGO, new JSONStorableBool("", selected), 255f, 60f, creatorName, 30, anchorPreset: AnchorPresets.middleLeft, rescaleText: false, checkBoxSize: 40f);
                if (missingCreatorNames.Contains(creatorName))
                {
                    toggle.backgroundColor = Color.red;
                    toggle.textColor = Color.white;
                }
                else if (selected) toggle.backgroundColor = Color.cyan;
                else toggle.backgroundColor = BrowserUI.lightGrey;

                if (missingCreatorNames.Contains(creatorName)) ApplyHelp(toggle, "There are no currently available resources that are packaged by the Creator '"+creatorName+"'");
                else ApplyHelp(toggle, "Select this Package Creator for the '" + selectedCombiTagsRow.clc.logicConditionJSEnum.displayVal + "' condition.");

                toggle.toggle.onValueChanged.AddListener(delegate
                {
                    bool selected2 = selectedCombiTagsRow.clc.varCreators.Contains(creatorName);

                    if (!selected2)
                    {
                        selectedCombiTagsRow.clc.UseVARCreator(creatorName);
                        RefreshCombiTagDraftStatusButtons();
                        toggle.backgroundColor = Color.cyan;
                    }
                    else
                    {
                        selectedCombiTagsRow.clc.RemoveUseVARCreator(creatorName);
                        RefreshCombiTagDraftStatusButtons();
                        toggle.backgroundColor = BrowserUI.lightGrey;
                    }
                    selectedCombiTagsRow.UpdateTagListText();
                });
            }
        }

        private HashSet<string> cachedAddonPackagesFolderNames= null;
        private void RefreshVARFolderSelectionList()
        {
            if (cachedAddonPackagesFolderNames == null)
            {
                string fileExt = "";
                if (selectedCombiTagsRow.clc.logicConditionJSEnum.val == CombiLogicConditionOptions.varInFolder || selectedCombiTagsRow.clc.logicConditionJSEnum.val == CombiLogicConditionOptions.varNotInFolder) fileExt = ".var";
                cachedAddonPackagesFolderNames = Utils.GetLocalFolderNamesRecursive("AddonPackages", fileExt);
            }

            var unFilteredFolderNames = new List<string>(cachedAddonPackagesFolderNames);
            unFilteredFolderNames.Sort();

            var filteredFolderNames = unFilteredFolderNames;

            if (combiTagFilterJSS.inputField.text != "") filteredFolderNames = unFilteredFolderNames.Where(folderName => folderName.ToLower().Contains(combiTagFilterJSS.inputField.text)).ToList();

            currentFilteredFolderCount = filteredFolderNames.Count;
            maxFolderPage = (currentFilteredFolderCount / folderPerPage) + 1;
            if (folderPage > maxFolderPage) folderPage = maxFolderPage;

            int startIndex = (folderPage - 1) * folderPerPage;
            var folderNames = filteredFolderNames;

            if (filteredFolderNames.Count > 0) folderNames = filteredFolderNames.GetRange(startIndex, Mathf.Min(filteredFolderNames.Count - startIndex, folderPerPage));

            folderCountJSS.val = folderPage.ToString() + " of " + maxFolderPage.ToString();

            foreach (string folderName in folderNames)
            {
                if (combiTagFilterJSS.inputField.text != "" && !folderName.ToLower().Contains(combiTagFilterJSS.inputField.text)) continue;

                var rowGO = UI.AddChildGOImage(combiLogicListTagSelectionContentGO, new Color(0, 0, 0, 0), AnchorPresets.hStretchTop, 255f, 120f, Vector2.zero);
                combiTagSelectionRows.Add(rowGO);

                bool selected = selectedCombiTagsRow.clc.varFolders.Contains(folderName);

                var folderDisplayName = folderName;
                if (folderName != "AddonPackages")
                {
                    folderDisplayName = folderName.Substring(13);
                    if (folderDisplayName.Length>30) folderDisplayName = "..."+folderDisplayName.Substring(folderDisplayName.Length-30);
                }

                if (selectedCombiTagsRow.clc.logicConditionJSEnum.val == CombiLogicConditionOptions.varUnderFolder || selectedCombiTagsRow.clc.logicConditionJSEnum.val == CombiLogicConditionOptions.varNotUnderFolder)
                {
                    var button = UI.CreateUIButton(rowGO, 255f, 120f, folderDisplayName, 30, anchorPreset: AnchorPresets.middleLeft, rescaleText: false);
                    ApplyHelp(button, "Select this AddonPackages folder '" + folderName + "' (and all of it's sub-folders) for the '" + selectedCombiTagsRow.clc.logicConditionJSEnum.displayVal + "' condition.");

                    if (selected)
                    {
                        button.buttonColor = Color.cyan;
                        lastTagsButtonActivate = button;
                    }
                    else button.buttonColor = BrowserUI.lightGrey;
                    button.button.onClick.AddListener(delegate
                    {
                        bool selected2 = selectedCombiTagsRow.clc.varFolders.Contains(folderName);
                        if (!selected2)
                        {
                            if (lastTagsButtonActivate != null) lastTagsButtonActivate.buttonColor = BrowserUI.lightGrey;
                            selectedCombiTagsRow.clc.UseVARFolder(folderName,true);
                            RefreshCombiTagDraftStatusButtons();
                            button.buttonColor = Color.cyan;
                            lastTagsButtonActivate = button;
                        }
                        else
                        {
                            selectedCombiTagsRow.clc.RemoveUseVARFolder(folderName);
                            RefreshCombiTagDraftStatusButtons();
                            button.buttonColor = BrowserUI.lightGrey;
                        }
                        selectedCombiTagsRow.UpdateTagListText();
                    });
                }
                else
                {
                    var toggle = UI.CreateToggle(rowGO, new JSONStorableBool("", selected), 255f, 120f, folderDisplayName, 30, anchorPreset: AnchorPresets.middleLeft, rescaleText: false, checkBoxSize: 40f);

                    if (selected) toggle.backgroundColor = Color.cyan;
                    else toggle.backgroundColor = BrowserUI.lightGrey;

                    ApplyHelp(toggle, "Select the AddonPackages sub-folders '" + folderName + "' for the '" + selectedCombiTagsRow.clc.logicConditionJSEnum.displayVal + "' condition.");

                    toggle.toggle.onValueChanged.AddListener(delegate
                    {
                        bool selected2 = selectedCombiTagsRow.clc.varFolders.Contains(folderName);

                        if (!selected2)
                        {
                            selectedCombiTagsRow.clc.UseVARFolder(folderName);
                            RefreshCombiTagDraftStatusButtons();
                            toggle.backgroundColor = Color.cyan;
                        }
                        else
                        {
                            selectedCombiTagsRow.clc.RemoveUseVARFolder(folderName);
                            RefreshCombiTagDraftStatusButtons();
                            toggle.backgroundColor = BrowserUI.lightGrey;
                        }
                        selectedCombiTagsRow.UpdateTagListText();
                    });
                }

            }
        }

        private Dictionary<int,HashSet<string>> cachedLocalFolderNamesByResourceType = new Dictionary<int, HashSet<string>>();
        private Dictionary<string, HashSet<string>> cachedLocalFolderNamesByAtomPresetResource = new Dictionary<string, HashSet<string>>();

        private int folderPage = 1;
        private const int folderPerPage = 50;
        private int maxFolderPage = 1;
        private int currentFilteredFolderCount = 0;
        
        private JSONStorableString folderCountJSS = new JSONStorableString("folderCount", "");

        private HashSet<string> GetLocalFolderNames (int resourceType, string atomPresetResourceType)
        {
            HashSet<string> folderNames = new HashSet<string>();

            string fileExt = "";
            if (selectedCombiTagsRow.clc.logicConditionJSEnum.val == CombiLogicConditionOptions.inLocalFolder || selectedCombiTagsRow.clc.logicConditionJSEnum.val == CombiLogicConditionOptions.notInLocalFolder) fileExt = ResourceType.GetPrimaryFileExtension(resourceType);

            foreach (string rootFolder in Const.GetResourceRootFolderNames(resourceType, atomPresetResourceType))
            {
                folderNames.UnionWith(Utils.GetLocalFolderNamesRecursive(rootFolder, fileExt));
            }

            return folderNames;
        }

        private void RefreshLocalFolderSelectionList()
        {
            HashSet<string> localFolderNames;

            if (resourceType != ResourceType.presetAtom)
            {
                if (!cachedLocalFolderNamesByResourceType.ContainsKey(resourceType)) cachedLocalFolderNamesByResourceType.Add(resourceType, GetLocalFolderNames(resourceType,atomPresetResourceType));
                else if (cachedLocalFolderNamesByResourceType[resourceType] == null) cachedLocalFolderNamesByResourceType[resourceType] = GetLocalFolderNames(resourceType, atomPresetResourceType);
                localFolderNames = cachedLocalFolderNamesByResourceType[resourceType];
            }
            else
            {
                if (!cachedLocalFolderNamesByAtomPresetResource.ContainsKey(atomPresetResourceType)) cachedLocalFolderNamesByAtomPresetResource.Add(atomPresetResourceType, GetLocalFolderNames(resourceType, atomPresetResourceType));
                else if (cachedLocalFolderNamesByAtomPresetResource[atomPresetResourceType] == null) cachedLocalFolderNamesByAtomPresetResource[atomPresetResourceType] = GetLocalFolderNames(resourceType, atomPresetResourceType);
                localFolderNames = cachedLocalFolderNamesByAtomPresetResource[atomPresetResourceType];
            }

            var unFilteredFolderNames = new List<string>(localFolderNames);
            unFilteredFolderNames.Sort();

            var filteredFolderNames = unFilteredFolderNames;

            if (combiTagFilterJSS.inputField.text != "") filteredFolderNames = unFilteredFolderNames.Where(folderName => folderName.ToLower().Contains(combiTagFilterJSS.inputField.text)).ToList();

            currentFilteredFolderCount = filteredFolderNames.Count;
            maxFolderPage = (currentFilteredFolderCount / folderPerPage) + 1;
            if (folderPage>maxFolderPage) folderPage = maxFolderPage;

            int startIndex = (folderPage - 1) * folderPerPage;
            var folderNames = filteredFolderNames   ;

            if (filteredFolderNames.Count > 0) folderNames = filteredFolderNames.GetRange(startIndex, Mathf.Min(filteredFolderNames.Count- startIndex, folderPerPage));

            folderCountJSS.val = folderPage.ToString() + " of " + maxFolderPage.ToString() ;           

            foreach (string folderName in folderNames)
            {
                var rowGO = UI.AddChildGOImage(combiLogicListTagSelectionContentGO, new Color(0, 0, 0, 0), AnchorPresets.hStretchTop, 255f, 120f, Vector2.zero);
                combiTagSelectionRows.Add(rowGO);

                bool selected = selectedCombiTagsRow.clc.localFolders.Contains(folderName);

                var folderDisplayName = folderName;
                foreach (string rootFolder in Const.GetResourceRootFolderNames(resourceType, atomPresetResourceType))
                {
                    if (folderName.StartsWith(rootFolder) && folderName!=rootFolder) folderDisplayName=folderName.Substring(rootFolder.Length);
                    if (folderDisplayName.Length > 30) folderDisplayName = "..." + folderDisplayName.Substring(folderDisplayName.Length - 30);
                }             

                if (selectedCombiTagsRow.clc.logicConditionJSEnum.val == CombiLogicConditionOptions.underLocalFolder || selectedCombiTagsRow.clc.logicConditionJSEnum.val == CombiLogicConditionOptions.notUnderLocalFolder)
                {
                    var button = UI.CreateUIButton(rowGO, 255f, 120f, folderDisplayName, 30, anchorPreset: AnchorPresets.middleLeft, rescaleText: false);
                    ApplyHelp(button, "Select this folder '" + folderName + "' (and all of it's sub-folders) for the '" + selectedCombiTagsRow.clc.logicConditionJSEnum.displayVal + "' condition.");

                    if (selected)
                    {
                        button.buttonColor = Color.cyan;
                        lastTagsButtonActivate = button;
                    }
                    else button.buttonColor = BrowserUI.lightGrey;
                    button.button.onClick.AddListener(delegate
                    {
                        bool selected2 = selectedCombiTagsRow.clc.localFolders.Contains(folderName);
                        if (!selected2)
                        {
                            if (lastTagsButtonActivate != null) lastTagsButtonActivate.buttonColor = BrowserUI.lightGrey;
                            selectedCombiTagsRow.clc.UseLocalFolder(folderName, true);
                            RefreshCombiTagDraftStatusButtons();
                            button.buttonColor = Color.cyan;
                            lastTagsButtonActivate = button;
                        }
                        else
                        {
                            selectedCombiTagsRow.clc.RemoveUseLocalFolder(folderName);
                            RefreshCombiTagDraftStatusButtons();
                            button.buttonColor = BrowserUI.lightGrey;
                        }
                        selectedCombiTagsRow.UpdateTagListText();
                    });
                }
                else
                {
                    var toggle = UI.CreateToggle(rowGO, new JSONStorableBool("", selected), 255f, 120f, folderDisplayName, 30, anchorPreset: AnchorPresets.middleLeft, rescaleText: false, checkBoxSize: 40f);

                    if (selected) toggle.backgroundColor = Color.cyan;
                    else toggle.backgroundColor = BrowserUI.lightGrey;

                    ApplyHelp(toggle, "Select the folder '" + folderName + "' for the '" + selectedCombiTagsRow.clc.logicConditionJSEnum.displayVal + "' condition.");

                    toggle.toggle.onValueChanged.AddListener(delegate
                    {
                        bool selected2 = selectedCombiTagsRow.clc.localFolders.Contains(folderName);

                        if (!selected2)
                        {
                            selectedCombiTagsRow.clc.UseLocalFolder(folderName);
                            RefreshCombiTagDraftStatusButtons();
                            toggle.backgroundColor = Color.cyan;
                        }
                        else
                        {
                            selectedCombiTagsRow.clc.RemoveUseLocalFolder(folderName);
                            RefreshCombiTagDraftStatusButtons();
                            toggle.backgroundColor = BrowserUI.lightGrey;
                        }
                        selectedCombiTagsRow.UpdateTagListText();
                    });
                }

            }
        }

        private void UpdateDayChoices()
        {
            var clc = selectedCombiTagsRow.clc;
            int daysInMonth = DateTime.DaysInMonth(clc.dateYear,clc.dateMonth);
            List<string> dayChoices = new List<string>();
            for (int day = 1; day <= daysInMonth; day++)
            {
                dayChoices.Add(day.ToString());
            }
            dayJSSC.choices = dayChoices;
            if (!dayChoices.Contains(dayJSSC.val)) dayJSSC.val = dayChoices.Last();
        }

        protected IEnumerator ReparentPopup(UIDynamicPopup popupUID, GameObject parentGO)
        {
            yield return new WaitForEndOfFrame();

            if (popupUID.popup.visible)
            {
                popupUID.popup.popupPanel.transform.SetParent(parentGO.transform.parent.parent.parent.parent);
            }
        }

        private void RefreshDateSelection()
        {
            List<string> yearChoices = new List<string>();
            int currentYear = DateTime.Now.Year;
            for (int year = currentYear-10; year <= currentYear+10; year++)
            {
                yearChoices.Add(year.ToString());
            }
            yearJSSC.choices = yearChoices;
            monthJSSC.choices = new List<string>() { "1", "2", "3", "4","5","6","7","8","9","10","11","12" };
            monthJSSC.displayChoices = new List<string>() { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
            UpdateDayChoices();

            yearJSSC.valNoCallback = selectedCombiTagsRow.clc.dateYear.ToString();
            monthJSSC.valNoCallback = selectedCombiTagsRow.clc.dateMonth.ToString();
            dayJSSC.valNoCallback = selectedCombiTagsRow.clc.dateDay.ToString();

            var yearRowGO = UI.AddChildGOImage(combiLogicListTagSelectionContentGO, new Color(0, 0, 0, 0), AnchorPresets.hStretchTop, 255f, 120f, Vector2.zero);
            combiTagSelectionRows.Add(yearRowGO);
            var yearPopup = UI.CreateScrollablePopup(yearRowGO, yearJSSC, 255f, 120f,null,30,rescaleText:false,popupRowHeight:60f, disableSlider: true, popupFontSize:30, anchorPreset:AnchorPresets.middleLeft);
            yearPopup.popup.onOpenPopupHandlers += () =>
            {
                Globals.mvrScript.StartCoroutine(ReparentPopup(yearPopup, combiLogicListTagSelectionContentGO));
            };

            var monthRowGO = UI.AddChildGOImage(combiLogicListTagSelectionContentGO, new Color(0, 0, 0, 0), AnchorPresets.hStretchTop, 255f, 120f, Vector2.zero);
            combiTagSelectionRows.Add(monthRowGO);
            var monthPopup = UI.CreateScrollablePopup(monthRowGO, monthJSSC, 255f, 120f, null, 30, rescaleText: false, popupRowHeight: 60f, disableSlider: true, popupFontSize: 30, anchorPreset: AnchorPresets.middleLeft);
            monthPopup.popup.onOpenPopupHandlers += () =>
            {
                Globals.mvrScript.StartCoroutine(ReparentPopup(monthPopup, combiLogicListTagSelectionContentGO));
            };

            var dayRowGO = UI.AddChildGOImage(combiLogicListTagSelectionContentGO, new Color(0, 0, 0, 0), AnchorPresets.hStretchTop, 255f, 120f, Vector2.zero);
            combiTagSelectionRows.Add(dayRowGO);
            var dayPopup = UI.CreateScrollablePopup(dayRowGO, dayJSSC, 255f, 120f, null, 30, rescaleText: false, popupRowHeight: 60f,disableSlider:true, popupFontSize: 30, anchorPreset: AnchorPresets.middleLeft);
            dayPopup.popup.onOpenPopupHandlers += () =>
            {
                Globals.mvrScript.StartCoroutine(ReparentPopup(dayPopup, combiLogicListTagSelectionContentGO));
            };
        }

        private void RefreshLicenseSelectionList()
        {
            foreach (int licenseType in EnumManifest.GetEnumsList(LicenseType.enumManifestName))
            {

                if (licenseType == LicenseType.Any) continue;
                string licenseDisplayName = EnumManifest.GetDisplayVal(LicenseType.enumManifestName, licenseType);

                var rowGO = UI.AddChildGOImage(combiLogicListTagSelectionContentGO, new Color(0, 0, 0, 0), AnchorPresets.hStretchTop, 255f, 60f, Vector2.zero);
                combiTagSelectionRows.Add(rowGO);

                bool selected = selectedCombiTagsRow.clc.varLicenses.Contains(licenseType);

                var toggle = UI.CreateToggle(rowGO, new JSONStorableBool("", selected), 255f, 60f, licenseDisplayName, 30, anchorPreset: AnchorPresets.middleLeft, rescaleText: false, checkBoxSize: 40f);
                if (selected) toggle.backgroundColor = Color.cyan;
                else toggle.backgroundColor = BrowserUI.lightGrey;
                ApplyHelp(toggle, "Select this License Type for the '" + selectedCombiTagsRow.clc.logicConditionJSEnum.displayVal + "' condition.");

                toggle.toggle.onValueChanged.AddListener(delegate
                {
                    bool selected2 = selectedCombiTagsRow.clc.varLicenses.Contains(licenseType);

                    if (!selected2)
                    {
                        selectedCombiTagsRow.clc.UseLicense(licenseType);
                        RefreshCombiTagDraftStatusButtons();
                        toggle.backgroundColor = Color.cyan;
                    }
                    else
                    {
                        selectedCombiTagsRow.clc.RemoveUseLicense(licenseType);
                        RefreshCombiTagDraftStatusButtons();
                        toggle.backgroundColor = BrowserUI.lightGrey;
                    }
                    selectedCombiTagsRow.UpdateTagListText();
                });
            }
        }


        private void RefreshCombiTagSelectionList()
        {
            CombiTag currentCombiTag = TagManifest.GetTag(resourceType, atomPresetResourceType, combiTagsJSSC.val) as CombiTag;
            foreach (Tag tag in TagManifest.GetTags(resourceType, atomPresetResourceType))
            {
                if (selectedCombiTagsRow.clc.logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAnyBodyRegion && tag.tagCategory != TagCategory.bodyRegion) continue;

                if (selectedCombiTagsRow.clc.logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAnyClothingType && tag.tagCategory != TagCategory.clothingType) continue;

                if (selectedCombiTagsRow.clc.logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAnyUserTag && tag.tagCategory != TagCategory.userDefined && tag.tagCategory != TagCategory.combi) continue;


                if ((tag.tagCategory != TagCategory.combi || !currentCombiTag.usedByCombiTagsHieararchy.Contains(tag as CombiTag)) && tag != currentCombiTag)
                {
                    if (combiTagFilterJSS.inputField.text == "" || tag.tagName.Contains(combiTagFilterJSS.inputField.text))
                    {
                        var rowGO = UI.AddChildGOImage(combiLogicListTagSelectionContentGO, new Color(0, 0, 0, 0), AnchorPresets.hStretchTop, 255f, 60f, Vector2.zero);
                        combiTagSelectionRows.Add(rowGO);

                        bool selected = selectedCombiTagsRow.clc.usesTags.Contains(tag);

                        string tagName = tag.tagName;
                        if (tag.tagCategory == TagCategory.combi) tagName = "<i>" + tag.tagName + "</i>";

                        if (selectedCombiTagsRow.clc.logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWith)
                        {
                            var button = UI.CreateUIButton(rowGO, 255f, 60f, tagName, 30, anchorPreset: AnchorPresets.middleLeft, rescaleText: false);
                            ApplyHelp(button, "Select this tag for the '"+ selectedCombiTagsRow.clc.logicConditionJSEnum.displayVal+"' condition.");

                            if (selected)
                            {
                                button.buttonColor = Color.cyan;
                                lastTagsButtonActivate = button;
                            }
                            else button.buttonColor = BrowserUI.lightGrey;
                            button.button.onClick.AddListener(delegate
                            {
                                bool selected2 = selectedCombiTagsRow.clc.usesTags.Contains(tag);
                                if (!selected2)
                                {
                                    if (lastTagsButtonActivate != null) lastTagsButtonActivate.buttonColor = BrowserUI.lightGrey;
                                    selectedCombiTagsRow.clc.UseTag(tag, true);
                                    RefreshCombiTagDraftStatusButtons();
                                    button.buttonColor = Color.cyan;
                                    lastTagsButtonActivate = button;
                                }
                                else
                                {
                                    selectedCombiTagsRow.clc.RemoveUseTag(tag);
                                    RefreshCombiTagDraftStatusButtons();
                                    button.buttonColor = BrowserUI.lightGrey;
                                }
                                selectedCombiTagsRow.UpdateTagListText();
                            });
                        }
                        else
                        {

                            var toggle = UI.CreateToggle(rowGO, new JSONStorableBool("", selected), 255f, 60f, tagName, 30, anchorPreset: AnchorPresets.middleLeft, rescaleText: false, checkBoxSize: 40f);
                            if (selected) toggle.backgroundColor = Color.cyan;
                            else toggle.backgroundColor = BrowserUI.lightGrey;
                            ApplyHelp(toggle, "Select this tag for the '" + selectedCombiTagsRow.clc.logicConditionJSEnum.displayVal + "' condition.");

                            toggle.toggle.onValueChanged.AddListener(delegate
                            {
                                bool selected2 = selectedCombiTagsRow.clc.usesTags.Contains(tag);

                                if (!selected2)
                                {
                                    selectedCombiTagsRow.clc.UseTag(tag);
                                    RefreshCombiTagDraftStatusButtons();
                                    toggle.backgroundColor = Color.cyan;
                                }
                                else
                                {
                                    selectedCombiTagsRow.clc.RemoveUseTag(tag);
                                    RefreshCombiTagDraftStatusButtons();
                                    toggle.backgroundColor = BrowserUI.lightGrey;
                                }
                                selectedCombiTagsRow.UpdateTagListText();
                            });
                        }

                    }

                }

            }
        }

        public void RefreshCombiConditionsLogicOperand(CombiTagConditionRowUI clcUIexception)
        {
            foreach(var rowUI in combiTagConditionRows)
            {
                if (rowUI != clcUIexception)
                {
                    rowUI.RefreshUI();
                }
            }
        }

        public void RefreshCombiTagDraftStatusButtons()
        {
            if ((draftCombiTag!=null && draftCombiTag.isModfied && combiTagMode == CombTagUIMode.normal) || combiTagMode == CombTagUIMode.deleteConfirm)
            {
                if (draftCombiTagApplyButton!=null) draftCombiTagApplyButton.button.gameObject.SetActive(true);
                if (draftCombiTagCancelButton != null) draftCombiTagCancelButton.button.gameObject.SetActive(true);
                if (combiTagNewButton != null) combiTagNewButton.button.interactable = false;
                if (combiTagDeleteButton != null) combiTagDeleteButton.button.interactable = false;
                if (combiTagRenameButton != null) combiTagRenameButton.button.interactable = false;
                if (combiTagSelectPopup != null) combiTagSelectPopup.popup.topButton.interactable = false;
            }
            else
            {
                if (draftCombiTagApplyButton != null) draftCombiTagApplyButton.button.gameObject.SetActive(false);
                if (draftCombiTagCancelButton != null) draftCombiTagCancelButton.button.gameObject.SetActive(false);
                if (combiTagNewButton != null) combiTagNewButton.button.interactable = true;
                if (combiTagDeleteButton != null) combiTagDeleteButton.button.interactable = true;
                if (combiTagRenameButton != null) combiTagRenameButton.button.interactable = true;
                if (combiTagSelectPopup != null) combiTagSelectPopup.popup.topButton.interactable = true;
            }
        }

        private void AddCombiTagCallback()
        {
            var tagNames = TagManifest.GetTagNames(resourceType, atomPresetResourceType);

            int index = 1;
            while (tagNames.Contains("combi tag" + index.ToString())) index++;


            var newCombiTag = TagManifest.CreateCombiTag("combi tag" + index.ToString(), resourceType, atomPresetResourceType);
            Settings.SaveSettings();

            combiTagMode = CombTagUIMode.normal;
            combiTagsJSSC.val = newCombiTag.tagName;
        }

        private void DeleteCombiTagCallback()
        {
            combiTagMode = CombTagUIMode.deleteConfirm;
            RefreshGroupUI(2);
        }

        private void RenameCombiTagCallback()
        {
            combiTagMode = CombTagUIMode.rename;
            RefreshGroupUI(2);
        }

        private void ConfirmDeleteCombiTagCallback()
        {
            combiTagMode = CombTagUIMode.normal;
            TagManifest.DeleteTag(combiTagsJSSC.val, resourceType, atomPresetResourceType);
            Settings.SaveSettings();

            RefreshGroupUI(2);
        }

        private void CancelDeleteCombiTagCallback()
        {
            combiTagMode = CombTagUIMode.normal;
            RefreshGroupUI(2);
        }

        private void CreateUITagSettings(string resourceTypeDisplayName)
        {
            defaultHelpText = "The Tags tab allows '" + resourceTypeDisplayName + "' tags to be modified and deleted. Individual tags can be renamed, deleted or marked as a 'Hide' tag (making all tagged resource hidden). There is also an option to mark tags as 'Preferred'. Preferred tags will be listed at the top of the Tag filter selection drop down.";
            helpTextJSS.val = defaultHelpText;

            helpTextCustomUI.InitUI(CustomRowWidth.full);

            var helpTextBox = helpTextCustomUI.AddTextBox(helpTextJSS, 1050f, 200f, 0f, 0f);
            helpTextBox.textColor = Color.white;
            helpTextBox.backgroundColor = new Color(0.4f, 0.5f, 0.2f, 0.4f);

            tagSettingsFilterJSS.val = "";
            var tagSettingsFilterTextBox = resourceSettingsCustomUI.AddTextEntryBox(tagSettingsFilterJSS, 400f, 50f, 30, 0f, 5f, TextAnchor.MiddleLeft, InputField.LineType.SingleLine, "Enter text to filter Tags list...", 0f, 0f, TagFilterNameChangeCallback);
            ApplyHelp(tagSettingsFilterTextBox, "Enter a search string to filter the list of Tags and allow quicker selection from a large number of tags.");
            var tagSettingsBoxGO = resourceSettingsCustomUI.CreateVScrollableContent(AnchorPresets.topLeft, 1050f, 775f, new Vector2(0f, -50f), 20f, 3f);
            tagSettingsListContentGO = tagSettingsBoxGO.transform.Find("scrollableContentViewportGO/vScrollableContentGO").gameObject;
            RefreshTagSettingsList();

            RegisterCustomUIElement(2, helpTextCustomUI);
        }

        private void CreateUICreatorSettings(string resourceTypeDisplayName)
        {
            defaultHelpText = "The Creators tab allows the Creators of '" + resourceTypeDisplayName + "' resources to be designated as 'Hidden' (making all resources by that creator hidden) or 'Preferred' (making the Creator name appear at the top of filter lists).";
            helpTextJSS.val = defaultHelpText;

            helpTextCustomUI.InitUI(CustomRowWidth.full);

            var helpTextBox = helpTextCustomUI.AddTextBox(helpTextJSS, 1050f, 200f, 0f, 0f);
            helpTextBox.textColor = Color.white;
            helpTextBox.backgroundColor = new Color(0.4f, 0.5f, 0.2f, 0.4f);

            creatorSettingsFilterJSS.val = "";
            var creatorSettingsFilterTextBox = resourceSettingsCustomUI.AddTextEntryBox(creatorSettingsFilterJSS, 400f, 50f, 30, 0f, 5f, TextAnchor.MiddleLeft, InputField.LineType.SingleLine, "Enter text to filter Creators list...", 0f, 0f, CreatorFilterNameChangeCallback);
            ApplyHelp(creatorSettingsFilterTextBox, "Enter a search string to filter the list of Creators and allow quicker selection from a large number of Creators.");
            var creatorSettingsBoxGO = resourceSettingsCustomUI.CreateVScrollableContent(AnchorPresets.topLeft, 1050f, 720f, new Vector2(0f, -50f), 20f, 3f);
            creatorSettingsListContentGO = creatorSettingsBoxGO.transform.Find("scrollableContentViewportGO/vScrollableContentGO").gameObject;
            RefreshCreatorSettingsList();
            var creatorSettingsSortOrderPopup = resourceSettingsCustomUI.AddPopup(Settings.GetResourceSettings(resourceType, atomPresetResourceType).creatorFilterOrderJSEC, null, 1050f, 50f, 0f, -775f,labelProportion:0.7f);
            ApplyHelp(creatorSettingsSortOrderPopup, "Select the order that VAR Creators are listed in the Creator filter drop down. If 'default' is selected then the order will be determined by the setting under 'Global Settings > General'.");

            RegisterCustomUIElement(2, helpTextCustomUI);
        }

        private void CombiTagFilterNameChangeCallback(string tagName)
        {
            string tagNameMod = TagManifest.NormalizeTagName(tagName);

            combiTagFilterJSS.inputField.text = tagNameMod;
            folderPage = 1;
            RefreshCombiTagSelectionUI();
        }

        private void TagFilterNameChangeCallback(string tagName)
        {
            string tagNameMod = TagManifest.NormalizeTagName(tagName);

            tagSettingsFilterJSS.inputField.text = tagNameMod;
            RefreshTagSettingsList();
        }

        private void CreatorFilterNameChangeCallback(string creatorName)
        {
            RefreshCreatorSettingsList();
        }

        private void ClearTagSettingsList()
        {
            foreach (var tsr in tagSettingRows)
            {
                tsr.ClearUI();
            }
            tagSettingRows.Clear();
        }
        private void ClearCreatorSettingsList()
        {
            foreach (var csr in creatorSettingRows)
            {
                csr.ClearUI();
            }
            creatorSettingRows.Clear();
        }

        private void RefreshTagSettingsList()
        {
            ClearTagSettingsList();
            var tagsList = TagManifest.GetTags(resourceType, atomPresetResourceType);

            foreach (Tag tag in tagsList)
            {
                if (tagSettingsFilterJSS.inputField.text == "" || tag.tagName.Contains(tagSettingsFilterJSS.inputField.text))
                {
                    var tagSettingRow = new TagSettingsRowUI(tagSettingsListContentGO, tag,this);
                    tagSettingRows.Add(tagSettingRow);
                }

            }
        }

        private void RefreshCreatorSettingsList()
        {
            ClearCreatorSettingsList();
            var creators = CreatorManifest.GetAllCreators(resourceType, atomPresetResourceType);

            var creatorsList = creators.OrderBy(o =>o.creatorName);

            foreach (CreatorManifestEntry creator in creatorsList)
            {
                if (creatorSettingsFilterJSS.inputField.text == "" || creator.creatorName.ToLower().Contains(creatorSettingsFilterJSS.inputField.text.ToLower())) 
                {
                    var creatorSettingRow = new CreatorSettingsRowUI(creatorSettingsListContentGO, creator, this);
                    creatorSettingRows.Add(creatorSettingRow);
                }
            }
        }


        private void ClearCombiTagConditionsList()
        {
            foreach (var ctcr in combiTagConditionRows)
            {
                ctcr.ClearUI();
            }
            combiTagConditionRows.Clear();
            if (addCombiRowGO!=null) GameObject.Destroy(addCombiRowGO);
        }

        private void CombiTagSelectCallback(string value)
        {
            CreateDraftCombiTag(value);
            selectedCombiTagsRow = null;
            RefreshGroupUI(2);


        }

        private void CreateDraftCombiTag(string combiTagValue)
        {
            var combiTag = TagManifest.GetTag(resourceType, atomPresetResourceType, combiTagValue) as CombiTag;

            draftCombiTag = new CombiTag("", resourceType, atomPresetResourceType);
            draftCombiTag.CopyFrom(combiTag);
        }

        private void ApplyCombiCallback()
        {
            var combiTag = TagManifest.GetTag(resourceType, atomPresetResourceType, combiTagsJSSC.val) as CombiTag;

            var beforeTags = combiTag.usesTags;
            var afterTags = draftCombiTag.usesTags;

            var tagsToRemove = beforeTags.Except(afterTags);
            var tagsToAdd = afterTags.Except(beforeTags);


            combiTag.CopyFrom(draftCombiTag);

            foreach (var tag in tagsToRemove)
            {
                tag.RemoveUsedBy(combiTag);
            }
            foreach (var tag in tagsToAdd)
            {
                tag.AddUsedBy(combiTag);
            }

            draftCombiTag.isModfied = false;
            selectedCombiTagsRow = null;

            foreach (var rvge in ResourceManifest.GetAllResourceGroups(resourceType, atomPresetResourceType))
            {
                rvge.SetTag(combiTag);
//                TagManifest.AssessCombiTagState(rvge, combiTag);
            }

            Settings.SaveSettings();

            if (BrowserAssist.mainBrowserUI.resourceActionsUI.currentActionUITabType == ActionUITabType.tagsTab) BrowserAssist.mainBrowserUI.resourceActionsUI.getCurrentActionUITab.RefreshUI();
            if (resourceType==BrowserAssist.mainBrowserUI.browserFiltersUI.currentResourceType) BrowserAssist.mainBrowserUI.RefreshFilterResults();
            if (BrowserAssist.callbackBrowserUI.isActive && resourceType == BrowserAssist.callbackBrowserUI.browserFiltersUI.currentResourceType) BrowserAssist.callbackBrowserUI.RefreshFilterResults();
            RefreshGroupUI(2);
        }
        private void CancelCombiCallback()
        {
            var combiTag = TagManifest.GetTag(resourceType, atomPresetResourceType, combiTagsJSSC.val) as CombiTag;
            draftCombiTag.CopyFrom(combiTag);
            draftCombiTag.isModfied = false;
            selectedCombiTagsRow = null;
            RefreshGroupUI(2);
        }

        private void RefreshCombiTagConditionsList()
        {
            
            ClearCombiTagConditionsList();
            var combiLogicStatements = draftCombiTag.combiLogicConditions;
         
            int rowIndex = 1;
            foreach (CombiLogicCondition cl in combiLogicStatements)
            {
                var combiTagConditionsRow = new CombiTagConditionRowUI(combiLogicListContentGO, cl, rowIndex, this);
                combiTagConditionRows.Add(combiTagConditionsRow);
                rowIndex++;

            }
            addCombiRowGO = UI.AddChildGOImage(combiLogicListContentGO, new Color(0f,0f,0f,0f), AnchorPresets.hStretchTop, 705f, 60f, Vector2.zero);
            var addButton = UI.CreateUIButton(addCombiRowGO, 120f, 60f, "Add...",30,0f,0f,AnchorPresets.middleRight,false,false);
            addButton.button.onClick.AddListener(delegate
            {
                draftCombiTag.AddCombiLogicCondition();
                selectedCombiTagsRow = null;
                RefreshGroupUI(2);
            });
            ApplyHelp(addButton, "Add a new logic condition to the currently selected Combi-Tag.");
        }

        public void DeleteCombiTagConditionRow(CombiLogicCondition clc)
        {
            draftCombiTag.RemoveCombiLogicConditions(clc);
            RefreshGroupUI(2);
        }
        public void SelectTagsActive(CombiTagConditionRowUI rowUI)
        {
            if (selectedCombiTagsRow != null && selectedCombiTagsRow.tagNameListButton!=null) selectedCombiTagsRow.tagNameListButton.buttonColor = BrowserUI.lightGrey;
            selectedCombiTagsRow = rowUI;
            cachedAddonPackagesFolderNames = null;
            if (cachedLocalFolderNamesByResourceType.ContainsKey(resourceType)) cachedLocalFolderNamesByResourceType[resourceType] = null;
            if (cachedLocalFolderNamesByAtomPresetResource.ContainsKey(atomPresetResourceType)) cachedLocalFolderNamesByAtomPresetResource[atomPresetResourceType] = null;

            RefreshCombiTagSelectionUI();
        }

        private void CreateUIDefaultFilterSettings(string resourceTypeDisplayName)
        {
            defaultHelpText = "The Default Filters tab determines the Browser Filter settings that will be applied for '"+resourceTypeDisplayName+"' resources when the Browser Assist plugin is loaded.";
            helpTextJSS.val = defaultHelpText;

            helpTextCustomUI.InitUI(CustomRowWidth.full);

            var helpTextBox = helpTextCustomUI.AddTextBox(helpTextJSS, 1050f, 200f, 0f, 0f);
            helpTextBox.textColor = Color.white;
            helpTextBox.backgroundColor = new Color(0.4f, 0.5f, 0.2f, 0.4f);

            var selectModePopup = resourceSettingsCustomUI.AddPopup(resourceSettings.defaultFilterSettings.selectModeJSEnum, null, 520, 50, 0, 0, labelProportion: 0.6f, labelOverride: "Default Select Mode");
            ApplyHelp(selectModePopup, "This setting will determine the Select Mode that is applied when the Browser Assist plugin is loaded.");

            var tagPopupUID = resourceSettingsCustomUI.AddPopup(resourceSettings.defaultFilterSettings.tagsFilterJSMultiEnum, null, 520, 50, 530, 0, labelProportion: 0.6f, labelOverride: "Default Tag Filter", panelHeight: 600f);
            tagPopupUID.popup.onOpenPopupHandlers += resourceSettings.defaultFilterSettings.RefreshTagFilters;
            ApplyHelp(tagPopupUID, "This setting will determine the Tag Filter that is applied when the Browser Assist plugin is loaded.");

            var showHiddenToggle =resourceSettingsCustomUI.AddToggle(resourceSettings.defaultFilterSettings.showHiddenJSB, 300, 50, 0, -60, "Show Hidden");
            ApplyHelp(showHiddenToggle, "This setting will determine if the 'Show Hidden' filter is applied when the Browser Assist plugin is loaded.");
            var onlyFavToggle = resourceSettingsCustomUI.AddToggle(resourceSettings.defaultFilterSettings.onlyFavouritesJSB, 300, 50, 310, -60, "Only Favorites");
            ApplyHelp(onlyFavToggle, "This setting will determine if the 'Only Favorites' filter is applied when the Browser Assist plugin is loaded.");
            var onlyNewToggle = resourceSettingsCustomUI.AddToggle(resourceSettings.defaultFilterSettings.onlyNewJSB, 300, 50, 620, -60, "Only New");
            ApplyHelp(onlyNewToggle, "This setting will determine if the 'Only New' filter is applied when the Browser Assist plugin is loaded.");

            var creatorPopupUID = resourceSettingsCustomUI.AddPopup(resourceSettings.defaultFilterSettings.creatorJSMultiEnum, null, 720, 100, 0, -120, labelProportion: 0.6f, labelOverride: "Default Creator / Local filter: Packages or Local content", panelHeight: 800f);
            creatorPopupUID.popup.onOpenPopupHandlers += resourceSettings.defaultFilterSettings.RefreshCreatorFilters;
            ForceUIRefreshOnValChange(resourceSettings.defaultFilterSettings.creatorJSMultiEnum, 2);
            ApplyHelp(creatorPopupUID, "This setting will determine the 'Creator' filter that is applied when the Browser Assist plugin is loaded.");

            if (resourceSettings.defaultFilterSettings.creatorJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.mainValFlag || resourceSettings.defaultFilterSettings.creatorJSMultiEnum.valTopEnum == CreatorFilterChoices.allVARs)
            {
                var licensePopup = resourceSettingsCustomUI.AddPopup(resourceSettings.defaultFilterSettings.licenseFilterJSEnum, null, 320, 100, 730, -120, labelProportion: 0.6f, labelOverride: "Default VAR License");
                ApplyHelp(licensePopup, "This setting will determine the 'VAR License' filter that is applied when the Browser Assist plugin is loaded.");
            }
            else if (resourceSettings.defaultFilterSettings.resourceCategory == ResourceCategory.clothing)
            {
                var clothingCreatorPopup = resourceSettingsCustomUI.AddPopup(resourceSettings.defaultFilterSettings.clothingCreatorJSME, null, 320, 100, 730, -120, labelProportion: 0.4f, labelOverride: "Clothing Creator", panelHeight: 800f);
                clothingCreatorPopup.popup.onOpenPopupHandlers += resourceSettings.defaultFilterSettings.RefreshClothingCreatorFilters;
                ApplyHelp(clothingCreatorPopup, "This setting will determine the 'Clothing Creator' filter that is applied when the Browser Assist plugin is loaded.");
            }
            else if (resourceSettings.defaultFilterSettings.creatorJSMultiEnum.valTopEnum == CreatorFilterChoices.localOnly)
            {
                var folderTextBox = resourceSettingsCustomUI.AddTextBox(resourceSettings.defaultFilterSettings.folderFilterJSS, 320, 150, 730, -120, TextAnchor.UpperLeft);
                ApplyHelp(folderTextBox, "This is the currently selected local folder that will be applied as filter when Browser Assist is loaded.");
                var selectFolderButton = resourceSettingsCustomUI.AddButton(320, 80, 730, -270, "Select default local folder...", GetDefaultLocalFolder);
                ApplyHelp(selectFolderButton, "This button will open a folder browser to select a local folder that will be applied as filter when Browser Assist is loaded.");
                var subFoldersToggle =resourceSettingsCustomUI.AddToggle(resourceSettings.defaultFilterSettings.inludeSubFoldersJSB, 320, 80, 730, -370, "Include Sub-Folders by default", 50);
                ApplyHelp(subFoldersToggle, "This setting will determine if the 'Include sub folders' filter is applied when the Browser Assist plugin is loaded.");
            }
            var sortModePopup = resourceSettingsCustomUI.AddPopup(resourceSettings.defaultFilterSettings.sortModeJSEnum, null, 400, 80, 0, -230, popupRowHeight: 44f);
            ApplyHelp(sortModePopup, "This setting will determine the Sort order that will be selected for browser results when the Browser Assist plugin is loaded.");

            var resetButton = resourceSettingsCustomUI.AddButton(450f, 120f, 300f, -500f, "Set '" + resourceTypeDisplayName + "' default Filter values to the current '" + resourceTypeDisplayName + "' Filter settings", ResetFilterSettings);
            resetButton.buttonColor = Color.yellow;
            ApplyHelp(resetButton, "This button will copy all the current filter settings that are active in the Browser Assist window and apply them as the default values for this resource type ("+resourceTypeDisplayName+") when the Browser Assist plugin is loaded.");


            RegisterCustomUIElement(2, helpTextCustomUI);
        }

        private void GetDefaultLocalFolder()
        {
            SuperController.singleton.GetDirectoryPathDialog(FolderSelectedCallback,Const.GetResourceRootFolderNames(resourceType,atomPresetResourceType).First(),null,false);
        }

        private void FolderSelectedCallback(string folderPath)
        {
            resourceSettings.defaultFilterSettings.folderFilterJSS.val = folderPath;
        }
        private void ResetFilterSettings()
        {
            var currentBrowserResourceFilterSettings = BrowserAssist.mainBrowserUI.resourceBrowserUI.browserFiltersUI.GetFilterSettings(resourceType,atomPresetResourceType);

            resourceSettings.defaultFilterSettings.RefreshTagFilters();

            resourceSettings.defaultFilterSettings.selectModeJSEnum.val = currentBrowserResourceFilterSettings.selectModeJSEnum.val;
            resourceSettings.defaultFilterSettings.tagsFilterJSMultiEnum.val = currentBrowserResourceFilterSettings.tagsFilterJSMultiEnum.val;
            resourceSettings.defaultFilterSettings.showHiddenJSB.valNoCallback = currentBrowserResourceFilterSettings.showHiddenJSB.val;
            resourceSettings.defaultFilterSettings.onlyFavouritesJSB.valNoCallback = currentBrowserResourceFilterSettings.onlyFavouritesJSB.val;
            resourceSettings.defaultFilterSettings.onlyNewJSB.valNoCallback = currentBrowserResourceFilterSettings.onlyNewJSB.val;
            resourceSettings.defaultFilterSettings.RefreshCreatorFilters();
            resourceSettings.defaultFilterSettings.creatorJSMultiEnum.val = currentBrowserResourceFilterSettings.creatorJSMultiEnum.val;
            resourceSettings.defaultFilterSettings.RefreshLicenseFilters();
            resourceSettings.defaultFilterSettings.licenseFilterJSEnum.valNoCallback = currentBrowserResourceFilterSettings.licenseFilterJSEnum.val;
            resourceSettings.defaultFilterSettings.folderFilterJSS.valNoCallback = currentBrowserResourceFilterSettings.folderFilterJSS.val;
            resourceSettings.defaultFilterSettings.inludeSubFoldersJSB.valNoCallback = currentBrowserResourceFilterSettings.inludeSubFoldersJSB.val;
            resourceSettings.defaultFilterSettings.selectedFolderFilter = currentBrowserResourceFilterSettings.selectedFolderFilter;


            resourceSettings.defaultFilterSettings.sortModeJSEnum.valNoCallback = currentBrowserResourceFilterSettings.sortModeJSEnum.val;
            Settings.SaveSettings();
        }

        private void ResourceSettingsTabSelected(int tabRef)
        {
            if (tabRef != currentResourceSelectionSettingsTab)
            {
                currentResourceSelectionSettingsTab = tabRef;
                RefreshGroupUI(2);

                
            }
        }

    }



    public class CombiTagConditionRowUI
    {
        private GameObject rowGO;
        public CombiLogicCondition clc { get; private set; }
        private int rowIndex;

        private UIDynamicPopup tagOperandPopup;
        private UIDynamicPopup logicConditionPopup;
        public UIDynamicButton tagNameListButton { get; private set; }
        public UIDynamicButton rowDeleteButton;
        private GameObject thatAreTextBoxGO;

        private GameObject parentGO;

        private ResourceSettingsUITab parentUI;

        public CombiTagConditionRowUI(GameObject parentGO, CombiLogicCondition cls, int rowIndex,  ResourceSettingsUITab parentUI)
        {
            this.clc = cls;
            this.rowIndex = rowIndex;
            this.parentGO = parentGO;
            this.parentUI = parentUI;
            rowGO = UI.AddChildGOImage(parentGO, Color.grey, AnchorPresets.hStretchTop, 715f, 120f, Vector2.zero);

            CreateUI();
        }

        public void ClearUI()
        {
            if (rowGO != null) GameObject.Destroy(rowGO);
        }

        public void RefreshUI()
        {
            if (logicConditionPopup != null) GameObject.Destroy(logicConditionPopup.gameObject);
            if (tagNameListButton != null) GameObject.Destroy(tagNameListButton.gameObject);
            if (tagOperandPopup != null) GameObject.Destroy(tagOperandPopup.gameObject);
            if (thatAreTextBoxGO != null) GameObject.Destroy(thatAreTextBoxGO);
            if (rowDeleteButton !=null) GameObject.Destroy(rowDeleteButton.gameObject);
            CreateUI();
        }

        private void CreateUI()
        {
            if (rowIndex == 1) thatAreTextBoxGO = UI.CreateBasicTextBox(rowGO, "that are", 100f, 60f, 30, 0f, 0f, AnchorPresets.topLeft, false, 20, HorizontalWrapMode.Overflow, null, false);
            else if (rowIndex == 2)
            {
                tagOperandPopup = UI.CreatePopup(rowGO, clc.parentCombiTag.logicOperandJSEnum, 115f, 60f, null, 30, 0f, 0f, AnchorPresets.topLeft, labelProportion: 0f, rescaleText: false, popupRowHeight: 60f);
                parentUI.ApplyHelp(tagOperandPopup,"The logical operation that will combine multiple conditions for this CombiTag. This is with 'AND' or 'OR'.");
                tagOperandPopup.popup.onOpenPopupHandlers += () =>
                {
                    Globals.mvrScript.StartCoroutine(ReparentPopup(tagOperandPopup));
                };
                tagOperandPopup.popup.onValueChangeHandlers += (string val) =>
                {                    
                    parentUI.RefreshCombiTagDraftStatusButtons();
                    parentUI.RefreshCombiConditionsLogicOperand(this);

                    parentUI.SelectTagsActive(this);
                    RefreshUI();
                };
            }
            else thatAreTextBoxGO = UI.CreateBasicTextBox(rowGO, clc.parentCombiTag.logicOperandJSEnum.displayVal, 100f, 60f, 30, 0f, 0f, AnchorPresets.topLeft, false, 20, HorizontalWrapMode.Overflow, null, false,TextAnchor.MiddleCenter);

            logicConditionPopup = UI.CreateScrollablePopup(rowGO, clc.logicConditionJSEnum, 330f, 60f, "", 30, 120f, 0f, 400f, false, true, AnchorPresets.topLeft, 0f, 40f, 30, false,20f);
            parentUI.ApplyHelp(logicConditionPopup, "Select a condition that will determine whether the Combi-tag is applied to a resource.");


            logicConditionPopup.popup.onOpenPopupHandlers += () =>
            {
                Globals.mvrScript.StartCoroutine(ReparentPopup(logicConditionPopup));
            };

            if (rowIndex > 1)
            {
                rowDeleteButton = UI.CreateUIButton(rowGO, 40f, 40f, "X", 34, 10f, 10f, AnchorPresets.bottomLeft, rescaleText: false);
                parentUI.ApplyHelp(rowDeleteButton, "Remove this condition from the list of conditions applied to this Combi-tag.");
                rowDeleteButton.buttonColor = Color.red;
                rowDeleteButton.textColor = Color.white;
                rowDeleteButton.button.onClick.AddListener(delegate
                {
                    parentUI.DeleteCombiTagConditionRow(clc);
                });
            }

            logicConditionPopup.popup.onValueChangeHandlers += (string val) =>
            {
                if (clc.logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWith) clc.UseOnlyOneTag();
                if (clc.logicConditionJSEnum.val == CombiLogicConditionOptions.varUnderFolder || clc.logicConditionJSEnum.val == CombiLogicConditionOptions.varNotUnderFolder) clc.UseOnlyOneVARFolder();
                if (clc.logicConditionJSEnum.val == CombiLogicConditionOptions.underLocalFolder || clc.logicConditionJSEnum.val == CombiLogicConditionOptions.notUnderLocalFolder) clc.UseOnlyOneLocalFolder();
                parentUI.SelectTagsActive(this);
                parentUI.RefreshCombiTagDraftStatusButtons();
                parentUI.RefreshCombiTagSelectionUI();
               
                RefreshUI();
            };
            if (clc.isMultiConfigurableCondition)
            {
                tagNameListButton = UI.CreateUIButton(rowGO, 240f, 120f, "", 30, -10f, 0, AnchorPresets.topRight, true, false);
                if (clc.isTaggableCondition) parentUI.ApplyHelp(tagNameListButton, "The list of tags that this condition is assessed against. Activate the button and select tags in the right hand list to modify.");
                else if (clc.isLicenseCondition) parentUI.ApplyHelp(tagNameListButton, "The list of package licenses that this condition is assessed against. Activate the button and select license types in the right hand list to modify.");
                else if (clc.isCreatorCondition) parentUI.ApplyHelp(tagNameListButton, "The list of package Creators that this condition is assessed against. Activate the button and select the Creator Names in the right hand list to modify.");
                else if (clc.isVARFolderCondition) parentUI.ApplyHelp(tagNameListButton, "The list of AddonPackage folders that this condition is assessed against. Activate the button and select the Folder names in the right hand list to modify.");

                if (parentUI.selectedCombiTagsRow == this)
                {
                    tagNameListButton.buttonColor = Color.cyan;
                }
                else
                {
                    tagNameListButton.buttonColor = BrowserUI.lightGrey;
                }


                tagNameListButton.button.onClick.AddListener(delegate
                {                    
                    parentUI.SelectTagsActive(this);
                    tagNameListButton.buttonColor = Color.cyan;
                    UpdateTagListText();
                });

                UpdateTagListText();
            }

        }

        public void UpdateTagListText()
        {
            string emptyListLabel = "";
            string listLabel = "";
            int listCount = 0;

            if (clc.isTaggableCondition)
            {
                emptyListLabel = "Select tags...";
                listLabel = clc.GetTagNameList();
                listCount = clc.usesTags.Count;
            }
            else if (clc.isLicenseCondition)
            {
                emptyListLabel = "Select licenses...";
                listLabel = clc.GetLicenseTypeNameList();
                listCount = clc.varLicenses.Count;
            }
            else if (clc.isCreatorCondition)
            {
                emptyListLabel = "Select Creator Names...";
                listLabel = clc.GetCreatorNameList();
                listCount = clc.varCreators.Count;
            }
            else if (clc.isVARFolderCondition)
            {
                emptyListLabel = "Select AddonPackage SubFolders...";
                listLabel = clc.GetVARFolderListDescription();
                listCount = clc.varFolders.Count;
            }
            else if (clc.isLocalFolderCondition)
            {
                emptyListLabel = "Select Local Folders...";
                listLabel = clc.GetLocalFolderList();
                listCount = clc.localFolders.Count;
            }
            else if (clc.isDateCondition)
            {
                emptyListLabel = "Select date...";
                listLabel = clc.GetDateLabel();
                listCount = 1;
            }

            if (listCount == 0)
            {
                tagNameListButton.buttonText.alignment = TextAnchor.MiddleCenter;
                tagNameListButton.buttonText.text = emptyListLabel;
                tagNameListButton.buttonText.fontSize = 30;
            }
            else
            {
                tagNameListButton.buttonText.alignment = TextAnchor.UpperLeft;
                tagNameListButton.buttonText.text = listLabel;
                tagNameListButton.buttonText.fontSize = 28;
            }

        }

        protected IEnumerator ReparentPopup(UIDynamicPopup popupUID)
        {
            yield return new WaitForEndOfFrame();

            if (popupUID.popup.visible)
            {
                popupUID.popup.popupPanel.transform.SetParent(parentGO.transform.parent.parent.parent.parent);
            }
        }

    }

    public class TagSettingsRowUI
    {
        private GameObject rowGO;
        private Tag tag;
        private ResourceSettingsUITab parentUI;

        private const int normalMode = 0;
        private const int renameMode =1;
        private const int deleteMode =2;

        private int rowMode = normalMode;
        private int resourceType ;

        private string atomPresetResourceType;

        private GameObject tagNameTextBoxGO;
        private UIDynamicButton tagRenameUIDButton;
        private UIDynamicButton tagDeleteUIDButton;
        private UIDynamicButton tagDeleteConfirmUIDButton;
        private UIDynamicButton tagDeleteCancelUIDButton;

        private UIDynamicToggle tagHideToggle;
        private UIDynamicToggle preferredTagToggle;
        private UIDynamicTextField tagNameEditTextBox;
        private GameObject tagNameWarningTextBoxGO;

        private JSONStorableString newTagNameJSS;
        private JSONStorableString newTagNameWarningJSS;
        public TagSettingsRowUI(GameObject parentGO, Tag tag, ResourceSettingsUITab parentUI)
        {
            newTagNameJSS = new JSONStorableString("newTagName","");
            newTagNameWarningJSS = new JSONStorableString("newTagNameWarning", "");
            rowGO = UI.AddChildGOImage(parentGO, Color.gray, AnchorPresets.hStretchTop, 1030f, 60f, Vector2.zero);
            this.tag = tag;
            resourceType = tag.resourceType;
            atomPresetResourceType = tag.atomPresetResourceType;
            this.parentUI = parentUI;

            CreateUI();
        }

        public void ClearUI()
        {
            GameObject.Destroy(rowGO);
        }

        private void RefreshUI()
        {
            if (tagNameTextBoxGO != null) GameObject.Destroy(tagNameTextBoxGO);
            if (tagRenameUIDButton!=null) GameObject.Destroy(tagRenameUIDButton.gameObject);
            if (tagHideToggle != null) GameObject.Destroy(tagHideToggle.gameObject);
            if (preferredTagToggle != null) GameObject.Destroy(preferredTagToggle.gameObject);
            if (tagNameEditTextBox != null) GameObject.Destroy(tagNameEditTextBox.gameObject);
            if (tagNameWarningTextBoxGO != null) GameObject.Destroy(tagNameWarningTextBoxGO);

            if (tagDeleteUIDButton != null) GameObject.Destroy(tagDeleteUIDButton.gameObject);
            if (tagDeleteConfirmUIDButton != null) GameObject.Destroy(tagDeleteConfirmUIDButton.gameObject);
            if (tagDeleteCancelUIDButton != null) GameObject.Destroy(tagDeleteCancelUIDButton.gameObject);
            CreateUI();
        }

        private void CreateUI()
        {
            if (rowMode != renameMode)
            {
                string tagType = "Tag";
                string tagName = tag.tagName;
                bool userTag = tag.tagCategory == TagCategory.userDefined || tag.tagCategory == TagCategory.combi;
                if (tag.tagCategory == TagCategory.combi)
                {
                    tagType = "Combi Tag";
                    tagName = "<i>"+tag.tagName+ "</i>";
                }
                else if (tag.tagCategory != TagCategory.userDefined)
                {
                    tagType = "System Tag";
                }

                tagNameTextBoxGO = UI.CreateBasicTextBox(rowGO, tagName, 400f, 60f, 30, 0f, 0f, AnchorPresets.middleLeft, true, 18, HorizontalWrapMode.Overflow, rescaleText: false);
                
                if (!userTag) parentUI.ApplyHelp(tagNameTextBoxGO, "The current name of this " + tagType + ". System Tags are predefined BrowserAssist tags that can't be modified.");
                else parentUI.ApplyHelp(tagNameTextBoxGO, "The current name of this " + tagType + ".");

                if (rowMode != deleteMode)
                {
                    if (userTag)
                    {
                        tagRenameUIDButton = UI.CreateUIButton(rowGO, 150f, 60f, "Rename..", 26, 400f, 0f, AnchorPresets.middleLeft, false, rescaleText: false);
                        tagRenameUIDButton.button.onClick.AddListener(delegate {
                            rowMode = renameMode;
                            newTagNameJSS.valNoCallback = tag.tagName;
                            RefreshUI();
                        });
                        parentUI.ApplyHelp(tagRenameUIDButton, "Alter the name of the '" + tag.tagName + "' " + tagType + ".");

                        tagHideToggle = UI.CreateToggle(rowGO, tag.hideTaggedResourcesJSB, 175f, 60f, "Hide resources", 26, 550f, 0f, AnchorPresets.middleLeft, 40f, rescaleText: false);
                        parentUI.ApplyHelp(tagHideToggle, "Mark this " + tagType + " as a 'Hide' tag. Once activated, any resources tagged with this Tag will be treated as a hidden resource within BrowserAssist.");
                    }
                    

                    preferredTagToggle = UI.CreateToggle(rowGO, tag.preferredTagJSB, 175f, 60f, "Preferred tag", 26, 725f, 0f, AnchorPresets.middleLeft, 40f, rescaleText: false);
                    parentUI.ApplyHelp(preferredTagToggle, "Mark this "+tagType+" as a 'Preferred' tag. Once activated, preferred tags will be listed at the top of the Tag Filter drop down.");

                    if (tag.tagCategory == TagCategory.userDefined)
                    {
                        string deleteLabel = "Delete...";
                        Color deleteButtonColor = Color.red;
                        Color deleteButtonTextColor = Color.white;
                        if (TagManifest.GetTaggedResourcesCount(tag) == 0)
                        {
                            deleteLabel = "Purge";
                            deleteButtonColor = Color.yellow;
                            deleteButtonTextColor = Color.black;

                        }
                        tagDeleteUIDButton = UI.CreateUIButton(rowGO, 130f, 60f, deleteLabel, 26, 900f, 0f, AnchorPresets.middleLeft, false, rescaleText: false);
                        if (deleteLabel == "Purge")
                        {
                            parentUI.ApplyHelp(tagDeleteUIDButton, "No resources are currently tagged with this Tag. This button will instantly remove the tag from the list of available tags. Note that when Browser Assist is re-started any unused Tags are automatically purged.");
                        }
                        else
                        {
                            parentUI.ApplyHelp(tagDeleteUIDButton, "This button will untag any resources (or combi-tags) and remove the tag from the list of available tags.");
                        }

                        tagDeleteUIDButton.textColor = deleteButtonTextColor;
                        tagDeleteUIDButton.buttonColor = deleteButtonColor;
                        tagDeleteUIDButton.button.onClick.AddListener(delegate {
                            if (TagManifest.GetTaggedResourcesCount(tag) == 0)
                            {
                                TagManifest.DeleteTag(tag);
                                if (BrowserAssist.mainBrowserUI.resourceActionsUI.currentActionUITabType == ActionUITabType.tagsTab) BrowserAssist.mainBrowserUI.resourceActionsUI.getCurrentActionUITab.RefreshUI();
                                if (BrowserAssist.callbackBrowserUI != null) BrowserAssist.callbackBrowserUI.resourceActionsUI.getCurrentActionUITab.RefreshUI();
                                parentUI.RefreshGroupUI(2);
                            }
                            else
                            {
                                rowMode = deleteMode;
                                RefreshUI();
                            }
                        });
                    }

                }
                else
                {
                    tagDeleteConfirmUIDButton = UI.CreateUIButton(rowGO, 450f, 60f, "Confirm Delete Tag from "+ TagManifest.GetTaggedResourcesCount(tag).ToString()+" resources", 26, 400f, 0f, AnchorPresets.middleLeft, false, rescaleText: false);
                    parentUI.ApplyHelp(tagDeleteConfirmUIDButton, "This button will instantly untag any resources (or combi-tags) that use this tag and remove the tag from the list of available tags.");

                    tagDeleteConfirmUIDButton.textColor = Color.white;
                    tagDeleteConfirmUIDButton.buttonColor = Color.red;
                    tagDeleteConfirmUIDButton.button.onClick.AddListener(delegate {
                        TagManifest.DeleteTag(tag);
                        if (BrowserAssist.mainBrowserUI.resourceActionsUI.currentActionUITabType == ActionUITabType.tagsTab) BrowserAssist.mainBrowserUI.resourceActionsUI.getCurrentActionUITab.RefreshUI();
                        parentUI.RefreshGroupUI(2);
                    });

                    tagDeleteCancelUIDButton = UI.CreateUIButton(rowGO, 180f, 60f, "Cancel Delete", 26, 0f, 0f, AnchorPresets.middleRight, false, rescaleText: false);
                    parentUI.ApplyHelp(tagDeleteCancelUIDButton, "This button will cancel the tag deletion process.");
                    tagDeleteCancelUIDButton.buttonColor = Color.green;
                    tagDeleteCancelUIDButton.button.onClick.AddListener(delegate {
                        rowMode = normalMode;

                        RefreshUI();
                    });
                }
                
            }
            else
            {
                tagNameEditTextBox = UI.CreateTextBox(rowGO, newTagNameJSS,400f,60f,30,0f,0f,AnchorPresets.middleLeft,true,InputField.LineType.SingleLine,"",0f,0f, NewTagNameChangeCallback, TextAnchor.MiddleLeft, NewTagNameCallback,false);
                var inputComponent = tagNameEditTextBox.gameObject.GetComponent<InputField>();
                inputComponent.ActivateInputField();
                newTagNameWarningJSS.val = "";
                tagNameWarningTextBoxGO = UI.CreateBasicTextBox(rowGO, newTagNameWarningJSS, 620f, 60f, 30, 400f, 0f, AnchorPresets.middleLeft, true, 20, HorizontalWrapMode.Overflow, Color.gray, false) ;
            }    
            
        }

        private void NewTagNameCallback(string newTagName)
        {
            TagManifest.RenameTag(tag, newTagName);
            rowMode = normalMode;
            Settings.SaveSettings();
            RefreshUI();

            if (BrowserAssist.mainBrowserUI.resourceActionsUI.currentActionUITabType == ActionUITabType.tagsTab) BrowserAssist.mainBrowserUI.resourceActionsUI.getCurrentActionUITab.RefreshUI();
        }

        private void NewTagNameChangeCallback(string tagName)
        {
            string tagNameMod = TagManifest.NormalizeTagName(tagName);
            newTagNameJSS.inputField.text = tagNameMod;
            if (tagNameMod == "") newTagNameWarningJSS.val = "Blank tag names are not permitted";
            else if (tagNameMod != tag.tagName && TagManifest.GetTagNameCount(resourceType, atomPresetResourceType, tagNameMod) > 0) newTagNameWarningJSS.val = "Tag name already exists";
            else if (tagNameMod =="hide") newTagNameWarningJSS.val = "Tag name 'hide' is not permitted";
            else if (tagNameMod == "favorite") newTagNameWarningJSS.val = "Tag name 'favorite' is not permitted";
            else if (tagNameMod.EndsWith(" ")) newTagNameWarningJSS.val = "Tag names can not end with a space character";
            else newTagNameWarningJSS.val = "";
        }
    }

    public class CreatorSettingsRowUI
    {
        private GameObject rowGO;
        private CreatorManifestEntry cme;
        private ResourceSettingsUITab parentUI;

        private int resourceType;
        private string atomPresetResourceType;

        private GameObject creatorNameTextBoxGO;

        private UIDynamicToggle creatorHideToggle;
        private UIDynamicToggle preferredCreatorToggle;

        private JSONStorableBool creatorHideJSB;
        private JSONStorableBool creatorPreferredJSB;

        public CreatorSettingsRowUI(GameObject parentGO, CreatorManifestEntry cme, ResourceSettingsUITab parentUI)
        {
            rowGO = UI.AddChildGOImage(parentGO, Color.gray, AnchorPresets.hStretchTop, 1030f, 60f, Vector2.zero);
            this.cme = cme;
            resourceType = parentUI.resourceType;
            atomPresetResourceType = parentUI.atomPresetResourceType;
            this.parentUI = parentUI;

            creatorHideJSB = new JSONStorableBool("creatorHide", cme.IsHiddenResources(resourceType,atomPresetResourceType), CreatorHideCallback);
            creatorPreferredJSB = new JSONStorableBool("creatorPreferred", cme.IsPreferredCreator(resourceType,atomPresetResourceType), CreatorPreferredCallback);

            CreateUI();
        }

        private void CreatorHideCallback(bool value)
        {
            cme.SetHiddenResources(resourceType,atomPresetResourceType,value);
        }
        private void CreatorPreferredCallback(bool value)
        {
            cme.SetPreferredCreator(resourceType, atomPresetResourceType, value);
        }

        public void ClearUI()
        {
            GameObject.Destroy(rowGO);
        }

        private void RefreshUI()
        {
            if (creatorNameTextBoxGO != null) GameObject.Destroy(creatorNameTextBoxGO);
            if (creatorHideToggle != null) GameObject.Destroy(creatorHideToggle.gameObject);
            if (preferredCreatorToggle != null) GameObject.Destroy(preferredCreatorToggle.gameObject);

            CreateUI();
        }

        private void CreateUI()
        {
            string creatorName = cme.creatorName;

            creatorNameTextBoxGO = UI.CreateBasicTextBox(rowGO, creatorName, 400f, 60f, 30, 0f, 0f, AnchorPresets.middleLeft, true, 18, HorizontalWrapMode.Overflow, rescaleText: false);

            parentUI.ApplyHelp(creatorNameTextBoxGO, "The name of the VAR Package Creator.");

            creatorHideToggle = UI.CreateToggle(rowGO, creatorHideJSB, 175f, 60f, "Hide resources", 26, 550f, 0f, AnchorPresets.middleLeft, 40f, rescaleText: false);
            parentUI.ApplyHelp(creatorHideToggle, "Mark this Creator as a 'Hidden' Creator. Once activated, any resources in packages from the Creator will be treated as a hidden resource within BrowserAssist.");

            preferredCreatorToggle = UI.CreateToggle(rowGO, creatorPreferredJSB, 175f, 60f, "Preferred Creator", 26, 725f, 0f, AnchorPresets.middleLeft, 40f, rescaleText: false);
            parentUI.ApplyHelp(preferredCreatorToggle, "Mark this Creator as a 'Preferred' Creator. Once activated, preferred Creators will be listed at the top of the Creator Filter drop down.");
         }

    }

}
