using UnityEngine;
using System.Collections.Generic;
using System.Linq;
using UnityEngine.UI;


namespace JayJayWon
{


    public class ResourceSelections
    {
        private Dictionary<int, ResourceTypeSelections> resourceTypeSelections = new Dictionary<int, ResourceTypeSelections>();

        private Dictionary<string, ResourceTypeSelections> atomPresetResourceTypeSelections = new Dictionary<string, ResourceTypeSelections>();

        private GameObject resourceSelectionsScrollableBoxGO;
        private GameObject resourceSelectionsBoxGO;
        private GameObject resourceSelectionScrollableViewportContentGO;

        private Color defaultButtonColor;
        private bool allFocus = false;

        private BrowserFiltersUI browserFiltersUI;
        private BrowserUI browserUI;

        private List<GameObject> resourceSelectionRows = new List<GameObject>();

        private Dictionary<ResourceVersionGroupEntry, UIDynamicButton> selectedRVGEButtons = new Dictionary<ResourceVersionGroupEntry, UIDynamicButton>();

        const int resourceSelectionBoxHeightMin = 20;
        const int resourceSelectionBoxHeightMax = 55;
        const int resourceSelectionBoxWidthMin = 441;
        const int resourceSelectionBoxWidthMax = 530;

        public int resourceSelectionBoxHeight
        {
            get
            {
                if (browserFiltersUI == null || browserFiltersUI.selectModeJSEnum == null || browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi) return resourceSelectionBoxHeightMin;
                return resourceSelectionBoxHeightMax;
            }
        }
        public int resourceSelectionBoxWidth
        {
            get
            {
                if (browserFiltersUI == null || browserFiltersUI.selectModeJSEnum == null || browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi) return resourceSelectionBoxWidthMin;
                return resourceSelectionBoxWidthMax;
            }
        }

        private int resourceType { get { return browserFiltersUI.currentResourceType; } }
        private string atomPresetResourceType { get { return browserFiltersUI.currentAtomPresetResourceType; } }

        private int resourceCategory {get{ return ResourceType.GetResourceCategory(resourceType);} }



        private ResourceTypeSelections resourceTypeSelection
        {
            get
            {
                if (resourceType != ResourceType.presetAtom)
                {
                    if (resourceTypeSelections.ContainsKey(resourceType)) return resourceTypeSelections[resourceType];
                }
                else if (atomPresetResourceTypeSelections.ContainsKey(atomPresetResourceType)) return  atomPresetResourceTypeSelections[atomPresetResourceType];
                return null;
            }
        }
        private List<ResourceVersionGroupEntry> selectedRVGEs
        {
            get
            {
                var resourceTypeSelection = this.resourceTypeSelection;

                if (resourceTypeSelection != null)
                {
                    if (browserFiltersUI.selectModeJSEnum.val == BrowserSelectMode.multi) return resourceTypeSelection.selectedRVGEs;
                    else
                    {
                        var selectedRVGEs = new List<ResourceVersionGroupEntry>();
                        var firstSelectedRVGE = resourceTypeSelection.GetFirstSelectedRVGE();
                        if (firstSelectedRVGE != null) selectedRVGEs.Add(firstSelectedRVGE);

                        return selectedRVGEs;
                    }
                }
                return null;
            }
        }

        public ResourceSelections(BrowserUI _browserUI)
        {
            browserUI = _browserUI;
            browserFiltersUI = _browserUI.browserFiltersUI;

        }
        public void InitUI(GameObject _resourceSelectionsBoxGO)
        {
            resourceSelectionsBoxGO = _resourceSelectionsBoxGO;
            resourceSelectionsScrollableBoxGO = UI.CreateVScrollableContent(resourceSelectionsBoxGO, BrowserUI.darkBlueGrey, AnchorPresets.vhStretch, resourceSelectionBoxWidth, resourceSelectionBoxHeight, new Vector2(0f, 0f));
            resourceSelectionScrollableViewportContentGO = resourceSelectionsScrollableBoxGO.transform.Find("scrollableContentViewportGO/vScrollableContentGO").gameObject;

            RePopulateSelectionsBox();

        }

        public void SetAllFocus()
        {
            allFocus= true;
            foreach (var rvge in selectedRVGEs)
            {
                selectedRVGEButtons[rvge].buttonColor = Color.cyan;
            }
        }

        public void SetSingleFocus()
        {
            allFocus= false;
            var selectedRVGEs = this.selectedRVGEs;
            if (selectedRVGEs != null)
            {
                foreach (var rvge in selectedRVGEs.Skip(1))
                {
                    selectedRVGEButtons[rvge].buttonColor = defaultButtonColor;
                }
            }
            
        }

        private void ClearResourceSelectionRows()
        {
            foreach (var rowGO in resourceSelectionRows) GameObject.Destroy(rowGO);
            resourceSelectionRows.Clear();
        }
        public void RePopulateSelectionsBox()
        {


            ClearResourceSelectionRows();
            selectedRVGEButtons.Clear();

            if (browserUI != null && browserUI.resourceBrowserUI != null) browserUI.resourceBrowserUI.SetMultiSelectClear();



            if (resourceTypeSelection!=null)
            {

                GameObject currentRowGO = CreateSelectResourceRow();

                float currentRowResourcesWidth = 0f;

                

                if (selectedRVGEs.Count() > 1) browserUI.resourceBrowserUI.selectedResourceCountTextJSS.val = selectedRVGEs.Count().ToString();
                else browserUI.resourceBrowserUI.selectedResourceCountTextJSS.val = "";

                foreach (var resourceVersionGroupEntry in selectedRVGEs)
                {
                    float nextButtonWidth = GetRVGEButtonWidth(resourceVersionGroupEntry);
                    float cancelButtonWidth = 0f;
                    if (selectedRVGEs.Count > 1 && browserFiltersUI.selectModeJSEnum.val == BrowserSelectMode.multi) cancelButtonWidth = 18f;

                    if (currentRowResourcesWidth + nextButtonWidth + cancelButtonWidth > resourceSelectionBoxWidth)
                    {
                        currentRowGO = CreateSelectResourceRow();
                        currentRowResourcesWidth = 0f;
                    }
                    var resourceButton = UI.CreateUIButton(currentRowGO, nextButtonWidth, 16f, resourceVersionGroupEntry.resourceDisplayName, 20, currentRowResourcesWidth, 0, AnchorPresets.middleLeft);
                    selectedRVGEButtons.Add(resourceVersionGroupEntry, resourceButton);
                    resourceButton.buttonImage.sprite = null;
                    if (resourceCategory != ResourceCategory.clothing) resourceButton.button.interactable = false;
                    else
                    {
                        resourceButton.button.interactable = true;
                        resourceButton.button.onClick.AddListener(delegate
                        {
                            browserUI.resourceActionsUI.primaryActionUITab. ActivateClothingUI(resourceVersionGroupEntry);
                        });
                    }
                    resourceButton.buttonText.resizeTextForBestFit = true;
                    resourceButton.buttonText.resizeTextMinSize = 15;
                    resourceButton.buttonText.resizeTextMaxSize = 21;

                    if (resourceVersionGroupEntry == selectedRVGEs.First())
                    {
                        defaultButtonColor = new Color(resourceButton.buttonColor.r, resourceButton.buttonColor.g, resourceButton.buttonColor.b, resourceButton.buttonColor.a);
                        resourceButton.buttonColor = Color.cyan;

                    }
                    else if (allFocus) resourceButton.buttonColor = Color.cyan;

                    ColorBlock colors = resourceButton.button.colors;
                    colors.disabledColor = BrowserUI.lightGrey;
                    resourceButton.button.colors = colors;
                    SelectedResourceButtonPointerEnterExit click = resourceButton.button.gameObject.AddComponent<SelectedResourceButtonPointerEnterExit>();
                    click.Init(resourceVersionGroupEntry, browserUI.helpTextJSS);

                    currentRowResourcesWidth += nextButtonWidth;

                    if (cancelButtonWidth > 0f)
                    {
                        var cancelButton = UI.CreateUIButton(currentRowGO, 16f, 16f, "X", 22, currentRowResourcesWidth, 0, AnchorPresets.middleLeft);
                        cancelButton.buttonImage.sprite = null;
                        cancelButton.buttonColor = Color.cyan;
                        currentRowResourcesWidth += 18f;
                        SelectedResourceButtonPointerEnterExit click2 = cancelButton.button.gameObject.AddComponent<SelectedResourceButtonPointerEnterExit>();
                        click2.Init(resourceVersionGroupEntry, browserUI.helpTextJSS);
                        cancelButton.button.onClick.AddListener(delegate
                        {
                            DeSelectRVGE(resourceVersionGroupEntry);
                        });
                    }
                    currentRowResourcesWidth += 3f;
                }
                if (selectedRVGEs.Count < 2 && !browserUI.isVREnabled && browserUI.browserFiltersUI.selectModeJSEnum.val == BrowserSelectMode.multi)
                {
                    var guideTextJSS = new JSONStorableString("guideText", "<i>Hold <Ctrl> key to add or remove a clicked resource to these selections. Use left <Shift> to select multiple contiguous resources prior to the clicked resource. Use right <Shift> to select multiple contiguous resources after the clicked resource. Use <Alt> key to change the last selected resource.</i>");
                    var textBox = UI.CreateBasicTextBox(currentRowGO, guideTextJSS, resourceSelectionBoxWidth - currentRowResourcesWidth, 40, 20, currentRowResourcesWidth, 0f, AnchorPresets.topLeft, backgoundColor: new Color(0f, 0f, 0f, 0f));

                }
            }
            if (browserUI.resourceActionsUI != null)
            {
                browserUI.resourceActionsUI.primaryActionUITab.deleteConfirmationActive = false;
            }
        }

        private GameObject CreateSelectResourceRow()
        {
            GameObject currentRowGO = UI.AddChildGOImage(resourceSelectionScrollableViewportContentGO, new Color(0f, 0.8f, 0f, 0f), AnchorPresets.hStretchTop, resourceSelectionBoxWidth, 20f, Vector2.zero);
            resourceSelectionRows.Add(currentRowGO);

            return currentRowGO;
        }

        private float GetRVGEButtonWidth(ResourceVersionGroupEntry rvge)
        {
            if (browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi) return Mathf.Clamp(rvge.resourceDisplayName.Length * 7, 80f, 400f);
            return Mathf.Clamp(rvge.resourceDisplayName.Length * 6, 40f, 200f);

        }
        public ResourceTypeSelections GetResourceTypeSelections(int resourceType, string atomPresetResourceType)
        {
            if (resourceType == ResourceType.presetAtom)
            {
                if (!atomPresetResourceTypeSelections.ContainsKey(atomPresetResourceType)) atomPresetResourceTypeSelections.Add(atomPresetResourceType, new ResourceTypeSelections(browserUI));
                return atomPresetResourceTypeSelections[atomPresetResourceType];
            }
            else
            {
                if (!resourceTypeSelections.ContainsKey(resourceType)) resourceTypeSelections.Add(resourceType, new ResourceTypeSelections(browserUI));
                return resourceTypeSelections[resourceType];
            }
        }
        public int SelectionCount(int resourceType, string atomPresetResourceType)
        {
            return GetResourceTypeSelections(resourceType,atomPresetResourceType).selectedRVGEs.Count;               
        }
        public bool IsSelected(ResourceVersionGroupEntry rvge, int selectionMode, UIDynamicButton updatedButton)
        {
            return GetResourceTypeSelections(rvge.resourceType, rvge.atomPresetResourceType).IsSelected(rvge, selectionMode, updatedButton);
                
        }

        public void ForcePrimarySelection(ResourceVersionGroupEntry rvge, UIDynamicButton button, int selectionMode)
        {
            GetResourceTypeSelections(rvge.resourceType, rvge.atomPresetResourceType).ForcePrimarySelection(rvge, button, selectionMode);

            RePopulateSelectionsBox();

            if (browserUI.resourceActionsUI != null) browserUI.resourceActionsUI.RefreshUIResourceSelectionChange();
        }

        public void SingleSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button)
        {
            GetResourceTypeSelections(rvge.resourceType, rvge.atomPresetResourceType).SingleSelectRVGE(rvge, button);

            RePopulateSelectionsBox();

            if (browserUI.resourceActionsUI != null) browserUI.resourceActionsUI.RefreshUIResourceSelectionChange();
        }

        public void LeftShiftSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button, List<ResourceVersionGroupEntry> currentPageRVGEs, List<UIDynamicButton> currentPageButtons)
        {
            GetResourceTypeSelections(rvge.resourceType, rvge.atomPresetResourceType).LeftShiftSelectRVGE(rvge, button, currentPageRVGEs, currentPageButtons);

            RePopulateSelectionsBox();
            if (browserUI.resourceActionsUI != null) browserUI.resourceActionsUI.RefreshUIResourceSelectionChange();
        }
        public void RightShiftSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button, List<ResourceVersionGroupEntry> currentPageRVGEs, List<UIDynamicButton> currentPageButtons)
        {
            GetResourceTypeSelections(rvge.resourceType, rvge.atomPresetResourceType).RightShiftSelectRVGE(rvge, button, currentPageRVGEs, currentPageButtons);

            RePopulateSelectionsBox();
            if (browserUI.resourceActionsUI != null) browserUI.resourceActionsUI.RefreshUIResourceSelectionChange();
        }

        public void ToggleSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button)
        {
            GetResourceTypeSelections(rvge.resourceType, rvge.atomPresetResourceType).ToggleSelectRVGE(rvge, button);

            RePopulateSelectionsBox();
            if (browserUI.resourceActionsUI != null) browserUI.resourceActionsUI.RefreshUIResourceSelectionChange();
        }

        public void UpdateDisplayName(ResourceVersionGroupEntry rvge)
        {
            if (selectedRVGEButtons.ContainsKey(rvge))
            {
                selectedRVGEButtons[rvge].label = rvge.resourceDisplayName;
            }

        }
        public void ClearSelectedRVGEButtons(int resourceType, string atomPresetResourceType)
        {
            GetResourceTypeSelections(resourceType, atomPresetResourceType).ClearSelectedRVGEButtons();
        }

        public void DeSelectRVGE(ResourceVersionGroupEntry rvge)
        {
            GetResourceTypeSelections(rvge.resourceType, rvge.atomPresetResourceType).DeSelectRVGE(rvge);

            RePopulateSelectionsBox();
            if (browserUI.resourceActionsUI != null) browserUI.resourceActionsUI.RefreshUIResourceSelectionChange();
        }

        public void DeSelectAll(int resourceType, string atomPresetResourceType, bool keepFirst = true)
        {
            GetResourceTypeSelections(resourceType, atomPresetResourceType).DeSelectAll(keepFirst);

            RePopulateSelectionsBox();
            if (browserUI.resourceActionsUI != null) browserUI.resourceActionsUI.RefreshUIResourceSelectionChange();
        }

        public void SelectAll(int resourceType, string atomPresetResourceType, List<ResourceVersionGroupEntry> currentFilterResults)
        {
            GetResourceTypeSelections(resourceType, atomPresetResourceType).SelectAll(currentFilterResults);

            RePopulateSelectionsBox();
            if (browserUI.resourceActionsUI != null) browserUI.resourceActionsUI.RefreshUIResourceSelectionChange();
        }

        public void SwitchSelectMode(int resourceType, string atomPresetResourceType, int selectionMode)
        {
            GetResourceTypeSelections(resourceType, atomPresetResourceType).SwitchSelectMode(selectionMode);

            RePopulateSelectionsBox();
        }

        public ResourceVersionGroupEntry GetFirstSelectedRVGE(int resourceType, string atomPresetResourceType)
        {
            return GetResourceTypeSelections(resourceType, atomPresetResourceType).GetFirstSelectedRVGE();

        }

        public int IsSelectionTagged(int resourceType, string atomPresetResourceType, int selectMode, Tag tag)
        {
            var resourceTypeSelection = GetResourceTypeSelections(resourceType, atomPresetResourceType);
            if (selectMode != BrowserSelectMode.multi) return resourceTypeSelection.IsFirstTagged(tag);
            return resourceTypeSelection.IsTagged(tag);
        }

        public int IsSelectionFavourite(int resourceType, string atomPresetResourceType, int selectMode)
        {
            var resourceTypeSelection = GetResourceTypeSelections(resourceType, atomPresetResourceType);
            if (selectMode != BrowserSelectMode.multi) return resourceTypeSelection.isFirstFavourite;
            return resourceTypeSelection.isFavourite;
        }
        public int IsSelectionNew(int resourceType, string atomPresetResourceType, int selectMode)
        {
            var resourceTypeSelection = GetResourceTypeSelections(resourceType, atomPresetResourceType);
            if (selectMode != BrowserSelectMode.multi) return resourceTypeSelection.isFirstNew;
            return resourceTypeSelection.isNew;
        }
        public int IsSelectionHidden(int resourceType, string atomPresetResourceType, int selectMode)
        {
            var resourceTypeSelection = GetResourceTypeSelections(resourceType, atomPresetResourceType);
            if (selectMode != BrowserSelectMode.multi) return resourceTypeSelection.isFirstHidden;
            return resourceTypeSelection.isHidden;
        }
        public void SetSelectionTag(int resourceType, string atomPresetResourceType, int selectMode, Tag tag, int tagTriState)
        {
            var resourceTypeSelection = GetResourceTypeSelections(resourceType, atomPresetResourceType);
            if (selectMode != BrowserSelectMode.multi) resourceTypeSelection.SetFirstTagged(tag, tagTriState);
            else resourceTypeSelection.SetTagged(tag, tagTriState);
        }
        public void SetSelectionFavourite(int resourceType, string atomPresetResourceType, int selectMode, int favTriState)
        {
            var resourceTypeSelection = GetResourceTypeSelections(resourceType, atomPresetResourceType);
            if (selectMode != BrowserSelectMode.multi) resourceTypeSelection.isFirstFavourite = favTriState;
            else resourceTypeSelection.isFavourite = favTriState;
        }

        public void SetSelectionHidden(int resourceType, string atomPresetResourceType, int selectMode, int hiddenTriState)
        {
            var resourceTypeSelection = GetResourceTypeSelections(resourceType, atomPresetResourceType);
            if (selectMode != BrowserSelectMode.multi) resourceTypeSelection.isFirstHidden = hiddenTriState;
            else resourceTypeSelection.isHidden = hiddenTriState;
        }
        public void SetSelectionNew(int resourceType, string atomPresetResourceType, int selectMode, int hiddenTriState)
        {
            var resourceTypeSelection = GetResourceTypeSelections(resourceType, atomPresetResourceType);
            if (selectMode != BrowserSelectMode.multi) resourceTypeSelection.isFirstNew = hiddenTriState;
            else resourceTypeSelection.isNew = hiddenTriState;
        }

        public void RefreshRVGESelections()
        {
            foreach (var resourceTypeSelectionsKVP in resourceTypeSelections)
            {
                resourceTypeSelectionsKVP.Value.RefreshRVGESelections();
            }

            foreach (var resourceTypeSelectionsKVP in atomPresetResourceTypeSelections)
            {
                resourceTypeSelectionsKVP.Value.RefreshRVGESelections();
            }
        }
    }

    public class ResourceTypeSelections
    {
        private Dictionary<ResourceVersionGroupEntry, UIDynamicButton> selectedRVGEButtons = new Dictionary<ResourceVersionGroupEntry, UIDynamicButton>();

        public List<ResourceVersionGroupEntry> selectedRVGEs = new List<ResourceVersionGroupEntry>();

        private List<ResourceVersionGroupEntry> randomSelectedRVGEs = new List<ResourceVersionGroupEntry>();
        private ResourceVersionGroupEntry lastRandomRVGE = null;

        private ResourceBrowserUI resourceBrowserUI { get { return browserUI.resourceBrowserUI; } }
        private BrowserUI browserUI;

        public int isHidden
        {
            get
            {
                bool hiddenFound = false;
                bool notHiddenFound = false;
                foreach (var rvge in selectedRVGEs)
                {
                    if (rvge.isHidden) hiddenFound = true;
                    else notHiddenFound = true;

                    if (hiddenFound && notHiddenFound) return TriState.indeterminate;
                }

                if (hiddenFound) return TriState.active;
                return TriState.inactive;
            }
            set
            {
                if (value == TriState.indeterminate) return;
                bool hide = (value == TriState.active);
                foreach (var rvge in selectedRVGEs)
                {
                    rvge.isHidden = hide;
                    resourceBrowserUI.SetHideFavColorTags(rvge);
                }
            }
        }

        public int isFirstHidden
        {
            get
            {
                if (selectedRVGEs.Count > 0 && GetFirstSelectedRVGE().isHidden)
                {
                    return TriState.active;
                }
                return TriState.inactive;
            }
            set
            {
                if (value == TriState.indeterminate) return;
                bool hide = (value == TriState.active);
                var rvge = GetFirstSelectedRVGE();

                rvge.isHidden = hide;
                resourceBrowserUI.SetHideFavColorTags(rvge);
            }
        }

        public int isFavourite
        {
            get
            {
                bool favFound = false;
                bool notFavFound = false;
                foreach (var rvge in selectedRVGEs)
                {
                    if (rvge.isFavorite) favFound = true;
                    else notFavFound = true;

                    if (favFound && notFavFound) return TriState.indeterminate;
                }

                if (favFound) return TriState.active;
                return TriState.inactive;
            }
            set
            {
                if (value == TriState.indeterminate) return;
                bool fav = (value == TriState.active);
                foreach (var rvge in selectedRVGEs)
                {
                    rvge.isFavorite = fav;
                    resourceBrowserUI.SetHideFavColorTags(rvge);
                }
            }
        }

        public bool containsLocal
        {
            get
            {
                foreach (var rvge in selectedRVGEs)
                {
                    if (rvge.localFileResource) return true;
                }
                return false;
            }
        }
        public int isNew
        {
            get
            {
                bool newFound = false;
                bool notNewFound = false;
                foreach (var rvge in selectedRVGEs)
                {
                    if (rvge.isNew) newFound = true;
                    else notNewFound = true;

                    if (newFound && notNewFound) return TriState.indeterminate;
                }

                if (newFound) return TriState.active;
                return TriState.inactive;
            }
            set
            {
                if (value == TriState.indeterminate) return;
                bool resNew = (value == TriState.active);
                foreach (var rvge in selectedRVGEs)
                {
                    rvge.isNew = resNew;
                    resourceBrowserUI.SetNewIcon(rvge);
                }
            }
        }

        public int isFirstFavourite
        {
            get
            {
                if (selectedRVGEs.Count > 0 && GetFirstSelectedRVGE().isFavorite) return TriState.active;
                return TriState.inactive;
            }
            set
            {
                if (value == TriState.indeterminate) return;
                bool fav = (value == TriState.active);
                var rvge = GetFirstSelectedRVGE();
                rvge.isFavorite = fav;
                resourceBrowserUI.SetHideFavColorTags(rvge);
            }
        }

        public int isFirstNew
        {
            get
            {
                if (selectedRVGEs.Count > 0 && GetFirstSelectedRVGE().isNew) return TriState.active;
                return TriState.inactive;
            }
            set
            {
                if (value == TriState.indeterminate) return;
                bool resNew = (value == TriState.active);
                var rvge = GetFirstSelectedRVGE();
                rvge.isNew = resNew;
                resourceBrowserUI.SetNewIcon(rvge);
            }
        }

        public ResourceTypeSelections(BrowserUI _browserUI)
        {
            browserUI = _browserUI;
        }

        public void SwitchSelectMode(int selectMode)
        {
            if (selectedRVGEs.Count > 0)
            {
                UIDynamicButton firstButton = null;

                if (selectedRVGEButtons.ContainsKey(selectedRVGEs.First()))
                    firstButton = selectedRVGEButtons[selectedRVGEs.First()];
                foreach (var rvgeButton in selectedRVGEButtons)
                {
                    if (selectMode == BrowserSelectMode.multi || rvgeButton.Value == firstButton) ResourceBrowserUI.SelectColor(rvgeButton.Value);
                    else ResourceBrowserUI.DeSelectColor(rvgeButton.Value);
                }

            }

        }

        public void ClearSelectedRVGEButtons()
        {
            selectedRVGEButtons.Clear();
        }

        public void RightShiftSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button, List<ResourceVersionGroupEntry> currentPageRVGEs, List<UIDynamicButton> currentPageButtons)
        {
            int selectionIndex = currentPageButtons.FindIndex(x => x == button);
            int altIndex = currentPageButtons.Count - 1;

            bool selectToggle = !selectedRVGEs.Contains(rvge);

            for (int searchIndex = selectionIndex + 1; searchIndex < currentPageButtons.Count; searchIndex++)
            {
                if (selectToggle && selectedRVGEs.Contains(currentPageRVGEs[searchIndex]))
                {
                    altIndex = searchIndex - 1;
                    break;
                }
                if (!selectToggle && !selectedRVGEs.Contains(currentPageRVGEs[searchIndex]))
                {
                    altIndex = searchIndex - 1;
                    break;
                }
            }

            for (int toggleIndex = altIndex; toggleIndex >= selectionIndex; toggleIndex--)
            {
                ToggleSelectRVGE(currentPageRVGEs[toggleIndex], currentPageButtons[toggleIndex]);
            }

        }

        public void LeftShiftSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button, List<ResourceVersionGroupEntry> currentPageRVGEs, List<UIDynamicButton> currentPageButtons)
        {
            int selectionIndex = currentPageButtons.FindIndex(x => x == button);
            int altIndex = 0;

            bool selectToggle = !selectedRVGEs.Contains(rvge);

            for (int searchIndex = selectionIndex - 1; searchIndex >= 0; searchIndex--)
            {
                if (selectToggle && selectedRVGEs.Contains(currentPageRVGEs[searchIndex]))
                {
                    altIndex = searchIndex + 1;
                    break;
                }
                if (!selectToggle && !selectedRVGEs.Contains(currentPageRVGEs[searchIndex]))
                {
                    altIndex = searchIndex + 1;
                    break;
                }
            }

            for (int toggleIndex = altIndex; toggleIndex <= selectionIndex; toggleIndex++)
            {
                ToggleSelectRVGE(currentPageRVGEs[toggleIndex], currentPageButtons[toggleIndex]);
            }

        }
        public void ToggleSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button)
        {
            if (!selectedRVGEs.Contains(rvge)) SelectRVGE(rvge, button);
            else DeSelectRVGE(rvge);
        }
        private void SelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button)
        {
            if (selectedRVGEs.Contains(rvge)) selectedRVGEs.Remove(rvge);
            selectedRVGEs.Insert(0, rvge);

            if (button != null)
            {
                ResourceBrowserUI.SelectColor(button);
                if (!selectedRVGEButtons.ContainsKey(rvge)) selectedRVGEButtons.Add(rvge, button);
                else selectedRVGEButtons[rvge] = button;
            }

            if (browserUI.resourceActionsUI != null) browserUI.resourceActionsUI.primaryActionUITab.RefreshTargetAtomValue();

            randomSelectedRVGEs.Clear();
        }
        public void ForcePrimarySelection(ResourceVersionGroupEntry rvge, UIDynamicButton button,int selectionMode)
        {
            if (selectionMode != BrowserSelectMode.multi || selectedRVGEs.Count<2) SingleSelectRVGE(rvge, button);
            else SelectRVGE(rvge, button);           

        }

        public void SingleSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button)
        {
            if (selectedRVGEs.Count > 0)
            {
                ResourceVersionGroupEntry firstRVGE = selectedRVGEs.First();
                if (selectedRVGEButtons.ContainsKey(firstRVGE)) ResourceBrowserUI.DeSelectColor(selectedRVGEButtons[firstRVGE]);

                selectedRVGEs.Remove(firstRVGE);
                selectedRVGEButtons.Remove(firstRVGE);
            }
            SelectRVGE(rvge, button);

        }
        private void InternalDeSelectRVGE(ResourceVersionGroupEntry rvge)
        {
            selectedRVGEs.Remove(rvge);
            if (selectedRVGEButtons.ContainsKey(rvge))
            {
                UIDynamicButton button = selectedRVGEButtons[rvge];
                selectedRVGEButtons.Remove(rvge);
                if (button != null) ResourceBrowserUI.DeSelectColor(button);
            }
        }
        public void DeSelectRVGE(ResourceVersionGroupEntry rvge)
        {

            InternalDeSelectRVGE(rvge);
            randomSelectedRVGEs.Clear();
        }

        public ResourceVersionGroupEntry GetFirstSelectedRVGE()
        {
            if (selectedRVGEs.Count > 0) return selectedRVGEs.First();
            return null;
        }

        public void DeSelectAll(bool keepFirst)
        {
            ResourceVersionGroupEntry firstRVGE = null;

            if (selectedRVGEs.Count > 0) firstRVGE = selectedRVGEs.First();

            UIDynamicButton firstRVGEButton = null;
            foreach (var kvp in selectedRVGEButtons)
            {
                if (kvp.Key != firstRVGE)
                {
                    if (kvp.Value != null) ResourceBrowserUI.DeSelectColor(kvp.Value);
                }
                else firstRVGEButton = kvp.Value;
            }
            selectedRVGEs.Clear();
            selectedRVGEButtons.Clear();
            if (firstRVGE != null && keepFirst)
            {
                selectedRVGEs.Add(firstRVGE);
                selectedRVGEButtons.Add(firstRVGE, firstRVGEButton);
            }

        }

        public void SelectAll(List<ResourceVersionGroupEntry> currentFilterResults)
        {

            foreach (var rvge in currentFilterResults)
            {
                SelectRVGE(rvge, resourceBrowserUI.GetUIDButton(rvge));
            }
        }

        public bool IsSelected(ResourceVersionGroupEntry rvge, int selectionMode, UIDynamicButton updatedButton)
        {
            if (!selectedRVGEs.Contains(rvge)) return false;
            if (!selectedRVGEButtons.ContainsKey(rvge)) selectedRVGEButtons.Add(rvge, updatedButton);
            else selectedRVGEButtons[rvge] = updatedButton;

            ResourceVersionGroupEntry firstRVGE = selectedRVGEs.First();
            if (rvge == firstRVGE || selectionMode == BrowserSelectMode.multi) return true;
            return false;
        }

        public int IsFirstTagged(Tag tag)
        {
            if (selectedRVGEs.Count > 0 && GetFirstSelectedRVGE().IsTagged(tag)) return TriState.active;
            return TriState.inactive;
        }

        public int IsTagged(Tag tag)
        {
            bool tagFound = false;
            bool notTagFound = false;
            foreach (var rvge in selectedRVGEs)
            {
                if (rvge.IsTagged(tag)) tagFound = true;
                else notTagFound = true;

                if (tagFound && notTagFound) return TriState.indeterminate;
            }

            if (tagFound) return TriState.active;
            return TriState.inactive;
        }
        public void SetTagged(Tag tag, int value)
        {
            if (value == TriState.indeterminate) return;
            bool tagValue = (value == TriState.active);
            foreach (var rvge in selectedRVGEs)
            {
                rvge.SetTag(tag, tagValue);
            }

        }
        public int GetFavouriteCount()
        {
            int count = 0;
            foreach (var rvge in selectedRVGEs)
            {
                if (rvge.isFavorite) count++;
            }

            return count;
        }
        public int GetHiddenCount()
        {
            int count = 0;
            foreach (var rvge in selectedRVGEs)
            {
                if (rvge.isHidden) count++;
            }

            return count;
        }
        public int GetNewCount()
        {
            int count = 0;
            foreach (var rvge in selectedRVGEs)
            {
                if (rvge.isNew) count++;
            }

            return count;
        }
        public int GetTaggedCount(Tag tag)
        {
            int count = 0;
            foreach (var rvge in selectedRVGEs)
            {
                if (rvge.IsTagged(tag)) count++;
            }

            return count;
        }

        public int GetSelectedCount()
        {
            return selectedRVGEs.Count();
        }

        public void SetFirstTagged(Tag tag, int value)
        {
            if (value == TriState.indeterminate) return;
            bool tagValue = (value == TriState.active);
            GetFirstSelectedRVGE().SetTag(tag, tagValue);
        }


        public ResourceVersionGroupEntry GetNextRandomRVGE()
        {
            if (randomSelectedRVGEs.Count == 0)
            {
                randomSelectedRVGEs = selectedRVGEs.OrderBy(item => UnityEngine.Random.Range(0, 1f)).ToList();
                if (lastRandomRVGE!=null) randomSelectedRVGEs.Remove(lastRandomRVGE);
            }
            if (randomSelectedRVGEs.Count > 0)
            {
                ResourceVersionGroupEntry nextRVGE = randomSelectedRVGEs.First();

                lastRandomRVGE = nextRVGE;
                randomSelectedRVGEs.Remove(nextRVGE);
                return nextRVGE;
            }
            return null;
        }

        public void RefreshRVGESelections()
        {
            var rvges = new HashSet<ResourceVersionGroupEntry>(selectedRVGEs);
            foreach (ResourceVersionGroupEntry rvge in rvges)
            {
                string rvgeKey = rvge.rvgeKey;
                var newRVGE = ResourceManifest.GetRVGE(rvgeKey);

                UIDynamicButton button = null;

                if (selectedRVGEButtons.ContainsKey(rvge)) button = selectedRVGEButtons[rvge];

                DeSelectRVGE(rvge);
                if (newRVGE != null) SelectRVGE(newRVGE, button);
            }
            

        }
    }

}
