﻿
using Battlehub.RTCommon;
using MVR.FileManagementSecure;
using SimpleJSON;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Security.AccessControl;
using static TabbedUIBuilder;


namespace JayJayWon
{
    public class Tag
    {
        protected string _tagName;
        public JSONStorableEnumStringChooser tagCategoryJSEnum;
        protected int _resourceType;

        protected string _atomPresetResourceType;

        protected HashSet<CombiTag> _usedByCombiTags = new HashSet<CombiTag>();

        public JSONStorableBool hideTaggedResourcesJSB;
        public JSONStorableBool preferredTagJSB;


        protected void CopyFrom (Tag tag)
        {
            tagCategoryJSEnum.valNoCallback = tag.tagCategoryJSEnum.val;
            _tagName = tag._tagName;
            _atomPresetResourceType = tag._atomPresetResourceType;
            _resourceType = tag._resourceType;
            hideTaggedResourcesJSB.valNoCallback = tag.hideTaggedResourcesJSB.val;
            preferredTagJSB.valNoCallback = tag.preferredTagJSB.val;

            _usedByCombiTags.Clear();
            foreach(CombiTag combiTag in tag._usedByCombiTags)
            {
                _usedByCombiTags.Add(combiTag);
            }
        }
        public string tagName { get
            {
                return _tagName;
            }
        }

        public int tagCategory
        {
            get
            {
                return tagCategoryJSEnum.val;
            }
        }

        public int resourceType
        {
            get
            {
                return _resourceType;
            }
        }
        public int resourceCategory
        {
            get
            {
                return ResourceType.GetResourceCategory(resourceType);
            }
        }


        public string atomPresetResourceType
        {
            get
            {
                return _atomPresetResourceType;
            }
        }

        public void AddUsedBy(CombiTag combiTag)
        {
            _usedByCombiTags.Add(combiTag);
        }

        public void RemoveUsedBy(CombiTag combiTag)
        {
            _usedByCombiTags.Remove(combiTag);
        }

        public HashSet<CombiTag> usedByCombiTagsHieararchy
        {
            get
            {
                HashSet<CombiTag> recursiveUsedByCombiTags = new HashSet<CombiTag>(_usedByCombiTags);

                foreach (var usedCombiTag in _usedByCombiTags)
                {
                    recursiveUsedByCombiTags.UnionWith(usedCombiTag.usedByCombiTagsHieararchy);
                }

                return recursiveUsedByCombiTags;
            }
        }

        public HashSet<CombiTag> usedByCombiTags
        {
            get
            {
                return _usedByCombiTags;
            }
        }


        public Tag(string newTagName, int tagCategory, int newTagResourceType, string atomPresetResourceType)
        {
            tagCategoryJSEnum = new JSONStorableEnumStringChooser("tagCategory", TagCategory.enumManifestName, tagCategory, "Tag Category");
            hideTaggedResourcesJSB = new JSONStorableBool("hideTaggedResources", false, HideTaggedResourceCallback);
            preferredTagJSB = new JSONStorableBool("preferredTag", false, PreferredTagCallback);

            _tagName = newTagName;
            _resourceType = newTagResourceType;
            _atomPresetResourceType = atomPresetResourceType;
        }

        public void RenameTag(string newName)
        {
            _tagName = newName;
        }

        private void SaveRefreshOnChange()
        {
            Settings.SaveSettings();
            if (BrowserAssist.mainBrowserUI.browserFiltersUI.currentResourceType == resourceType) BrowserAssist.mainBrowserUI.RefreshFilterResults();
            if (BrowserAssist.callbackBrowserUI != null && BrowserAssist.callbackBrowserUI.browserFiltersUI.currentFilterSettings != null)
            {
                if (BrowserAssist.callbackBrowserUI.browserFiltersUI.currentResourceType == resourceType) BrowserAssist.callbackBrowserUI.RefreshFilterResults();
            }
        }

        protected void HideTaggedResourceCallback(bool hide)
        {
            SaveRefreshOnChange();
        }

        protected void PreferredTagCallback(bool hide)
        {
            SaveRefreshOnChange();
        }
    }

    public class CombiTag : Tag
    {
        public HashSet<Tag> usesTags
        {
            get
            {
                HashSet<Tag> result = new HashSet<Tag>();
                foreach (var condition in _combiLogicConditions)
                {
                    if (condition.isTaggableCondition) result.UnionWith(condition.usesTags);
                }
                return result;
            }
        }
        public bool usesNew
        {
            get
            {
                foreach (var condition in _combiLogicConditions)
                {
                    if (condition.usesNew) return true;
                }
                return false;
            }
        }

        public bool usesHidden
        {
            get
            {
                foreach (var condition in _combiLogicConditions)
                {
                    if (condition.usesHidden) return true;
                }
                return false;
            }
        }

        public bool usesFavorite
        {
            get
            {
                foreach (var condition in _combiLogicConditions)
                {
                    if (condition.usesFavorite) return true;
                }
                return false;
            }
        }

        public JSONStorableEnumStringChooser logicOperandJSEnum;

        protected bool _modified = false;
        public bool isModfied { get { return _modified; }
            set { _modified = value; } }

        protected List<CombiLogicCondition> _combiLogicConditions = new List<CombiLogicCondition>();

        public IEnumerable<CombiLogicCondition> combiLogicConditions { get { return _combiLogicConditions.Select(o => o); } }

        public void AddCombiLogicCondition()
        {
            _combiLogicConditions.Add(new CombiLogicCondition(this));
            _modified=true;
        }

        public void RemoveCombiLogicConditions(CombiLogicCondition clc)
        {
            _combiLogicConditions.Remove(clc);
            _modified=true;
        }

        public CombiTag (CombiTag combiTag) : base (combiTag.tagName, TagCategory.combi, combiTag.resourceType, combiTag.atomPresetResourceType )
        {
            logicOperandJSEnum = new JSONStorableEnumStringChooser("logicOperand", LogicOperands.enumManifestName, LogicOperands.and, "", LogicOperandCallback);
            CopyFrom(combiTag);
        }

        public CombiTag(string newTagName, int newTagResourceType, string atomPresetResourceType) : base(newTagName, TagCategory.combi, newTagResourceType, atomPresetResourceType)
        {
            logicOperandJSEnum = new JSONStorableEnumStringChooser("logicOperand", LogicOperands.enumManifestName, LogicOperands.and, "",LogicOperandCallback);

            AddCombiLogicCondition();
        }

        public CombiTag(int newTagResourceType, string atomPresetResourceType, JSONClass jc) : base("", TagCategory.combi, newTagResourceType, atomPresetResourceType)
        {
            logicOperandJSEnum = new JSONStorableEnumStringChooser("logicOperand", LogicOperands.enumManifestName, LogicOperands.and, "", LogicOperandCallback);

            _tagName = jc["tagName"];
            logicOperandJSEnum.RestoreFromJSON(jc);

            JSONArray clcJA = jc["combiLogicCondition"].AsArray;

            foreach (JSONClass clcJC  in clcJA)
            {
                _combiLogicConditions.Add(new CombiLogicCondition(this, clcJC));
            }

        }

        public void Init()
        {
            foreach (var clc in _combiLogicConditions)
            {
                clc.Init();
            }
        }

        public JSONClass GetJSON()
        {
            var jc = new JSONClass();
            jc["tagName"] = tagName;
            logicOperandJSEnum.StoreJSON(jc,forceStore:true);

            var clcJA = new JSONArray();

            foreach (var clc in combiLogicConditions)
            {
                clcJA.Add(clc.GetJSON());
            }

            jc["combiLogicCondition"] = clcJA;

            return jc;
        }

        public void LogicOperandCallback(int val)
        {
            isModfied = true;
        }

        public void CopyFrom(CombiTag sourceTag)
        {
            base.CopyFrom(sourceTag);
            logicOperandJSEnum.val = sourceTag.logicOperandJSEnum.val;
            _combiLogicConditions.Clear();

            foreach (var combiLogicCondition in sourceTag.combiLogicConditions)
            {
                _combiLogicConditions.Add(new CombiLogicCondition(this, combiLogicCondition));
            }

            _modified = false; 
        }

        public bool GetState(ResourceVersionGroupEntry rvge)
        {
            int validCLCcount = 0;
            foreach(CombiLogicCondition combiLogicCondition in _combiLogicConditions)
            {
                if (combiLogicCondition.isValidCondition)
                {
                    bool clcState = combiLogicCondition.GetState(rvge);
                    validCLCcount++;
                    if (logicOperandJSEnum.val == LogicOperands.or && clcState) return true;
                    if (logicOperandJSEnum.val == LogicOperands.and && !clcState) return false;
                }


            }

            if (validCLCcount==0) return true;

            return logicOperandJSEnum.val == LogicOperands.and;
        }

        public string description
        {
            get
            {
                string descriptionBuild = "";
                int validCLCcount = 0;
                foreach (CombiLogicCondition combiLogicCondition in _combiLogicConditions)
                {
                    if (combiLogicCondition.isValidCondition)
                    {
                        validCLCcount++;
                        if (validCLCcount > 1)
                        {
                            if (logicOperandJSEnum.val == LogicOperands.or) descriptionBuild += " or";
                            if (logicOperandJSEnum.val == LogicOperands.and) descriptionBuild += " and";
                        }
                        descriptionBuild += combiLogicCondition.description;
                    }
                }

                if (validCLCcount == 0) return "Select Tag Filter conditions...";

                return descriptionBuild;
            }

        }


        public void DeleteTagUsage(Tag tag)
        {
            foreach (var clc in combiLogicConditions)
            {
                clc.DeleteTagUsage(tag);
            }
        }
    }

    public class CombiLogicCondition
    {
        protected HashSet<Tag> _usesTags = new HashSet<Tag>();

        private HashSet<string> usesTagsToInit = new HashSet<string>();

        protected HashSet<int> _varLicenses = new HashSet<int>();
        protected HashSet<string> _varCreators = new HashSet<string>();
        protected HashSet<string> _varFolders = new HashSet<string>();
        protected HashSet<string> _localFolders = new HashSet<string>();
        protected int _day = DateTime.Now.Day;
        protected int _month = DateTime.Now.Month;
        protected int _year = DateTime.Now.Year;

        public ReadOnlyCollection<Tag> usesTags
        {
            get
            {
                return _usesTags.ToList().AsReadOnly();
            }
        }

        public ReadOnlyCollection<int> varLicenses
        {
            get
            {
                return _varLicenses.ToList().AsReadOnly();
            }
        }

        public ReadOnlyCollection<string> varCreators
        {
            get
            {
                return _varCreators.ToList().AsReadOnly();
            }
        }

        public ReadOnlyCollection<string> varFolders
        {
            get
            {
                return _varFolders.ToList().AsReadOnly();
            }
        }

        public ReadOnlyCollection<string> localFolders
        {
            get
            {
                return _localFolders.ToList().AsReadOnly();
            }
        }

        public int dateDay
        {
            get { return _day; }
            set
            {
                if (DateTime.DaysInMonth(_year, _month) >= value)
                {
                    _day = value;
                    parentCombiTag.isModfied = true;
                }
                else SuperController.LogError("BA.CombiLogicCondition.dateDay: Attempt to set invalid Day for current month and year.");
            }
        }

        public int dateMonth
        {
            get { return _month; }
            set
            {
                _month = value;
                if (_day > DateTime.DaysInMonth(_year, _month)) _day = DateTime.DaysInMonth(_year, _month);
                parentCombiTag.isModfied = true;
            }
        }

        public int dateYear
        {
            get { return _year; }
            set {
                _year = value;
                if (_day > DateTime.DaysInMonth(_year, _month)) _day = DateTime.DaysInMonth(_year, _month);
                parentCombiTag.isModfied = true;
            }
        }

        public bool usesHidden
        {
            get
            {
                return (logicConditionJSEnum.val == CombiLogicConditionOptions.hidden || logicConditionJSEnum.val == CombiLogicConditionOptions.notHidden);
            }
        }

        public bool usesFavorite
        {
            get
            {
                return (logicConditionJSEnum.val == CombiLogicConditionOptions.favorite || logicConditionJSEnum.val == CombiLogicConditionOptions.notFavorite);
            }
        }

        public bool usesNew
        {
            get
            {
                return (logicConditionJSEnum.val == CombiLogicConditionOptions.New || logicConditionJSEnum.val == CombiLogicConditionOptions.notNew);
            }
        }

        public CombiTag parentCombiTag { get; protected set; }
        public JSONStorableEnumStringChooser logicConditionJSEnum;

        public bool isMultiConfigurableCondition { get {
                return (isTaggableCondition || isLicenseCondition || isCreatorCondition || isVARFolderCondition || isLocalFolderCondition || isDateCondition);
            }
        }

        public bool isMultiConfigurationFilterable
        {
            get
            {
                return (isTaggableCondition || isCreatorCondition || isVARFolderCondition || isLocalFolderCondition);
            }
        }


        public bool isTaggableCondition { get 
            {
                return (logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWith || logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAll || logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAny || logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAnyBodyRegion || logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAnyClothingType || logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAnyUserTag ||  logicConditionJSEnum.val == CombiLogicConditionOptions.notTaggedWithAny);
            }
        }

        public bool isLicenseCondition
        {
            get
            {
                return (logicConditionJSEnum.val == CombiLogicConditionOptions.varLicense || logicConditionJSEnum.val == CombiLogicConditionOptions.notVarLicense);
            }
        }

        public bool isCreatorCondition
        {
            get
            {
                return (logicConditionJSEnum.val == CombiLogicConditionOptions.notPackagedByCreator || logicConditionJSEnum.val == CombiLogicConditionOptions.packagedByCreator);
            }
        }

        public bool isVARFolderCondition
        {
            get
            {
                return (logicConditionJSEnum.val == CombiLogicConditionOptions.varInFolder || logicConditionJSEnum.val == CombiLogicConditionOptions.varNotInFolder || logicConditionJSEnum.val == CombiLogicConditionOptions.varUnderFolder || logicConditionJSEnum.val == CombiLogicConditionOptions.varNotUnderFolder);
            }
        }

        public bool isLocalFolderCondition
        {
            get
            {
                return (logicConditionJSEnum.val == CombiLogicConditionOptions.inLocalFolder || logicConditionJSEnum.val == CombiLogicConditionOptions.notInLocalFolder || logicConditionJSEnum.val == CombiLogicConditionOptions.underLocalFolder || logicConditionJSEnum.val == CombiLogicConditionOptions.notUnderLocalFolder);
            }
        }

        public bool isDateCondition
        {
            get
            {
                return (logicConditionJSEnum.val == CombiLogicConditionOptions.createdAfter || logicConditionJSEnum.val == CombiLogicConditionOptions.createdBefore || logicConditionJSEnum.val == CombiLogicConditionOptions.updatedBefore || logicConditionJSEnum.val == CombiLogicConditionOptions.updatedAfter);
            }
        }

        public string GetTagNameList()
        {
            string usesTagsNameList="";

            if (isTaggableCondition)
            {
                foreach (Tag tag in usesTags)
                {
                    string tagName = tag.tagName;
                    if (tag.tagCategory == TagCategory.combi) tagName = "<i>" + tag.tagName + "</i>";
                    if (usesTagsNameList == "") usesTagsNameList = tagName;
                    else usesTagsNameList = usesTagsNameList + ", " + tagName;
                }
            }
            return usesTagsNameList;
        }

        public string GetLicenseTypeNameList()
        {
            string licenseTypeNameList = "";

            if (isLicenseCondition)
            {
                foreach (int licenseType in _varLicenses) 
                {
                    string licenseName = EnumManifest.GetDisplayVal(LicenseType.enumManifestName, licenseType);
                    if (licenseTypeNameList == "") licenseTypeNameList = licenseName;
                    else licenseTypeNameList = licenseTypeNameList + ", " + licenseName;
                }
            }
            return licenseTypeNameList;
        }

        public string GetCreatorNameList()
        {
            string creatorNameList = "";

            if (isCreatorCondition)
            {
                foreach (string creatorName in _varCreators)
                {
                    if (creatorNameList == "") creatorNameList = creatorName;
                    else creatorNameList = creatorNameList + ", " + creatorName;
                }
            }
            return creatorNameList;
        }

        public string GetVARFolderListDescription()
        {
            string varFolderNameList = "";

            if (isVARFolderCondition)
            {
                foreach (string varFolderName in _varFolders)
                {
                    string folderName = varFolderName;
                    if (folderName != "AddonPackages") folderName = varFolderName.Substring(13);
                    if (varFolderNameList == "") varFolderNameList = folderName;
                    else varFolderNameList = varFolderNameList + ", " + folderName;
                }
            }
            return varFolderNameList;
        }

        public string GetLocalFolderList()
        {
            string localFolderNameList = "";

            if (isLocalFolderCondition)
            {
                foreach (string localFolderName in _localFolders)
                {
                    if (localFolderNameList == "") localFolderNameList = localFolderName;
                    else localFolderNameList = localFolderNameList + ", " + localFolderName;
                }

            }
            return localFolderNameList;
        }

        public string GetDateLabel()
        {
            DateTime date = new DateTime(_year,_month,_day);
            return date.ToLongDateString();
        }

        public CombiLogicCondition(CombiTag parentCombiTag)
        {
            this.parentCombiTag = parentCombiTag;
            List<int> exlcusions = new List<int>();
            exlcusions.Add(CombiLogicConditionOptions.everything);
            if (parentCombiTag.resourceCategory != ResourceCategory.clothing)
            {
                exlcusions.Add(CombiLogicConditionOptions.taggedWithAnyClothingType);
                exlcusions.Add(CombiLogicConditionOptions.taggedWithAnyBodyRegion);
                exlcusions.Add(CombiLogicConditionOptions.taggedWithAnyUserTag);
            }
            else
            {
                exlcusions.Add(CombiLogicConditionOptions.inLocalFolder);
                exlcusions.Add(CombiLogicConditionOptions.notInLocalFolder);
                exlcusions.Add(CombiLogicConditionOptions.underLocalFolder);
                exlcusions.Add(CombiLogicConditionOptions.notUnderLocalFolder);
            }
            logicConditionJSEnum = new JSONStorableEnumStringChooser("logicCondition", CombiLogicConditionOptions.enumManifestName, CombiLogicConditionOptions.taggedWith, "", LogicConditionCallback, exlcusions);
        }

        public CombiLogicCondition(CombiTag parentCombiTag, JSONClass jc): this (parentCombiTag)
        {
            logicConditionJSEnum.valNoCallback = EnumManifest.GetEnumValFromStoreVal(CombiLogicConditionOptions.enumManifestName, jc["logicCondition"]);

            if (isTaggableCondition)
            {
                var usesTagsJA = jc["usesTags"].AsArray;
                foreach (JSONClass usesTagJC in usesTagsJA)
                {
                    int tagCategory = EnumManifest.GetEnumValFromStoreVal(TagCategory.enumManifestName, usesTagJC["tagCategory"]);
                    string tagName = usesTagJC["tagName"];

                    if (tagCategory != TagCategory.combi && TagManifest.GetTagNameCount(parentCombiTag.resourceType, parentCombiTag.atomPresetResourceType, tagName) > 0)
                    {
                        Tag tag = TagManifest.GetTag(parentCombiTag.resourceType, parentCombiTag.atomPresetResourceType, tagName);
                        UseTag(tag);
                        tag.AddUsedBy(parentCombiTag);
                        
                    }
                    else  usesTagsToInit.Add(tagName);

                }
            }

            if (isLicenseCondition)
            {
                var licensesJA = jc["varLicenses"].AsArray;
                foreach (JSONNode licenseTypeJN in licensesJA) _varLicenses.Add(EnumManifest.GetEnumValFromStoreVal(LicenseType.enumManifestName, licenseTypeJN));

            }
            if (isCreatorCondition)
            {
                var creators = CreatorManifest.GetAllCreatorNames(parentCombiTag.resourceType, parentCombiTag.atomPresetResourceType);
                var creatorsJA = jc["varCreators"].AsArray;
                foreach (JSONNode creatorJN in creatorsJA)
                {
                    string creatorName = creatorJN;
//                    if (creators.Contains(creatorName)) _varCreators.Add(creatorName);
                    _varCreators.Add(creatorName);
                }
            }
            if (isVARFolderCondition)
            {
                var varFoldersJA = jc["varFolders"].AsArray;
                foreach (JSONNode folderJN in varFoldersJA)
                {
                    _varFolders.Add(folderJN);
                }
            }

            if (isLocalFolderCondition)
            {
                var localFoldersJA = jc["localFolders"].AsArray;
                foreach (JSONNode folderJN in localFoldersJA)
                {
                    _localFolders.Add(folderJN);
                }
            }

            if (isDateCondition)
            {
                _year = jc["dateYear"].AsInt;
                _month = jc["dateMonth"].AsInt;
                _day = jc["dateDay"].AsInt;
            }
        }

        public void Init()
        {
            foreach (string usesTagName in usesTagsToInit)
            {
                if (TagManifest.GetTagNameCount(parentCombiTag.resourceType, parentCombiTag.atomPresetResourceType, usesTagName) >0)
                {
                    Tag tag = TagManifest.GetTag(parentCombiTag.resourceType, parentCombiTag.atomPresetResourceType, usesTagName);
                    UseTag(tag);
                    tag.AddUsedBy(parentCombiTag);
                }
                
            }
            usesTagsToInit.Clear();
        }

        public bool isValidCondition
        {
            get
            {
                if (isTaggableCondition && _usesTags.Count == 0) return false;
                if (isLicenseCondition && _varLicenses.Count==0) return false;
                if (isCreatorCondition && _varCreators.Count == 0) return false;
                if (isVARFolderCondition && _varFolders.Count == 0) return false;
                if (isLocalFolderCondition && _localFolders.Count == 0) return false;

                return true;
            }
        }

        public string description
        {
            get
            {
                string descriptionBuild="";
                string usesDescription = " (";
                if (isTaggableCondition && _usesTags.Count == 0) return "";
                if (isLicenseCondition && _varLicenses.Count == 0) return "";
                if (isTaggableCondition && _usesTags.Count > 0) usesDescription += GetTagNameList();
                if (isLicenseCondition && _varLicenses.Count > 0) usesDescription += GetLicenseTypeNameList();
                if (isCreatorCondition && _varCreators.Count > 0) usesDescription += GetCreatorNameList();
                if (isVARFolderCondition && _varFolders.Count > 0) usesDescription += GetVARFolderListDescription();
                if (isLocalFolderCondition && _localFolders.Count > 0) usesDescription += GetLocalFolderList();

                usesDescription += ")";

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWith)
                    descriptionBuild = " tagged with" + usesDescription;

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAll) descriptionBuild = " tagged with all of"+usesDescription;

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAny) descriptionBuild = " tagged with any of" + usesDescription;

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAnyBodyRegion) descriptionBuild = " on body regions" + usesDescription;

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAnyClothingType) descriptionBuild = " of clothing types" + usesDescription;

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAnyUserTag) descriptionBuild = " tagged with any of" + usesDescription;

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.notTagged) descriptionBuild = " not tagged";

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.notTaggedWithAny) descriptionBuild = " not tagged with any of" + usesDescription; ;

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAnything) descriptionBuild = " tagged with anything"; ;

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.hidden) descriptionBuild = " hidden";

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.notHidden) descriptionBuild = " not hidden";

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.favorite) descriptionBuild = " favorite";

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.notFavorite) descriptionBuild = " not favorite";

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.New) descriptionBuild = " new";

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.notNew) descriptionBuild = " not new";

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.varLicense) descriptionBuild = " in VAR licensed with";
                if (logicConditionJSEnum.val == CombiLogicConditionOptions.notVarLicense) descriptionBuild = " in VAR not licensed with";

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.packagedByCreator) descriptionBuild = " package by Creators";
                if (logicConditionJSEnum.val == CombiLogicConditionOptions.notPackagedByCreator) descriptionBuild = " packaged but not by Creators";

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.inLocalFolder) descriptionBuild = " in local folders";
                if (logicConditionJSEnum.val == CombiLogicConditionOptions.notInLocalFolder) descriptionBuild = " not in local folders";
                if (logicConditionJSEnum.val == CombiLogicConditionOptions.underLocalFolder) descriptionBuild = " under local folder";
                if (logicConditionJSEnum.val == CombiLogicConditionOptions.notUnderLocalFolder) descriptionBuild = " not under local folder";

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.varInFolder) descriptionBuild = " within VAR in folders";
                if (logicConditionJSEnum.val == CombiLogicConditionOptions.varNotInFolder) descriptionBuild = " within VAR not in folders";
                if (logicConditionJSEnum.val == CombiLogicConditionOptions.varUnderFolder) descriptionBuild = " within VAR under folder";
                if (logicConditionJSEnum.val == CombiLogicConditionOptions.varNotUnderFolder) descriptionBuild = " within VAR not under folder";

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.createdBefore) descriptionBuild = " created before";
                if (logicConditionJSEnum.val == CombiLogicConditionOptions.createdAfter) descriptionBuild = " created after";
                if (logicConditionJSEnum.val == CombiLogicConditionOptions.updatedBefore) descriptionBuild = " updated before";
                if (logicConditionJSEnum.val == CombiLogicConditionOptions.updatedAfter) descriptionBuild = " updated after";

                return descriptionBuild;
            }
        }

        public bool GetState(ResourceVersionGroupEntry rvge)
        {
            if (isTaggableCondition && _usesTags.Count==0) return false;
            if (isLicenseCondition && _varLicenses.Count == 0) return false;
            if (isCreatorCondition && _varCreators.Count == 0) return false;
            if (isVARFolderCondition && _varFolders.Count == 0) return false;
            if (isLocalFolderCondition && _localFolders.Count == 0) return false;

            var rvgeTags = new HashSet<Tag>(TagManifest.GetRVGETags(rvge));
            rvgeTags.Remove(parentCombiTag);        

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWith)  return (rvgeTags.Contains(_usesTags.First()));

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAll) return (_usesTags.All(s => rvgeTags.Contains(s)));

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAny || logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAnyBodyRegion ||logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAnyClothingType || logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAnyUserTag) return (_usesTags.Intersect(rvgeTags).Count() >0);

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.notTagged) return (rvgeTags.Count() ==0);

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.notTaggedWithAny) return (_usesTags.Intersect(rvgeTags).Count() == 0);

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAnything) return (rvgeTags.Count() > 0);

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.hidden) return (rvge.isHidden || rvge.isTagHidden || rvge.isCreatorHidden);

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.notHidden) return (!rvge.isHidden && !rvge.isTagHidden && !rvge.isCreatorHidden);

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.favorite) return (rvge.isFavorite);

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.notFavorite) return (!rvge.isFavorite);

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.New) return (rvge.isNew);

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.notNew) return (!rvge.isNew);

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.varLicense) return !rvge.localFileResource && _varLicenses.Contains((rvge as PackagedRVGE).varLicenseType);

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.notVarLicense) return !rvge.localFileResource && !_varLicenses.Contains((rvge as PackagedRVGE).varLicenseType);

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.packagedByCreator) return !rvge.localFileResource && _varCreators.Contains(rvge.creatorName);
            if (logicConditionJSEnum.val == CombiLogicConditionOptions.notPackagedByCreator) return !rvge.localFileResource && !_varCreators.Contains(rvge.creatorName);

            if (isLocalFolderCondition && rvge.localFileResource)
            {
                if (logicConditionJSEnum.val == CombiLogicConditionOptions.inLocalFolder) return _localFolders.Contains(rvge.folderPath);
                if (logicConditionJSEnum.val == CombiLogicConditionOptions.notInLocalFolder) return !_localFolders.Contains(rvge.folderPath);
                if (logicConditionJSEnum.val == CombiLogicConditionOptions.underLocalFolder && _localFolders.Count > 0) return rvge.folderPath.StartsWith(_localFolders.First());
                if (logicConditionJSEnum.val == CombiLogicConditionOptions.notUnderLocalFolder && _localFolders.Count > 0) return !rvge.folderPath.StartsWith(_localFolders.First());
            }

            if (isVARFolderCondition && !rvge.localFileResource)
            {
                PackagedRVGE packagedRVGE = (rvge as PackagedRVGE);
                string varFolder = FileManagerSecure.GetDirectoryName(packagedRVGE.vpvge.GetVPME(packagedRVGE.highestVARVersion).varFilePath);
                if (logicConditionJSEnum.val == CombiLogicConditionOptions.varInFolder) return _varFolders.Contains(varFolder);
                if (logicConditionJSEnum.val == CombiLogicConditionOptions.varNotInFolder) return !_varFolders.Contains(varFolder);
                if (logicConditionJSEnum.val == CombiLogicConditionOptions.varUnderFolder && _varFolders.Count>0) return varFolder.StartsWith(_varFolders.First());
                if (logicConditionJSEnum.val == CombiLogicConditionOptions.varNotUnderFolder && _varFolders.Count > 0) return !varFolder.StartsWith(_varFolders.First());
            }

            if (isDateCondition)
            {
                DateTime dateTime = new DateTime(_year, _month, _day);
                DateTime createdDate = new DateTime(rvge.firstCataloguedDateTime.Year, rvge.firstCataloguedDateTime.Month, rvge.firstCataloguedDateTime.Day);
                DateTime updatedDate = new DateTime(rvge.lastCataloguedDateTime.Year, rvge.lastCataloguedDateTime.Month, rvge.lastCataloguedDateTime.Day);

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.createdBefore) return createdDate < dateTime;
                if (logicConditionJSEnum.val == CombiLogicConditionOptions.createdAfter) return createdDate > dateTime;
                if (logicConditionJSEnum.val == CombiLogicConditionOptions.updatedBefore) return updatedDate < dateTime;
                if (logicConditionJSEnum.val == CombiLogicConditionOptions.updatedAfter) return updatedDate > dateTime;
            }
            

            return false;
        }

        public JSONClass GetJSON()
        {
            var jc = new JSONClass();
            jc["logicCondition"] = logicConditionJSEnum.storeVal;

            if (isTaggableCondition)
            {
                var usesTagsJA = new JSONArray();
                foreach (Tag usesTag in _usesTags)
                {
                    var usesTagJC = new JSONClass();
                    usesTagJC["tagCategory"] = usesTag.tagCategoryJSEnum.storeVal;
                    usesTagJC["tagName"] = usesTag.tagName;

                    usesTagsJA.Add(usesTagJC);
                }
                jc["usesTags"] = usesTagsJA;

            }

            if (isLicenseCondition)
            {
                var licensesJA = new JSONArray();

                foreach (int licenseType in _varLicenses) licensesJA[-1] = EnumManifest.GetStoreVal(LicenseType.enumManifestName,licenseType);
                jc["varLicenses"] = licensesJA;
            }

            if (isCreatorCondition)
            {
                var creatorsJA = new JSONArray();

                foreach (string creatorName in _varCreators) creatorsJA[-1] = creatorName;
                jc["varCreators"] = creatorsJA;
            }

            if (isVARFolderCondition)
            {
                var varFoldersJA = new JSONArray();

                foreach (string varFolderName in _varFolders) varFoldersJA[-1] = varFolderName;
                jc["varFolders"] = varFoldersJA;
            }

            if (isLocalFolderCondition)
            {
                var localFoldersJA = new JSONArray();

                foreach (string localFolderName in _localFolders) localFoldersJA[-1] = localFolderName;
                jc["localFolders"] = localFoldersJA;
            }

            if (isDateCondition)
            {
                jc["dateDay"].AsInt = _day;
                jc["dateMonth"].AsInt = _month;
                jc["dateYear"].AsInt = _year;
            }


            return jc;
        }
        public CombiLogicCondition(CombiTag parentCombiTag, CombiLogicCondition copyFrom) : this (parentCombiTag)
        {
            logicConditionJSEnum.valNoCallback = copyFrom.logicConditionJSEnum.val;
            foreach (Tag tag in copyFrom._usesTags)
            {
                _usesTags.Add(tag);
            }

            foreach (int licenseType in copyFrom._varLicenses)
            {
                _varLicenses.Add(licenseType);
            }

            foreach (string creatorName in copyFrom._varCreators)
            {
                _varCreators.Add(creatorName);
            }

            foreach (string varFolderName in copyFrom._varFolders)
            {
                _varFolders.Add(varFolderName);
            }

            foreach (string localFolderName in copyFrom._localFolders)
            {
                _localFolders.Add(localFolderName);
            }

            _day = copyFrom._day;
            _month = copyFrom._month;
            _year = copyFrom._year;
        }
        private void LogicConditionCallback(int value)
        {
            parentCombiTag.isModfied = true;
        }

        public void UseOnlyOneTag()
        {
            if (_usesTags.Count > 1)
            {
                Tag keepTag = _usesTags.First();
                _usesTags.Clear();
                UseTag(keepTag);
            }
        }

        public void UseTag(Tag tag, bool replace=false)
        {
            if (replace)  _usesTags.Clear();
            
            _usesTags.Add(tag);
            parentCombiTag.isModfied = true;
        }

        public void RemoveUseTag(Tag tag)
        {
            _usesTags.Remove(tag);
            parentCombiTag.isModfied = true;
        }
        public void DeleteTagUsage(Tag tag)
        {
            _usesTags.Remove(tag);
        }

        public void UseLicense(int licenseType)
        {
            _varLicenses.Add(licenseType);
            parentCombiTag.isModfied = true;
        }

        public void RemoveUseLicense(int licenseType)
        {
            _varLicenses.Remove(licenseType);
            parentCombiTag.isModfied = true;
        }

        public void UseVARCreator(string creatorName)
        {
            _varCreators.Add(creatorName);
            parentCombiTag.isModfied = true;
        }

        public void RemoveUseVARCreator(string creatorName)
        {
            _varCreators.Remove(creatorName);
            parentCombiTag.isModfied = true;
        }

        public void UseVARFolder(string folderName, bool replace = false)
        {
            if (replace) _varFolders.Clear();
            _varFolders.Add(folderName);
            parentCombiTag.isModfied = true;
        }

        public void RemoveUseVARFolder(string folderName)
        {
            _varFolders.Remove(folderName);
            parentCombiTag.isModfied = true;
        }

        public void UseOnlyOneVARFolder()
        {
            if (_varFolders.Count > 1)
            {
                string keepFolder = _varFolders.First();
                _varFolders.Clear();
                UseVARFolder(keepFolder);
            }
        }

        public void UseLocalFolder(string folderName, bool replace = false)
        {
            if (replace) _localFolders.Clear();
            _localFolders.Add(folderName);
            parentCombiTag.isModfied = true;
        }

        public void RemoveUseLocalFolder(string folderName)
        {
            _localFolders.Remove(folderName);
            parentCombiTag.isModfied = true;
        }

        public void UseOnlyOneLocalFolder()
        {
            if (_localFolders.Count > 1)
            {
                string keepFolder = _localFolders.First();
                _localFolders.Clear();
                UseLocalFolder(keepFolder);
            }
        }

    }


    public static class TagManifest
    {
        private static HashSet<Tag> tags = new HashSet<Tag>();

        private static Dictionary<ResourceVersionGroupEntry, HashSet<Tag>> rvgeTags = new Dictionary<ResourceVersionGroupEntry, HashSet<Tag>>();
        private static Dictionary<Tag, HashSet<ResourceVersionGroupEntry>> tagsRVGEs = new Dictionary<Tag, HashSet<ResourceVersionGroupEntry>>();

        private static Dictionary<int, Dictionary<string, Tag>> tagNamesToTagByResourceType = new Dictionary<int, Dictionary<string, Tag>>();

        private static Dictionary<string, Dictionary<string, Tag>> tagNamesToTagByAtomPresetResourceType = new Dictionary<string, Dictionary<string, Tag>>();

        private static Dictionary<int, HashSet<Tag>> categoryTagMap = new Dictionary<int, HashSet<Tag>>();

        private static Dictionary<int, HashSet<Tag>> resourceTypeTagMap = new Dictionary<int, HashSet<Tag>>();
        private static Dictionary<string, HashSet<Tag>> atomPresetResourceTypeTagMap = new Dictionary<string, HashSet<Tag>>();

        private static Dictionary<int, HashSet<CombiTag>> resourceTypeCombiTagMap = new Dictionary<int, HashSet<CombiTag>>();
        private static Dictionary<string, HashSet<CombiTag>> atomPresetResourceTypeCombiTagMap = new Dictionary<string, HashSet<CombiTag>>();

        private static Dictionary<int, Dictionary<int,HashSet<ResourceVersionGroupEntry>>> tagCategoryToResourceByResourceType = new Dictionary<int, Dictionary<int, HashSet<ResourceVersionGroupEntry>>> ();

        private static Dictionary<string, Dictionary<int, HashSet<ResourceVersionGroupEntry>>> tagCategoryToResourceByAtomPresetResourceType = new Dictionary<string, Dictionary<int, HashSet<ResourceVersionGroupEntry>>>();

        static TagManifest()
        {
            foreach (int resourceType in ResourceType.GetResourceTypesByTargetCategory(ResourceTargetCategory.personAtomTargetable))
            {
                CreateTag("male", TagCategory.gender, resourceType,"Person");
                CreateTag("female", TagCategory.gender, resourceType,"Person");
            }
            CreateTag("session", TagCategory.pluginType, ResourceType.plugins,"");
            CreateTag("scene", TagCategory.pluginType, ResourceType.plugins,"");

            foreach (int resourceType in ResourceType.GetResourceTypesByCategory(ResourceCategory.clothing))
            {
                foreach (string tagName in Const.BodyRegionTagNames)
                {
                    CreateTag(tagName, TagCategory.bodyRegion, resourceType, "Person");
                }

                if (resourceType == ResourceType.femaleClothingItems)
                {
                    foreach (string tagName in Const.FemaleClothingTypeTagNames)
                    {
                        CreateTag(tagName, TagCategory.clothingType, resourceType, "Person");
                    }
                }
                if (resourceType == ResourceType.maleClothingItems)
                {
                    foreach (string tagName in Const.MaleClothingTypeTagNames)
                    {
                        CreateTag(tagName, TagCategory.clothingType, resourceType, "Person");
                    }
                }

            }
        }

        public static int GetTaggedResourcesCount(Tag tag)
        {
            return tagsRVGEs[tag].Count();
        }
        public static string NormalizeTagName(string tagName)
        {
            string tagNameMod = tagName.ToLower();
            if (tagName.Length > 0)
            {
                char lastCharacter = tagNameMod.Last();
                if (tagNameMod.Length > 25 || tagNameMod.EndsWith("  ")|| lastCharacter==',' )
                {
                    string tagNameLessOneChar = tagName.Remove(tagName.Length - 1, 1);
                    tagNameMod = tagNameLessOneChar.ToLower();

                }
            }

            return tagNameMod;
        }
/*
        private static HashSet<CombiTag> assesedDependentCombiTabs = new HashSet<CombiTag>();
        public static void AssessCombiTagState(ResourceVersionGroupEntry rvge, CombiTag combiTag)
        {
            
            assesedDependentCombiTabs.Clear();

            bool combiTagState = GetCombiTagState(rvge, combiTag);
            rvge.SetTag(combiTag, combiTagState);
            assesedDependentCombiTabs.Clear();
        }

        private static bool GetCombiTagState(ResourceVersionGroupEntry rvge, CombiTag combiTag)
        {
            var dependentCombiTags = combiTag.usesCombiTagsHierarchy;
            foreach( var dependentCombiTag in dependentCombiTags)
            {
                if (!assesedDependentCombiTabs.Contains(dependentCombiTag))
                {
                    bool dependentCombiTagState = GetCombiTagState(rvge, dependentCombiTag);
                    rvge.SetTag(dependentCombiTag, dependentCombiTagState);
                    assesedDependentCombiTabs.Add(dependentCombiTag);
                }
            }

            return combiTag.GetState(rvge);

        }
*/
        public static bool RenameTag(Tag tag, string newName)
        {
            if (newName != "hide" && newName != "favorite" && !newName.EndsWith(" ") && newName != "" && !GetTags(tag.resourceType, tag.atomPresetResourceType).Select(x => x.tagName).Contains(newName))
            {
                
                tagNamesToTagByResourceType[tag.resourceType].Remove(tag.tagName);
                tagNamesToTagByResourceType[tag.resourceType].Add(newName, tag);
                if (tag.resourceType == ResourceType.presetAtom)
                {
                    tagNamesToTagByAtomPresetResourceType[tag.atomPresetResourceType].Remove(tag.tagName);
                    tagNamesToTagByAtomPresetResourceType[tag.atomPresetResourceType].Add(newName,tag);
                }

                tag.RenameTag(newName);

                if (tagsRVGEs.ContainsKey(tag))
                {
                    foreach (var rvge in tagsRVGEs[tag])
                    {
                        rvge.SetUserDataPersist();
                    }
                }

                if (tag.usedByCombiTags.Count > 0) Settings.SaveSettings();

                return true;
            }
            return false;
        }

        public static int GetTagNameCount(int resourceType, string atomPresetResourceType, string tagName)
        {
            if (resourceType == ResourceType.presetAtom)
            {
                if (!tagNamesToTagByAtomPresetResourceType.ContainsKey(atomPresetResourceType)) return 0;
                return tagNamesToTagByAtomPresetResourceType[atomPresetResourceType].Where(x => x.Key == tagName).ToList().Count();
            }

            return tagNamesToTagByResourceType[resourceType].Where(x => x.Key == tagName).ToList().Count();
        }

        public static CombiTag CreateCombiTag(string name, int resourceType, string atomPresetResourceType)
        {
            if (tags.Any(t => t.tagName == name && t.resourceType == resourceType)) return null;

            CombiTag newTag = new CombiTag(name, resourceType, atomPresetResourceType);
            RegisterNewTag(newTag);

            if (!resourceTypeCombiTagMap.ContainsKey(resourceType)) resourceTypeCombiTagMap.Add(resourceType, new HashSet<CombiTag>());
            resourceTypeCombiTagMap[resourceType].Add(newTag);

            return newTag;
        }

        public static JSONArray GetCombiTagsJSON(int resourceType, string atomPresetResourceType)
        {
            var combiTagsJA = new JSONArray();

            if (resourceType != ResourceType.presetAtom)
            {
                if (resourceTypeCombiTagMap.ContainsKey(resourceType))
                {
                    foreach (var combiTag in resourceTypeCombiTagMap[resourceType])
                    {
                        var combiTagJC = combiTag.GetJSON();
                        combiTagsJA.Add(combiTagJC);
                    }
                }
            }
            else
            {
                if (atomPresetResourceTypeCombiTagMap.ContainsKey(atomPresetResourceType))
                {
                    foreach (var combiTag in atomPresetResourceTypeCombiTagMap[atomPresetResourceType])
                    {
                        var combiTagJC = combiTag.GetJSON();
                        combiTagsJA.Add(combiTagJC);
                    }
                }
            }
            return combiTagsJA;
        }
        public static JSONArray GetHiddenTagsJSON(int resourceType, string atomPresetResourceType)
        {
            var hiddenTagsJA = new JSONArray();
            if (resourceType != ResourceType.presetAtom)
            {
                if (resourceTypeTagMap.ContainsKey(resourceType))
                {
                    foreach (var tag in resourceTypeTagMap[resourceType])
                    {
                        if (tag.hideTaggedResourcesJSB.val) hiddenTagsJA.Add(tag.tagName);

                    }
                }
            }
            else
            {
                if (atomPresetResourceTypeTagMap.ContainsKey(atomPresetResourceType))
                {
                    foreach (var tag in atomPresetResourceTypeTagMap[atomPresetResourceType])
                    {
                        if (tag.hideTaggedResourcesJSB.val) hiddenTagsJA.Add(tag.tagName);

                    }
                }
            }
                
            return hiddenTagsJA;
        }

        public static JSONArray GetPreferredTagsJSON(int resourceType, string atomPresetResourceType)
        {
            var preferredTagsJA = new JSONArray();
            if (resourceType != ResourceType.presetAtom)
            {
                if (resourceTypeTagMap.ContainsKey(resourceType))
                {
                    foreach (var tag in resourceTypeTagMap[resourceType])
                    {
                        if (tag.preferredTagJSB.val) preferredTagsJA.Add(tag.tagName);

                    }
                }
            }
            else
            {
                if (atomPresetResourceTypeTagMap.ContainsKey(atomPresetResourceType))
                {
                    foreach (var tag in atomPresetResourceTypeTagMap[atomPresetResourceType])
                    {
                        if (tag.preferredTagJSB.val) preferredTagsJA.Add(tag.tagName);

                    }
                }
            }

            return preferredTagsJA;
        }


        public static void RestoreHiddenTagsJSON(int resourceType, string atomPresetResourceType, JSONArray tagsJA)
        {
            foreach (JSONNode node in tagsJA)
            {
                if (resourceType == ResourceType.presetAtom)
                {
                    if (tagNamesToTagByAtomPresetResourceType.ContainsKey(atomPresetResourceType) && tagNamesToTagByAtomPresetResourceType[atomPresetResourceType].ContainsKey(node))
                    {
                        var tag = GetTag(resourceType, atomPresetResourceType, node);
                        tag.hideTaggedResourcesJSB.valNoCallback = true;
                    }

                }
                else
                {
                    if (tagNamesToTagByResourceType.ContainsKey(resourceType) && tagNamesToTagByResourceType[resourceType].ContainsKey(node))
                    {
                        var tag = GetTag(resourceType, atomPresetResourceType, node);
                        tag.hideTaggedResourcesJSB.valNoCallback = true;
                    }
                }
                
                
            }
        }

        public static void RestorePreferredTagsJSON(int resourceType, string atomPresetResourceType, JSONArray tagsJA)
        {
            foreach (JSONNode node in tagsJA)
            {
                if (resourceType == ResourceType.presetAtom)
                {
                    if (tagNamesToTagByAtomPresetResourceType.ContainsKey(atomPresetResourceType) && tagNamesToTagByAtomPresetResourceType[atomPresetResourceType].ContainsKey(node))
                    {
                        var tag = GetTag(resourceType, atomPresetResourceType, node);
                        tag.preferredTagJSB.valNoCallback = true;
                    }

                }
                else
                {
                    if (tagNamesToTagByResourceType.ContainsKey(resourceType) && tagNamesToTagByResourceType[resourceType].ContainsKey(node))
                    {
                        var tag = GetTag(resourceType, atomPresetResourceType, node);
                        tag.preferredTagJSB.valNoCallback = true;
                    }
                }


            }
        }


        public static void RestoreCombiTagsJSON(int resourceType, string atomPresetResourceType, JSONArray combiTagsJA)
        {
            if (resourceType == ResourceType.presetAtom)
            {
                foreach (JSONClass jc in combiTagsJA)
                {
                    CombiTag newTag = new CombiTag(resourceType, atomPresetResourceType, jc);
                    RegisterNewTag(newTag);

                    if (!atomPresetResourceTypeCombiTagMap.ContainsKey(atomPresetResourceType)) atomPresetResourceTypeCombiTagMap.Add(atomPresetResourceType, new HashSet<CombiTag>());
                    atomPresetResourceTypeCombiTagMap[atomPresetResourceType].Add(newTag);
                }
            }
            else
            {
                foreach (JSONClass jc in combiTagsJA)
                {
                    CombiTag newTag = new CombiTag(resourceType, "", jc);
                    RegisterNewTag(newTag);

                    if (!resourceTypeCombiTagMap.ContainsKey(resourceType)) resourceTypeCombiTagMap.Add(resourceType, new HashSet<CombiTag>());
                    resourceTypeCombiTagMap[resourceType].Add(newTag);
                }
                if (resourceTypeCombiTagMap.ContainsKey(resourceType))
                {
                    foreach (CombiTag combiTag in resourceTypeCombiTagMap[resourceType])
                    {
                        combiTag.Init();
                    }
                }
            }        

            foreach (CombiTag combiTag in GetCombiTags(resourceType, atomPresetResourceType))
            {
                foreach (var rvge in ResourceManifest.GetAllResourceGroups(resourceType, atomPresetResourceType))  rvge.SetTag(combiTag);
            }

        }

        public static void DeleteTag(Tag tag)
        {
            if (tag.tagCategory == TagCategory.userDefined || tag.tagCategory == TagCategory.combi)
            {
                if (tagsRVGEs.ContainsKey(tag))
                {
                    var taggedRVGEs = new HashSet<ResourceVersionGroupEntry> (tagsRVGEs[tag]);
                    foreach (var rvge in taggedRVGEs)
                    {
                        rvge.SetTag(tag, false);
                        
                    }
                }

                if (tag.usedByCombiTags.Count > 0)
                {
                    foreach (var combiTag in tag.usedByCombiTags)
                    {
                        combiTag.DeleteTagUsage(tag);
                    }
                    Settings.SaveSettings();
                }


                categoryTagMap[tag.tagCategory].Remove(tag);
                resourceTypeTagMap[tag.resourceType].Remove(tag);
                if (tag.resourceType == ResourceType.presetAtom) atomPresetResourceTypeTagMap[tag.atomPresetResourceType].Remove(tag);

                if (tag.tagCategory == TagCategory.combi)
                {
                    resourceTypeCombiTagMap[tag.resourceType].Remove(tag as CombiTag);
                    if (tag.resourceType == ResourceType.presetAtom) atomPresetResourceTypeCombiTagMap[tag.atomPresetResourceType].Remove(tag as CombiTag);
                }
                var tagNamesToTag = tagNamesToTagByResourceType[tag.resourceType];               
                tagNamesToTag.Remove(tag.tagName);
                if (tag.resourceType == ResourceType.presetAtom)
                {
                    var tagNamesToTag2 = tagNamesToTagByAtomPresetResourceType[tag.atomPresetResourceType];
                    tagNamesToTag2.Remove(tag.tagName);
                }

                tags.Remove(tag);

            }
            else SuperController.LogError("BA.TagManifest.DeleteTag: Attempt to delete tag that is not a user defined tag.");
        }
        public static void DeleteTag(string tagName, int resourceType, string atomPresetResourceType)
        {
            Tag tag = GetTag(resourceType, atomPresetResourceType, tagName);

            DeleteTag(tag);
        }

        private static void RegisterNewTag(Tag newTag)
        {
            tags.Add(newTag);
            tagsRVGEs.Add(newTag, new HashSet<ResourceVersionGroupEntry> ());

            if (!categoryTagMap.ContainsKey(newTag.tagCategory)) categoryTagMap.Add(newTag.tagCategory, new HashSet<Tag>());
            categoryTagMap[newTag.tagCategory].Add(newTag);

            if (!resourceTypeTagMap.ContainsKey(newTag.resourceType)) resourceTypeTagMap.Add(newTag.resourceType, new HashSet<Tag>());
            resourceTypeTagMap[newTag.resourceType].Add(newTag);

            if (!tagCategoryToResourceByResourceType.ContainsKey(newTag.resourceType)) tagCategoryToResourceByResourceType.Add(newTag.resourceType, new Dictionary<int, HashSet<ResourceVersionGroupEntry>>());
            var tagCategoryToResource = tagCategoryToResourceByResourceType[newTag.resourceType];
            if (!tagCategoryToResource.ContainsKey(newTag.tagCategory)) tagCategoryToResource.Add(newTag.tagCategory, new HashSet<ResourceVersionGroupEntry>());

            if (!tagNamesToTagByResourceType.ContainsKey(newTag.resourceType)) tagNamesToTagByResourceType.Add(newTag.resourceType, new Dictionary<string, Tag>());

            var tagNamesToTag = tagNamesToTagByResourceType[newTag.resourceType];
            if (!tagNamesToTag.ContainsKey(newTag.tagName)) tagNamesToTag.Add(newTag.tagName, newTag);
            else tagNamesToTag[newTag.tagName] = newTag;

            if (newTag.resourceType == ResourceType.presetAtom)
            {
                if (!atomPresetResourceTypeTagMap.ContainsKey(newTag.atomPresetResourceType)) atomPresetResourceTypeTagMap.Add(newTag.atomPresetResourceType, new HashSet<Tag>());
                atomPresetResourceTypeTagMap[newTag.atomPresetResourceType].Add(newTag);

                if (!tagCategoryToResourceByAtomPresetResourceType.ContainsKey(newTag.atomPresetResourceType)) tagCategoryToResourceByAtomPresetResourceType.Add(newTag.atomPresetResourceType, new Dictionary<int, HashSet<ResourceVersionGroupEntry>>());
                var tagCategoryToResource2 = tagCategoryToResourceByAtomPresetResourceType[newTag.atomPresetResourceType];
                if (!tagCategoryToResource2.ContainsKey(newTag.tagCategory)) tagCategoryToResource2.Add(newTag.tagCategory, new HashSet<ResourceVersionGroupEntry>());

                if (!tagNamesToTagByAtomPresetResourceType.ContainsKey(newTag.atomPresetResourceType)) tagNamesToTagByAtomPresetResourceType.Add(newTag.atomPresetResourceType, new Dictionary<string, Tag>());
                var tagNamesToTag2 = tagNamesToTagByAtomPresetResourceType[newTag.atomPresetResourceType];
                if (!tagNamesToTag2.ContainsKey(newTag.tagName)) tagNamesToTag2.Add(newTag.tagName, newTag);
                else tagNamesToTag2[newTag.tagName] = newTag;
            }
        }

        public static Tag CreateTag(string name, int tagCategory, int resourceType, string atomPresetResourceType)
        {
            int resourceCategory = ResourceType.GetResourceCategory(resourceType);
            if ((resourceCategory == ResourceCategory.preset && resourceType != ResourceType.presetAtom) || resourceCategory == ResourceCategory.clothing) atomPresetResourceType = "Person";
            else if (resourceType != ResourceType.presetAtom) atomPresetResourceType = "";

            if (tags.Any(t => t.tagName == name && t.resourceType == resourceType && t.atomPresetResourceType==atomPresetResourceType)) return null;

            Tag newTag = new Tag(name, tagCategory, resourceType, atomPresetResourceType);

            RegisterNewTag(newTag);

            return newTag;
        }

        public static HashSet<Tag> GetRVGETags(ResourceVersionGroupEntry rvge)
        {
            if (!rvgeTags.ContainsKey(rvge)) return new HashSet<Tag>();
            return rvgeTags[rvge];
        }

        public static List<Tag> GetTags(int resourceType, string atomPresetResourceType)
        {
            if (resourceType != ResourceType.presetAtom)
            {
                if (!resourceTypeTagMap.ContainsKey(resourceType)) resourceTypeTagMap.Add(resourceType, new HashSet<Tag>());
                return resourceTypeTagMap[resourceType].OrderBy(t => t.tagCategory).ThenBy(t => t.tagName).ToList();
            }
            else
            {
                if (!atomPresetResourceTypeTagMap.ContainsKey(atomPresetResourceType)) atomPresetResourceTypeTagMap.Add(atomPresetResourceType, new HashSet<Tag>());
                return atomPresetResourceTypeTagMap[atomPresetResourceType].OrderBy(t => t.tagCategory).ThenBy(t => t.tagName).ToList();
            }
            
        }

        public static HashSet<CombiTag> GetCombiTags(int resourceType, string atomPresetResourceType)
        {
            if (resourceType != ResourceType.presetAtom)
            {
                if (!resourceTypeCombiTagMap.ContainsKey(resourceType)) resourceTypeCombiTagMap.Add(resourceType, new HashSet<CombiTag>());
                return resourceTypeCombiTagMap[resourceType];
            }
            else
            {
                if (!atomPresetResourceTypeCombiTagMap.ContainsKey(atomPresetResourceType)) atomPresetResourceTypeCombiTagMap.Add(atomPresetResourceType, new HashSet<CombiTag>());
                return atomPresetResourceTypeCombiTagMap[atomPresetResourceType];
            }
                
        }
        public static List<Tag> GetTags(int resourceType, string atomPresetResourceType, int tagCategory)
        {
            if (resourceType != ResourceType.presetAtom)
            {
                if (!resourceTypeTagMap.ContainsKey(resourceType)) resourceTypeTagMap.Add(resourceType, new HashSet<Tag>());
                return resourceTypeTagMap[resourceType].Where(x => x.tagCategory == tagCategory).OrderBy(t => t.tagName).ToList();
            }
            else
            {
                if (!atomPresetResourceTypeTagMap.ContainsKey(atomPresetResourceType)) atomPresetResourceTypeTagMap.Add(atomPresetResourceType, new HashSet<Tag>());
                return atomPresetResourceTypeTagMap[atomPresetResourceType].Where(x => x.tagCategory == tagCategory).OrderBy(t => t.tagName).ToList();
            }
                
        }

        private static List<string> GetOrderedTagNames (HashSet<Tag> tagSet)
        {
            var preferredTags = new List<string>();
            var normalTags = new List<string>();

            foreach (var tag in tagSet)
            {
                if (tag.preferredTagJSB.val) preferredTags.Add(tag.tagName);
                else normalTags.Add(tag.tagName);
            }

            preferredTags.Sort();
            normalTags.Sort();

            preferredTags.AddRange(normalTags);
            return preferredTags;
        }
        public static List<string> GetTagNames(int resourceType, string atomPresetResourceType)
        {
            if (resourceType != ResourceType.presetAtom)
            {
                if (!tagNamesToTagByResourceType.ContainsKey(resourceType)) tagNamesToTagByResourceType.Add(resourceType, new Dictionary<string, Tag>());

                return GetOrderedTagNames(new HashSet<Tag>(tagNamesToTagByResourceType[resourceType].Values));

            }
            else
            {
                if (!tagNamesToTagByAtomPresetResourceType.ContainsKey(atomPresetResourceType)) tagNamesToTagByAtomPresetResourceType.Add(atomPresetResourceType, new Dictionary<string, Tag>());

                return GetOrderedTagNames(new HashSet<Tag>(tagNamesToTagByAtomPresetResourceType[atomPresetResourceType].Values));
            }
                
        }

        public static HashSet<int> GetTagCategories(int resourceType, string atomPresetResourceType)
        {
            var tagCategoryToResource = new Dictionary<int, HashSet<ResourceVersionGroupEntry>>();

            if (resourceType != ResourceType.presetAtom)
            {
                if (!tagCategoryToResourceByResourceType.ContainsKey(resourceType)) tagCategoryToResourceByResourceType.Add(resourceType, new Dictionary<int, HashSet<ResourceVersionGroupEntry>>());
                tagCategoryToResource= tagCategoryToResourceByResourceType[resourceType];
            }
            else
            {
                if (!tagCategoryToResourceByAtomPresetResourceType.ContainsKey(atomPresetResourceType)) tagCategoryToResourceByAtomPresetResourceType.Add(atomPresetResourceType, new Dictionary<int, HashSet<ResourceVersionGroupEntry>>());
                tagCategoryToResource = tagCategoryToResourceByAtomPresetResourceType[atomPresetResourceType];
            }

            var tagCategories = new HashSet<int>();
            foreach (var kvp in tagCategoryToResource)
            {
                if (kvp.Value.Count>0) tagCategories.Add(kvp.Key);
            }

            return tagCategories;
        }

        public static Tag GetTag(int resourceType, string atomPresetResourceType, string tagName)
        {
            if (resourceType != ResourceType.presetAtom)
            {
                if (!tagNamesToTagByResourceType[resourceType].ContainsKey(tagName)) return null;
                return tagNamesToTagByResourceType[resourceType][tagName];
            }
            else
            {
                if (!tagNamesToTagByAtomPresetResourceType.ContainsKey(atomPresetResourceType) || !tagNamesToTagByAtomPresetResourceType[atomPresetResourceType].ContainsKey(tagName)) return null;
                return tagNamesToTagByAtomPresetResourceType[atomPresetResourceType][tagName];
            }
        }

        public static bool IsRVGETagged(ResourceVersionGroupEntry rvge, Tag tag)
        {
            if (!rvgeTags.ContainsKey(rvge)) return false;

            if (rvgeTags[rvge].Contains(tag)) return true;

            return false;
        }

        public static bool IsRVGENotTagged(ResourceVersionGroupEntry rvge)
        {
            if (!rvgeTags.ContainsKey(rvge)) return true;

            if (rvgeTags[rvge].Count==0) return true;

            return false;
        }

        public static bool IsRVGENotUserTagged(ResourceVersionGroupEntry rvge)
        {
            if (!rvgeTags.ContainsKey(rvge)) return true;

            if (rvgeTags[rvge].Where(t=>t.tagCategory==TagCategory.userDefined).Count() == 0) return true;

            return false;
        }


        public static void SetRVGETagged(ResourceVersionGroupEntry rvge, Tag tag, bool state)
        {           
            if (tag != null)
            {
                if (state)
                {
                    if (!rvgeTags.ContainsKey(rvge)) rvgeTags.Add(rvge, new HashSet<Tag>());
                    rvgeTags[rvge].Add(tag);
                    if (!tagsRVGEs.ContainsKey(tag)) tagsRVGEs.Add(tag, new HashSet<ResourceVersionGroupEntry>());
                    tagsRVGEs[tag].Add(rvge);


                    tagCategoryToResourceByResourceType[tag.resourceType][tag.tagCategory].Add(rvge);
                    if (rvge.resourceType == ResourceType.presetAtom)
                    {
                        tagCategoryToResourceByAtomPresetResourceType[tag.atomPresetResourceType][tag.tagCategory].Add(rvge);
                    }
                }
                else
                {
                    if (!rvgeTags.ContainsKey(rvge)) return;
                    rvgeTags[rvge].Remove(tag);
                    tagsRVGEs[tag].Remove(rvge);

                    tagCategoryToResourceByResourceType[tag.resourceType][tag.tagCategory].Remove(rvge);
                    if (rvge.resourceType == ResourceType.presetAtom)
                    {
                        tagCategoryToResourceByAtomPresetResourceType[tag.atomPresetResourceType][tag.tagCategory].Remove(rvge);
                    }
                }
            }
            
        }

        public static void SetRVGETagged(ResourceVersionGroupEntry rvge, string tagName, bool state)
        {
            Tag tag =null;

            if (rvge.resourceType == ResourceType.presetAtom)
            {
                if (!tagNamesToTagByAtomPresetResourceType.ContainsKey(rvge.atomPresetResourceType) ||
                    !tagNamesToTagByAtomPresetResourceType[rvge.atomPresetResourceType].ContainsKey(tagName)) CreateTag(tagName, TagCategory.userDefined, rvge.resourceType, rvge.atomPresetResourceType);
                tag = tagNamesToTagByAtomPresetResourceType[rvge.atomPresetResourceType][tagName];
            }
            else
            {
                if (!tagNamesToTagByResourceType[rvge.resourceType].ContainsKey(tagName)) CreateTag(tagName, TagCategory.userDefined, rvge.resourceType, rvge.atomPresetResourceType);
                tag = tagNamesToTagByResourceType[rvge.resourceType][tagName];
            }
            
            SetRVGETagged(rvge, tag, state);
        }

    }
}
