using System;
using UnityEngine;
using System.Collections.Generic;
using System.Linq;
using UnityEngine.UI;
using UnityEngine.EventSystems;


namespace JayJayWon
{

    public static class UI
    {
        public static GameObject CreateHScrollableContent(GameObject parentGO, Color backgroundColor, int anchorPreset, float horizontalSize, float verticalSize, Vector2 anchoredPositionOffset, float spacing =0f)
        {
            GameObject scrollableContentGO = AddChildGOImage(parentGO, backgroundColor, anchorPreset, horizontalSize, verticalSize, anchoredPositionOffset);

            GameObject buttonListViewportGO = CreateHScrollableContentViewPort(scrollableContentGO, horizontalSize - 10f, verticalSize, spacing);
            buttonListViewportGO.name = "scrollableContentViewportGO";

            GameObject buttonListScrollbarGO = CreateScrollBar(scrollableContentGO, horizontalSize, 10f, 0f, 0f, Scrollbar.Direction.LeftToRight);
            buttonListScrollbarGO.name = "scrollBar";

            ScrollRect scrollRect = scrollableContentGO.AddComponent<ScrollRect>();
            GameObject buttonListContentGO = buttonListViewportGO.transform.Find("hScrollableContentGO").gameObject;
            scrollRect.content = buttonListContentGO.GetComponent<RectTransform>();
            scrollRect.horizontal = true;
            scrollRect.vertical = false;

            scrollRect.movementType = ScrollRect.MovementType.Clamped;
            scrollRect.viewport = buttonListViewportGO.GetComponent<RectTransform>();
            scrollRect.horizontalScrollbar = buttonListScrollbarGO.GetComponent<Scrollbar>();
            scrollRect.horizontalScrollbarVisibility = ScrollRect.ScrollbarVisibility.AutoHideAndExpandViewport;

            return (scrollableContentGO);
        }


        public static GameObject CreateVScrollableContent(GameObject parentGO, Color backgroundColor, int anchorPreset, float horizontalSize, float verticalSize, Vector2 anchoredPositionOffset, float scrollBarWidth = 10f, float spacing =0f)
        {
            GameObject scrollableContentGO = AddChildGOImage(parentGO, backgroundColor, anchorPreset, horizontalSize, verticalSize, anchoredPositionOffset);

            GameObject buttonListViewportGO = CreateVScrollableContentViewPort(scrollableContentGO, horizontalSize - scrollBarWidth, verticalSize, spacing);
            buttonListViewportGO.name = "scrollableContentViewportGO";

            GameObject buttonListScrollbarGO = CreateScrollBar(scrollableContentGO, scrollBarWidth, verticalSize, 0f, 0f, Scrollbar.Direction.BottomToTop);
            buttonListScrollbarGO.name = "scrollBar";

            ScrollRect scrollRect = scrollableContentGO.AddComponent<ScrollRect>();
            GameObject buttonListContentGO = buttonListViewportGO.transform.Find("vScrollableContentGO").gameObject;
            scrollRect.content = buttonListContentGO.GetComponent<RectTransform>();
            scrollRect.horizontal = false;
            scrollRect.vertical = true;

            scrollRect.movementType = ScrollRect.MovementType.Clamped;
            scrollRect.viewport = buttonListViewportGO.GetComponent<RectTransform>();
            scrollRect.verticalScrollbar = buttonListScrollbarGO.GetComponent<Scrollbar>();
            scrollRect.verticalScrollbarVisibility = ScrollRect.ScrollbarVisibility.AutoHideAndExpandViewport;

            return (scrollableContentGO);
        }

        private static GameObject CreateScrollBar(GameObject parentGO, float width, float height, float xAnchorOffset, float yAnchorOffset, Scrollbar.Direction scrollbarDirection)
        {
            Sprite backgroundSprite = Resources.FindObjectsOfTypeAll<Sprite>().First(sprite => sprite.name == "Background");
            Sprite uiSprite = Resources.FindObjectsOfTypeAll<Sprite>().First(sprite => sprite.name == "UISprite");

            int anchorPreset = AnchorPresets.vStretchRight;
            if (scrollbarDirection == Scrollbar.Direction.LeftToRight || scrollbarDirection == Scrollbar.Direction.RightToLeft) anchorPreset = AnchorPresets.hStretchBottom;

            GameObject scrollbarGO = AddChildGOImage(parentGO, new Color(1f, 1f, 1f), anchorPreset, width, height, Vector2.zero, backgroundSprite);
            Scrollbar scrollbar = scrollbarGO.AddComponent<Scrollbar>();
            scrollbar.SetDirection(scrollbarDirection, false);
            scrollbarGO.SetActive(true);

            GameObject scrollbarSlidingAreaGO = new GameObject();
            scrollbarSlidingAreaGO.AddComponent<RectTransform>();
            scrollbarSlidingAreaGO.transform.SetParent(scrollbarGO.transform, false);
            scrollbarSlidingAreaGO.SetActive(true);
            ResizeRecTransform(scrollbarSlidingAreaGO, width, height, 0f, 0f, anchorPreset);

            GameObject scrollbarHandleGO = AddChildGOImage(scrollbarSlidingAreaGO, new Color(1f, 1f, 1f), AnchorPresets.centre, 3f, 3f, Vector2.zero, uiSprite);

            scrollbar.targetGraphic = scrollbarHandleGO.GetComponent<UnityEngine.UI.Image>();
            scrollbar.handleRect = scrollbarHandleGO.GetComponent<RectTransform>();

            return (scrollbarGO);
        }

        private static GameObject CreateHScrollableContentViewPort(GameObject parentGO, float width, float height, float spacing = 0f)
        {
            GameObject scrollableContentViewportGO = AddChildGOImage(parentGO, new Color(1f, 0f, 0f, 0f), AnchorPresets.hStretchTop, width, height, Vector2.zero);
            //           Mask blvpMask = scrollableContentViewportGO.AddComponent<Mask>();
            //           blvpMask.showMaskGraphic = false;
            RectMask2D altMask = scrollableContentViewportGO.AddComponent<RectMask2D>();

            GameObject scrollableContentGO = AddChildGOImage(scrollableContentViewportGO, new Color(0f, 0f, 1f, 0f), AnchorPresets.hStretchTop, 200f, 0f, Vector2.zero, null, false);
            scrollableContentGO.name = "hScrollableContentGO";

            HorizontalLayoutGroup hlg = scrollableContentGO.AddComponent<HorizontalLayoutGroup>();
            hlg.childAlignment = TextAnchor.UpperLeft;
            hlg.childForceExpandHeight = true;
            hlg.childForceExpandWidth = true;
            hlg.childControlHeight = false;
            hlg.childControlWidth = false;
            hlg.spacing = spacing;

            ContentSizeFitter csf = scrollableContentGO.AddComponent<ContentSizeFitter>();
            csf.horizontalFit = ContentSizeFitter.FitMode.MinSize;
            csf.verticalFit = ContentSizeFitter.FitMode.MinSize;

            return (scrollableContentViewportGO);
        }


        private static GameObject CreateVScrollableContentViewPort(GameObject parentGO, float width, float height, float spacing = 0f)
        {
            GameObject scrollableContentViewportGO = AddChildGOImage(parentGO, new Color(1f, 0f, 0f,0f), AnchorPresets.vStretchLeft, width, height, Vector2.zero);
 //           Mask blvpMask = scrollableContentViewportGO.AddComponent<Mask>();
 //           blvpMask.showMaskGraphic = false;
            RectMask2D altMask = scrollableContentViewportGO.AddComponent<RectMask2D>();

            GameObject scrollableContentGO = AddChildGOImage(scrollableContentViewportGO, new Color(0f, 0f, 1f,0f), AnchorPresets.vStretchLeft, 0f, 200f, Vector2.zero, null, false);
            scrollableContentGO.name = "vScrollableContentGO";

            VerticalLayoutGroup vlg = scrollableContentGO.AddComponent<VerticalLayoutGroup>();
            vlg.childAlignment = TextAnchor.UpperLeft;
            vlg.childForceExpandHeight = true;
            vlg.childForceExpandWidth = true;
            vlg.childControlHeight = false;
            vlg.childControlWidth = false;
            vlg.spacing = spacing;

            ContentSizeFitter csf = scrollableContentGO.AddComponent<ContentSizeFitter>();
            csf.horizontalFit = ContentSizeFitter.FitMode.MinSize;
            csf.verticalFit = ContentSizeFitter.FitMode.MinSize;

            return (scrollableContentViewportGO);
        }

        public static void AddImageComponent(GameObject go, Color color, Sprite sprite = null, bool enabled = true)
        {
            Image image = go.AddComponent<Image>();
            image.color = color;
            
            if (sprite!=null) image.sprite = sprite;
            image.type = Image.Type.Sliced;
            image.enabled = enabled;
            
        }

        public static GameObject AddChildImage(Transform parentTransform, Color color, Vector2 anchorMin, Vector2 anchorMax, float horizontalSize, float verticalSize, Vector2 anchoredPositionOffset, Vector2 pivot, Sprite sprite = null, bool enabled = true)
        {
            GameObject go = new GameObject();
            go.transform.SetParent(parentTransform, false);

            AddImageComponent(go, color, sprite, enabled);
            RectTransform rt = go.GetComponent<RectTransform>();
            rt.sizeDelta = Vector2.zero;
            rt.anchorMin = anchorMin;
            rt.anchorMax = anchorMax;
            rt.pivot = pivot;
            if (verticalSize != 0f) rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, verticalSize);
            if (horizontalSize != 0f) rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, horizontalSize);

            rt.anchoredPosition = anchoredPositionOffset;

            return go;
        }

        public static GameObject AddChildImage(GameObject parentGO, Color color, Vector2 anchorMin, Vector2 anchorMax, float horizontalSize, float verticalSize, Vector2 anchoredPositionOffset, Vector2 pivot, Sprite sprite = null, bool enabled = true)
        {
            return AddChildImage(parentGO.transform, color, anchorMin, anchorMax, horizontalSize, verticalSize, anchoredPositionOffset, pivot, sprite, enabled);
        }
        public static GameObject AddChildGOImage(GameObject parentGO, Color color, int anchorPreset, float horizontalSize, float verticalSize, Vector2 anchoredPositionOffset, Sprite sprite = null, bool enabled = true)
        {
            return AddChildImage(parentGO.transform, color, AnchorPresets.GetAnchorMin(anchorPreset), AnchorPresets.GetAnchorMax(anchorPreset), horizontalSize, verticalSize, anchoredPositionOffset, AnchorPresets.GetPivot(anchorPreset), sprite, enabled);
        }

        public static GameObject AddChildImage(Transform parentTransform, Color color, int anchorPreset, float horizontalSize, float verticalSize, Vector2 anchoredPositionOffset, Sprite sprite = null, bool enabled = true)
        {
            return AddChildImage(parentTransform, color, AnchorPresets.GetAnchorMin(anchorPreset), AnchorPresets.GetAnchorMax(anchorPreset), horizontalSize, verticalSize, anchoredPositionOffset, AnchorPresets.GetPivot(anchorPreset), sprite, enabled);
        }

        public static GameObject CreateBasicTextBox(GameObject parentGO, string textString, float width, float height, int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool resizeTextForBestFit = false, int minFontSize = 20, HorizontalWrapMode horizontalWrapMode = HorizontalWrapMode.Wrap, Color? backgoundColor = null, bool rescaleText =true, TextAnchor textAnchor =TextAnchor.MiddleLeft, Color? textColor = null)
        {
            GameObject imageGO = AddChildGOImage(parentGO, backgoundColor ?? Color.gray, anchorPreset, width, height, new Vector2(anchorXOffset, anchorYOffset));
            var textGo = new GameObject("text");
            textGo.transform.SetParent(imageGO.transform, false);
            textGo.AddComponent<RectTransform>();
            Text text = textGo.AddComponent<Text>();
            Font selectedFont = (Font)Resources.GetBuiltinResource(typeof(Font), "Arial.ttf");
            text.font = selectedFont;
            ReformatText(textGo, width - 4f, height, fontSize, 2f, 0, AnchorPresets.centre, textAnchor, resizeTextForBestFit, rescaleText);

            text.horizontalOverflow = horizontalWrapMode;
            //           text.resizeTextForBestFit = resizeTextForBestFit;

            if (resizeTextForBestFit)
            {
                text.resizeTextMaxSize = fontSize;
                text.resizeTextMinSize = minFontSize;
            }
            else text.fontSize = fontSize;

            text.text = textString;

            text.color = textColor ?? Color.white;

            return imageGO;
        }

        public static GameObject CreateBasicTextBox(GameObject parentGO, JSONStorableString jss, float width, float height, int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool resizeTextForBestFit = false, int minFontSize = 20, HorizontalWrapMode horizontalWrapMode = HorizontalWrapMode.Wrap, Color? backgoundColor=null, bool rescaleText = true)
        {
            GameObject textImageGO = CreateBasicTextBox( parentGO, jss.val, width, height, fontSize , anchorXOffset, anchorYOffset, anchorPreset, resizeTextForBestFit, minFontSize, horizontalWrapMode, backgoundColor, rescaleText);

            GameObject textGO = textImageGO.transform.Find("text").gameObject;

            jss.RegisterText(textGO.GetComponent<Text>());

            return textImageGO;
        }

        public static UIDynamicTextField CreateTextBox(GameObject parentGO, JSONStorableString jss, float width, float height, int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool input = false, InputField.LineType inputFieldType = InputField.LineType.MultiLineNewline, string placeHolderText = "", float hScrollBarHeight = 0f, float vScrollBarWidth = 10f, Action<string> textEntryChangeCallback = null, TextAnchor textAlignment = TextAnchor.MiddleLeft, Action<string> textEntryEndCallback=null, bool rescaleText = true,Color ? backgoundColor = null)
        {
            Transform transform = GameObject.Instantiate<Transform>(Globals.mvrScript.manager.configurableTextFieldPrefab);

            GameObject textFieldGO = transform.gameObject;
            UIDynamicTextField textFieldUID = textFieldGO.transform.GetComponent<UIDynamicTextField>();
            transform.SetParent(parentGO.transform, false);

            ResizeRecTransform(textFieldGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);
            if (input)
            {
                textFieldUID.backgroundColor = Color.white;
                var inputComponent = textFieldUID.gameObject.AddComponent<InputField>();

                inputComponent.textComponent = textFieldUID.UItext;
                inputComponent.lineType = inputFieldType;
                inputComponent.caretWidth = 2;

                inputComponent.textComponent.resizeTextForBestFit = false;
                jss.RegisterInputField(inputComponent);

                if (placeHolderText != "")
                {
                    var placeholderGo = new GameObject("placeHolderText");
                    placeholderGo.transform.SetParent(textFieldUID.gameObject.transform, false);

                    var placeholderRect = placeholderGo.AddComponent<RectTransform>();
                    Font selectedFont = (Font)Resources.GetBuiltinResource(typeof(Font), "Arial.ttf");

                    UnityEngine.UI.Text placeholderUIText = placeholderGo.AddComponent<UnityEngine.UI.Text>();
                    placeholderUIText.raycastTarget = false;
                    placeholderUIText.font = selectedFont;
                    placeholderUIText.color = Color.gray;
                    placeholderUIText.fontStyle = FontStyle.Italic;

                    ReformatText(placeholderGo, width - vScrollBarWidth, height - hScrollBarHeight, fontSize, 0f, 0f, AnchorPresets.middleLeft, TextAnchor.MiddleLeft, rescaleText:rescaleText);

                    if (jss.val == "") placeholderUIText.text = placeHolderText;

                    inputComponent.onValueChanged.AddListener((text) =>
                    {
                        InputTextFieldValueChanged(inputComponent, placeholderUIText, jss, placeHolderText, textEntryChangeCallback);
                    });
                  
                }
                else
                {
                    inputComponent.onValueChanged.AddListener((text) =>
                    {
                        if (textEntryChangeCallback!=null) textEntryChangeCallback(text);
                    });
                }
                inputComponent.onEndEdit.AddListener((text) =>
                {
                    if (textEntryEndCallback != null) textEntryEndCallback(text);
                });

                
            }
            else
            {
                jss.dynamicText = textFieldUID;
            }

            textFieldUID.height = height - hScrollBarHeight;

            GameObject scrollViewGO = textFieldGO.transform.Find("Scroll View").gameObject;
            ReformatScrollView(scrollViewGO, width, vScrollBarWidth, height, hScrollBarHeight);

            Color bkColor = new Color(0, 0f, 0f, 0f);
            if (backgoundColor != null) bkColor = (Color) backgoundColor;

            GameObject contentGO = textFieldGO.transform.Find("Scroll View/Viewport/Content").gameObject;
            ResizeRecTransform(contentGO, width, height, anchorPreset: GetAnchorPreset(textAlignment));
            AddChildGOImage(contentGO, bkColor, GetAnchorPreset(textAlignment), width, height, new Vector2(0f, 0f));

            GameObject textContentGO = textFieldGO.transform.Find("Scroll View/Viewport/Content/Text").gameObject;
            ReformatText(textContentGO, width - vScrollBarWidth, height - hScrollBarHeight, fontSize, 2f, 0f, anchorPreset: GetAnchorPreset(textAlignment), textAlginment: textAlignment, rescaleText: rescaleText);

            if (input && inputFieldType != InputField.LineType.SingleLine)
            {
                var csf = jss.inputField.textComponent.gameObject.AddComponent<ContentSizeFitter>();
                csf.verticalFit = ContentSizeFitter.FitMode.PreferredSize;
                jss.inputField.textComponent.horizontalOverflow = HorizontalWrapMode.Wrap;
                jss.inputField.textComponent.verticalOverflow = VerticalWrapMode.Truncate;
            }

            return textFieldUID;
        }



        private static int GetAnchorPreset(TextAnchor textAnchor)
        {
            if (textAnchor == TextAnchor.UpperLeft) return AnchorPresets.topLeft;
            if (textAnchor == TextAnchor.UpperCenter) return AnchorPresets.topMiddle;
            if (textAnchor == TextAnchor.UpperRight) return AnchorPresets.topRight;

            if (textAnchor == TextAnchor.MiddleLeft) return AnchorPresets.middleLeft;
            if (textAnchor == TextAnchor.MiddleCenter) return AnchorPresets.centre;
            if (textAnchor == TextAnchor.UpperRight) return AnchorPresets.middleRight;

            if (textAnchor == TextAnchor.LowerLeft) return AnchorPresets.bottomLeft;
            if (textAnchor == TextAnchor.LowerCenter) return AnchorPresets.bottomMiddle;
            if (textAnchor == TextAnchor.LowerRight) return AnchorPresets.bottomRight;

            return -1;
        }
        private static void InputTextFieldValueChanged(InputField inputField, Text placeholderUIText, JSONStorableString jss, string placeHolderText, Action<string> textEntryChangeCallback)
        {
            if (inputField.text == "") placeholderUIText.text = placeHolderText;
            else placeholderUIText.text = "";

            if (textEntryChangeCallback != null) textEntryChangeCallback(inputField.text);
        }

        public static UIDynamicButton CreateUIButton(GameObject parentGO, float width, float height, string label = "", int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool clearSprite = false, bool rescaleText = true)
        {
            Transform buttonTransform = UnityEngine.Object.Instantiate(Globals.mvrScript.manager.configurableButtonPrefab.transform, parentGO.transform, false);

            GameObject buttonGO = buttonTransform.gameObject;
            buttonTransform.SetParent(parentGO.transform, false);

            UIDynamicButton buttonUID = buttonGO.transform.GetComponent<UIDynamicButton>();
            Transform textTransform = buttonUID.button.transform.Find("Text");
            ReformatText(textTransform.GetComponent<RectTransform>().gameObject, width, height, fontSize, 0, 0, AnchorPresets.centre,rescaleText:rescaleText);

            buttonUID.label = label;
            buttonUID.buttonText.fontSize = fontSize;

            if (clearSprite)
            {
                Image buttonImage = buttonGO.transform.GetComponent<Image>();
                buttonImage.sprite = null;
            }

            ResizeRecTransform(buttonGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            return (buttonUID);

        }

        public static UIDynamicSlider CreateSlider(GameObject parentGO, JSONStorableFloat jsf, float width, float height, string label = null, int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, float quickButtonHeight =0f, float labelHeight =0f, float rangeWidth=0f, bool hideRange=false, bool hideRangeLabel = false, bool rescaleText =true)
        {
            Transform transform = GameObject.Instantiate<Transform>(Globals.mvrScript.manager.configurableSliderPrefab);

            GameObject sliderGO = transform.gameObject;

            
            UIDynamicSlider sliderUID = sliderGO.transform.GetComponent<UIDynamicSlider>();
            if (label != null) sliderUID.label = label;
            jsf.RegisterSlider(sliderUID.slider);

            transform.SetParent(parentGO.transform, false);
            ResizeRecTransform(sliderGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            float sliderHeight = height - quickButtonHeight - labelHeight;
            float sliderWidth = width - rangeWidth;
            if (labelHeight == 0f) sliderWidth = width - (rangeWidth * 2);
            if (hideRange) sliderWidth = sliderWidth + rangeWidth;

            GameObject subSliderGO = sliderGO.transform.Find("Slider").gameObject;
            ResizeRecTransform(subSliderGO, sliderWidth, sliderHeight, 0f, -labelHeight, AnchorPresets.topLeft);

            float qbWidth = width / 9f;
            float qbGroupWidth = qbWidth * 4f;
            GameObject quickButtonsGroupGO = sliderGO.transform.Find("QuickButtonsGroup").gameObject;
            ResizeRecTransform(quickButtonsGroupGO, width, quickButtonHeight, 0f, 0f, AnchorPresets.bottomMiddle);
                       
            GameObject quickButtonsLeftGO = sliderGO.transform.Find("QuickButtonsGroup/QuickButtonsLeft").gameObject;
            ResizeRecTransform(quickButtonsLeftGO, qbGroupWidth, quickButtonHeight, 0f, 0f, AnchorPresets.bottomLeft);
            
            ResizeSliderQuickButton("M4Button", quickButtonsLeftGO, qbWidth, quickButtonHeight, qbWidth*0);
            ResizeSliderQuickButton("M3Button", quickButtonsLeftGO, qbWidth, quickButtonHeight, qbWidth*1);
            ResizeSliderQuickButton("M2Button", quickButtonsLeftGO, qbWidth, quickButtonHeight, qbWidth*2);
            ResizeSliderQuickButton("M1Button", quickButtonsLeftGO, qbWidth, quickButtonHeight, qbWidth*3);

            GameObject quickButtonsRightGO = sliderGO.transform.Find("QuickButtonsGroup/QuickButtonsRight").gameObject;
            ResizeRecTransform(quickButtonsRightGO, qbGroupWidth, quickButtonHeight, 0f, 0f, AnchorPresets.bottomRight);

            ResizeSliderQuickButton("P1Button", quickButtonsRightGO, qbWidth, quickButtonHeight, qbWidth * 0);
            ResizeSliderQuickButton("P2Button", quickButtonsRightGO, qbWidth, quickButtonHeight, qbWidth * 1);
            ResizeSliderQuickButton("P3Button", quickButtonsRightGO, qbWidth, quickButtonHeight, qbWidth * 2);
            ResizeSliderQuickButton("P4Button", quickButtonsRightGO, qbWidth, quickButtonHeight, qbWidth * 3);

            GameObject quickButtonsDefaultGO = sliderGO.transform.Find("DefaultValueButton").gameObject;
            ReformatUnityButton(quickButtonsDefaultGO, qbWidth, quickButtonHeight, 14, 0, 0, AnchorPresets.bottomMiddle, true);

            GameObject textGO = sliderGO.transform.Find("Text").gameObject;
            ReformatText(textGO, width * 0.7f, labelHeight, fontSize, 0, 0, AnchorPresets.topLeft, TextAnchor.MiddleLeft, false, rescaleText: rescaleText);
            if (labelHeight == 0) textGO.SetActive(false);

            GameObject inputGO = sliderGO.transform.Find("ValueInputField").gameObject;
            if (labelHeight > 0f)
            {
                ResizeRecTransform(inputGO, width * 0.3f, labelHeight, 0f, 0f, AnchorPresets.topRight);
                GameObject inputPlaceHolderGO = inputGO.transform.Find("Placeholder").gameObject;
                ResizeRecTransform(inputPlaceHolderGO, width * 0.3f, labelHeight, 0f, 0f, AnchorPresets.topRight);
                GameObject inputTextGO = inputGO.transform.Find("Text").gameObject;
                ReformatText(inputTextGO, width * 0.3f, labelHeight, fontSize, 0, 0, AnchorPresets.topRight, TextAnchor.MiddleRight, false,rescaleText);
            }
            else
            {
                float inputOffset = rangeWidth;
                if (hideRange) inputOffset = 0f;
                ResizeRecTransform(inputGO, rangeWidth, sliderHeight, -inputOffset, 0f, AnchorPresets.topRight);
                GameObject inputPlaceHolderGO = inputGO.transform.Find("Placeholder").gameObject;
                ResizeRecTransform(inputPlaceHolderGO, rangeWidth, sliderHeight, 0f, 0f, AnchorPresets.topRight);
                GameObject inputTextGO = inputGO.transform.Find("Text").gameObject;
                ReformatText(inputTextGO, rangeWidth, sliderHeight, fontSize, -0f, 0, AnchorPresets.topRight, TextAnchor.MiddleCenter, false,rescaleText);
            }

            float rangeButtonHeight = sliderHeight / 2f;
            if (hideRangeLabel) rangeButtonHeight = sliderHeight;

            GameObject rangeGO = sliderGO.transform.Find("RangePanel").gameObject;
            ResizeRecTransform(rangeGO, rangeWidth*.95f, sliderHeight * .95f, 0f, -labelHeight, AnchorPresets.topRight);
            GameObject rangeTextGO = rangeGO.transform.Find("RangeText").gameObject;
            ReformatText(rangeTextGO, rangeWidth * .95f, sliderHeight / 2f, fontSize, 0, 0, AnchorPresets.topRight, TextAnchor.MiddleCenter, true, rescaleText);
            if (hideRangeLabel) rangeTextGO.SetActive(false);
            GameObject x10RangeButtonGO = rangeGO.transform.Find("10XRangeButton").gameObject;
            ReformatUnityButton(x10RangeButtonGO, rangeWidth/2f, rangeButtonHeight*0.8f, 20, 0, 1f, AnchorPresets.bottomRight,true,resizeTextForBestFit:false);
            GameObject x01RangeButtonGO = rangeGO.transform.Find("P1XRangeButton").gameObject;
            ReformatUnityButton(x01RangeButtonGO, rangeWidth / 2f, rangeButtonHeight * 0.8f, 20, 0, 1f, AnchorPresets.bottomLeft, true, resizeTextForBestFit: false);
            if (hideRange) rangeGO.SetActive(false);

            return sliderUID;
        }

        private static void ResizeSliderQuickButton(string transformTag, GameObject parentGO, float width, float height, float xOffset)
        {
            GameObject quickButtonsLeftMxGO = parentGO.transform.Find(transformTag).gameObject;
            ReformatUnityButton(quickButtonsLeftMxGO, width, height, 14, xOffset, 0, AnchorPresets.bottomLeft,true,resizeTextForBestFit:true);
        }

        public static UIDynamicToggle CreateToggle(GameObject parentGO, JSONStorableBool jsb, float width, float height, string label = null, int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, float checkBoxSize = 20f, bool rescaleText=true)
        {
            Transform transform = GameObject.Instantiate<Transform>(Globals.mvrScript.manager.configurableTogglePrefab);

            GameObject toggleGO = transform.gameObject;
            UIDynamicToggle toggleUID = toggleGO.transform.GetComponent<UIDynamicToggle>();
            if (label!=null) toggleUID.label = label;
            jsb.RegisterToggle(toggleUID.toggle);

            transform.SetParent(parentGO.transform, false);

            ResizeRecTransform(toggleGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            GameObject labelTextGO = toggleGO.transform.Find("Label").gameObject;
            ReformatText(labelTextGO, width - checkBoxSize, height, fontSize, 0f, 0f, AnchorPresets.middleRight, textAlginment: TextAnchor.MiddleLeft,rescaleText:rescaleText);

            GameObject checkBoxGO = toggleGO.transform.Find("Background").gameObject;
            ResizeRecTransform(checkBoxGO, checkBoxSize, checkBoxSize, 0f, 0f, AnchorPresets.middleLeft);

            GameObject checkMarkGO = toggleGO.transform.Find("Background/Checkmark").gameObject;
            ResizeRecTransform(checkMarkGO, checkBoxSize, checkBoxSize, 0f, 0f, AnchorPresets.centre);

            return toggleUID;
        }


        public static UIDynamicPopup CreatePopup(GameObject parentGO, JSONStorableStringChooser jssc, float width, float height, string label = null, int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool popupAbove = false, float labelProportion = 0.4f, float popupRowHeight = 15f, bool rescaleText=true)
        {
            Transform transform = GameObject.Instantiate<Transform>(Globals.mvrScript.manager.configurablePopupPrefab);

            GameObject popupGO = transform.gameObject;
            UIDynamicPopup uiPopup = popupGO.transform.GetComponent<UIDynamicPopup>();
           
            uiPopup.popup.selectColor = Color.cyan;

            jssc.RegisterPopup(uiPopup.popup);
            if (label != null) uiPopup.popup.label = label;

            transform.SetParent(parentGO.transform, false);

            ResizeRecTransform(popupGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            GameObject dropDownButtonGO = transform.Find("Button").gameObject;
            ReformatUnityButton(dropDownButtonGO, width * (1 - labelProportion), height * .9f, fontSize, 0f, 0f, AnchorPresets.middleRight, rescaleText:rescaleText);

            GameObject textGO = transform.Find("Text").gameObject;
            ReformatText(textGO, width * (labelProportion), height * .9f, fontSize, 0f, 0f, AnchorPresets.middleLeft,rescaleText:rescaleText);

            GameObject popupPanelGO = transform.Find("PopupPanel").gameObject;
            
            if (Globals.isVREnabled) popupPanelGO.transform.Translate(new Vector3(0, 0, 0.06f));

            if (popupAbove) ReformatPopupPanel(popupPanelGO, width * (1 - labelProportion), 300f, fontSize, 0f, height, anchorPreset: AnchorPresets.bottomRight, popupRowHeight: popupRowHeight, rescaleText: rescaleText);
            else ReformatPopupPanel(popupPanelGO, width * (1 - labelProportion), 300f, fontSize, 0f, -height, anchorPreset: AnchorPresets.topRight, popupRowHeight: popupRowHeight, rescaleText: rescaleText);

            return uiPopup;
        }
        public static UIDynamicPopup CreateScrollablePopup(GameObject parentGO, JSONStorableStringChooser jssc, float width, float height, string label = null, int labelFontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, float popupHeight = 300f, bool popupAbove = false, bool disableSlider = false, int anchorPreset = AnchorPresets.none, float labelProportion = 0.4f, float popupRowHeight = 15f,int popupFontSize = 20, bool rescaleText =true, float scrollBarWidth = 10f)
        {
            Transform transform = GameObject.Instantiate<Transform>(Globals.mvrScript.manager.configurableScrollablePopupPrefab);

            GameObject popupGO = transform.gameObject;
            UIDynamicPopup uiPopup = popupGO.transform.GetComponent<UIDynamicPopup>();
            
            jssc.RegisterPopup(uiPopup.popup);
            if (label != null) uiPopup.popup.label = label;

            float sliderHeight = 10f;
            if (disableSlider) sliderHeight = 0f;

            uiPopup.popup.selectColor = Color.cyan;
            transform.SetParent(parentGO.transform, false);

            ResizeRecTransform(popupGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            GameObject dropDownButtonGO = transform.Find("Button").gameObject;
            ReformatUnityButton(dropDownButtonGO, width * (1 - labelProportion), (height * .9f) - sliderHeight, popupFontSize, -2f, 0f, AnchorPresets.middleRight, true,rescaleText:rescaleText);

            GameObject textGO = transform.Find("Text").gameObject;
            ReformatText(textGO, width * (labelProportion), height * .9f, labelFontSize, 0f, 0f, AnchorPresets.middleLeft, rescaleText: rescaleText);

            GameObject popupPanelGO = transform.Find("PopupPanel").gameObject;
            GameObject scrollViewGO = popupPanelGO.transform.Find("Scroll View").gameObject;

            ScrollRect scrollRect = scrollViewGO.GetComponent<ScrollRect>();
            scrollRect.verticalScrollbarVisibility = ScrollRect.ScrollbarVisibility.AutoHideAndExpandViewport;

            if (popupAbove) ReformatPopupPanel(popupPanelGO, width * (1 - labelProportion), popupHeight, popupFontSize, 0f, height, anchorPreset: AnchorPresets.bottomRight, popupRowHeight: popupRowHeight, rescaleText: rescaleText, scrollBarWidth: scrollBarWidth);
            else ReformatPopupPanel(popupPanelGO, width * (1 - labelProportion), popupHeight, popupFontSize, 0f, -height, anchorPreset: AnchorPresets.topRight, popupRowHeight: popupRowHeight, rescaleText: rescaleText, scrollBarWidth: scrollBarWidth);

            GameObject sliderGO = transform.Find("Slider").gameObject;
            ReformatScrollPopupSlider(sliderGO, width * (1 - labelProportion), sliderHeight, 0f, 0f, anchorPreset: AnchorPresets.topRight, clearSprite: false);

            return uiPopup;
        }


        public static void RecursiveTrans(Transform transform, string spaces="")
        {

            foreach (Transform childTrans in transform)
            {
                RectTransform rt = transform.gameObject.GetComponent<RectTransform>();


                SuperController.LogMessage(spaces + childTrans.name);
                string newSpaces = spaces + "    ";
                RecursiveTrans(childTrans, newSpaces);
            }
        }

        public static UIDynamicPopup CreateFilterablePopup(GameObject parentGO, JSONStorableStringChooser jssc, float width, float height, string label = "", int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, float popupHeight = 300f, int anchorPreset = AnchorPresets.none, bool rescaleText = true)
        {
            float labelProportion = 0.3f;
            Transform transform = GameObject.Instantiate<Transform>(Globals.mvrScript.manager.configurableFilterablePopupPrefab);

            GameObject popupGO = transform.gameObject;

            UIDynamicPopup uiPopup = popupGO.transform.GetComponent<UIDynamicPopup>();
            uiPopup.label = label;
            uiPopup.popup.selectColor = Color.cyan;
            jssc.RegisterPopup(uiPopup.popup);

            transform.SetParent(parentGO.transform, false);

            ResizeRecTransform(popupGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            GameObject dropDownButtonGO = transform.Find("Button").gameObject;
            ReformatUnityButton(dropDownButtonGO, width * (1 - labelProportion), height * 0.6f, fontSize, 0f, 0f, AnchorPresets.bottomRight, false);

            GameObject labelTextGO = transform.Find("Text").gameObject;
            ReformatText(labelTextGO, width * labelProportion, height, fontSize, 0f, 0f, AnchorPresets.middleLeft, rescaleText: rescaleText);

            GameObject inputFieldGO = transform.Find("InputField").gameObject;
            ReformatFilterPopupInputField(inputFieldGO, width * (1 - labelProportion), height * 0.4f, fontSize, 0f, 0f, AnchorPresets.topRight);

            GameObject popupPanelGO = transform.Find("PopupPanel").gameObject;
            ReformatPopupPanel(popupPanelGO, width * (1 - labelProportion), popupHeight, fontSize, 0f, -height, anchorPreset: AnchorPresets.topRight, rescaleText: rescaleText);
            GameObject scrollViewGO = popupPanelGO.transform.Find("Scroll View").gameObject;

            ScrollRect scrollRect = scrollViewGO.GetComponent<ScrollRect>();
            scrollRect.verticalScrollbarVisibility = ScrollRect.ScrollbarVisibility.AutoHideAndExpandViewport;
            return uiPopup;
        }
        private static void ReformatPopupPanel(GameObject popupPanelGO, float width, float height, int fontSize = 12, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, float popupRowHeight = 15f, bool rescaleText = true, float scrollBarWidth = 10f)
        {
            ResizeRecTransform(popupPanelGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            GameObject buttonTemplateGO = popupPanelGO.transform.Find("PopupValueButtonTemplate").gameObject;
            ReformatUnityButton(buttonTemplateGO, width - scrollBarWidth, popupRowHeight, fontSize, clearSprite: true, anchorPreset: AnchorPresets.none, rescaleText: rescaleText);

            if (popupPanelGO.transform.Find("Scroll View") != null)
            {
                ReformatScrollView(popupPanelGO.transform.Find("Scroll View").gameObject, width, scrollBarWidth, height);
            }
        }

        private static void ReformatScrollView(GameObject scrollViewGO, float width, float vScrollBarWidth, float height = 25, float hScrollBarHeight = 0)
        {
            GameObject viewPortGO = scrollViewGO.transform.Find("Viewport").gameObject;
            ResizeRecTransform(viewPortGO, width-vScrollBarWidth, height- hScrollBarHeight);

            RectTransform viewPortRT = viewPortGO.transform.GetComponent<RectTransform>();
            viewPortRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height - hScrollBarHeight / 2);
            viewPortRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width - vScrollBarWidth / 2);

            if (scrollViewGO.transform.Find("Scrollbar Vertical") != null)
            {
                GameObject scrollBarVerticalGO = scrollViewGO.transform.Find("Scrollbar Vertical").gameObject;
                Scrollbar sb = scrollBarVerticalGO.transform.GetComponent<Scrollbar>();
                sb.enabled = vScrollBarWidth > 0f;
                
                RectTransform scrollBarRT = scrollBarVerticalGO.transform.GetComponent<RectTransform>();
                scrollBarRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, vScrollBarWidth);
                scrollBarRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
            }
            if (scrollViewGO.transform.Find("Scrollbar Horizontal") != null)
            {
                GameObject scrollBarVerticalGO = scrollViewGO.transform.Find("Scrollbar Horizontal").gameObject;
                Scrollbar sb = scrollBarVerticalGO.transform.GetComponent<Scrollbar>();
                sb.enabled = hScrollBarHeight > 0f;
                RectTransform scrollBarRT = scrollBarVerticalGO.transform.GetComponent<RectTransform>();
                scrollBarRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, hScrollBarHeight);
                scrollBarRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width  );

            }


        }

        private static void ReformatScrollPopupSlider(GameObject sliderGO, float width, float height, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool clearSprite = true)
        {
            ResizeRecTransform(sliderGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            GameObject handleGO = sliderGO.transform.Find("Handle Slide Area/Handle").gameObject;
            Image handleImage = handleGO.transform.GetComponent<Image>();
            if (clearSprite) handleImage.sprite = null;
        }

        private static void ReformatImage(GameObject imageGO, float width, float height, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool clearSprite = true)
        {
            ResizeRecTransform(imageGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);
            Image popupImage = imageGO.transform.GetComponent<Image>();
            if (clearSprite) popupImage.sprite = null;
            popupImage.color = new Color(0.9f, 0.9f, 0.9f);
        }
        private static void ReformatFilterPopupInputField(GameObject inputFieldGO, float width, float height, int fontSize = 12, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool rescaleText = true)
        {
            Image popupImage = inputFieldGO.transform.GetComponent<Image>();
            popupImage.sprite = null;

            ResizeRecTransform(inputFieldGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            GameObject buttonGO = inputFieldGO.transform.Find("Button").gameObject;
            ReformatUnityButton(buttonGO, 15f, height, fontSize - 4, -40f, clearSprite: false, textAlignment: TextAnchor.MiddleLeft);

            GameObject countTextGO = inputFieldGO.transform.Find("CountText").gameObject;
            ReformatText(countTextGO, 40f, height, fontSize - 4, anchorPreset: AnchorPresets.middleRight, rescaleText: rescaleText);

            GameObject countTextBGGO = inputFieldGO.transform.Find("CountTextBG").gameObject;
            ReformatImage(countTextBGGO, 40f, height, anchorPreset: AnchorPresets.middleRight, clearSprite: true);
            GameObject maxCountBGGO = inputFieldGO.transform.Find("MaxIndicator").gameObject;
            ReformatImage(maxCountBGGO, 40f, height, anchorPreset: AnchorPresets.middleRight, clearSprite: true);

            GameObject inputTextGO = inputFieldGO.transform.Find("Text").gameObject;
            ReformatText(inputTextGO, width - 55f, height, fontSize, 0f, 0f, AnchorPresets.middleLeft, textAlginment: TextAnchor.MiddleLeft, rescaleText: rescaleText);

            GameObject placeholderTextGO = inputFieldGO.transform.Find("Placeholder").gameObject;
            ReformatText(placeholderTextGO, width - 55f, height, fontSize, 0f, 0f, AnchorPresets.middleLeft, textAlginment: TextAnchor.MiddleLeft, rescaleText: rescaleText);
        }

        public static void ReformatText(GameObject textGO, float width, float height, int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, TextAnchor textAlginment = TextAnchor.MiddleCenter, bool resizeTextForBestFit = false, bool rescaleText = true)
        {

            RectTransform textRT = textGO.transform.GetComponent<RectTransform>();

            if (rescaleText)
            {
                textRT.localScale = new Vector3(0.5f, 0.5f, 0.5f);
                width = width * 2;
                height = height * 2;
            }

            ResizeRecTransform(textGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            Text text = textGO.transform.GetComponent<Text>();
            text.alignment = textAlginment;
            text.fontSize = fontSize;
            text.resizeTextForBestFit = resizeTextForBestFit;
        }
        private static void ReformatUnityButton(GameObject buttonGO, float width, float height, int fontSize = 12, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool clearSprite = true, TextAnchor textAlignment = TextAnchor.MiddleCenter, bool resizeTextForBestFit = false, bool rescaleText=true)
        {
            ResizeRecTransform(buttonGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            if (clearSprite)
            {
                Image buttonImage = buttonGO.transform.GetComponent<Image>();
                buttonImage.sprite = null;
            }

            Transform textTransform = buttonGO.transform.Find("Text");
            ReformatText(textTransform.gameObject, width - 4f, height, fontSize, 2f, 0f, AnchorPresets.centre, textAlignment,resizeTextForBestFit, rescaleText);
        }

        public static void ResizeRecTransform(GameObject go, float width, float height, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none)
        {
            RectTransform rt = go.GetComponent<RectTransform>();
            ResizeRecTransform(rt, width, height, anchorXOffset, anchorYOffset, anchorPreset);

        }

        public static void ResizeRecTransform(RectTransform rt, float width, float height, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none)
        {
            rt.sizeDelta = Vector2.zero;
            if (anchorPreset != AnchorPresets.none)
            {
                
                rt.anchorMin = AnchorPresets.GetAnchorMin(anchorPreset);
                rt.anchorMax = AnchorPresets.GetAnchorMax(anchorPreset);
                rt.pivot = AnchorPresets.GetPivot(anchorPreset);
            }
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
            rt.anchoredPosition = new Vector2(anchorXOffset, anchorYOffset);
        }

        public static void ResizeWidth(GameObject go, float width)
        {
            RectTransform rt = go.GetComponent<RectTransform>();
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);
        }

        public static void ResizeHeight(GameObject go, float height)
        {
            RectTransform rt = go.GetComponent<RectTransform>();
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
        }

        public static void ResetBottomOffset(GameObject go, float offset)
        {
            RectTransform rt = go.GetComponent<RectTransform>();
            rt.offsetMin = new Vector2(rt.offsetMin.x, offset);
        }

        public static void ResetTopOffset(GameObject go, float offset)
        {
            RectTransform rt = go.GetComponent<RectTransform>();
            rt.offsetMax = new Vector2(rt.offsetMax.x, offset);
        }

        public static void ResetAnchorPosition(GameObject go, float xPosition, float yPosition)
        {
            RectTransform rt = go.GetComponent<RectTransform>();
            rt.anchoredPosition = new Vector2(xPosition, yPosition);
        }
        public static void ResetAnchorType(GameObject go, int anchorPreset)
        {
            RectTransform rt = go.GetComponent<RectTransform>();
            rt.anchorMin = AnchorPresets.GetAnchorMin(anchorPreset);
            rt.anchorMax = AnchorPresets.GetAnchorMax(anchorPreset);
            rt.pivot = AnchorPresets.GetPivot(anchorPreset);
        }

    }

    public class TriStateToggle:UIDynamic
    {
        private JSONStorableBool toggleJSB;

        public UIDynamicToggle toggleUID { get; private set; }

        public Toggle toggle { get { return toggleUID.toggle; } }

        public Color backgroundColor { get { return toggleUID.backgroundColor; } set { toggleUID.backgroundColor = value; } }
        public Color textColor { get { return toggleUID.textColor; } set { toggleUID.textColor = value; } }

        public string label { get { return toggleUID.label; } set { toggleUID.label = value; } }

        public Text labelText { get { return toggleUID.labelText; } set { toggleUID.labelText = value; } }

        public new float height { get { return toggleUID.height; } set { toggleUID.height = value; } }

        public TriStateToggle (JSONStorableBool _toggleJSB, UIDynamicToggle _toggleUID)
        {
            toggleJSB = _toggleJSB;
            toggleUID = _toggleUID;
        }

        public void RefreshToggle(int triState)
        {
            var backgroundTrans = toggle.transform.Find("Background");
            var image = backgroundTrans.gameObject.GetComponent<Image>();
            if (triState != TriState.inactive && toggle.interactable) image.color = Color.cyan;
            else image.color = Color.white;

            if (triState == TriState.active) toggleJSB.valNoCallback = true;
            else toggleJSB.valNoCallback = false;
        }

    }

    public static class AnchorPresets
    {
        public const int none = 0;
        public const int topLeft = 1;
        public const int topMiddle = 2;
        public const int topRight = 3;
        public const int middleLeft = 4;
        public const int centre = 5;
        public const int middleRight = 6;
        public const int bottomLeft = 7;
        public const int bottomMiddle = 8;
        public const int bottomRight = 9;

        public const int vStretchMiddle = 11;
        public const int hStretchMiddle = 12;
        public const int vStretchRight = 13;
        public const int vStretchLeft = 14;
        public const int hStretchTop = 15;
        public const int hStretchBottom = 16;
        public const int vhStretch = 17;

        public static Vector2 GetAnchorMin(int anchorPreset)
        {
            Vector2 anchorMin = new Vector2();

            if (anchorPreset == topLeft) anchorMin = new Vector2(0f, 1f);
            if (anchorPreset == topMiddle) anchorMin = new Vector2(0.5f, 1f);
            if (anchorPreset == topRight) anchorMin = new Vector2(1f, 1f);
            if (anchorPreset == middleLeft) anchorMin = new Vector2(0f, 0.5f);
            if (anchorPreset == centre) anchorMin = new Vector2(0.5f, 0.5f);
            if (anchorPreset == middleRight) anchorMin = new Vector2(1f, 0.5f);
            if (anchorPreset == bottomLeft) anchorMin = new Vector2(0f, 0f);
            if (anchorPreset == bottomMiddle) anchorMin = new Vector2(0.5f, 0f);
            if (anchorPreset == bottomRight) anchorMin = new Vector2(1f, 0f);

            if (anchorPreset == vStretchMiddle) anchorMin = new Vector2(0.5f, 0f);
            if (anchorPreset == vStretchRight) anchorMin = new Vector2(1f, 0f);
            if (anchorPreset == vStretchLeft) anchorMin = new Vector2(0f, 0f);
            if (anchorPreset == hStretchMiddle) anchorMin = new Vector2(0f, 0.5f);
            if (anchorPreset == hStretchTop) anchorMin = new Vector2(0f, 1f);
            if (anchorPreset == hStretchBottom) anchorMin = new Vector2(0f, 0f);
            if (anchorPreset == vhStretch) anchorMin = new Vector2(0f, 0f);

            return anchorMin;
        }

        public static Vector2 GetAnchorMax(int anchorPreset)
        {
            Vector2 anchorMax = new Vector2();

            if (anchorPreset == topLeft) anchorMax = new Vector2(0f, 1f);
            if (anchorPreset == topMiddle) anchorMax = new Vector2(0.5f, 1f);
            if (anchorPreset == topRight) anchorMax = new Vector2(1f, 1f);
            if (anchorPreset == middleLeft) anchorMax = new Vector2(0f, 0.5f);
            if (anchorPreset == centre) anchorMax = new Vector2(0.5f, 0.5f);
            if (anchorPreset == middleRight) anchorMax = new Vector2(1f, 0.5f);
            if (anchorPreset == bottomLeft) anchorMax = new Vector2(0f, 0f);
            if (anchorPreset == bottomMiddle) anchorMax = new Vector2(0.5f, 0f);
            if (anchorPreset == bottomRight) anchorMax = new Vector2(1f, 0f);

            if (anchorPreset == vStretchMiddle) anchorMax = new Vector2(0.5f, 1f);
            if (anchorPreset == vStretchRight) anchorMax = new Vector2(1f, 1f);
            if (anchorPreset == vStretchLeft) anchorMax = new Vector2(0f, 1f);
            if (anchorPreset == hStretchMiddle) anchorMax = new Vector2(1f, 0.5f);
            if (anchorPreset == hStretchTop) anchorMax = new Vector2(1f, 1f);
            if (anchorPreset == hStretchBottom) anchorMax = new Vector2(1f, 0f);
            if (anchorPreset == vhStretch) anchorMax = new Vector2(1f, 1f);

            return anchorMax;
        }

        public static Vector2 GetPivot(int anchorPreset)
        {
            Vector2 pivot = new Vector2();

            if (anchorPreset == topLeft) pivot = new Vector2(0f, 1f);
            if (anchorPreset == topMiddle) pivot = new Vector2(0.5f, 1f);
            if (anchorPreset == topRight) pivot = new Vector2(1f, 1f);
            if (anchorPreset == middleLeft) pivot = new Vector2(0f, 0.5f);
            if (anchorPreset == centre) pivot = new Vector2(0.5f, 0.5f);
            if (anchorPreset == middleRight) pivot = new Vector2(1f, 0.5f);
            if (anchorPreset == bottomLeft) pivot = new Vector2(0f, 0f);
            if (anchorPreset == bottomMiddle) pivot = new Vector2(0.5f, 0f);
            if (anchorPreset == bottomRight) pivot = new Vector2(1f, 0f);

            if (anchorPreset == vStretchMiddle) pivot = new Vector2(0.5f, 1f);
            if (anchorPreset == hStretchMiddle) pivot = new Vector2(0f, 0.5f);
            if (anchorPreset == vStretchRight) pivot = new Vector2(1f, 1f);
            if (anchorPreset == vStretchLeft) pivot = new Vector2(0f, 1f);
            if (anchorPreset == hStretchTop) pivot = new Vector2(0f, 1f);
            if (anchorPreset == hStretchBottom) pivot = new Vector2(0f, 0f);
            if (anchorPreset == vhStretch) pivot = new Vector2(0f, 1f);

            return pivot;
        }

    }

    public class MainBrowserEnterExit : MonoBehaviour, IPointerEnterHandler, IPointerExitHandler
    {
        public void OnPointerEnter(PointerEventData eventData)
        {
            Globals.browserPointerActive= true;
        }

        public void OnPointerExit(PointerEventData eventData)
        {
            Globals.browserPointerActive = false;
        }
    }

    public class PrimaryActionPluginPresetsPointerEnterExitHelp : MonoBehaviour, IPointerEnterHandler, IPointerExitHandler
    {
        private ResourceTypeSelections resourceTypeSelections;
        private JSONStorableString helpTextJSS;
        private bool sessionPluginPreset = false;
        public void Init(JSONStorableString _helpTextJSS, ResourceTypeSelections _resourceTypeSelections, bool sessionPluginPreset)
        {
            helpTextJSS = _helpTextJSS;
            resourceTypeSelections = _resourceTypeSelections;
            this.sessionPluginPreset = sessionPluginPreset;
        }
        public void OnPointerEnter(PointerEventData eventData)
        {
            if (resourceTypeSelections != null)
            {
                string pluginPresetType = "Scene";
                if (sessionPluginPreset) pluginPresetType = "Session";

                ResourceVersionGroupEntry rvge = resourceTypeSelections.GetFirstSelectedRVGE();
                if (rvge != null)
                {
                    helpTextJSS.val = "Load the currently selected preset '"+rvge.resourceFullPathReference+"' as a "+pluginPresetType+" Plugin Preset.";

                    return;
                }
            }
            helpTextJSS.val = "Select a Resource to Load Presets";
        }

        public void OnPointerExit(PointerEventData eventData)
        {
            helpTextJSS.val = "";
        }
    }


    public class PrimaryActionPointerEnterExitHelp : MonoBehaviour, IPointerEnterHandler, IPointerExitHandler
    {
        private ResourceTypeSelections resourceTypeSelections;
        private JSONStorableString helpTextJSS;
        private ResourcePrimaryActionUI resourcePrimaryActionUI;
        public void Init(JSONStorableString _helpTextJSS, ResourceTypeSelections _resourceTypeSelections, ResourcePrimaryActionUI _resourcePrimaryActionUI)
        {
            helpTextJSS = _helpTextJSS;
            resourceTypeSelections = _resourceTypeSelections;
            resourcePrimaryActionUI = _resourcePrimaryActionUI;
        }
        public void OnPointerEnter(PointerEventData eventData)
        {
            if (resourceTypeSelections != null)
            {
                ResourceVersionGroupEntry rvge = resourceTypeSelections.GetFirstSelectedRVGE();
                if (rvge != null)
                {
                    if (rvge.resourceType == ResourceType.scene) helpTextJSS.val = "Load Scene: '" + rvge.resourceFullPathReference + "'";
                    else if (rvge.resourceCategory == ResourceCategory.preset)
                    {
                        if (resourcePrimaryActionUI.targetAtomJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.mainValFlag) helpTextJSS.val = "Load Preset'" + rvge.resourceFullPathReference + "' to Atom '" + resourcePrimaryActionUI.targetAtomJSMultiEnum.mainVal + "'";
                        else helpTextJSS.val = "Create a target atom to apply preset '" + rvge.resourceFullPathReference + "'";
                    }
                    else if (rvge.resourceCategory == ResourceCategory.plugins) helpTextJSS.val = "Load Plugin: '" + rvge.resourceFullPathReference + "'";
                    return;
                }               
            }
            helpTextJSS.val = "Select Resource to Load";
        }

        public void OnPointerExit(PointerEventData eventData)
        {
            helpTextJSS.val = "";
        }
    }
    public class SetSingleSelectionsEnter : MonoBehaviour, IPointerEnterHandler, IPointerExitHandler
    {

        private ResourceSelections resourceSelectionsUI = null;
        public void Init(ResourceSelections resourceSelectionsUI)
        {
            this.resourceSelectionsUI = resourceSelectionsUI;

        }

        public void OnPointerEnter(PointerEventData eventData)
        {
            if (resourceSelectionsUI != null)
            {
                resourceSelectionsUI.SetSingleFocus();
            }

        }

        public void OnPointerExit(PointerEventData eventData)
        {
            resourceSelectionsUI.SetAllFocus();
        }
    }

    public class SetAllSelectionsEnter : MonoBehaviour, IPointerEnterHandler, IPointerExitHandler
    {

        private ResourceSelections resourceSelectionsUI =null;
        public void Init(ResourceSelections resourceSelectionsUI)
        {
            this.resourceSelectionsUI  = resourceSelectionsUI;

        }

        public void OnPointerEnter(PointerEventData eventData)
        {
            if (resourceSelectionsUI != null)
            {
                resourceSelectionsUI.SetAllFocus();
            }

        }

        public void OnPointerExit(PointerEventData eventData)
        {
            resourceSelectionsUI.SetSingleFocus();
        }
    }

    public class FavPointerEnterExitHelp : MonoBehaviour, IPointerEnterHandler, IPointerExitHandler
    {
        private ResourceTypeSelections resourceSelections;
        private BrowserFiltersUI browserFiltersUI;
        private JSONStorableString helpTextJSS;
        public void Init(JSONStorableString _helpTextJSS, ResourceTypeSelections _resourceSelections, BrowserFiltersUI _browserFiltersUI)
        {
            helpTextJSS = _helpTextJSS;
            resourceSelections = _resourceSelections;
            browserFiltersUI = _browserFiltersUI;
        }
        public void OnPointerEnter(PointerEventData eventData)
        {
            helpTextJSS.val = "Set selected Resource(s) as Favourites.";
            if (browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi || resourceSelections.isFavourite == TriState.active) helpTextJSS.val = "Unset selected Resource(s) as Favourites.";
            else if (resourceSelections.isFavourite == TriState.indeterminate)
            {
                int selectedResourceCount = resourceSelections.GetSelectedCount();
                int notFavouriteResourcesCount = selectedResourceCount - resourceSelections.GetFavouriteCount();
                helpTextJSS.val = helpTextJSS.val + " WARNING: The Favourite state of " + notFavouriteResourcesCount.ToString() + " resource(s) out of " + selectedResourceCount.ToString() + " selected resources will be changed. This can't be easily reversed without manually identifying which resource(s) were not previously set as Favourites.";

                Image image = helpTextJSS.text.transform.parent.gameObject.GetComponent<Image>();
                image.color = Color.red;
            }
        }

        public void OnPointerExit(PointerEventData eventData)
        {
            helpTextJSS.val = "";

            Image image = helpTextJSS.text.transform.parent.gameObject.GetComponent<Image>();
            image.color = Color.grey;
        }
    }
    public class NewPointerEnterExitHelp : MonoBehaviour, IPointerEnterHandler, IPointerExitHandler
    {
        private ResourceTypeSelections resourceSelections;
        private BrowserFiltersUI browserFiltersUI;
        private JSONStorableString helpTextJSS;
        public void Init(JSONStorableString _helpTextJSS, ResourceTypeSelections _resourceSelections, BrowserFiltersUI _browserFiltersUI)
        {
            helpTextJSS = _helpTextJSS;
            resourceSelections = _resourceSelections;
            browserFiltersUI = _browserFiltersUI;
        }
        public void OnPointerEnter(PointerEventData eventData)
        {
            helpTextJSS.val = "Set selected Resource(s) as New or Updated.";
            if (browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi || resourceSelections.isNew == TriState.active) helpTextJSS.val = "Unset selected Resource(s) as New or Updated.";
            else if (resourceSelections.isNew == TriState.indeterminate)
            {
                int selectedResourceCount = resourceSelections.GetSelectedCount();
                int notNewResourcesCount = selectedResourceCount - resourceSelections.GetNewCount();
                helpTextJSS.val = helpTextJSS.val + " WARNING: The New state of " + notNewResourcesCount.ToString() + " resource(s) out of " + selectedResourceCount.ToString() + " selected resources will be changed. This can't be easily reversed without manually identifying which resource(s) were not previously set as New.";

                Image image = helpTextJSS.text.transform.parent.gameObject.GetComponent<Image>();
                image.color = Color.red;
            }
        }

        public void OnPointerExit(PointerEventData eventData)
        {
            helpTextJSS.val = "";

            Image image = helpTextJSS.text.transform.parent.gameObject.GetComponent<Image>();
            image.color = Color.grey;
        }
    }
    public class HiddenPointerEnterExitHelp : MonoBehaviour, IPointerEnterHandler, IPointerExitHandler
    {
        private ResourceTypeSelections resourceSelections;
        private BrowserFiltersUI browserFiltersUI;
        private JSONStorableString helpTextJSS;
        public void Init(JSONStorableString _helpTextJSS, ResourceTypeSelections _resourceSelections, BrowserFiltersUI _browserFiltersUI)
        {
            helpTextJSS = _helpTextJSS;
            resourceSelections = _resourceSelections;
            browserFiltersUI = _browserFiltersUI;
        }
        public void OnPointerEnter(PointerEventData eventData)
        {
            helpTextJSS.val = "Set selected Resource(s) as Hidden.";
            if (browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi || resourceSelections.isHidden == TriState.active) helpTextJSS.val = "Unset selected Resource(s) as Hidden.";
            else if (resourceSelections.isHidden == TriState.indeterminate)
            {
                int selectedResourceCount = resourceSelections.GetSelectedCount();
                int notHiddenResourcesCount = selectedResourceCount - resourceSelections.GetHiddenCount();
                helpTextJSS.val = helpTextJSS.val + " WARNING: The Hidden state of " + notHiddenResourcesCount.ToString() + " resource(s) out of " + selectedResourceCount.ToString() + " selected resources will be changed. This can't be easily reversed without manually identifying which resource(s) were not previously set as Hidden.";

                Image image = helpTextJSS.text.transform.parent.gameObject.GetComponent<Image>();
                image.color = Color.red;
            }
        }

        public void OnPointerExit(PointerEventData eventData)
        {
            helpTextJSS.val = "";

            Image image = helpTextJSS.text.transform.parent.gameObject.GetComponent<Image>();
            image.color = Color.grey;
        }
    }

    public class TagPointerEnterExitHelp : MonoBehaviour, IPointerEnterHandler, IPointerExitHandler
    {
        private Tag baTag;
        private ResourceTypeSelections resourceSelections;
        private BrowserFiltersUI browserFiltersUI;
        private JSONStorableString helpTextJSS;
        public void Init(Tag _tag, JSONStorableString _helpTextJSS, ResourceTypeSelections _resourceSelections, BrowserFiltersUI _browserFiltersUI)
        {
            helpTextJSS = _helpTextJSS;
            baTag = _tag;
            resourceSelections = _resourceSelections;
            browserFiltersUI = _browserFiltersUI;
        }
        public void OnPointerEnter(PointerEventData eventData)
        {
            if (baTag.tagCategory != TagCategory.combi)
            {
                helpTextJSS.val = "Set Tag '" + baTag.tagName + "' for the selected Resources.";
                if ((browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi && resourceSelections.IsFirstTagged(baTag)==TriState.active)  || resourceSelections.IsTagged(baTag) == TriState.active) helpTextJSS.val = "Unset Tag '" + baTag.tagName + "' for the selected Resources.";
                else if (browserFiltersUI.selectModeJSEnum.val == BrowserSelectMode.multi && resourceSelections.IsTagged(baTag) == TriState.indeterminate)
                {
                    int selectedResourceCount = resourceSelections.GetSelectedCount();
                    int notTaggedResourcesCount = selectedResourceCount - resourceSelections.GetTaggedCount(baTag);
                    helpTextJSS.val = helpTextJSS.val + " WARNING: The tagged state of " + notTaggedResourcesCount.ToString() + " resource(s) out of " + selectedResourceCount.ToString() + " selected resources will be changed. This can't be easily reversed without manually identifying which resource(s) were not previously tagged.";

                    Image image = helpTextJSS.text.transform.parent.gameObject.GetComponent<Image>();
                    image.color = Color.red;

                    return;
                }

                if (browserFiltersUI.currentResourceCategory == ResourceCategory.clothing) helpTextJSS.val = helpTextJSS.val + " BA Tags will be synced with VAM Clothing tags on VAM restart. VAM Tag changes can be synced using the Rescan button. Clothing Tag changes in both VAM and BA without syncing may result in conflicts.";
            }
            else
            {
                helpTextJSS.val = "Combi tags are automatically set based on the rules defined by the user in the Browser Assist Plugin UI.";
            }

        }

        public void OnPointerExit(PointerEventData eventData)
        {
            helpTextJSS.val = "";

            Image image = helpTextJSS.text.transform.parent.gameObject.GetComponent<Image>();
            image.color = Color.grey;
        }
    }

    public class GeneralPointerEnterExitHelp : MonoBehaviour, IPointerEnterHandler, IPointerExitHandler
    {
        private string helpTextOnEnter="";
        private string helpTextOnExit="";
        private JSONStorableString helpTextJSS;
        private bool warning = false;
        public void Init(string helpTextOnEnter, JSONStorableString _helpTextJSS, bool warning, string helpTextOnExit = "")
        {
            helpTextJSS = _helpTextJSS;
            this.helpTextOnEnter = helpTextOnEnter;
            this.helpTextOnExit = helpTextOnExit;
            this.warning = warning;
        }
        public void OnPointerEnter(PointerEventData eventData)
        {

            helpTextJSS.val = helpTextOnEnter;

            if (helpTextJSS.text != null)
            {
                Image image = helpTextJSS.text.transform.parent.gameObject.GetComponent<Image>();
                if (image != null)
                {
                    if (warning) image.color = Color.red;
                    else image.color = Color.gray;
                }
            }
        }

        public void OnPointerExit(PointerEventData eventData)
        {
            helpTextJSS.val = helpTextOnExit;

            if (helpTextJSS.text != null)
            {
                Image image = helpTextJSS.text.transform.parent.gameObject.GetComponent<Image>();
                if (image != null) image.color = Color.gray;
            }
        }
    }

    public class LaunchButtonPointerEnterExitHelp : MonoBehaviour, IPointerEnterHandler, IPointerExitHandler
    {

        public void OnPointerEnter(PointerEventData eventData)
        {
            SuperController.singleton.SetHelpHUDText("Launch BrowserAssist by JayJayWon");
        }

        public void OnPointerExit(PointerEventData eventData)
        {
            SuperController.singleton.SetHelpHUDText("");
        }
    }


    public class ResourceButtonPointerEnterExitClick : MonoBehaviour, IPointerEnterHandler, IPointerExitHandler, IPointerClickHandler, IPointerDownHandler, IPointerUpHandler
    {
        private UIDynamicButton button;
        private Dictionary<UIDynamicButton, ResourceVersionGroupEntry> rvgeDict;
        private JSONStorableString helpTextJSS;
        private BrowserUI browserUI;

        private bool firstLeftClickInitiated = false;
        private float firstLeftClickTime;
        private JSONStorableBool rvgeHiddenJSB ;
        private JSONStorableBool rvgeFavJSB ;
        private JSONStorableBool rvgeNewJSB;
        private UIDynamicToggle rvgeHiddenToggle;
        private UIDynamicToggle rvgeFavToggle;
        private UIDynamicToggle rvgeNewToggle;

        private float pointerDownTimeFloat;
        private bool longPressDetectActive = false;

        private ResourceVersionGroupEntry rvge { get { return rvgeDict[button]; } }
        private bool togglesVisible = false;
        public void Init(UIDynamicButton _button, Dictionary<UIDynamicButton, ResourceVersionGroupEntry> _rvgeDict, JSONStorableString _helpText, BrowserUI _browserUI)
        {
            button = _button;
            rvgeDict = _rvgeDict;
            helpTextJSS = _helpText;
            browserUI = _browserUI;
            rvgeHiddenJSB = new JSONStorableBool("rvgeHidden", false, HiddenCallback);
            rvgeFavJSB = new JSONStorableBool("rvgeFav", false,FavCallback);
            rvgeNewJSB = new JSONStorableBool("rvgeNew", false, NewCallback);
        }
        private void NewCallback(bool val)
        {
            rvge.isNew = val;
            browserUI.resourceBrowserUI.SetNewIcon(rvge);
            browserUI.resourceActionsUI.RefreshUIResourceSelectionChange();
        }
        private void HiddenCallback(bool val)
        {
            rvge.isHidden = val;
            browserUI.resourceBrowserUI.SetHideFavColorTags(rvge);
            browserUI.resourceActionsUI.RefreshUIResourceSelectionChange();

            if (!browserUI.browserFiltersUI.showHiddenJSB.val)
            {
                browserUI.RefreshFilterResults();
                HideToggles();
            }
        }

        private void FavCallback(bool val)
        {
            rvge.isFavorite = val;
            browserUI.resourceBrowserUI.SetHideFavColorTags(rvge);
            browserUI.resourceActionsUI.RefreshUIResourceSelectionChange();

            if (browserUI.browserFiltersUI.onlyFavouritesJSB.val)
            {
                browserUI.RefreshFilterResults();
                HideToggles();  
            }
        }
        public void OnPointerEnter(PointerEventData eventData)
        {
            if (rvgeDict.ContainsKey(button))
            {
                if (browserUI.isVREnabled)
                {
                    if (browserUI.browserFiltersUI.selectModeJSEnum.val == BrowserSelectMode.instant)
                        helpTextJSS.val = rvge.resourceFullPathReference + "\nClick to load. Click and hold for Hide & Favorite.";
                    else helpTextJSS.val = rvge.resourceFullPathReference + "\nSingle click to select. Double click to load. Click and hold for Hide & Favorite.";
                }
                else
                {
                    if (browserUI.browserFiltersUI.selectModeJSEnum.val == BrowserSelectMode.instant)
                        helpTextJSS.val = rvge.resourceFullPathReference + "\nCick to load. Right click for Hide & Favorite.";
                    else helpTextJSS.val = rvge.resourceFullPathReference + "\nSingle click to select. Double click to load. Right click for Hide & Favorite.";
                }
                
            }
            Image image = helpTextJSS.text.transform.parent.gameObject.GetComponent<Image>();
            image.color = Color.gray;
        }

        public void OnPointerExit(PointerEventData eventData)
        {
            helpTextJSS.val = "";
            firstLeftClickInitiated = false;
            if (togglesVisible) HideToggles();
            Image image = helpTextJSS.text.transform.parent.gameObject.GetComponent<Image>();
            image.color = Color.gray;
            longPressDetectActive = false;
        }

        public virtual void OnPointerClick(PointerEventData eventData)
        {
            if (rvgeDict.ContainsKey(button))
            {
                if (browserUI.browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.instant && eventData.button == PointerEventData.InputButton.Left)
                {
                    if (!firstLeftClickInitiated)
                    {
                        firstLeftClickInitiated = true;
                        firstLeftClickTime =Time.unscaledTime*1000;
                        return;
                    }
                    float interval = (Time.unscaledTime*1000) - firstLeftClickTime;
                    if (interval < 350f)
                    {
                        browserUI.resourceActionsUI.primaryActionUITab.Activate(rvge);
                        longPressDetectActive = false;
                        firstLeftClickInitiated = false;
                    }
                    else firstLeftClickTime = Time.unscaledTime * 1000;
                }
                else if (eventData.button == PointerEventData.InputButton.Right)
                {
                    ToggleFavHideActivation();
                }
            }
            
        }

        private void ToggleFavHideActivation()
        {
            if (!togglesVisible)
            {
                rvgeHiddenJSB.valNoCallback = rvge.isHidden;
                rvgeFavJSB.valNoCallback = rvge.isFavorite;
                rvgeNewJSB.valNoCallback = rvge.isNew;

                rvgeHiddenToggle = UI.CreateToggle(button.button.gameObject, rvgeHiddenJSB, button.height * 0.8f, 20f, "Hide", 20, 0f, 56f, AnchorPresets.bottomMiddle);
                rvgeFavToggle = UI.CreateToggle(button.button.gameObject, rvgeFavJSB, button.height * 0.8f, 20f, "Favorite", 20, 0f, 36f, AnchorPresets.bottomMiddle);

                GeneralPointerEnterExitHelp contextHelpFav = rvgeFavToggle.toggle.gameObject.AddComponent<GeneralPointerEnterExitHelp>();
                GeneralPointerEnterExitHelp contextHelpHide = rvgeHiddenToggle.toggle.gameObject.AddComponent<GeneralPointerEnterExitHelp>();

                if (!Globals.addonPackageFilePrefsFolderAccess || !Globals.customFolderAccess || !Globals.scenesFolderAccess)
                {
                    rvgeFavToggle.toggle.interactable = false;
                    rvgeHiddenToggle.toggle.interactable = false;
                    
                    contextHelpFav.Init("Enable BA Sym Links using .bat file provided in PluginData folder to allow toggling the VAM Favorite state from within Browser Assist. See BA VAMHub page for detailed instructions.", helpTextJSS,false);
                    
                    contextHelpHide.Init("Enable BA Sym Links using .bat file provided in PluginData folder to allow toggling the VAM Hidden state from within Browser Assist. See BA VAMHub page for detailed instructions.", helpTextJSS, false);
                }
                else
                {
                    contextHelpFav.Init("Toggle setting the resource as a Favourite", helpTextJSS, false);

                    contextHelpHide.Init("Toggle setting the resource as Hidden", helpTextJSS, false);
                }


                string newLabel = "New";
                if (rvge.isMultiVARVersion) newLabel = "Updated";
                rvgeNewToggle = UI.CreateToggle(button.button.gameObject, rvgeNewJSB, button.height * 0.8f, 20f, newLabel, 20, 0f, 16f, AnchorPresets.bottomMiddle);
                GeneralPointerEnterExitHelp contextHelpNew = rvgeNewToggle.toggle.gameObject.AddComponent<GeneralPointerEnterExitHelp>();
                contextHelpNew.Init("Toggle setting the resource as New/Updated", helpTextJSS, false);


                togglesVisible = true;
            }
            else HideToggles();
        }
        private void HideToggles()
        {
            if (rvgeHiddenToggle!=null) Destroy(rvgeHiddenToggle.gameObject);
            if (rvgeFavToggle != null)  Destroy(rvgeFavToggle.gameObject);
            if (rvgeNewToggle != null) Destroy(rvgeNewToggle.gameObject);
            togglesVisible = false;
        }

        public void OnPointerDown(PointerEventData eventData)
        {
            if (browserUI.browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.instant) longPressDetectActive = true;
            pointerDownTimeFloat = Time.unscaledTime * 1000;
        }

        public void OnPointerUp(PointerEventData eventData)
        {
            float timeDown = (Time.unscaledTime * 1000) - pointerDownTimeFloat;

            if (timeDown > 500f && longPressDetectActive) ToggleFavHideActivation();
            longPressDetectActive = false;
        }
    }

    public class ClothingItemPresetPointerEnterExitClick : MonoBehaviour, IPointerEnterHandler, IPointerExitHandler, IPointerClickHandler, IPointerDownHandler, IPointerUpHandler
    {
        private UIDynamicButton button;

        private JSONStorableString helpTextJSS;
        private BrowserUI browserUI;

        private bool firstLeftClickInitiated = false;
        private float firstLeftClickTime;
        private JSONStorableBool rvgeHiddenJSB;
        private JSONStorableBool rvgeFavJSB;

        private UIDynamicToggle rvgeHiddenToggle;
        private UIDynamicToggle rvgeFavToggle;

        private float pointerDownTimeFloat;
        private bool longPressDetectActive = false;

        private ResourceVersionGroupEntry rvge ;
        private bool togglesVisible = false;
        private ResourceClothingPresetsActionUI resourceClothingPresetsActionUI;

        private ResourceClothingPresetDetailUI resourceClothingPresetDetailUI;

        public void Init(UIDynamicButton _button, ResourceVersionGroupEntry _rvge, JSONStorableString _helpText, BrowserUI _browserUI, ResourceClothingPresetsActionUI resourceClothingPresetsActionUI, ResourceClothingPresetDetailUI resourceClothingPresetDetailUI)
        {
            button = _button;
            rvge = _rvge;
            helpTextJSS = _helpText;
            browserUI = _browserUI;
            rvgeHiddenJSB = new JSONStorableBool("rvgeHidden", false, HiddenCallback);
            rvgeFavJSB = new JSONStorableBool("rvgeFav", false, FavCallback);
            this.resourceClothingPresetsActionUI = resourceClothingPresetsActionUI;
            this.resourceClothingPresetDetailUI = resourceClothingPresetDetailUI;
        }
        private void HiddenCallback(bool val)
        {
            rvge.isHidden = val;
            HideToggles();
            resourceClothingPresetsActionUI.RefreshUI();
 //           SetHideFavColorTags();
        }

        private void FavCallback(bool val)
        {
            rvge.isFavorite = val;

            HideToggles();
            resourceClothingPresetsActionUI.RefreshUI();
//            SetHideFavColorTags();
        }

        public void SetHideFavColorTags()
        {
            Transform hiddenImageTransform = button.transform.Find("HiddenIconImage");
            Transform favImageTransform = button.transform.Find("FavIconImage");
            Transform primaryIconTransform = button.transform.Find("PrimaryIcon");

            favImageTransform.gameObject.SetActive(rvge.isFavorite);
            hiddenImageTransform.gameObject.SetActive(rvge.isHidden || rvge.isTagHidden || rvge.isCreatorHidden);
            primaryIconTransform.gameObject.SetActive(resourceClothingPresetsActionUI.firstSelectedRVGE.primaryClothingItemPresetRVGE == rvge);
        }

        public void OnPointerEnter(PointerEventData eventData)
        {
            helpTextJSS.val = rvge.resourceFullPathReference + "\nSingle click to select. Double click to load. Click and hold for Hide & Favorite.";

            Image image = helpTextJSS.text.transform.parent.gameObject.GetComponent<Image>();
            image.color = Color.gray;
        }

        public void OnPointerExit(PointerEventData eventData)
        {
            helpTextJSS.val = "";
            firstLeftClickInitiated = false;
            if (togglesVisible) HideToggles();
            Image image = helpTextJSS.text.transform.parent.gameObject.GetComponent<Image>();
            image.color = Color.gray;

            longPressDetectActive = false;
        }

        public virtual void OnPointerClick(PointerEventData eventData)
        {
            if (button!=null)
            {
                
                if (eventData.button == PointerEventData.InputButton.Left)
                {
                    if (!firstLeftClickInitiated)
                    {
                        firstLeftClickInitiated = true;
                        firstLeftClickTime = Time.unscaledTime * 1000;
                        resourceClothingPresetDetailUI.ToggleUIOpen(rvge, this);
                        return;
                    }
                    float interval = (Time.unscaledTime * 1000) - firstLeftClickTime;
                    if (interval < 350f)
                    {
                        resourceClothingPresetsActionUI.ClothingItemPresetActivate(rvge);
                        longPressDetectActive = false;
                        firstLeftClickInitiated = false;
                    }
                    else
                    {
                        firstLeftClickTime = Time.unscaledTime * 1000;
                        resourceClothingPresetDetailUI.ToggleUIOpen(rvge, this);
                    }
                }
                else if (eventData.button == PointerEventData.InputButton.Right)
                {
                    ToggleFavHideActivation();
                }
            }

        }

        private void ToggleFavHideActivation()
        {
            if (!togglesVisible)
            {
                rvgeHiddenJSB.valNoCallback = rvge.isHidden;
                rvgeFavJSB.valNoCallback = rvge.isFavorite;

                rvgeHiddenToggle = UI.CreateToggle(button.button.gameObject, rvgeHiddenJSB, button.height, 20f, "Hide", 20, 0f, 0f, AnchorPresets.topMiddle);
                rvgeFavToggle = UI.CreateToggle(button.button.gameObject, rvgeFavJSB, button.height, 20f, "Favorite", 20, 0f, -20f, AnchorPresets.topMiddle);

                GeneralPointerEnterExitHelp contextHelpFav = rvgeFavToggle.toggle.gameObject.AddComponent<GeneralPointerEnterExitHelp>();
                GeneralPointerEnterExitHelp contextHelpHide = rvgeHiddenToggle.toggle.gameObject.AddComponent<GeneralPointerEnterExitHelp>();

                if (!Globals.addonPackageFilePrefsFolderAccess || !Globals.customFolderAccess || !Globals.scenesFolderAccess)
                {
                    rvgeFavToggle.toggle.interactable = false;
                    rvgeHiddenToggle.toggle.interactable = false;

                    contextHelpFav.Init("Enable BA Sym Links using .bat file provided in PluginData folder to allow toggling the VAM Favorite state from within Browser Assist. See BA VAMHub page for detailed instructions.", helpTextJSS, false);

                    contextHelpHide.Init("Enable BA Sym Links using .bat file provided in PluginData folder to allow toggling the VAM Hidden state from within Browser Assist. See BA VAMHub page for detailed instructions.", helpTextJSS, false);
                }
                else
                {
                    contextHelpFav.Init("Toggle setting the resource as a Favourite", helpTextJSS, false);

                    contextHelpHide.Init("Toggle setting the resource as Hidden", helpTextJSS, false);
                }


                togglesVisible = true;
            }
            else HideToggles();
        }
        private void HideToggles()
        {
            if (rvgeHiddenToggle != null) Destroy(rvgeHiddenToggle.gameObject);
            if (rvgeFavToggle != null) Destroy(rvgeFavToggle.gameObject);
            togglesVisible = false;
        }

        public void OnPointerDown(PointerEventData eventData)
        {
            longPressDetectActive = true;
            pointerDownTimeFloat = Time.unscaledTime * 1000;
        }

        public void OnPointerUp(PointerEventData eventData)
        {
            float timeDown = (Time.unscaledTime * 1000) - pointerDownTimeFloat;

            if (timeDown > 500f && longPressDetectActive) ToggleFavHideActivation();
            longPressDetectActive = false;
        }
    }


    public class SelectedResourceButtonPointerEnterExit : MonoBehaviour, IPointerEnterHandler, IPointerExitHandler
    {
        private ResourceVersionGroupEntry rvge;
        private JSONStorableString helpText;
        public void Init(ResourceVersionGroupEntry _rvge, JSONStorableString _helpText)
        {
            rvge = _rvge;
            helpText = _helpText;
        }
        public void OnPointerEnter(PointerEventData eventData)
        {
            helpText.val = rvge.resourceFullPathReference;

        }

        public void OnPointerExit(PointerEventData eventData)
        {
            helpText.val = "";
        }

    }


    public class ButtonPointerDrag : MonoBehaviour, IBeginDragHandler, IDragHandler
    {
        private Canvas mainCanvas;


        private float distance;
        private Camera controllerCamera;

        public void Init(Canvas _mainCanvas)
        {
            mainCanvas = _mainCanvas;
        }
        public void OnBeginDrag(PointerEventData eventData)
        {
            RectTransform rt = transform.GetComponent<RectTransform>();
            Vector3[] corners = new Vector3[4];
            rt.GetWorldCorners(corners);
            Vector3 buttonPos = (corners[2] + corners[0]) / 2;

            if (eventData.enterEventCamera == SuperController.singleton.leftControllerCamera)
            {
                controllerCamera = SuperController.singleton.leftControllerCamera;               
                distance = Vector3.Distance(buttonPos, controllerCamera.transform.position);
            }
            else if (eventData.enterEventCamera == SuperController.singleton.rightControllerCamera)
            {
                controllerCamera = SuperController.singleton.rightControllerCamera;
                distance = Vector3.Distance(buttonPos, controllerCamera.transform.position);
            }
                
        }

        public void OnDrag(PointerEventData eventData)
        {
            RectTransform rt = transform.GetComponent<RectTransform>();
            Vector3[] corners = new Vector3[4];
            rt.GetWorldCorners(corners);
            Vector3 buttonPos = (corners[2] + corners[0]) / 2;

            Ray controllerRay = new Ray(controllerCamera.transform.position, controllerCamera.transform.forward);

            Vector3 buttonToCanvasOffset = buttonPos - mainCanvas.transform.position;

             Vector3 rayPoint = controllerRay.GetPoint(distance);
            //Move the GameObject when you drag it
            mainCanvas.transform.position = rayPoint - buttonToCanvasOffset;
        }
    }

}
