﻿using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using MVR.FileManagementSecure;
using UnityEngine.UI;
using Leap.Unity.Query;

namespace JayJayWon
{
    public class CallbackBrowserUI :BrowserUI
    {
        public override int resourceActionBoxHeight { get; } = 0;

        protected JSONStorableString fileNameCallbackJSS;
        protected JSONStorableBool keepOpenJSB = new JSONStorableBool("keepOpen", false);

        protected int baKeepOpenOption;
        protected Vector3 vrBrowserPosition;
        private UIDynamicToggle keepOpenToggle;
        public CallbackBrowserUI():base()
        {
            isCallbackBrowser = true;
        }

        protected override void InitBrowserUI()
        {
            base.InitBrowserUI();
            selectedResourceGO.SetActive(false);

            if (baKeepOpenOption!=BAKeepOpenOptionEnum.unavailable)
            {
                keepOpenToggle = UI.CreateToggle(backgroundBoxGO, keepOpenJSB, 90f, 20f, "Keep Open", 20, 5f, 30f, AnchorPresets.bottomLeft);

            }
            

        }
        protected override void SetVRPosition()
        {
            if (attachedVAMUI)
            {
                SuperController.singleton.activeUI = SuperController.ActiveUI.None;
                CloseUIAssistUI();
            }
            else mainCanvas.transform.position = vrBrowserPosition;
        }

        public override void SetInactive()
        {
            base.SetInactive();
            if (fileNameCallbackJSS != null) fileNameCallbackJSS.val = "";
            fileNameCallbackJSS=null;
            CallbackResourceBrowserUI callbackResourceBrowserUI = resourceBrowserUI as CallbackResourceBrowserUI;
            callbackResourceBrowserUI.InitCallback(null, BAKeepOpenOptionEnum.unavailable, null);
        }

        protected override void CreateResourceActionsUI()
        {
            base.CreateResourceActionsUI();
            resourceActionsUI.SetActive(false);
        }

        protected override void CreateBrowserFiltersUI()
        {
            browserFiltersUI = new CallbackBrowserFiltersUI(this);
        }

        protected override void CreateResourceBrowserUI()
        {
            resourceBrowserUI = new CallbackResourceBrowserUI(this, resourceBrowserGO);
        }

        public void InitCallback(ResourceBrowserFilterSettings filter, string browserTitle, JSONStorableString fileNameCallbackJSS, int baKeepOpenOption, Vector3 vrBrowserPosition)
        {
            CallbackBrowserFiltersUI callbackBrowserFiltersUI = browserFiltersUI as CallbackBrowserFiltersUI;
            CallbackResourceBrowserUI callbackResourceBrowserUI = resourceBrowserUI as CallbackResourceBrowserUI;
            if (fileNameCallbackJSS != null) fileNameCallbackJSS.val = "";
            this.fileNameCallbackJSS = fileNameCallbackJSS;
            this.baKeepOpenOption= baKeepOpenOption;
            this.vrBrowserPosition = vrBrowserPosition;
            callbackBrowserFiltersUI.InitCallback(filter, browserTitle, filtersBoxGO);

            if (baKeepOpenOption == BAKeepOpenOptionEnum.unavailable) keepOpenToggle.gameObject.SetActive(false);
            else
            {
                keepOpenToggle.gameObject.SetActive(true);
                if (baKeepOpenOption == BAKeepOpenOptionEnum.availableInactive) keepOpenJSB.val = false;
                else keepOpenJSB.val = true;
            }
            callbackResourceBrowserUI.InitCallback(fileNameCallbackJSS, baKeepOpenOption, keepOpenJSB);

            resourceSelectionsUI.DeSelectAll (filter.resourceType,filter.atomPresetResourceType,false);
        }

        protected override void InitBrowserFiltersUI()
        {
            // Do nothing
        }
    }

    public class BrowserUI
    {
        public static Color darkBlueGrey = new Color(0.4f, 0.4f, 0.6f, 1f);
        public static Color lightBlueGrey = new Color(0.5f, 0.5f, 0.7f, 1f);
        public static Color lightGrey = new Color(0.9f, 0.9f, 0.9f, 1f);
        public static Color mediumGrey = new Color(0.5f, 0.5f, 0.5f, 1f);
        public static Color darkGrey = new Color(0.2f, 0.2f, 0.2f, 1f);

        const int filterBoxHeight = 50;
        const int boxSpacerHeight = 3;
        public static int edgeSpacerSize = 2;
        const int infoTextBoxHeight = 22;
        public virtual int resourceActionBoxHeight { get; } = 90;
        private Image fileBrowserImage;
        private Image hubBrowserImage;
        private CanvasScaler mainCanvasScalar;

        public bool isCallbackBrowser { get; protected set; } = false;
        public BrowserUI()
        {
            GameObject canvasObject = new GameObject();
            mainCanvas = canvasObject.AddComponent<Canvas>();
            mainCanvas.pixelPerfect = false;

            if (isVREnabled) mainCanvas.renderMode = RenderMode.WorldSpace;
            else mainCanvas.renderMode = RenderMode.ScreenSpaceOverlay;

            mainCanvasScalar = canvasObject.AddComponent<CanvasScaler>();
            GraphicRaycaster rayCaster = canvasObject.AddComponent<GraphicRaycaster>();

            mainCanvas.sortingOrder = -3;

            SuperController.singleton.AddCanvas(mainCanvas);

            if (isVREnabled)
            {
                mainCanvasScalar.dynamicPixelsPerUnit = 4f;

                RefreshVRScale();
            }
            else
            {
                mainCanvasScalar.dynamicPixelsPerUnit = 2f;
                mainCanvasScalar.uiScaleMode = CanvasScaler.ScaleMode.ScaleWithScreenSize;
                RefreshDesktopScale();

            }
            CreateBrowserFiltersUI();

            resourceSelectionsUI = new ResourceSelections(this);
            helpTextJSS = new JSONStorableString("helpText", "");

            

            InitBrowserUI();

            var fileBrowserTransform = SuperController.singleton.worldUI.Find("Canvas/FileBrowserWorld/MainPanel");
            var hubBrowserTransform = SuperController.singleton.worldUI.Find("Canvas/HubBrowsePanel");
            fileBrowserImage = fileBrowserTransform.GetComponent<Image>();
            hubBrowserImage = hubBrowserTransform.GetComponent<Image>();

            resourceBrowserUI.RefactorButtonRows((int)Settings.maxButtonsPerPage, 6);
            SetInactive();
        }

        protected virtual void CreateBrowserFiltersUI()
        {
            browserFiltersUI = new BrowserFiltersUI(this);
            
        }

        public float worldScaleAdjust
        {
            get
            {
                if (isVREnabled) return SuperController.singleton.worldScale;
                else return 1f;
            }
        }

        public void SetLocalScale(bool attachedToVAMUI)
        {            
            if (attachedToVAMUI && isVREnabled)
            {
                mainCanvas.transform.localScale = new Vector3(Const._localCanvasScale * .6f * worldScaleAdjust, Const._localCanvasScale * .6f * worldScaleAdjust, Const._localCanvasScale * .6f * worldScaleAdjust);
            }
            else mainCanvas.transform.localScale = new Vector3(Const._localCanvasScale * worldScaleAdjust, Const._localCanvasScale * worldScaleAdjust, Const._localCanvasScale * worldScaleAdjust);
        }
        public void RefreshDesktopScale()
        {
            if (!isVREnabled)  mainCanvasScalar.referenceResolution = new Vector2(1440 / Settings.uiScaleJSF.val, 900 / Settings.uiScaleJSF.val);
        }

        private void RefreshVRScale()
        {
            SetLocalScale(attachedVAMUI);

        }

        public int screenHeight { get { if (isVREnabled) return 600;
                else return Screen.height; } }

        private int resourceBrowserBottomOffset { get
            {
                return edgeSpacerSize + infoTextBoxHeight + resourceActionBoxHeight + resourceSelectionsUI.resourceSelectionBoxHeight + (boxSpacerHeight * 3);
            }
        }
        public int resourceBrowserTopOffset
        {
            get
            {
                return edgeSpacerSize + filterBoxHeight + boxSpacerHeight - filterBoxVerticalOffset;
            }
        }

        public int resourceBrowserHeight { get
            {
                return screenHeight - resourceBrowserBottomOffset - resourceBrowserTopOffset;
            } }

        public int selectedResourcesVOffset
        {
            get
            {
                return edgeSpacerSize + infoTextBoxHeight + resourceActionBoxHeight + (boxSpacerHeight * 2);
            }
        }

        public float browserScreenWidth { get
            {
                if (isVREnabled) return 650f;
                if (currentDesktopDisplayMode == DesktopDisplayMode.rightDocked || currentDesktopDisplayMode == DesktopDisplayMode.leftDocked) return 650f;
                if (currentDesktopDisplayMode == DesktopDisplayMode.fullWidth) return 1440 / Settings.uiScaleJSF.val*0.9f;
                    return -1f;
            } }

        protected float tagBoxHorizontalOffset { get
            {
                if (isVREnabled) return -(browserScreenWidth/2f)-60f;
                if (currentDesktopDisplayMode == DesktopDisplayMode.rightDocked) return -browserScreenWidth-5f;
                if (currentDesktopDisplayMode == DesktopDisplayMode.fullWidth) return 0;
                if (currentDesktopDisplayMode == DesktopDisplayMode.leftDocked) return browserScreenWidth + 5f; ;
                return 0f;
            } }

        protected float backgroundBoxHorizontalOffset
        {
            get
            {
                if (currentDesktopDisplayMode == DesktopDisplayMode.rightDocked) return  -5f;
                if (currentDesktopDisplayMode == DesktopDisplayMode.fullWidth) return 0;
                if (currentDesktopDisplayMode == DesktopDisplayMode.leftDocked) return  5f; ;
                return 0f;
            }
        }
        protected bool isQuickFilterVisible { get
            {
                return (PC.patreonContentEnabled && Settings.enableQuickFiltersJSB.val && !isCallbackBrowser);
            } }

        protected int quickFilterBoxHeight
        {
            get
            {
                if (QuickFilterSettings.qfsMinimized.val && quickFiltersUI!=null && quickFiltersUI.currentQFSeditMode==null) return QuickFiltersUI.qfsButtonsHeight;
                return QuickFiltersUI.buttonRowCount * QuickFiltersUI.qfsButtonsHeight;
            }
        }

        protected int filterBoxVerticalOffset
        {
            get
            {
                if (!isQuickFilterVisible) return -edgeSpacerSize;
                else return -edgeSpacerSize- quickFilterBoxHeight;
            }
        }

        public int tagBoxAnchorPreset
        {
            get
            {
                if (isVREnabled) return AnchorPresets.middleRight;
                if (currentDesktopDisplayMode == DesktopDisplayMode.rightDocked) return AnchorPresets.vStretchRight;
                if (currentDesktopDisplayMode == DesktopDisplayMode.fullWidth) return AnchorPresets.vStretchLeft;
                if (currentDesktopDisplayMode == DesktopDisplayMode.leftDocked) return AnchorPresets.vStretchLeft;
                return -1;
            }
        }

        public int backgroundBoxAnchorPreset { get
            {
                if (isVREnabled) return AnchorPresets.centre;
                if (currentDesktopDisplayMode == DesktopDisplayMode.rightDocked) return AnchorPresets.vStretchRight;
                if (currentDesktopDisplayMode == DesktopDisplayMode.fullWidth) return AnchorPresets.vStretchMiddle;
                if (currentDesktopDisplayMode == DesktopDisplayMode.leftDocked) return AnchorPresets.vStretchLeft;
                return -1;
            } }

        protected virtual void InitBrowserUI()
        {
            float vSize = 0f;
            if (isVREnabled) vSize = screenHeight;

            backgroundBoxGO = UI.AddChildImage(parentTransform: mainCanvas.transform,
                color: new Color(0.5f, 0.5f, 0.5f, 0.95f),
                anchorPreset: backgroundBoxAnchorPreset,
                horizontalSize: browserScreenWidth,
                verticalSize: vSize,
                anchoredPositionOffset: new Vector2(backgroundBoxHorizontalOffset, 0f));

            sideTagBoxGO= UI.AddChildImage(parentTransform: mainCanvas.transform,
                color: new Color(0.5f, 0.5f, 0.5f, 0.95f),
                anchorPreset: tagBoxAnchorPreset,
                horizontalSize: 150f,
                verticalSize: vSize,
                anchoredPositionOffset: new Vector2( tagBoxHorizontalOffset, 0f));
            sideTagBoxGO.SetActive(false);
            if (isVREnabled) sideTagBoxGO.transform.Rotate(new Vector3(0, -45, 0));

            resourceBrowserGO = UI.CreateVScrollableContent(backgroundBoxGO, new Color(0.8f, 0.8f, 0.8f, 0.8f), AnchorPresets.vStretchMiddle, browserScreenWidth-10f, resourceBrowserHeight, new Vector2(0f, filterBoxVerticalOffset - filterBoxHeight- boxSpacerHeight));

            if (!isCallbackBrowser) resourceBrowserGO.AddComponent<MainBrowserEnterExit>();

            selectedResourceGO = UI.AddChildGOImage(parentGO:backgroundBoxGO,
                color: darkBlueGrey,
                anchorPreset: AnchorPresets.bottomMiddle,
                horizontalSize: resourceSelectionsUI.resourceSelectionBoxWidth,
                verticalSize: resourceSelectionsUI.resourceSelectionBoxHeight,
                anchoredPositionOffset: new Vector2((resourceSelectionsUI.resourceSelectionBoxWidth -640f)/2, selectedResourcesVOffset));

            resourceActionsGO = UI.AddChildGOImage(parentGO: backgroundBoxGO,
                color: darkBlueGrey,
                anchorPreset: AnchorPresets.bottomMiddle,
                horizontalSize: 640f,
                verticalSize: resourceActionBoxHeight,
                anchoredPositionOffset: new Vector2(0f, infoTextBoxHeight+edgeSpacerSize+boxSpacerHeight));

            filtersBoxGO = UI.AddChildGOImage(parentGO: backgroundBoxGO,
                color: darkBlueGrey,
                anchorPreset: AnchorPresets.topMiddle,
                horizontalSize: 640f,
                verticalSize: filterBoxHeight,
                anchoredPositionOffset: new Vector2(0f, filterBoxVerticalOffset));

            quickFiltersGO = UI.AddChildGOImage(parentGO: backgroundBoxGO,
                color: darkBlueGrey,
                anchorPreset: AnchorPresets.topMiddle,
                horizontalSize: 640f,
                verticalSize: quickFilterBoxHeight,
                anchoredPositionOffset: new Vector2(0f, -edgeSpacerSize));

            cancelButtonUID = UI.CreateUIButton(backgroundBoxGO, 20f, 20f, "<b>X</b>", 28, (650f / 2) - 15f, edgeSpacerSize+1f, anchorPreset: AnchorPresets.bottomMiddle) ;
            cancelButtonUID.buttonImage.sprite = null;
            cancelButtonUID.button.onClick.AddListener(delegate {
                SetInactive();
            });
            ApplyHelp(cancelButtonUID, "Close the Browser Assist window.");

            if (isVREnabled)
            {
                repositionUI = UI.CreateUIButton(backgroundBoxGO, 25f, 25f, "<b>O</b>", 34, 1, 2, anchorPreset: AnchorPresets.bottomLeft);
                repositionUI.button.onClick.AddListener(delegate {                   
                    if (Settings.dockToVAMUIJSB.val && Globals.isVREnabled)
                    {
                        attachedVAMUI = false;
                        SuperController.singleton.HideMainHUD();
                        SetActive();
                    }
                    });
                var component = repositionUI.button.gameObject.AddComponent<ButtonPointerDrag>();
                component.Init(mainCanvas);
            }
            else
            {
                leftExpandContractButtonUID = UI.CreateUIButton(backgroundBoxGO, 20f, 20f, "<b>|<</b>", 28, -(640f / 2) + 10f, edgeSpacerSize+1f, anchorPreset: AnchorPresets.bottomMiddle);
                leftExpandContractButtonUID.buttonImage.sprite = null;
                leftExpandContractButtonUID.button.onClick.AddListener(delegate {
                    if (currentDesktopDisplayMode == DesktopDisplayMode.leftDocked)
                    {
                        FullWidthUI();
                    }
                    else
                    {
                        LeftDockUI();                        
                    }
                    
                });
                rightExpandContractButtonUID = UI.CreateUIButton(backgroundBoxGO, 20f, 20f, "<b><<</b>", 28, -(640f / 2) + 32f, edgeSpacerSize+1f, anchorPreset: AnchorPresets.bottomMiddle);
                rightExpandContractButtonUID.buttonImage.sprite = null;
                rightExpandContractButtonUID.button.onClick.AddListener(delegate {
                    if (currentDesktopDisplayMode == DesktopDisplayMode.rightDocked)
                    {
                        FullWidthUI();

                    }
                    else
                    {
                        RightDockUI();
                    }
                    
                });
            }
            infoTextBoxGO = UI.CreateBasicTextBox(backgroundBoxGO, helpTextJSS, 574f, infoTextBoxHeight, 22, 12f, edgeSpacerSize, AnchorPresets.bottomMiddle, true, 18);

            if (PC.patreonContentEnabled) quickFiltersUI = new QuickFiltersUI(this, quickFiltersGO);
            RefreshQuickFilterPanelVisibility();

            InitBrowserFiltersUI();
            resourceSelectionsUI.InitUI(selectedResourceGO);
            CreateResourceActionsUI();

            CreateResourceBrowserUI();
        }
        protected virtual void InitBrowserFiltersUI()
        {
            browserFiltersUI.InitUI(filtersBoxGO);
        }
        public void LeftDockUI()
        {
            currentDesktopDisplayMode = DesktopDisplayMode.leftDocked;
            leftExpandContractButtonUID.label = "<b>>></b>";
            ApplyHelp(leftExpandContractButtonUID, "Expand the Browser Assist window across the screen width.");

            rightExpandContractButtonUID.label = "<b>>|</b>";
            ApplyHelp(rightExpandContractButtonUID, "Dock the Browser Assist window on the right of the screen.");
            RefreshBrowserUIDimensions();
            RepositionBrowserBottomUIElements();
            browserFiltersUI.SetFolderBrowserInActive();
        }
        public void RightDockUI()
        {
            currentDesktopDisplayMode = DesktopDisplayMode.rightDocked;

            rightExpandContractButtonUID.label = "<b><<</b>";
            ApplyHelp(rightExpandContractButtonUID, "Expand the Browser Assist window across the screen width.");

            leftExpandContractButtonUID.label = "<b>|<</b>";
            ApplyHelp(leftExpandContractButtonUID, "Dock the Browser Assist window on the left of the screen.");
            RefreshBrowserUIDimensions();
            RepositionBrowserBottomUIElements();
            browserFiltersUI.SetFolderBrowserInActive();
        }
        public void FullWidthUI()
        {
            currentDesktopDisplayMode = DesktopDisplayMode.fullWidth;
            leftExpandContractButtonUID.label = "<b>|<</b>";
            ApplyHelp(leftExpandContractButtonUID, "Dock the Browser Assist window on the left of the screen.");

            rightExpandContractButtonUID.label = "<b>>|</b>";
            ApplyHelp(rightExpandContractButtonUID, "Dock the Browser Assist window on the right of the screen.");
            RefreshBrowserUIDimensions();
            RepositionBrowserBottomUIElements();
            browserFiltersUI.SetFolderBrowserInActive();
        }
        protected virtual void CreateResourceActionsUI()
        {
            resourceActionsUI = new ResourceActionsUI(this, resourceActionsGO);
            resourceActionsUI.RefreshUI();
        }

        protected virtual void CreateResourceBrowserUI()
        {
            resourceBrowserUI = new ResourceBrowserUI(this, resourceBrowserGO);
        }

        private void RepositionBrowserBottomUIElements()
        {
            if (currentDesktopDisplayMode == DesktopDisplayMode.leftDocked)
            {
                UI.ResetAnchorPosition(rightExpandContractButtonUID.gameObject, (640f / 2) - 10f, edgeSpacerSize + 1f);
                UI.ResetAnchorPosition(leftExpandContractButtonUID.gameObject, (640f / 2) - 32f, edgeSpacerSize + 1f);
                UI.ResetAnchorPosition(cancelButtonUID.gameObject, -(650f / 2) + 15f, edgeSpacerSize + 1f);

                UI.ResetAnchorPosition(infoTextBoxGO, -12f, edgeSpacerSize);
            }
            else
            {
                UI.ResetAnchorPosition(rightExpandContractButtonUID.gameObject, -(640f / 2) +32f, edgeSpacerSize + 1f);
                UI.ResetAnchorPosition(leftExpandContractButtonUID.gameObject, -(640f / 2) + 10f, edgeSpacerSize + 1f);
                UI.ResetAnchorPosition(cancelButtonUID.gameObject, (650f / 2) - 15f, edgeSpacerSize + 1f);
                UI.ResetAnchorPosition(infoTextBoxGO, 12f, edgeSpacerSize);
            }
        }
        private void RefreshBrowserUIDimensions()
        {
            UI.ResizeWidth(backgroundBoxGO, browserScreenWidth);
            UI.ResetAnchorType(backgroundBoxGO, backgroundBoxAnchorPreset);
            UI.ResetAnchorPosition(backgroundBoxGO, backgroundBoxHorizontalOffset, 0f);
            UI.ResizeWidth(resourceBrowserGO, browserScreenWidth - 10f);
            GameObject buttonListViewportGO = resourceBrowserGO.transform.Find("scrollableContentViewportGO").gameObject;
            UI.ResizeWidth(buttonListViewportGO, browserScreenWidth - 20f);
            UI.ResetAnchorType(sideTagBoxGO, tagBoxAnchorPreset);
            UI.ResetAnchorPosition(sideTagBoxGO, tagBoxHorizontalOffset, 0f);

            if (!Settings.sideTagBoxEnabledJSB.val || currentDesktopDisplayMode == DesktopDisplayMode.fullWidth || resourceActionsUI.currentActionUITabType!=ActionUITabType.tagsTab) sideTagBoxGO.SetActive(false);
            else sideTagBoxGO.SetActive(true);

            resourceBrowserUI.RepopulateButtons();        
        }

        public void RefreshQuickFilterPanelVisibility()
        {
            UI.ResetAnchorPosition(filtersBoxGO, 0f, filterBoxVerticalOffset);
            UI.ResetTopOffset(resourceBrowserGO, filterBoxVerticalOffset - filterBoxHeight - boxSpacerHeight);
            UI.ResizeHeight(quickFiltersGO, quickFilterBoxHeight);
            if (isQuickFilterVisible) quickFiltersGO.SetActive(true);
            else quickFiltersGO.SetActive(false);
        }

        public void RefreshBrowserUILayout()
        {       
            UI.ResizeHeight(selectedResourceGO, resourceSelectionsUI.resourceSelectionBoxHeight);
            UI.ResizeWidth(selectedResourceGO, resourceSelectionsUI.resourceSelectionBoxWidth);
            UI.ResetAnchorPosition(selectedResourceGO, (resourceSelectionsUI.resourceSelectionBoxWidth - 640f) / 2, selectedResourcesVOffset);
            resourceBrowserUI.RepositionNavButtons();

            UI.ResetBottomOffset(resourceBrowserGO, (edgeSpacerSize * 1) + infoTextBoxHeight + resourceActionBoxHeight + resourceSelectionsUI.resourceSelectionBoxHeight + (boxSpacerHeight * 3));
        }
        public bool isVisible { get { return mainCanvas.enabled; } set { mainCanvas.enabled = value; } }

        public bool isActive { get; private set; }


        private bool _attachedVAMUI = false;
        public bool attachedVAMUI { get { return _attachedVAMUI; } set {
                _attachedVAMUI = value;
                SetLocalScale(value);
                
            } }    
        public void ToggleActive()
        {
            if (mainCanvas.enabled) SetInactive();
            else SetActive();
        }

        protected virtual void SetVRPosition()
        {
            if (!attachedVAMUI)
            {
                Vector3 headPosition = SuperController.singleton.lookCamera.transform.position;
                if (SuperController.singleton.mainHUD == null || !SuperController.singleton.mainHUD.gameObject.activeSelf)
                {

                    Quaternion headFacingDirection = SuperController.singleton.lookCamera.transform.rotation;
                    Vector3 offsetPosition = new Vector3(0f, 0f, 2f * worldScaleAdjust);

                    Vector3 newPosition = headPosition + headFacingDirection * offsetPosition;
                    mainCanvas.transform.position = new Vector3(newPosition.x, headPosition.y, newPosition.z);
                }
                else
                {

                    Vector3 newTargetPos = SuperController.singleton.mainHUD.position + (SuperController.singleton.navigationRig.up * 1f) - (SuperController.singleton.mainHUD.forward * 0.6f);

                    Vector3 newPos = new Vector3(newTargetPos.x, headPosition.y, newTargetPos.z);
                    mainCanvas.transform.position = newPos;

                    SuperController.singleton.HideMainHUD();
                }
            }
            else
            {
                SuperController.singleton.activeUI = SuperController.ActiveUI.None;
                CloseUIAssistUI();
            }
        }
        public static void CloseUIAssistUI()
        {
            var sessionPlugins = PluginUtils.GetSessionPlugins("JayJayWon.UIAssist");
            foreach (var plugin in sessionPlugins)
            {
                plugin.SendMessage("OnCloseVAMUIAttachedUI", SendMessageOptions.DontRequireReceiver);
            }
        }

        private void ClothingWarningAlertAccept()
        {
            Settings.acceptClothingTagsWarningJSB.val = true;
        }
        private void SymLinksWarningAlertAccept()
        {
            Settings.acceptSymLinksWarningJSB.val = true;
        }

        public void SetActive()
        {
            RefreshFilterResults();

            if (!isCallbackBrowser)
            {
                if (!Settings.acceptClothingTagsWarningJSB.val)
                {
#if VAM_GT_1_20_77_13
                    SuperController.singleton.Alert("Please be aware that Clothing Item tags can be synchronised between VAM and BrowserAssist - with some limitations. Clothing item tag changes in BrowserAsssist will only be synchronised with VAM upon a VAM hard reset or restart. Clothing item tag changes in VAM will sync with BrowserAssist upon a rescan (or plugin restart). Do not make changes for the same clothing item in both BrowserAssist and VAM without syncing in between.\n\n This message will only appear once.", ClothingWarningAlertAccept);
#else
                    SuperController.LogMessage("Please be aware that Clothing Item tags can be synchronised between VAM and BrowserAssist - with some limitations. Clothing item tag changes in BrowserAsssist will only be synchronised with VAM upon a VAM hard reset or restart. Clothing item tag changes in VAM will sync with BrowserAssist upon a rescan (or plugin restart). Do not make changes for the same clothing item in both BrowserAssist and VAM without syncing in between.\n\n This message will only appear once.");
#endif
                }

                if (!Settings.acceptSymLinksWarningJSB.val && (!Globals.scenesFolderAccess))
                {
#if VAM_GT_1_20_77_13
                    SuperController.singleton.Alert("Please be aware that many BrowserAssist features - such as deleting files or setting the Favorite or Hidden tags, are only available once the CreateBASymLinks.bat file has been run in the folder Saves\\PluginData\\JayJayWon\\BrowserAssist and BrowserAssist has been restarted. See the VAMHub for more details.\n\nWARNING: In theory, enabling this feature potentially allows a malicious VAM plugin to delete many files under the main VAM folder. Apply an appropriate level of caution and only source plugins from reliable and trusted creators.\n\n This message will only appear once.", SymLinksWarningAlertAccept);
#else
                    SuperController.LogMessage("Please be aware that many BrowserAssist features - such as deleting files or setting the Favorite or Hidden tags, are only available once the CreateBASymLinks.bat file has been run in the folder Saves\\PluginData\\JayJayWon\\BrowserAssist and BrowserAssist has been restarted. See the VAMHub for more details.\n\nWARNING: In theory, enabling this feature potentially allows a malicious VAM plugin to delete many files under the main VAM folder. Apply an appropriate level of caution and only source plugins from reliable and trusted creators.\n\n This message will only appear once.");
#endif
                }
            }

            bool mainHUDActive = SuperController.singleton.mainHUD.gameObject.activeSelf;

            if (!mainHUDActive || !Settings.dockToVAMUIJSB.val || !isVREnabled) attachedVAMUI = false;
            else attachedVAMUI = true;

            if (isVREnabled )
            {
                SetVRPosition();

            }
            else resourceBrowserUI.NudgeScrollBar();
            isVisible = true;
            isActive = true;
        }
        public virtual void SetInactive()
        {
            isVisible = false;
            isActive = false;
            Globals.browserPointerActive= false;
        }
        public void RefreshButtonLayout(bool forceSelectedResourceRefresh=false)
        {
            if (resourceBrowserUI!=null) resourceBrowserUI.RepopulateButtons(forceSelectedResourceRefresh);
        }

        private bool refreshOnNextUpdate = false;
        private bool forceSelectedResourceRefreshOnNextUpdate = false;

        public void RefreshFilterResults(bool forceSelectedResourceRefresh = false)
        {
            refreshOnNextUpdate = true;
            if (forceSelectedResourceRefresh) forceSelectedResourceRefreshOnNextUpdate =true;
        }

        private void InternalRefreshFilterResults()
        {
            if (refreshOnNextUpdate)
            {

                bool forceSelectedResourceRefresh = forceSelectedResourceRefreshOnNextUpdate;
                forceSelectedResourceRefreshOnNextUpdate = false;
                refreshOnNextUpdate = false;

                currentFilterResults = ResourceManifest.GetResourceGroups(browserFiltersUI.currentFilterSettings, Settings.resourceNameSearchModeJSESC.val);
                randomCurrentFilterResults.Clear();
                RefreshButtonLayout(forceSelectedResourceRefresh);
                if (resourceActionsUI != null) resourceActionsUI.RefreshUI();

                if (PC.patreonContentEnabled && quickFiltersUI.currentQFSeditMode!=null) QuickFilterSettings.ModifyQFS(browserFiltersUI.currentFilterSettings, quickFiltersUI.currentQFSeditMode);
            }
            
            
        }

        public void Update()
        {
            bool isVisibleAtFrameStart = isVisible;

            InternalRefreshFilterResults();

            resourceBrowserUI.Update();

            if (isVREnabled)
            {
                RefreshVRScale();

                if (Settings.dockToVAMUIJSB.val)
                {
                    bool mainHUDActive = SuperController.singleton.mainHUD.gameObject.activeSelf;

                    if (mainHUDActive && isActive)
                    {
                        if (SuperController.singleton.activeUI == SuperController.ActiveUI.None)
                        {
                            attachedVAMUI = true;
                            isVisible = true;
                            Transform mainHUDTransform = SuperController.singleton.mainHUD.transform;

                            Vector3 mainHUDRotationEuler = SuperController.singleton.mainHUD.transform.rotation.eulerAngles;
                            mainCanvas.transform.rotation = Quaternion.Euler(mainHUDRotationEuler.x, mainHUDRotationEuler.y, mainHUDRotationEuler.z);
                            mainCanvas.transform.Rotate(0f, -180f, 0f);

                            mainCanvas.transform.position = mainHUDTransform.position + (mainHUDTransform.up * 0.55f*worldScaleAdjust) + (mainHUDTransform.forward * -0f);
                        }
                        else
                        {
                            SetInactive();
                        }

                    }
                    else if (attachedVAMUI)
                    {
                        isVisible = false;
                    }
                }

                if (!Settings.dockToVAMUIJSB.val || !attachedVAMUI)
                {
                    Vector3 headPosition = SuperController.singleton.lookCamera.transform.position;
                    mainCanvas.transform.LookAt(headPosition, SuperController.singleton.navigationRig.up);
                    mainCanvas.transform.rotation *= Quaternion.Euler(0f, 180f, 0f);
                    
                }
                    
            }

            if (SuperController.singleton.screenshotCamera.enabled || SuperController.singleton.hiResScreenshotCamera.enabled)
            {
                if (isVisibleAtFrameStart)
                {
                    SetInactive();
                    inactiveDueToScreenShot = true;
                }
                
            }
            else if (inactiveDueToScreenShot)
            {
                SetActive();
                if (screenShotCallback != null)
                {
                    screenShotCallback(screenShotCallbackRVGE);
                    screenShotCallback= null;
                    screenShotCallbackRVGE= null;
                }

                inactiveDueToScreenShot = false;                
            }

            if (fileBrowserImage.gameObject.activeInHierarchy || hubBrowserImage.gameObject.activeInHierarchy || (!isVREnabled && !SuperController.singleton.MonitorCenterCamera.isActiveAndEnabled))
            {
                if (!isHidden && mainCanvas!=null && mainCanvas.enabled)
                {
                    SetInactive();
                    isHidden = true;
                }
                
            }
            else
            {
                if (isHidden)
                {
                    SetActive();
                    isHidden = false;
                }
            }

        }

        public delegate void ScreenShotCallback(ResourceVersionGroupEntry rvge);

        private ScreenShotCallback screenShotCallback =null;
        private ResourceVersionGroupEntry screenShotCallbackRVGE = null;

        public void SetScreenShotCallback(ScreenShotCallback screenShotCallback, ResourceVersionGroupEntry screenShotCallbackRVGE)
        {
            this.screenShotCallback = screenShotCallback;
            this.screenShotCallbackRVGE = screenShotCallbackRVGE;
        }

        public void OnDestroy()
        {
            if (SuperController.singleton != null)  SuperController.singleton.RemoveCanvas(mainCanvas);
            if (mainCanvas != null) mainCanvas.transform.SetParent(null, false);

            if (mainCanvas.gameObject != null) GameObject.Destroy(mainCanvas.gameObject);

            if (resourceBrowserUI!=null) resourceBrowserUI.OnDestory();
        }

        public void ApplyHelp(UIDynamicSlider sliderUID, string helpText)
        {
            ApplyHelp(sliderUID.slider.gameObject, helpText);
            ApplyHelp(sliderUID.sliderControl.gameObject, helpText);
            ApplyHelp(sliderUID.rangeAdjustGroup.gameObject, helpText);

            var inputTrans = sliderUID.transform.Find("ValueInputField");
            if (inputTrans != null) ApplyHelp(inputTrans.gameObject, helpText);
        }

        public void ApplyHelp(UIDynamicPopup popupUID, string helpText)
        {
            ApplyHelp(popupUID.popup.gameObject, helpText);
            ApplyHelp(popupUID.popup.popupPanel.gameObject, helpText);
        }
        public void ApplyHelp(UIDynamicToggle toggleUID, string helpText)
        {
            ApplyHelp(toggleUID.toggle.gameObject, helpText);
        }
        public void ApplyHelp(UIDynamicTextField textFieldUID, string helpText)
        {
            GameObject gameObject = textFieldUID.gameObject;
            if (textFieldUID.transform.GetComponent<InputField>() != null)
            {
                gameObject = textFieldUID.transform.GetComponent<InputField>().gameObject;
            }
            ApplyHelp(gameObject, helpText);
        }
        public void ApplyHelp(UIDynamicButton buttonUID, string helpText, bool warning = false)
        {
            ApplyHelp(buttonUID.button.gameObject, helpText, warning);
        }

        public void ClearHelp()
        {
            helpTextJSS.val = "";
            Image image = helpTextJSS.text.transform.parent.gameObject.GetComponent<Image>();
            image.color = Color.gray;
        }

        public void ApplyHelp(GameObject gameObject, string helpText, bool warning = false)
        {
            GeneralPointerEnterExitHelp contextHelp = gameObject.AddComponent<GeneralPointerEnterExitHelp>();
            contextHelp.Init(helpText, helpTextJSS, warning);
        }

        public void ApplyTagHelp(UIDynamicToggle toggleUID, string helpText)
        {
            ApplyHelp(toggleUID.toggle.gameObject, helpText);
        }

        public ResourceVersionGroupEntry GetNextRandomRVGEfromFilterResults()
        {
            if (randomCurrentFilterResults.Count == 0)
            {
                randomCurrentFilterResults = currentFilterResults.OrderBy(item => UnityEngine.Random.Range(0,1f)).ToList();
            }
            if (randomCurrentFilterResults.Count > 0)
            {
                ResourceVersionGroupEntry nextRVGE=null;
                if (randomCurrentFilterResults.Count ==1) nextRVGE = randomCurrentFilterResults.First();
                else
                {
                    while (nextRVGE==null || nextRVGE == lastRandomRVGE)
                    {
                        nextRVGE = randomCurrentFilterResults.First();
                    }
                }
                lastRandomRVGE = nextRVGE;
                randomCurrentFilterResults.Remove(nextRVGE);
                return nextRVGE;
            }
            return null;
        }

        public void SaveResourceScrollBarPosition()
        {

        }

        public int currentDesktopDisplayMode { get; private set; } = DesktopDisplayMode.rightDocked;

        protected GameObject backgroundBoxGO;

        public GameObject sideTagBoxGO { get; protected set; }
        protected GameObject filtersBoxGO;
        protected GameObject resourceBrowserGO;
        protected GameObject selectedResourceGO;
        private GameObject resourceActionsGO;
        private GameObject quickFiltersGO;

        private UIDynamicButton repositionUI;
        private UIDynamicButton leftExpandContractButtonUID;
        private UIDynamicButton rightExpandContractButtonUID;
        private UIDynamicButton cancelButtonUID;
        private GameObject infoTextBoxGO;

        public JSONStorableString helpTextJSS;
 

        public BrowserFiltersUI browserFiltersUI;
        public ResourceActionsUI resourceActionsUI;
        public ResourceBrowserUI resourceBrowserUI;
        public QuickFiltersUI quickFiltersUI;

        public List<ResourceVersionGroupEntry> currentFilterResults;
        private List<ResourceVersionGroupEntry> randomCurrentFilterResults = new List<ResourceVersionGroupEntry>();
        private ResourceVersionGroupEntry lastRandomRVGE = null;

        public int currentFilterResultsPage
        {
            get
            {
                int currentResourceType = browserFiltersUI.currentResourceType;
                if (!filterResultsPageByResourceType.ContainsKey(currentResourceType)) filterResultsPageByResourceType.Add(currentResourceType, 1);
                return filterResultsPageByResourceType[currentResourceType];
            }
            set
            {
                int currentResourceType = browserFiltersUI.currentResourceType;
                if (!filterResultsPageByResourceType.ContainsKey(currentResourceType)) filterResultsPageByResourceType.Add(currentResourceType, 1);
                filterResultsPageByResourceType[currentResourceType] = value;
            }
        }
        private Dictionary<int,int> filterResultsPageByResourceType = new Dictionary<int,int>();

        public ResourceSelections resourceSelectionsUI ;

        public Canvas mainCanvas;

        public bool isVREnabled { get { return Globals.isVREnabled; } }

        private bool isHidden = false;

        private bool inactiveDueToScreenShot = false;
    }


}
