﻿using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using MVR.FileManagementSecure;
using UnityEngine.UI;

namespace JayJayWon
{
    public class FolderBrowserUI
    {
        private GameObject folderBrowserBoxGO;
        private List<FolderRowUI> folderBrowserRows = new List<FolderRowUI>();
        private GameObject folderListContentGO;

        private int currentResourceType;
        private string currentAtomPresetResourceType;
        private int folderBrowserType;

        private BrowserFiltersUI browserFiltersUI;
        public BrowserUI browserUI;

        private UIDynamicButton closeButton;
        private UIDynamicButton cancelFolderMoveButton;       

        private UIDynamicToggle keepOpenToggle;

        public delegate void FolderSelectedCallback(Folder folder);

        private FolderSelectedCallback folderSelectedCallback;

        private GameObject folderOptionsGO;
        private FolderRowUI folderRowOptions;
        public Folder folderMove { get; private set; }
        public Folder folderDelete { get; private set; }
        private int folderBrowserUIMode = FolderBrowserUIMode.normal;

        public FolderBrowserUI(GameObject folderBrowserBoxGO, int folderBrowserType, FolderSelectedCallback callback, BrowserUI browserUI)
        {
            this.folderBrowserBoxGO = folderBrowserBoxGO;
            this.folderBrowserType = folderBrowserType;
            folderSelectedCallback = callback;
            folderListContentGO = folderBrowserBoxGO.transform.Find("scrollableContentViewportGO/vScrollableContentGO").gameObject;
            browserFiltersUI = browserUI.browserFiltersUI;
            this.browserUI = browserUI;
        }

        public void RefreshUI(int resourceType,string atomPresetResourceType)
        {
            ClearUI();
            bool fullWidthDesktopModeActive = !Globals.isVREnabled && browserUI.currentDesktopDisplayMode == DesktopDisplayMode.fullWidth;

            if (Settings.keepFolderBrowserOpenJSB.val && !fullWidthDesktopModeActive)
            {
                if (browserUI.currentDesktopDisplayMode == DesktopDisplayMode.rightDocked) folderBrowserBoxGO.transform.localPosition = new Vector2(-330f, -30f);
                else folderBrowserBoxGO.transform.localPosition = new Vector2(680f, -30f);
            }
            else folderBrowserBoxGO.transform.localPosition = new Vector2(320f, -30f);

            currentResourceType = resourceType;
            currentAtomPresetResourceType = atomPresetResourceType;

            foreach (Folder folder in FolderBrowser.GetFolderWithVisibleSubFolders(resourceType, atomPresetResourceType))
            {

                GameObject rowGO = UI.AddChildGOImage(folderListContentGO, new Color(0f, 0.8f, 0f, 0f), AnchorPresets.hStretchTop, 500f, 20f, Vector2.zero);
                folderBrowserRows.Add(new FolderRowUI(rowGO, folder, this, browserFiltersUI, folderSelectedCallback, folderBrowserUIMode));
            }
            closeButton = UI.CreateUIButton(folderBrowserBoxGO, 20f, 20f, "X", 28, 0, -20f, AnchorPresets.bottomRight);
            browserFiltersUI.parentBrowser.ApplyHelp(closeButton, "Close the Folder Browser window");
            closeButton.button.onClick.AddListener(delegate
            {
                folderBrowserBoxGO.SetActive(false);
            });


            if (folderBrowserUIMode!=FolderBrowserUIMode.moveFolder)
            {
                keepOpenToggle = UI.CreateToggle(folderBrowserBoxGO, Settings.keepFolderBrowserOpenJSB, 80f, 20f, "Keep Open",20,0f,-20f,AnchorPresets.bottomLeft);
                browserFiltersUI.parentBrowser.ApplyHelp(keepOpenToggle, "Keep folder browser open after each folder selection");
                keepOpenToggle.toggle.onValueChanged.AddListener(delegate
                {
                    RefreshUI(currentResourceType, currentAtomPresetResourceType);
                });
            }            

            cancelFolderMoveButton = UI.CreateUIButton(folderBrowserBoxGO, 100f, 20f, "Cancel Folder Move", 20, 0f, -20f, AnchorPresets.bottomLeft);
            browserFiltersUI.parentBrowser.ApplyHelp(cancelFolderMoveButton, "Cancel the Folder Move operation");
            cancelFolderMoveButton.buttonColor = Color.yellow;
            cancelFolderMoveButton.button.onClick.AddListener(delegate
            {
                SetNormalFolderMode();
            });
            if (folderBrowserUIMode!= FolderBrowserUIMode.moveFolder) cancelFolderMoveButton.gameObject.SetActive(false);

        }
        public void ClearUI()
        {
            foreach (FolderRowUI folderRow in folderBrowserRows)
            {
                folderRow.ClearUI();
            }
            folderBrowserRows.Clear();
            if (closeButton != null) GameObject.Destroy(closeButton.gameObject);
            if (keepOpenToggle != null) GameObject.Destroy(keepOpenToggle.gameObject);
            if (cancelFolderMoveButton != null) GameObject.Destroy(cancelFolderMoveButton.gameObject);
            folderRowOptions = null;
            if (folderOptionsGO != null) GameObject.Destroy(folderOptionsGO);
        }

        public void RefreshUICurrentResourceType()
        {
            RefreshUI(currentResourceType, currentAtomPresetResourceType);
        }

        public void SetNormalFolderMode()
        {
            folderBrowserUIMode = FolderBrowserUIMode.normal;
            RefreshUI(currentResourceType, currentAtomPresetResourceType);
            
        }

        private Folder moveToFolder;
        public void SelectMoveFolder(Folder moveToFolder)
        {
            string moveFolderPath = folderMove.GetFolderPath();

            this.moveToFolder = moveToFolder;
            FileManagerSecure.MoveFile(Const.SymLinksFolder + "\\" + moveFolderPath, Const.SymLinksFolder + "\\" + moveToFolder.GetFolderPath()+"\\"+folderMove.folderName, FolderMoveConfirmed, null, FolderMoveException);

            SetNormalFolderMode();
        }
        public void SelectDeleteFolder(Folder deleteFolder)
        {
            string deleteFolderPath = deleteFolder.GetFolderPath();

            Utils.DeleteFolder(deleteFolderPath, FolderMoveDeletionConfirmed, FolderMoveDeletionException);

            SetNormalFolderMode();
        }
       
        private void FolderMoveConfirmed()
        {
            FolderBrowser.MoveFolder(folderMove, moveToFolder);
        }
        private void FolderMoveException(Exception e)
        {
            SetNormalFolderMode();
            SuperController.LogMessage("BrowserAssist: Folder move failed. Did you perhaps have the folder open in Windows Explorer or does the folder name already exist in the new parent folder?");
            SuperController.LogError("BA.FolderBrowserUI.FolderMoveException: " + e);
        }

        private void FolderMoveDeletionConfirmed()
        {
            var folders = folderDelete.GetFolderAndAllSubFolders();
            FolderBrowser.PurgeFolder(folderDelete);

            if (folders.Contains(browserFiltersUI.selectedFolderFilter)) folderSelectedCallback(folderDelete.parentFolder);
            else folderSelectedCallback(browserFiltersUI.selectedFolderFilter);
        }
        private void FolderMoveDeletionException(Exception e)
        {
            SetNormalFolderMode();
            SuperController.LogMessage("BrowserAssist: Folder deletion failed. Did you perhaps have the folder open in Windows Explorer?");
            SuperController.LogError("BA.FolderBrowserUI.FolderMoveDeletionException: " + e);
        }

        public void ToggleOptionsMenu(FolderRowUI folderRow)
        {

            if (folderRow == folderRowOptions)
            {
                folderRowOptions = null;
                if (folderOptionsGO != null) GameObject.Destroy(folderOptionsGO);
            }
            else
            {
                folderRowOptions = null;
                if (folderOptionsGO != null) GameObject.Destroy(folderOptionsGO);

                folderOptionsGO = UI.AddChildGOImage(folderBrowserBoxGO, Color.white, AnchorPresets.topLeft, 103f, (folderRow.optionsMenuItems*20f)+1.5f, new Vector2(-100f,-20f));

                if (Globals.isVREnabled)
                {
                    folderOptionsGO.transform.position = folderRow.optionsButton.transform.position;

                    folderOptionsGO.transform.Translate(new Vector3(-0.18f, -0.015f, 0f));
                }
                else
                {
                    folderOptionsGO.transform.position = folderRow.optionsButton.transform.position + (folderRow.optionsButton.transform.up * -10f) + (folderRow.optionsButton.transform.right * -110f);
                }
                    

                float offset = 1.5f;
                if (folderRow.isRemoveRenameFolderOption)
                {
                    var renameFolderButton = UI.CreateUIButton(folderOptionsGO, 100f, 20f, "Rename Folder...",20, 0, -offset, anchorPreset:AnchorPresets.topMiddle,clearSprite:true);
                    renameFolderButton.button.onClick.AddListener(delegate
                    {
                        folderRow.RenameFolderStart();
                        
                    });
                    browserUI.ApplyHelp(renameFolderButton, "Change the name of the folder");

                    var deleteFolderButton = UI.CreateUIButton(folderOptionsGO, 100f, 20f, "Delete Folder...",20,0, -offset - 20f,AnchorPresets.topMiddle, clearSprite: true);
                    deleteFolderButton.buttonColor = Color.red;
                    deleteFolderButton.textColor = Color.white;
                    deleteFolderButton.button.onClick.AddListener(delegate
                    {
                        folderBrowserUIMode = FolderBrowserUIMode.deleteFolder;
                        folderDelete = folderRow.folder;
                        RefreshUICurrentResourceType();
                    });
                    browserUI.ApplyHelp(deleteFolderButton, "Delete the folder\nWARNING: This will delete all files in the folder and any sub folders",true);

                    var moveFolderButton = UI.CreateUIButton(folderOptionsGO, 100f, 20f, "Move Folder...", 20, 0, -offset - 40f, AnchorPresets.topMiddle, clearSprite: true);
                    offset = 60f;
                    moveFolderButton.button.onClick.AddListener(delegate
                    {
                        folderBrowserUIMode = FolderBrowserUIMode.moveFolder;
                        folderMove = folderRow.folder;
                        RefreshUICurrentResourceType();
                    });
                    browserUI.ApplyHelp(moveFolderButton, "Move the folder to become a subfolder of a different folder");
                }
                if (folderRow.isMoveSelectedOption)
                {
                    var moveSelectedButton = UI.CreateUIButton(folderOptionsGO, 100f, 20f, "Move Selected Resources to this Folder", 17, 0, -offset, anchorPreset: AnchorPresets.topMiddle, clearSprite: true);
                    moveSelectedButton.button.onClick.AddListener(delegate
                    {
                        MoveSelectedResources(folderRow);
                    });
                    browserUI.ApplyHelp(moveSelectedButton, "Move all of the currently selected resources (highlighted in cyan) to become a child of this folder: " + folderRow.folder.GetFolderPath());

                    offset += 20f;
                }
                if (folderRow.isCreateFolderOption)
                {
                    var createFolderButton = UI.CreateUIButton(folderOptionsGO, 100f, 20f, "Create Sub-folder", 20, 0, -offset, AnchorPresets.topMiddle, clearSprite: true);
                    createFolderButton.button.onClick.AddListener(delegate
                    {
                        folderRow.CreateSubFolder();
                        RefreshUICurrentResourceType();
                    });
                    browserUI.ApplyHelp(createFolderButton, "Create a new subfolder under the current folder");

                    offset += 20f;
                }
                if (folderRow.isExpandCollapseFolderOption)
                {
                    var expandFolderButton = UI.CreateUIButton(folderOptionsGO, 100f, 20f, "Expand all Sub-folders", 20, 0, -offset, AnchorPresets.topMiddle, clearSprite: true);
                    expandFolderButton.button.onClick.AddListener(delegate
                    {
                        folderRow.ExpandAll();
                        RefreshUICurrentResourceType();
                    });
                    browserUI.ApplyHelp(expandFolderButton, "Expand the folder tree hierarchy for this folder and all sub folders");

                    var collapseFolderButton = UI.CreateUIButton(folderOptionsGO, 100f, 20f, "Collapse all Sub-folders", 20, 0, -offset-20f, AnchorPresets.topMiddle, clearSprite: true);
                    collapseFolderButton.button.onClick.AddListener(delegate
                    {
                        folderRow.CollapseAll();
                        RefreshUICurrentResourceType();
                    });
                    browserUI.ApplyHelp(collapseFolderButton, "Collapse the folder tree hierarchy for this folder and all sub folders");
                }

                    folderRowOptions = folderRow;
            }
        }

        private void MoveSelectedResources (FolderRowUI folderRow)
        {
            foreach (ResourceVersionGroupEntry rvge in browserUI.resourceSelectionsUI.GetResourceTypeSelections(browserUI.browserFiltersUI.currentResourceType, browserUI.browserFiltersUI.currentAtomPresetResourceType).selectedRVGEs)
            {
                if (rvge.localFileResource && rvge.containingLocalFolder!=folderRow.folder)
                {
                    LocalRVGE localRVGE = rvge as LocalRVGE;

                    string sourceFolderPathFileName = Const.SymLinksFolder + "\\" + localRVGE.resourceFolderPathFileName;
                    string sourceFolderPathFileNameNoExt = Const.SymLinksFolder + "\\" + localRVGE.resourceFolderPathFileNameNoExt;

                    string targetPath = Const.SymLinksFolder + "\\" + folderRow.folder.GetFolderPath();

                    string targetFolderPathFileNameNoExt ;
                    string targetFolderPathFileName ;
                    bool existingFileTargetFileName = false;

                    if (FileManagerSecure.FileExists(targetPath+"\\"+rvge.resourceFileNameNoExt+"."+rvge.resourceFileNameExtension)) existingFileTargetFileName = true;

                    if (localRVGE.resourceCategory == ResourceCategory.preset)
                    {
                        targetFolderPathFileNameNoExt = targetPath + "\\Preset_" + Utils.GetUniquePresetName(targetPath, rvge.resourceFileNameNoExtNoPrefix);
                        targetFolderPathFileName = targetFolderPathFileNameNoExt + "." + localRVGE.resourceFileNameExtension;
                    }
                    else if (localRVGE.resourceCategory == ResourceCategory.sceneOrSessionPluginPresets)
                    {
                        targetFolderPathFileNameNoExt = targetPath + "\\Plugins_" + Utils.GetUniquePresetName(targetPath, rvge.resourceFileNameNoExtNoPrefix);
                        targetFolderPathFileName = targetFolderPathFileNameNoExt + "." + localRVGE.resourceFileNameExtension;
                    }
                    else
                    {
                        targetFolderPathFileNameNoExt = targetPath + "\\" + Utils.GetUniqueFileName(targetPath, localRVGE.resourceFileNameNoExt,localRVGE.resourceFileNameExtension);
                        targetFolderPathFileName = targetFolderPathFileNameNoExt + "." + localRVGE.resourceFileNameExtension;
                    }

                    try
                    {
                        FileManagerSecure.MoveFile(sourceFolderPathFileName, targetFolderPathFileName,false);
                        if (existingFileTargetFileName) SuperController.LogMessage("Filename '"+FileManagerSecure.GetFileName(sourceFolderPathFileName) +"' already exists in target folder. File has been renamed to '"+ FileManagerSecure.GetFileName(targetFolderPathFileName) + "'");
                    }
                    catch (Exception e){
                        SuperController.LogMessage("Move Selected Resources failed for '"+ sourceFolderPathFileName+"'. Do you perhaps have the source resource file open in a text editor? Or does a file with the same name exist in the target folder?");
                        SuperController.LogError("BA.FolderBrowserUI.MoveSelectedResources: " + e);

                        continue;
                    }

                    try
                    {
                        if (FileManagerSecure.FileExists(sourceFolderPathFileName+".fav"))
                        {
                            FileManagerSecure.MoveFile(sourceFolderPathFileName + ".fav", targetFolderPathFileName + ".fav");
                        }
                        
                    }
                    catch (Exception e)
                    {
                        SuperController.LogMessage("Move Selected Resources Favourite Tag failed for '\"+ sourceFolderPathFileName+\"'.");
                        SuperController.LogError("BA.FolderBrowserUI.MoveSelectedResources: " + e);
                    }

                    try
                    {
                        if (FileManagerSecure.FileExists(sourceFolderPathFileName + ".hide"))
                        {
                            FileManagerSecure.MoveFile(sourceFolderPathFileName + ".hide", targetFolderPathFileName + ".hide");
                        }

                    }
                    catch (Exception e)
                    {
                        SuperController.LogMessage("Move Selected Resources Hidden Tag failed for '\"+ sourceFolderPathFileName+\"'.");
                        SuperController.LogError("BA.FolderBrowserUI.MoveSelectedResources: " + e);
                    }


                    try
                    {
                        if (FileManagerSecure.FileExists(sourceFolderPathFileNameNoExt + ".jpg"))
                        {
                            FileManagerSecure.MoveFile(sourceFolderPathFileNameNoExt + ".jpg", targetFolderPathFileNameNoExt + ".jpg");
                        }

                    }
                    catch (Exception e)
                    {
                        SuperController.LogMessage("Move Selected Resources Thumbnail image failed for '\"+ sourceFolderPathFileName+\"'.");
                        SuperController.LogError("BA.FolderBrowserUI.MoveSelectedResources: " + e);
                    }


                    ResourceManifest.MoveRenameRVGELocalFolderPath(localRVGE, FileManagerSecure.GetFileName(targetFolderPathFileName), folderRow.folder);

                }
            }
            RefreshUICurrentResourceType();
            browserUI.RefreshFilterResults();
            browserUI.resourceActionsUI.RefreshUI();
        }


    }



    public class FolderRowUI
    {
        private GameObject folderRowGO;
        public Folder folder { get; private set; }
        private FolderBrowserUI parentFolderBrowserUI;

        public UIDynamicButton optionsButton { get; private set; }

        private UIDynamicButton folderButton;

        private UIDynamicTextField folderNameTextField;

        private JSONStorableString folderRenameJSS;

        private bool isRenameActive = false;
        private int folderUIMode;

        public FolderRowUI(GameObject folderRowGO, Folder folder, FolderBrowserUI parentFolderBrowserUI, BrowserFiltersUI browserFiltersUI, FolderBrowserUI.FolderSelectedCallback callback, int folderUIMode)
        {
            this.folderRowGO = folderRowGO;
            this.folder = folder;
            this.parentFolderBrowserUI = parentFolderBrowserUI;
            this.folderUIMode = folderUIMode;

            int subFolderCount = folder.subFolderCount;
            string expandedSymbol = "+";
            if (folder.expanded) expandedSymbol = "-";

            int depth = folder.GetDepth();

            folderRenameJSS = new JSONStorableString("folderRename", "", FolderRenameUpdate);

            if (subFolderCount > 0)
            {
                var expandButton = UI.CreateUIButton(folderRowGO, 20f, 20f, expandedSymbol, 28, depth * 10f, 0, AnchorPresets.middleLeft);
                browserFiltersUI.parentBrowser.ApplyHelp(expandButton, "Expand or collapse sub folders in the tree view for this folder");

                expandButton.button.onClick.AddListener(delegate
                {
                    folder.ToggleExpanded();
                    parentFolderBrowserUI.RefreshUICurrentResourceType();
                });

                if (folderUIMode == FolderBrowserUIMode.deleteFolder)
                {
                    expandButton.button.interactable = false;
                }
            }

            float buttonWidth = 150f;
            if (folder.isRootFolder) buttonWidth = 250f;

            folderButton = UI.CreateUIButton(folderRowGO, buttonWidth, 20f, "  " + folder.folderName, 20, (depth * 10f) + 20f, 0, AnchorPresets.middleLeft, false);

            if (browserFiltersUI.selectedFolderFilter == folder)
            {
                folderButton.buttonText.text = "<b>  " + folder.folderName + "</b>";
                folderButton.buttonColor = Color.cyan;
            }
            if (isMoveFolder) folderButton.buttonColor = Color.yellow;

            browserFiltersUI.parentBrowser.ApplyHelp(folderButton, "Select folder " + folder.GetFolderPath()+" and apply as a filter to the browser results");
            folderButton.buttonText.alignment = TextAnchor.MiddleLeft;
            if (folderUIMode == FolderBrowserUIMode.normal)
            {
                folderButton.button.onClick.AddListener(delegate
                {

                    callback(folder);
                    parentFolderBrowserUI.SetNormalFolderMode();
                });
            }
            
       

            folderNameTextField = UI.CreateTextBox(folderRowGO, folderRenameJSS, buttonWidth, 20f, 20, (depth * 10f) + 20f, 0, AnchorPresets.middleLeft, true, InputField.LineType.SingleLine,"",0f,0f);
            folderNameTextField.gameObject.SetActive(false);

            if (folderUIMode == FolderBrowserUIMode.normal && optionsMenuItems > 0 && !parentFolderBrowserUI.browserUI.isCallbackBrowser)
            {
                optionsButton = UI.CreateUIButton(folderRowGO, 20f, 20f, "v", 20, (depth * 10f) + 20f + buttonWidth, 0, AnchorPresets.middleLeft, false);
                optionsButton.buttonText.alignment = TextAnchor.MiddleCenter;
                optionsButton.button.onClick.AddListener(delegate
                {
                    if (!isRenameActive) parentFolderBrowserUI.ToggleOptionsMenu(this);
                    else RenameFolderEnd();
                });
                browserFiltersUI.parentBrowser.ApplyHelp(optionsButton, "Toggle folder operations menu for the folder " + folder.GetFolderPath());
            }
            if (folderUIMode == FolderBrowserUIMode.moveFolder)
            {
                if (isMoveFolder)
                {
                    optionsButton = UI.CreateUIButton(folderRowGO, 20f, 20f, "X", 20, (depth * 10f) + 20f + buttonWidth, 0, AnchorPresets.middleLeft, false);
                    optionsButton.buttonText.alignment = TextAnchor.MiddleCenter;
                    optionsButton.button.onClick.AddListener(delegate
                    {
                        parentFolderBrowserUI.SetNormalFolderMode();
                    });
                    browserFiltersUI.parentBrowser.ApplyHelp(optionsButton, "Cancel the Folder move operation");
                }
                else if (parentFolderBrowserUI.folderMove.parentFolder!=null && parentFolderBrowserUI.folderMove.parentFolder!=folder && !isMoveFolderSubFolder)
                {
                    optionsButton = UI.CreateUIButton(folderRowGO, 20f, 20f, ">", 20, (depth * 10f) + 20f + buttonWidth, 0, AnchorPresets.middleLeft, false);
                    optionsButton.buttonText.alignment = TextAnchor.MiddleCenter;
                    optionsButton.buttonColor = Color.yellow;
                    optionsButton.button.onClick.AddListener(delegate
                    {
                        parentFolderBrowserUI.SelectMoveFolder(folder);
                    });
                    browserFiltersUI.parentBrowser.ApplyHelp(optionsButton, "Move folder '" + parentFolderBrowserUI.folderMove.folderName + "' to a sub folder of "+folder.GetFolderPath());
                }
            }

            if (folderUIMode == FolderBrowserUIMode.deleteFolder)
            {
                if (isDeleteFolder)
                {
                    optionsButton = UI.CreateUIButton(folderRowGO, 20f, 20f, "X", 20, (depth * 10f) + 20f + buttonWidth, 0, AnchorPresets.middleLeft, false);
                    optionsButton.buttonText.alignment = TextAnchor.MiddleCenter;
                    optionsButton.buttonColor = Color.green;
                    optionsButton.button.onClick.AddListener(delegate
                    {
                        parentFolderBrowserUI.SetNormalFolderMode();
                    });
                    browserFiltersUI.parentBrowser.ApplyHelp(optionsButton, "Cancel the Folder delete operation");

                    folderButton.button.onClick.AddListener(delegate
                    {
                        parentFolderBrowserUI.SelectDeleteFolder(folder);
                        browserFiltersUI.parentBrowser.ClearHelp();
                        parentFolderBrowserUI.SetNormalFolderMode();
                    });
                    folderButton.buttonColor = Color.red;
                    folderButton.textColor = Color.white;
                    browserFiltersUI.parentBrowser.ApplyHelp(folderButton, "Confirm Delete folder " + folder.GetFolderPath()+"\nWARNING: This will permanently delete all content in the folder and any subfolders.",true);
                }
                else
                {
                    folderButton.button.interactable = false;
                }
            }


        }

        private string updatedFolderName;
        private void FolderRenameUpdate(string newFolderName)
        {
            string folderPath = folder.GetFolderPath();

            updatedFolderName = newFolderName;
            FileManagerSecure.MoveFile(Const.SymLinksFolder + "\\" + folderPath, FileManagerSecure.GetDirectoryName(Const.SymLinksFolder + "\\" + folderPath) + "\\" + newFolderName,FolderRenameConfirmed,null, FolderRenameException);
           
        }
        private void FolderRenameException(Exception e)
        {
            SuperController.LogMessage("BrowserAssist: Folder rename failed. Did you perhaps choose a folder name that already exists?");
        }
        private void FolderRenameConfirmed()
        {
            folder.RenameFolder(updatedFolderName);
            folderButton.label = "  " + folder.folderName;
            RenameFolderEnd();
        }

        public void RenameFolderEnd()
        {
            isRenameActive = false;
            folderButton.gameObject.SetActive(true);
            optionsButton.gameObject.SetActive(true);
            folderRenameJSS.valNoCallback = folder.folderName;
            folderNameTextField.gameObject.SetActive(false);
        }
        public void RenameFolderStart()
        {
            isRenameActive = true;
            folderNameTextField.gameObject.SetActive(true);
            folderRenameJSS.valNoCallback = folder.folderName;
            folderButton.gameObject.SetActive(false);
            optionsButton.gameObject.SetActive(false);
            
//            folderNameTextField.backgroundColor = Color.gray;

            var inputComponent = folderNameTextField.gameObject.GetComponent<InputField>();
            inputComponent.ActivateInputField();
            inputComponent.onEndEdit.AddListener(delegate { RenameFolderEnd(); });

            parentFolderBrowserUI.ToggleOptionsMenu(this);
        }
        public void CreateSubFolder()
        {
            folder.CreateSubFolder();
        }

        public void ExpandAll()
        {
            folder.ExpandFolderPlusChildren();
        }
        public void CollapseAll()
        {
            folder.CollapseFolderPlusChildren();
        }

        public int optionsMenuItems { get
            {
                int count = 0;
                if (isMoveFolder) return 0;
                if (isRemoveRenameFolderOption) count += 3;
                if (isCreateFolderOption) count += 1;
                if (isExpandCollapseFolderOption) count += 2;
                if (isMoveSelectedOption) count += 1;
                return count;
            } }
        private bool isMoveFolder { get { return (folderUIMode == FolderBrowserUIMode.moveFolder && parentFolderBrowserUI.folderMove == folder); } }

        private bool isDeleteFolder { get { return (folderUIMode == FolderBrowserUIMode.deleteFolder && parentFolderBrowserUI.folderDelete == folder); } }

        private bool isMoveFolderSubFolder { get { return (folderUIMode == FolderBrowserUIMode.moveFolder && folder.allContainingLocalFolders.Contains(parentFolderBrowserUI.folderMove)); } }
        private bool isWriteAccesible { get
            {
                if (folder.GetFolderPath().StartsWith("Saves\\") && !Globals.scenesFolderAccess) return false;
                if (folder.GetFolderPath().StartsWith("Custom\\") && !Globals.customFolderAccess) return false;
                
                return true;
            } }
        public bool isRemoveRenameFolderOption
        {
            get {
                if (folder.isRootFolder) return false;
                if (!isWriteAccesible) return false;
                return true;
            }
        }

        public bool isCreateFolderOption
        {
            get
            {
                if (!isWriteAccesible) return false;
                return true;
            }
        }

        public bool isExpandCollapseFolderOption
        {
            get
            {
                if (folder.subFolderCount==0) return false;
                return true;
            }
        }

        public bool isMoveSelectedOption
        {
            get
            {
                var browserUI = parentFolderBrowserUI.browserUI;
                var browserFiltersUI = browserUI.browserFiltersUI;

                if (!isWriteAccesible) return false;
                if (browserUI.resourceSelectionsUI==null) return false;
                return browserUI.resourceSelectionsUI.GetResourceTypeSelections(browserFiltersUI.currentResourceType, browserFiltersUI.currentAtomPresetResourceType).containsLocal;
            }
        }

        public void ClearUI()
        {
            GameObject.Destroy(folderRowGO);
        }
    }

    public class FolderBrowser
    {        
        private static Dictionary<int, HashSet<Folder>> rootFoldersByResource = new Dictionary<int, HashSet<Folder>>();

        private static Dictionary<string, HashSet<Folder>> rootFoldersByAtomPresetResource = new Dictionary<string, HashSet<Folder>>();

        private static Dictionary<string, Folder> folders = new Dictionary<string, Folder>();

        public static HashSet<Folder> GetFolderWithVisibleSubFolders(int resourceType, string atomPresetResourceType)
        {
            HashSet<Folder> visibleFolders = new HashSet<Folder>();

            if (resourceType != ResourceType.presetAtom)
            {
                if (!rootFoldersByResource.ContainsKey(resourceType))
                {
                    CreateRootFolders(resourceType, atomPresetResourceType);
                }

                foreach (Folder rootFolder in rootFoldersByResource[resourceType])
                {
                    visibleFolders.UnionWith(rootFolder.GetFolderWithVisibleSubFolders());
                }
            }
            else
            {
                if (!rootFoldersByAtomPresetResource.ContainsKey(atomPresetResourceType))
                {
                    CreateRootFolders(resourceType, atomPresetResourceType);
                }

                foreach (Folder rootFolder in rootFoldersByAtomPresetResource[atomPresetResourceType])
                {
                    visibleFolders.UnionWith(rootFolder.GetFolderWithVisibleSubFolders());
                }
            }
            

            return visibleFolders;
        }

        public static void MoveFolder(Folder folderToMove, Folder newParentFolder)
        {
            HashSet<Folder> foldersToMove = folderToMove.GetFolderAndAllSubFolders();

            foreach (Folder folder in foldersToMove)
            {
                string folderKey = folder.GetFolderPath();
                if (!Globals.caseSensitiveFileSystem) folderKey = folderKey.ToLower();

                folders.Remove(folderKey);
            }
            
            Folder oldParentFolder = folderToMove.parentFolder;        

            folderToMove.ChangeParentFolder(newParentFolder);           

            foreach (Folder folder in foldersToMove)
            {
                string folderKey = folder.GetFolderPath();
                if (!Globals.caseSensitiveFileSystem) folderKey = folderKey.ToLower();

                folders.Add(folderKey, folder);
            }

            oldParentFolder.RefreshSubFolders();
            newParentFolder.RefreshSubFolders();
        }

        public static void PurgeFolder(Folder folderToDelete)
        {
            HashSet<Folder> foldersToDelete = folderToDelete.GetFolderAndAllSubFolders();

            foreach (Folder folder in foldersToDelete)
            {
                string folderKey = folder.GetFolderPath();
                if (!Globals.caseSensitiveFileSystem) folderKey = folderKey.ToLower();
                folders.Remove(folderKey);
                folder.PurgeAllLocalRVGEs();
            }

            Folder oldParentFolder = folderToDelete.parentFolder;
            oldParentFolder.RefreshSubFolders();
        }

        private static void CreateRootFolders(int resourceType, string atomPresetResourceType)
        {
            if (resourceType != ResourceType.presetAtom)
            {
                if (!rootFoldersByResource.ContainsKey(resourceType)) rootFoldersByResource.Add(resourceType, new HashSet<Folder>());
                
            }
            else
            {
                if (!rootFoldersByAtomPresetResource.ContainsKey(atomPresetResourceType)) rootFoldersByAtomPresetResource.Add(atomPresetResourceType, new HashSet<Folder>());
            }
            
            foreach (string rootFolderPath in Const.GetResourceRootFolderNames(resourceType, atomPresetResourceType))
            {
                Folder folder = null;
                string folderKey = rootFolderPath;
                if (!Globals.caseSensitiveFileSystem) folderKey = folderKey.ToLower();
                if (folders.ContainsKey(folderKey)) folder = folders[folderKey];
                else
                {
                    folder = new Folder(rootFolderPath);
                    folders.Add(folderKey, folder);
                }

                if (resourceType != ResourceType.presetAtom) rootFoldersByResource[resourceType].Add(folder);
                else rootFoldersByAtomPresetResource[atomPresetResourceType].Add(folder);              
            }
        }

        private static void RescanAllRootSubFolders(int resourceType, string atomPresetResourceType)
        {
            HashSet<Folder> folders =null;
            if (resourceType != ResourceType.presetAtom) folders = rootFoldersByResource[resourceType];
            else folders = rootFoldersByAtomPresetResource[atomPresetResourceType];

            foreach (Folder rootFolder in folders)
            {
                rootFolder.RescanSubFolders();
            }
        }

        public static Folder GetFolder(string folderPath, int resourceType, string atomPresetResourceType, bool returnRootFolderIfMissing=false)
        {
            if ((resourceType!=ResourceType.presetAtom && !rootFoldersByResource.ContainsKey(resourceType))|| (resourceType == ResourceType.presetAtom && !rootFoldersByAtomPresetResource.ContainsKey(atomPresetResourceType)))
            {
                CreateRootFolders(resourceType, atomPresetResourceType);
            }

            string folderKey = folderPath;
            if (!Globals.caseSensitiveFileSystem) folderKey = folderKey.ToLower();

            if (folders.ContainsKey(folderKey)) return folders[folderKey];
            else
            {
                RescanAllRootSubFolders(resourceType, atomPresetResourceType);
                if (folders.ContainsKey(folderKey)) return folders[folderKey];

                if (!FileManagerSecure.DirectoryExists(folderKey) && returnRootFolderIfMissing) return rootFoldersByResource[resourceType].First();
                else throw new Exception("BA.FolderBrowser.GetFolder: failed to find existing resource folder " + folderPath);

            }
        }
        public static void RemoveFolder(Folder removeFolder)
        {
            string folderKey = removeFolder.GetFolderPath();
            if (!Globals.caseSensitiveFileSystem) folderKey = folderKey.ToLower();

            folders.Remove(folderKey);

        }
        public static Folder GetSubFolder(string subFolderName, Folder parentFolder)
        {
            string subFolderPath = parentFolder.GetFolderPath() + "\\" + subFolderName;
            string folderKey = subFolderPath;
            if (!Globals.caseSensitiveFileSystem) folderKey = folderKey.ToLower();

            if (folders.Keys.Contains(folderKey))
            {
                return folders[folderKey];
            }
            var newFolder = new Folder(parentFolder, subFolderName);
            folders.Add(folderKey, newFolder);
            return newFolder;
        }
    }

    public class Folder
    {
        public bool isRootFolder { get; private set; } = false;
        public bool expanded { get; private set; } = false;

        public Folder parentFolder { get; private set; } = null;

        public int subFolderCount { get { return subFolders.Keys.Count; } }

        public string folderName { get; private set; }

        private Dictionary<string, Folder> subFolders = new Dictionary<string, Folder>();

        private HashSet<ResourceVersionGroupEntry> localRVGEs = new HashSet<ResourceVersionGroupEntry>();

        public Folder(string rootFolderPath)
        {
            parentFolder = null;
            folderName = rootFolderPath;
            isRootFolder = true;
            expanded = true;
            RefreshSubFolders();
        }

        public void RegisterLocalRVGE(ResourceVersionGroupEntry rvge)
        {
            if (rvge.localFileResource)
            {
                localRVGEs.Add(rvge);
            }
            else SuperController.LogError("BA.Folder.RegisterLocalRVGE: Attempt to register non local RVGE with Folder");
        }

        public void UnRegisterLocalRVGE(ResourceVersionGroupEntry rvge)
        {
            localRVGEs.Remove(rvge);
        }

        public Folder(Folder parentFolder, string folderName)
        {
            this.parentFolder = parentFolder;
            this.folderName = folderName;

            RefreshSubFolders();
        }
        public HashSet<Folder> allContainingLocalFolders
        {
            get
            {
                HashSet<Folder> result = new HashSet<Folder>();

                result.Add(this);
                if (!isRootFolder) result.UnionWith(parentFolder.allContainingLocalFolders);
                return result;
            }
        }
        public int GetDepth()
        {
            if (isRootFolder) return 0;
            return parentFolder.GetDepth() + 1;
        }

        public string GetFolderPath()
        {
            if (isRootFolder) return folderName;
            return parentFolder.GetFolderPath() + "\\" + folderName;
        }

        public HashSet<Folder> GetFolderWithVisibleSubFolders()
        {
            var visibleFolders = new HashSet<Folder>();
            visibleFolders.Add(this);
            if (expanded)
            {
                foreach (var subFolder in subFolders)
                {
                    visibleFolders.UnionWith(subFolder.Value.GetFolderWithVisibleSubFolders()); ;
                }

            }

            return visibleFolders;
        }
        public HashSet<Folder> GetFolderWithllSubFolders()
        {
            var folders = new HashSet<Folder>();
            folders.Add(this);
            foreach (var subFolder in subFolders)
            {
                folders.UnionWith(subFolder.Value.GetFolderWithllSubFolders()); ;
            }

            return folders;
        }


        public void RefreshSubFolders()
        {
            if (FileManagerSecure.DirectoryExists(GetFolderPath()))
            {
                string[] subFolderPaths = new string[] { };

                string folderPath = GetFolderPath();

                try
                {
                    subFolderPaths = FileManagerSecure.GetDirectories(folderPath);
                }
                catch(Exception ex)
                {
                    SuperController.LogError("BA.Folder.RefreshSubFolders: Exception obtaining directories for '"+folderPath+"' - " + ex);
                }

                var newSubFolders = new Dictionary<string, Folder>();

                foreach (var subFolderPath in subFolderPaths)
                {
                    string subFolderName = subFolderPath.Substring
    (subFolderPath.LastIndexOf('\\') + 1);

                    if (subFolders.Keys.Contains(subFolderName)) newSubFolders.Add(subFolderName, subFolders[subFolderName]);
                    else newSubFolders.Add(subFolderName, FolderBrowser.GetSubFolder(subFolderName, this));
                }
                var missingFolders = subFolders.Values.Except(newSubFolders.Values) as HashSet<Folder>;

                if (missingFolders != null) RemoveSubFolders(missingFolders);
                subFolders = newSubFolders;
            }
            
        }
        public void RescanSubFolders()
        {
            RefreshSubFolders();
            foreach (var subFolder in subFolders)
            {
                subFolder.Value.RescanSubFolders();
            }
        }

        public void RenameFolder(string newFolderName)
        {
            folderName = newFolderName;
            UpdateRVGEFolderPaths();
        }

        public void ChangeParentFolder(Folder newParentFolder)
        {
            parentFolder = newParentFolder;
            UpdateRVGEFolderPaths();
        }

        public void PurgeAllLocalRVGEs()
        {
            foreach (LocalRVGE rvge in localRVGEs)
            {
                ResourceManifest.PurgeLocalRVGE(rvge);

            }
        }

        private void UpdateRVGEFolderPaths()
        {
            foreach(LocalRVGE rvge in localRVGEs)
            {
                ResourceManifest.ResetRVGEFolderPath(rvge);
                
            }
            foreach(var subFolderKVP in subFolders)
            {
                subFolderKVP.Value.UpdateRVGEFolderPaths();
            }
        }



        public HashSet<Folder> GetFolderAndAllSubFolders()
        {
            HashSet<Folder> result = new HashSet<Folder>();

            result.Add(this);
            foreach (var subFolderKVP in subFolders)
            {
                result.UnionWith(subFolderKVP.Value.GetFolderAndAllSubFolders());
            }
            return result;
        }
        private void RemoveSubFolders(HashSet<Folder> removeSubFolders)
        {
            foreach (Folder subFolder in removeSubFolders)
            {
                subFolder.RemoveFolder();
            }
        }
        private void RemoveFolder()
        {
            var allSubFolders = new HashSet<Folder>(subFolders.Values.ToArray());
            RemoveSubFolders(allSubFolders);
            FolderBrowser.RemoveFolder(this);
        }


        public void ToggleExpanded()
        {
            if (!expanded)
            {
                RefreshSubFolders();
            }
            expanded = !expanded;
        }

        public void ExpandParents()
        {

            if (!isRootFolder) parentFolder.ExpandFolderPlusParents();
        }

        public void ExpandFolderPlusParents()
        {
            expanded = true;
            if (!isRootFolder) parentFolder.ExpandFolderPlusParents();
        }

        public void ExpandFolderPlusChildren()
        {
            expanded = true;
            foreach (var subFolder in subFolders) subFolder.Value.ExpandFolderPlusChildren();
        }

        public void CollapseFolderPlusChildren()
        {
            expanded = false;
            foreach (var subFolder in subFolders) subFolder.Value.CollapseFolderPlusChildren();
        }
        public void CreateSubFolder()
        {
            expanded = true;
            string folderName = "New Folder";
            int folderNumber = 1;
            while (FileManagerSecure.DirectoryExists(Const.SymLinksFolder + "\\" + GetFolderPath() + "\\" + folderName))
            {
                folderName = "New Folder(" + folderNumber.ToString() + ")";
                folderNumber++;
            }
            FileManagerSecure.CreateDirectory(Const.SymLinksFolder + "\\" + GetFolderPath() + "\\" + folderName);
            RefreshSubFolders();
        }
    }
}
