using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine.UI;
using System.Security.AccessControl;
using MVR.FileManagementSecure;

namespace JayJayWon
{
    public class CallbackResourceBrowserUI : ResourceBrowserUI
    {
        private JSONStorableString fileNameCallbackJSS;
        private JSONStorableBool keepOpenJSB;
        private int baKeepOpenOption;
        public CallbackResourceBrowserUI(BrowserUI _parentBrowser, GameObject _parentBoxGO): base(_parentBrowser, _parentBoxGO)
        {

        }


        public void InitCallback(JSONStorableString fileNameCallbackJSS, int baKeepOpenOption, JSONStorableBool keepOpenJSB)
        {
            this.fileNameCallbackJSS = fileNameCallbackJSS;
            this.baKeepOpenOption = baKeepOpenOption;
            this.keepOpenJSB = keepOpenJSB;

        }

        protected override void SelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton buttonUID, bool suppressQuickActivate = false)
        {
            if (browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi)
            {
                parentBrowser.resourceSelectionsUI.SingleSelectRVGE(rvge, buttonUID);
                if (!suppressQuickActivate)
                {
                    if (fileNameCallbackJSS != null)
                    {
                        string normalizedFilePath = FileManagerSecure.NormalizePath(rvge.resourceFullPathReference);
                        fileNameCallbackJSS.val = normalizedFilePath;
                        
                    }
                    if (baKeepOpenOption == BAKeepOpenOptionEnum.unavailable || !keepOpenJSB.val)
                    {
                        fileNameCallbackJSS = null;
                        parentBrowser.SetInactive();
                    }
                }
                
            }
            else if (!parentBrowser.isVREnabled)
            {
                if (multiSelectKeyPressed) tempMultiMode = false;
                if (Input.GetKey(KeyCode.LeftShift)) parentBrowser.resourceSelectionsUI.LeftShiftSelectRVGE(rvge, buttonUID, currentPageRVGEs, currentPageButtons);
                else if (Input.GetKey(KeyCode.RightShift)) parentBrowser.resourceSelectionsUI.RightShiftSelectRVGE(rvge, buttonUID, currentPageRVGEs, currentPageButtons);
                else if (Input.GetKey(KeyCode.LeftControl) || Input.GetKey(KeyCode.RightControl)) parentBrowser.resourceSelectionsUI.ToggleSelectRVGE(rvge, buttonUID);
                else if (Input.GetKey(KeyCode.LeftAlt) || Input.GetKey(KeyCode.RightAlt)) parentBrowser.resourceSelectionsUI.SingleSelectRVGE(rvge, buttonUID);
                else
                {
                    parentBrowser.resourceSelectionsUI.DeSelectAll(rvge.resourceType, rvge.atomPresetResourceType);
                    parentBrowser.resourceSelectionsUI.SingleSelectRVGE(rvge, buttonUID);
                }
            }
            else parentBrowser.resourceSelectionsUI.ToggleSelectRVGE(rvge, buttonUID);
        }

    }
    public class ResourceBrowserUI
    {
        private GameObject parentBoxGO;
        protected BrowserUI parentBrowser;
        public BrowserFiltersUI browserFiltersUI;

        public JSONStorableString resourceCountTextJSS;
        public JSONStorableString selectedResourceCountTextJSS;

        private int resourceButtonSize
        {
            get
            {
                if (Globals.isVREnabled) return Settings.resourceButtonSizeVR;
                else return Settings.resourceButtonSizeDesktop;
            }
            set
            {
                if (Globals.isVREnabled) Settings.resourceButtonSizeVR = value;
                else Settings.resourceButtonSizeDesktop = value;
            }
        }
        const int resourceButtonlabelHeight = 16;

        private UIDynamicButton firstPageUIDButton;
        private UIDynamicButton prevPageUIDButton;
        private UIDynamicButton nextPageUIDButton;
        private UIDynamicButton lastPageUIDButton;
        private UIDynamicButton incResourceButtonSizesUIDButton;
        private UIDynamicButton decResourceButtonSizesUIDButton;
        private UIDynamicButton clearMultiSelectResourcesUIDButton;
        private UIDynamicButton selectAllResourcesUIDButton;
        private GameObject resourceCountTextBox;
        private GameObject selectedResourceCountTextBox;

        private Dictionary<UIDynamicButton, ResourceVersionGroupEntry> UIDButtonToRVGEDict = new Dictionary<UIDynamicButton, ResourceVersionGroupEntry>();
        private Dictionary<ResourceVersionGroupEntry, UIDynamicButton> RVGEToUIDButtonDict = new Dictionary<ResourceVersionGroupEntry, UIDynamicButton>();

        private List<GameObject> activeResourceRowGOs = new List<GameObject>();
        private List<GameObject> inactiveResourceRowGOs = new List<GameObject>();

        private ScrollRect scrollRect;

        private Dictionary<int, float> scrollBarPositionByResourceType = new Dictionary<int, float>();

        private float scrollBarPosition
        {
            get
            {

                int currentResourceType = browserFiltersUI.currentResourceType;
                if (!scrollBarPositionByResourceType.ContainsKey(currentResourceType)) scrollBarPositionByResourceType.Add(currentResourceType, 1);
                return scrollBarPositionByResourceType[currentResourceType];
            }
            set
            {
                int currentResourceType = browserFiltersUI.currentResourceType;
                if (!scrollBarPositionByResourceType.ContainsKey(currentResourceType)) scrollBarPositionByResourceType.Add(currentResourceType, 1);
                scrollBarPositionByResourceType[currentResourceType] = value;
            }
        }

        private List<UIDynamicButton> activeButtonUIDs = new List<UIDynamicButton>();
        private HashSet<UIDynamicButton> unparentedButtonUIDs = new HashSet<UIDynamicButton>();

        private Dictionary<GameObject, List<UIDynamicButton>> buttonUIDsByRow = new Dictionary<GameObject, List<UIDynamicButton>>();
        private List<ResourceVersionGroupEntry> currentFilterResults { get { return parentBrowser.currentFilterResults; } }

        public void OnDestory()
        {
            foreach (UIDynamicButton buttonUID in activeButtonUIDs)
            {
                GameObject.Destroy(buttonUID);

            }
            foreach (UIDynamicButton buttonUID in unparentedButtonUIDs)
            {
                GameObject.Destroy(buttonUID);

            }
        }

        private int currentFilterResultsPage { get { return parentBrowser.currentFilterResultsPage; } set { parentBrowser.currentFilterResultsPage = value; } }

        protected List<ResourceVersionGroupEntry> currentPageRVGEs = new List<ResourceVersionGroupEntry>();
        protected List<UIDynamicButton> currentPageButtons = new List<UIDynamicButton>();
        private float resouceButtonWidth
        {
            get
            {
                if (resourceButtonSize == 0) return 60;
                if (resourceButtonSize == 1) return 67;
                if (resourceButtonSize == 2) return 75.5f;
                if (resourceButtonSize == 3) return 87;
                if (resourceButtonSize == 4) return 102;
                if (resourceButtonSize == 5) return 123;
                if (resourceButtonSize == 6) return 154;
                return 123;
            }
        }
        private int buttonsPerRow
        {
            get
            {
                return (int)Math.Floor((parentBrowser.browserScreenWidth - 20f) / (float)(resouceButtonWidth + 3));
            }
        }
        private int selectedResourcesVOffset
        {
            get
            {
                if (browserFiltersUI == null || browserFiltersUI.selectModeJSEnum==null || browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi) return -23;
                else return -45;
            }
        }

        private float navButtonsHOffset
        {
            get
            {
                return parentBrowser.resourceSelectionsUI.resourceSelectionBoxWidth - (640 / 2) + BrowserUI.edgeSpacerSize + 7.5f;
            }
        }
        private float navButtonsVOffset
        {
            get
            {
                if (browserFiltersUI == null || browserFiltersUI.selectModeJSEnum ==null|| browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi) return selectedResourcesVOffset;
                else return selectedResourcesVOffset + 5;
            }
        }
        private float resourceCountHOffset
        {
            get
            {
                if (browserFiltersUI == null || browserFiltersUI.selectModeJSEnum == null || browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi) return navButtonsHOffset + 47f + (76f / 2) + 15f;
                else return navButtonsHOffset + (76f / 2);
            }
        }
        private int resourceCountVOffset
        {
            get
            {
                if (browserFiltersUI == null || browserFiltersUI.selectModeJSEnum == null || browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi) return selectedResourcesVOffset;
                else return selectedResourcesVOffset + 25;
            }
        }
        private int resoureButtonSizeHOffset
        {
            get
            {
                return (640 / 2) - 30;
            }
        }

        public void NudgeScrollBar()
        {
            if (scrollRect != null)
            {
                if (scrollBarPosition < 0.5f)
                {
                    scrollBarPosition += 0.1f;
                    scrollBarPosition -= 0.1f;
                }
                else
                {
                    scrollBarPosition -= 0.1f;
                    scrollBarPosition += 0.1f;
                }
            }

        }
        public ResourceBrowserUI(BrowserUI _parentBrowser, GameObject _parentBoxGO)
        {
            parentBrowser = _parentBrowser;
            parentBoxGO = _parentBoxGO;
            browserFiltersUI = parentBrowser.browserFiltersUI;
            scrollRect = parentBoxGO.GetComponent<ScrollRect>();
            scrollRect.verticalScrollbar.onValueChanged.AddListener(delegate
            {
                if (scrollRect.verticalScrollbar.size > 0.999f) scrollBarPosition = 1f;
                else scrollBarPosition = scrollRect.verticalNormalizedPosition;
            });
            scrollRect.verticalNormalizedPosition = scrollBarPosition;

            firstPageUIDButton = UI.CreateUIButton(parentBoxGO, 15f, 20f, "<b>|<</b>", 26, navButtonsHOffset, selectedResourcesVOffset, anchorPreset: AnchorPresets.bottomMiddle);
            firstPageUIDButton.buttonImage.sprite = null;
            firstPageUIDButton.button.onClick.AddListener(delegate {
                FirstPageResults();
            });
            parentBrowser.ApplyHelp(firstPageUIDButton, "Navigate to the first page of browser results");

            prevPageUIDButton = UI.CreateUIButton(parentBoxGO, 15f, 20f, "<b><</b>", 26, navButtonsHOffset + 15, selectedResourcesVOffset, anchorPreset: AnchorPresets.bottomMiddle);
            prevPageUIDButton.buttonImage.sprite = null;
            prevPageUIDButton.button.onClick.AddListener(delegate {
                PrevPageResults();
            });
            parentBrowser.ApplyHelp(prevPageUIDButton, "Navigate to the previous page of browser results");

            nextPageUIDButton = UI.CreateUIButton(parentBoxGO, 15f, 20f, "<b>></b>", 26, navButtonsHOffset + 30, selectedResourcesVOffset, anchorPreset: AnchorPresets.bottomMiddle);
            nextPageUIDButton.buttonImage.sprite = null;
            nextPageUIDButton.button.onClick.AddListener(delegate {
                NextPageResults();

            });
            parentBrowser.ApplyHelp(nextPageUIDButton, "Navigate to the next page of browser results");

            lastPageUIDButton = UI.CreateUIButton(parentBoxGO, 15f, 20f, "<b>>|</b>", 26, navButtonsHOffset + 45, selectedResourcesVOffset, anchorPreset: AnchorPresets.bottomMiddle);
            lastPageUIDButton.buttonImage.sprite = null;
            lastPageUIDButton.button.onClick.AddListener(delegate {
                LastPageResults();
            });
            parentBrowser.ApplyHelp(lastPageUIDButton, "Navigate to the last page of browser results");


            incResourceButtonSizesUIDButton = UI.CreateUIButton(parentBoxGO, 20f, 20f, "<b>+</b>", 26, resoureButtonSizeHOffset, selectedResourcesVOffset, anchorPreset: AnchorPresets.bottomMiddle);
            incResourceButtonSizesUIDButton.buttonImage.sprite = null;
            incResourceButtonSizesUIDButton.button.onClick.AddListener(delegate {
                IncreaseButtonSizes();
            });
            parentBrowser.ApplyHelp(incResourceButtonSizesUIDButton, "Increase the size of resource icons in the browser results window.");

            decResourceButtonSizesUIDButton = UI.CreateUIButton(parentBoxGO, 20f, 20f, "<b>-</b>", 26, resoureButtonSizeHOffset + 20, selectedResourcesVOffset, anchorPreset: AnchorPresets.bottomMiddle);
            decResourceButtonSizesUIDButton.buttonImage.sprite = null;
            decResourceButtonSizesUIDButton.button.onClick.AddListener(delegate {
                DecreaseButtonSizes();
            });
            parentBrowser.ApplyHelp(decResourceButtonSizesUIDButton, "Decrease the size of resource icons in the browser results window.");

            clearMultiSelectResourcesUIDButton = UI.CreateUIButton(parentBoxGO, 15f, 15f, "X", 22, navButtonsHOffset, selectedResourcesVOffset - 31f, anchorPreset: AnchorPresets.bottomMiddle);
            clearMultiSelectResourcesUIDButton.buttonImage.sprite = null;
            clearMultiSelectResourcesUIDButton.buttonColor = Color.cyan;
            clearMultiSelectResourcesUIDButton.button.onClick.AddListener(delegate {
                parentBrowser.resourceSelectionsUI.DeSelectAll(browserFiltersUI.currentResourceType, browserFiltersUI.currentAtomPresetResourceType);
            });
            clearMultiSelectResourcesUIDButton.gameObject.SetActive(false);
            parentBrowser.ApplyHelp(clearMultiSelectResourcesUIDButton, "Clear all resources (other than the very latest selected resource) from the selection window.");

            selectAllResourcesUIDButton = UI.CreateUIButton(parentBoxGO, 42f, 15f, "Select All", 18, navButtonsHOffset, navButtonsVOffset, anchorPreset: AnchorPresets.bottomRight);
            selectAllResourcesUIDButton.buttonImage.sprite = null;
            selectAllResourcesUIDButton.buttonColor = Color.cyan;
            selectAllResourcesUIDButton.button.onClick.AddListener(delegate {

                if (parentBrowser.isVREnabled || (!Input.GetKey(KeyCode.LeftControl)) && !Input.GetKey(KeyCode.RightControl)) parentBrowser.resourceSelectionsUI.DeSelectAll(browserFiltersUI.currentResourceType, browserFiltersUI.currentAtomPresetResourceType, false);
                parentBrowser.resourceSelectionsUI.SelectAll(browserFiltersUI.currentResourceType, browserFiltersUI.currentAtomPresetResourceType, currentFilterResults);
            });
            selectAllResourcesUIDButton.gameObject.SetActive(false);
            if (parentBrowser.isVREnabled) parentBrowser.ApplyHelp(selectAllResourcesUIDButton, "Select all resources from the the browser results (replacing existing selections).");
            else parentBrowser.ApplyHelp(selectAllResourcesUIDButton, "Select all resources from the the browser results (replacing existing selections). Hold the Ctrl Key whilst pressing 'Select All' to add all resources from the the browser results to the existing selections.");

            selectedResourceCountTextJSS = new JSONStorableString("selectedResourceCountText", "");
            selectedResourceCountTextBox = UI.CreateBasicTextBox(parentBoxGO, selectedResourceCountTextJSS, 43f, 15f, 22, -45, navButtonsVOffset, AnchorPresets.bottomRight, true, 16,backgoundColor:Color.cyan);
            parentBrowser.ApplyHelp(selectedResourceCountTextBox, "Count of currently selected resources.");
            selectedResourceCountTextJSS.text.color = Color.black;
            selectedResourceCountTextJSS.text.alignment = TextAnchor.MiddleCenter;

            resourceCountTextJSS = new JSONStorableString("resourceCountText", "");
            resourceCountTextBox = UI.CreateBasicTextBox(parentBoxGO, resourceCountTextJSS, 90f, 20f, 22, resourceCountHOffset, resourceCountVOffset, AnchorPresets.bottomMiddle, true, 18);
            parentBrowser.ApplyHelp(resourceCountTextBox, "Total count of resources from the broswer filter results and the range displayed on the current page of results.");
        }

        public void IncreaseButtonSizes()
        {
            if (incResourceButtonSizesUIDButton.button.enabled)
            {
                if (resourceButtonSize < 6) resourceButtonSize++;
                RepopulateButtons();
            }
        }

        public void DecreaseButtonSizes()
        {
            if (decResourceButtonSizesUIDButton.button.enabled)
            {
                if (resourceButtonSize > 0) resourceButtonSize--;
                RepopulateButtons();
            }
        }


        public void FirstPageResults()
        {
            if (firstPageUIDButton.button.enabled)
            {
                currentFilterResultsPage = 1;
                scrollRect.verticalNormalizedPosition = 1f;
                RepopulateButtons();
            }
            
        }

        public void PrevPageResults()
        {
            if (prevPageUIDButton.button.enabled)
            {
                currentFilterResultsPage--;
                scrollRect.verticalNormalizedPosition = 1f;
                RepopulateButtons();
            }
        }

        public void NextPageResults()
        {
            if (nextPageUIDButton.button.enabled)
            {
                scrollRect.verticalNormalizedPosition = 1f;
                currentFilterResultsPage++;
                RepopulateButtons();
            }
        }

        public void LastPageResults()
        {
            if (lastPageUIDButton.button.enabled)
            {
                currentFilterResultsPage = (int)Math.Ceiling(currentFilterResults.Count / Settings.maxButtonsPerPage);
                scrollRect.verticalNormalizedPosition = 1f;
                RepopulateButtons();
            }
        }


        public void UpdateDisplayName(ResourceVersionGroupEntry rvge)
        {
            if (RVGEToUIDButtonDict.ContainsKey(rvge))
            {
                RVGEToUIDButtonDict[rvge].label = rvge.resourceDisplayName;
            }
        }

        public void ReloadSprite(ResourceVersionGroupEntry rvge)
        {
            if (RVGEToUIDButtonDict.ContainsKey(rvge))
            {
                var button = RVGEToUIDButtonDict[rvge];
                Transform newImageTrans = button.button.transform.Find("ButtonSubImage");
                UnityEngine.UI.Image image = newImageTrans.GetComponent<UnityEngine.UI.Image>();
                image.sprite = rvge.resourceSprite;
            }
        }
        public static void DeSelectColor(UIDynamicButton buttonUID)
        {
            if (buttonUID != null) buttonUID.buttonColor = BrowserUI.lightGrey;
        }
        public static void SelectColor(UIDynamicButton buttonUID)
        {
            if (buttonUID != null) buttonUID.buttonColor = Color.cyan;
        }
        public void RepopulateButtons(bool forceSelectedResourceRefresh = false)
        {

            if (currentFilterResults==null) return;
            float newScrollBarPosition = scrollBarPosition;
            int buttonsPerPage = (int)Settings.maxButtonsPerPage;

            if ((currentFilterResultsPage - 1) * buttonsPerPage > currentFilterResults.Count) currentFilterResultsPage = 1;

            if (currentFilterResultsPage < 1) currentFilterResultsPage = 1;
            int buttonCount = Math.Min(currentFilterResults.Count, Math.Min(buttonsPerPage, currentFilterResults.Count - ((currentFilterResultsPage - 1) * buttonsPerPage)));

            RefactorButtonRows(buttonCount, buttonsPerRow);

            foreach (var kvp in UIDButtonToRVGEDict) kvp.Value.CancelButtonImageUpdate();

            UIDButtonToRVGEDict.Clear();
            RVGEToUIDButtonDict.Clear();

            int resourceIndexOffset = (currentFilterResultsPage - 1) * buttonsPerPage;

            ResourceVersionGroupEntry firstSelectedRVGE = parentBrowser.resourceSelectionsUI.GetFirstSelectedRVGE(browserFiltersUI.currentResourceType, browserFiltersUI.currentAtomPresetResourceType);

            if (firstSelectedRVGE == null || (browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi && forceSelectedResourceRefresh && !currentFilterResults.Contains(firstSelectedRVGE)))
            {
                if (currentFilterResults.Count != 0 && !parentBrowser.isCallbackBrowser) SelectRVGE(currentFilterResults[resourceIndexOffset], activeButtonUIDs[0], true);
            }

            if (currentFilterResultsPage == 1)
            {
                firstPageUIDButton.button.enabled = false;
                firstPageUIDButton.textColor = Color.white;
                prevPageUIDButton.button.enabled = false;
                prevPageUIDButton.textColor = Color.white;
            }
            else
            {
                firstPageUIDButton.button.enabled = true;
                firstPageUIDButton.textColor = Color.black;
                prevPageUIDButton.button.enabled = true;
                prevPageUIDButton.textColor = Color.black;
            }
            if (currentFilterResultsPage >= Math.Ceiling(currentFilterResults.Count / (float)buttonsPerPage))
            {
                nextPageUIDButton.button.enabled = false;
                nextPageUIDButton.textColor = Color.white;
                lastPageUIDButton.button.enabled = false;
                lastPageUIDButton.textColor = Color.white;
            }
            else
            {
                nextPageUIDButton.button.enabled = true;
                nextPageUIDButton.textColor = Color.black;
                lastPageUIDButton.button.enabled = true;
                lastPageUIDButton.textColor = Color.black;
            }

            if (resourceButtonSize == 0)
            {
                decResourceButtonSizesUIDButton.button.enabled = false;
                decResourceButtonSizesUIDButton.textColor = Color.white;
            }
            else
            {
                decResourceButtonSizesUIDButton.button.enabled = true;
                decResourceButtonSizesUIDButton.textColor = Color.black;
            }
            if (resourceButtonSize == 6)
            {
                incResourceButtonSizesUIDButton.button.enabled = false;
                incResourceButtonSizesUIDButton.textColor = Color.white;
            }
            else
            {
                incResourceButtonSizesUIDButton.button.enabled = true;
                incResourceButtonSizesUIDButton.textColor = Color.black;
            }

            resourceCountTextJSS.val = (resourceIndexOffset + 1).ToString() + " - " + (resourceIndexOffset + buttonCount).ToString() + " of " + currentFilterResults.Count.ToString();

            currentPageRVGEs.Clear();
            currentPageButtons.Clear();

            parentBrowser.resourceSelectionsUI.ClearSelectedRVGEButtons(browserFiltersUI.currentResourceType, browserFiltersUI.currentAtomPresetResourceType);

            for (int buttonId = buttonCount - 1; buttonId >= 0; buttonId--)
            {
                ResourceVersionGroupEntry rvge = currentFilterResults[buttonId + resourceIndexOffset];
                currentPageRVGEs.Insert(0, rvge);

                UIDynamicButton buttonUID = activeButtonUIDs[buttonId];
                currentPageButtons.Insert(0, buttonUID);

                UIDButtonToRVGEDict.Add(buttonUID, rvge);
                RVGEToUIDButtonDict.Add(rvge, buttonUID);
                if (parentBrowser.resourceSelectionsUI.IsSelected(rvge, browserFiltersUI.selectModeJSEnum.val, buttonUID)) buttonUID.buttonColor = Color.cyan;
                else DeSelectColor(buttonUID);

                SetHideFavColorTags(buttonUID.button, rvge);
                SetNewIcon(buttonUID.button, rvge);
                SetActiveResourceIcon(buttonUID.button, rvge);

                buttonUID.label = rvge.resourceDisplayName;

                buttonUID.button.gameObject.SetActive(true);

                buttonUID.button.onClick.RemoveAllListeners();
                buttonUID.button.onClick.AddListener(delegate {
                    SelectRVGE(rvge, buttonUID);
                });

                currentFilterResults[buttonId + resourceIndexOffset].LoadSprite(activeButtonUIDs[buttonId]);
            }
            for (int i = buttonCount; i < (activeResourceRowGOs.Count) * buttonsPerRow; i++)
            {
                activeButtonUIDs[i].label = "";
                activeButtonUIDs[i].button.gameObject.SetActive(false);
            }

            if (forceSelectedResourceRefresh)
            {
                Globals.mvrScript.StartCoroutine(SetScollBarPosition(newScrollBarPosition));
            }
        }

        public void RefreshRVGEThumbnail(ResourceVersionGroupEntry rvge)
        {
            rvge.resourceSprite = null;
            if (RVGEToUIDButtonDict.ContainsKey(rvge))
            {
                rvge.LoadSprite(RVGEToUIDButtonDict[rvge]);
            }
        }

        private IEnumerator SetScollBarPosition(float newScrollBarPosition)
        {
            yield return new WaitForEndOfFrame();

        }

        public void SetButtonLoading(ResourceVersionGroupEntry rvge)
        {
            if (RVGEToUIDButtonDict.ContainsKey(rvge))
            {
                Button button = RVGEToUIDButtonDict[rvge].button;
                Transform newIconTransform = button.transform.Find("LoadingIcon");
                newIconTransform.gameObject.SetActive(true);
            }

        }

        public void UnsetButtonLoading(ResourceVersionGroupEntry rvge)
        {
            if (RVGEToUIDButtonDict.ContainsKey(rvge))
            {
                Button button = RVGEToUIDButtonDict[rvge].button;
                Transform newIconTransform = button.transform.Find("LoadingIcon");
                newIconTransform.gameObject.SetActive(false);
            }
        }

        private void SetNewIcon(Button button, ResourceVersionGroupEntry rvge)
        {
            Transform newIconTransform = button.transform.Find("NewIcon");
            Transform updatedIconTransform = button.transform.Find("UpdatedIcon");

            if (rvge.isMultiVARVersion)
            {
                newIconTransform.gameObject.SetActive(false);
                updatedIconTransform.gameObject.SetActive(rvge.isNew);
            }
            else
            {
                updatedIconTransform.gameObject.SetActive(false);
                newIconTransform.gameObject.SetActive(rvge.isNew);
            }
        }

        private HashSet<Button> activeResourceButtons = new HashSet<Button>();

        private void SetActiveResourceIcon(Button button, ResourceVersionGroupEntry rvge)
        {
            Transform wornIconTransform = button.transform.Find("WornIcon");

            if (rvge.resourceCategory != ResourceCategory.clothing)
            {
                wornIconTransform.gameObject.SetActive(false);
                activeResourceButtons.Remove(button);
            }
            else
            {
                bool activeClothing = parentBrowser.resourceActionsUI.primaryActionUITab.IsClothingActive(rvge);
                wornIconTransform.gameObject.SetActive(activeClothing);

                if (activeClothing) activeResourceButtons.Add(button);
                else activeResourceButtons.Remove(button);
            }

        }

        public void SetActiveResourceIcon(ResourceVersionGroupEntry rvge)
        {
            if (RVGEToUIDButtonDict.ContainsKey(rvge))
            {
                SetActiveResourceIcon(RVGEToUIDButtonDict[rvge].button, rvge);
            }
        }

        public void ClearActiveResourceIcons()
        {
            foreach (Button button in activeResourceButtons)
            {
                Transform wornIconTransform = button.transform.Find("WornIcon");
                wornIconTransform.gameObject.SetActive(false);
            }
            activeResourceButtons.Clear();
        }



        public void SetNewIcon(ResourceVersionGroupEntry rvge)
        {
            if (RVGEToUIDButtonDict.ContainsKey(rvge))
            {
                SetNewIcon(RVGEToUIDButtonDict[rvge].button, rvge);
            }
        }

        private void SetHideFavColorTags(Button button, ResourceVersionGroupEntry rvge)
        {
            Transform hiddenImageTransform = button.transform.Find("HiddenIconImage");
            Transform favImageTransform = button.transform.Find("FavIconImage");

            favImageTransform.gameObject.SetActive(rvge.isFavorite);
            hiddenImageTransform.gameObject.SetActive(rvge.isHidden || rvge.isTagHidden || rvge.isCreatorHidden);
        }

        public void SetHideFavColorTags(ResourceVersionGroupEntry rvge)
        {
            if (RVGEToUIDButtonDict.ContainsKey(rvge))
            {
                SetHideFavColorTags(RVGEToUIDButtonDict[rvge].button, rvge);
            }
        }

        public UIDynamicButton GetUIDButton(ResourceVersionGroupEntry rvge)
        {
            if (RVGEToUIDButtonDict.ContainsKey(rvge)) return RVGEToUIDButtonDict[rvge];
            return null;
        }

        public void RefactorButtonRows(int requiredButtonCount, int requiredButtonsPerRow)
        {
            int rowsRequired = (int)Math.Ceiling((double)requiredButtonCount / requiredButtonsPerRow);
            GameObject buttonListContentGO = parentBoxGO.transform.Find("scrollableContentViewportGO/vScrollableContentGO").gameObject;

            int newRowsRequired = rowsRequired - activeResourceRowGOs.Count;

            if (buttonUIDsByRow.Count > 0)
            {
                int startingButtonsPerRow = buttonUIDsByRow.First().Value.Count();

                // Change in number of buttons per row
                if (startingButtonsPerRow != requiredButtonsPerRow)
                {
                    // Decrease in buttons perRow
                    if (startingButtonsPerRow > requiredButtonsPerRow)
                    {
                        // Remove spare buttons from every row (active or not)
                        foreach (var kvp in buttonUIDsByRow)
                        {
                            // Remove the excess buttons and add them to the unparented buttons
                            for (int i = requiredButtonsPerRow; i < startingButtonsPerRow; i++)
                            {
                                UIDynamicButton button = kvp.Value.Last();
                                kvp.Value.Remove(button);
                                unparentedButtonUIDs.Add(button);
                                button.gameObject.SetActive(false);
                                UIDButtonToRVGEDict.Remove(button);
                                activeButtonUIDs.Remove(button);
                            }
                        }
                    }
                    // Increase in buttons per Row
                    else if (startingButtonsPerRow < requiredButtonsPerRow)
                    {
                        HashSet<GameObject> recycledRows = new HashSet<GameObject>();
                        // Cycle through every row
                        foreach (var kvp in buttonUIDsByRow)
                        {
                            // Check row is not tagged for recycling
                            if (!recycledRows.Contains(kvp.Key))
                            {
                                // Cycle through missing button
                                for (int i = startingButtonsPerRow; i < requiredButtonsPerRow; i++)
                                {
                                    // Check if there are no spare unparented buttons available
                                    if (unparentedButtonUIDs.Count == 0)
                                    {
                                        // Attempt to recycle an inactiveRow if it exists
                                        if (inactiveResourceRowGOs.Count > 0)
                                        {
                                            var rowGO = inactiveResourceRowGOs.First();

                                            // Remove all buttons from the inactive row
                                            // and add them to the unparented buttons
                                            var buttonsList = buttonUIDsByRow[rowGO];
                                            foreach (var button in buttonsList)
                                            {
                                                unparentedButtonUIDs.Add(button);
                                                UIDButtonToRVGEDict.Remove(button);
                                            }
                                            buttonsList.Clear();
                                            recycledRows.Add(rowGO);
                                            inactiveResourceRowGOs.Remove(rowGO);

                                        }
                                        // If rows are due to be removed, then recycle a row
                                        // and add the buttons to the unparented buttons
                                        else if (newRowsRequired < 0)
                                        {
                                            GameObject rowGO = activeResourceRowGOs.Last();

                                            activeResourceRowGOs.Remove(rowGO);
                                            var buttonsList = buttonUIDsByRow[rowGO];
                                            foreach (UIDynamicButton button in buttonsList)
                                            {
                                                activeButtonUIDs.Remove(button);
                                                unparentedButtonUIDs.Add(button);
                                                UIDButtonToRVGEDict.Remove(button);
                                                button.gameObject.SetActive(false);
                                            }
                                            buttonsList.Clear();
                                            recycledRows.Add(rowGO);
                                            rowGO.SetActive(false);
                                            newRowsRequired++;
                                        }
                                    }
                                    // If unparentedButtons exist then take one and add it to the row
                                    if (unparentedButtonUIDs.Count > 0)
                                    {
                                        UIDynamicButton button = unparentedButtonUIDs.Last();
                                        unparentedButtonUIDs.Remove(button);
                                        kvp.Value.Add(button);
                                        button.gameObject.SetActive(true);
                                    }
                                    else
                                    {
                                        kvp.Value.Add(CreateResourceButtonInRow(kvp.Key, i));
                                    }
                                }
                            }
                        }
                        foreach (var rowGO in recycledRows) buttonUIDsByRow.Remove(rowGO);
                        recycledRows.Clear();
                    }

                    activeButtonUIDs.Clear();
                    foreach (var rowGO in activeResourceRowGOs) ResizeResourceButtonRow(rowGO, true);
                    foreach (var rowGO in inactiveResourceRowGOs) ResizeResourceButtonRow(rowGO);
                }
            }

            if (newRowsRequired > 0)
            {
                for (int i = 0; i < newRowsRequired; i++)
                {
                    if (inactiveResourceRowGOs.Count == 0)
                    {
                        GameObject rowGO = UI.AddChildGOImage(buttonListContentGO, new Color(0f, 0.8f, 0f, 0f), AnchorPresets.hStretchTop, parentBrowser.browserScreenWidth - 20f, resouceButtonWidth + resourceButtonlabelHeight + 2, Vector2.zero);
                        activeResourceRowGOs.Add(rowGO);
                        List<UIDynamicButton> buttonsInRow = new List<UIDynamicButton>();
                        buttonUIDsByRow.Add(rowGO, buttonsInRow);
                        for (int column = 0; column < requiredButtonsPerRow; column++)
                        {
                            UIDynamicButton resourceButton = CreateResourceButtonInRow(rowGO, column);
                            buttonsInRow.Add(resourceButton);
                            activeButtonUIDs.Add(resourceButton);
                        }
                    }
                    else
                    {
                        GameObject rowGO = inactiveResourceRowGOs.Last();
                        inactiveResourceRowGOs.Remove(rowGO);
                        rowGO.SetActive(true);
                        activeResourceRowGOs.Add(rowGO);
                        foreach (UIDynamicButton button in buttonUIDsByRow[rowGO])
                        {
                            activeButtonUIDs.Add(button);
                        }

                    }

                }
            }
            else if (newRowsRequired < 0)
            {
                for (int i = 0; i < Math.Abs(newRowsRequired); i++)
                {
                    GameObject rowGO = activeResourceRowGOs.Last();

                    activeResourceRowGOs.Remove(rowGO);
                    foreach (UIDynamicButton button in buttonUIDsByRow[rowGO]) activeButtonUIDs.Remove(button);
                    inactiveResourceRowGOs.Add(rowGO);
                    rowGO.SetActive(false);
                }
            }
        }
        private void ResizeResourceButtonRow(GameObject rowGO, bool addtoActiveButtons = false)
        {
            UI.ResizeHeight(rowGO, resouceButtonWidth + resourceButtonlabelHeight + 2);
            int column = 0;
            foreach (var button in buttonUIDsByRow[rowGO])
            {
                button.transform.SetParent(rowGO.transform);
                button.buttonImage.enabled = true;
                button.button.enabled = true;

                UI.ResizeRecTransform(button.gameObject, resouceButtonWidth, resouceButtonWidth + resourceButtonlabelHeight, (column * resouceButtonWidth) + 2 + (column * 3), 0, AnchorPresets.middleLeft);

                Transform textTransform = button.button.transform.Find("Text");
                UI.ResizeRecTransform(textTransform.gameObject, resouceButtonWidth * 2, (resourceButtonlabelHeight - 1) * 2, 0, 0, AnchorPresets.bottomLeft);

                Transform imageTransform = button.button.transform.Find("ButtonSubImage");
                UI.ResizeRecTransform(imageTransform.gameObject, resouceButtonWidth, resouceButtonWidth, 0, -2f, AnchorPresets.topMiddle);

                if (addtoActiveButtons) activeButtonUIDs.Add(button);

                column++;
            }
        }
        private UIDynamicButton CreateResourceButtonInRow(GameObject rowGO, int column)
        {
            UIDynamicButton button = UI.CreateUIButton(rowGO, resouceButtonWidth, resouceButtonWidth + resourceButtonlabelHeight, "", 20, (column * resouceButtonWidth) + 2 + (column * 3), 0, AnchorPresets.middleLeft);
            button.gameObject.SetActive(false);

            GameObject newImageGO = UI.AddChildGOImage(button.button.gameObject, Color.white, AnchorPresets.topMiddle, button.height - 18f, button.height - 18f, new Vector2(0f, -2f));
            newImageGO.name = "ButtonSubImage";

            GameObject hiddenIconImageGO = UI.AddChildGOImage(button.button.gameObject, Color.red, AnchorPresets.bottomRight, 7f, 7f, new Vector2(0f, 0f));
            hiddenIconImageGO.name = "HiddenIconImage";
            hiddenIconImageGO.SetActive(false);
            GameObject favIconImageGO = UI.AddChildGOImage(button.button.gameObject, new Color(0f, 0.7f, 0f, 1f), AnchorPresets.bottomRight, 7f, 7f, new Vector2(0f, 7.5f));
            favIconImageGO.name = "FavIconImage";
            favIconImageGO.SetActive(false);

            var newIconGO = UI.CreateBasicTextBox(button.button.gameObject, new JSONStorableString("newIcon", "NEW"), 30, 15, 20, -1, -2, anchorPreset: AnchorPresets.topRight, backgoundColor: Color.red);
            newIconGO.name = "NewIcon";
            newIconGO.SetActive(false);

            var updatedIconGO = UI.CreateBasicTextBox(button.button.gameObject, new JSONStorableString("updatedIcon", "UPDATED"), 56, 15, 20, -1, -2, anchorPreset: AnchorPresets.topRight, backgoundColor: Color.red);
            updatedIconGO.name = "UpdatedIcon";
            updatedIconGO.SetActive(false);

            var wornIconGO = UI.CreateBasicTextBox(button.button.gameObject, new JSONStorableString("wornIcon", "WORN"), 38, 15, 20, -1,  resourceButtonlabelHeight-1, anchorPreset: AnchorPresets.bottomRight, backgoundColor: new Color(0f,0.6f,0f));
            wornIconGO.name = "WornIcon";
            wornIconGO.SetActive(false);

            var loadingIconGO = UI.CreateBasicTextBox(button.button.gameObject, new JSONStorableString("loadingIcon", "LOADING..."), 60, 20, 21, 0, 0, anchorPreset: AnchorPresets.centre, backgoundColor: Color.blue);
            loadingIconGO.name = "LoadingIcon";
            loadingIconGO.SetActive(false);

            Transform textTransform = button.button.transform.Find("Text");
            UI.ResizeRecTransform(textTransform.GetComponent<RectTransform>(), resouceButtonWidth * 2, (resourceButtonlabelHeight - 1) * 2, 0, 0, AnchorPresets.bottomLeft);
            button.buttonText.color = Color.black;
            button.buttonImage.sprite = null;
            button.buttonText.resizeTextForBestFit = true;
            button.buttonText.resizeTextMinSize = 15;
            button.buttonText.resizeTextMaxSize = 21;

            //            activeButtonUIDs.Add(button);

            ResourceButtonPointerEnterExitClick context = button.button.gameObject.AddComponent<ResourceButtonPointerEnterExitClick>();
            context.Init(button, UIDButtonToRVGEDict, parentBrowser.helpTextJSS, parentBrowser);


            return button;
        }

        protected bool tempMultiMode = false;
        protected int previousSelectMode;

        protected bool multiSelectKeyPressed
        {
            get
            {
                if (!Globals.browserPointerActive) return false;
                return Input.GetKey(KeyCode.LeftShift) || Input.GetKey(KeyCode.RightShift) || Input.GetKey(KeyCode.LeftAlt) || Input.GetKey(KeyCode.RightAlt) || Input.GetKey(KeyCode.LeftControl) || Input.GetKey(KeyCode.RightControl);
            }
        }
        public void Update()
        {
            if (!parentBrowser.isVREnabled && !parentBrowser.isCallbackBrowser)
            {
                if (multiSelectKeyPressed)
                {
                    if (browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi)
                    {
                        previousSelectMode = browserFiltersUI.selectModeJSEnum.val;
                        browserFiltersUI.selectModeJSEnum.val = BrowserSelectMode.multi;
                        tempMultiMode = true;
                    }

                }
                else if (tempMultiMode)
                {
                    browserFiltersUI.selectModeJSEnum.val = previousSelectMode;
                    tempMultiMode = false;
                }
            }

        }

        protected virtual void SelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton buttonUID, bool suppressQuickActivate = false)
        {
            if (browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi)
            {
                parentBrowser.resourceSelectionsUI.SingleSelectRVGE(rvge, buttonUID);
                if (browserFiltersUI.selectModeJSEnum.val == BrowserSelectMode.instant && !suppressQuickActivate) parentBrowser.resourceActionsUI.primaryActionUITab.Activate(rvge);
            }
            else if (!parentBrowser.isVREnabled)
            {
                if (multiSelectKeyPressed) tempMultiMode = false;
                if (Input.GetKey(KeyCode.LeftShift)) parentBrowser.resourceSelectionsUI.LeftShiftSelectRVGE(rvge, buttonUID, currentPageRVGEs, currentPageButtons);
                else if (Input.GetKey(KeyCode.RightShift)) parentBrowser.resourceSelectionsUI.RightShiftSelectRVGE(rvge, buttonUID, currentPageRVGEs, currentPageButtons);
                else if (Input.GetKey(KeyCode.LeftControl) || Input.GetKey(KeyCode.RightControl)) parentBrowser.resourceSelectionsUI.ToggleSelectRVGE(rvge, buttonUID);
                else if (Input.GetKey(KeyCode.LeftAlt) || Input.GetKey(KeyCode.RightAlt)) parentBrowser.resourceSelectionsUI.SingleSelectRVGE(rvge, buttonUID);
                else
                {
                    parentBrowser.resourceSelectionsUI.DeSelectAll(rvge.resourceType, rvge.atomPresetResourceType);
                    parentBrowser.resourceSelectionsUI.SingleSelectRVGE(rvge, buttonUID);
                }
            }
            else parentBrowser.resourceSelectionsUI.ToggleSelectRVGE(rvge, buttonUID);
        }

        public void RepositionNavButtons()
        {
            UI.ResetAnchorPosition(firstPageUIDButton.gameObject, navButtonsHOffset, navButtonsVOffset);
            UI.ResetAnchorPosition(prevPageUIDButton.gameObject, navButtonsHOffset + 15, navButtonsVOffset);
            UI.ResetAnchorPosition(nextPageUIDButton.gameObject, navButtonsHOffset + 30, navButtonsVOffset);
            UI.ResetAnchorPosition(lastPageUIDButton.gameObject, navButtonsHOffset + 45, navButtonsVOffset);

            UI.ResetAnchorPosition(incResourceButtonSizesUIDButton.gameObject, resoureButtonSizeHOffset, navButtonsVOffset);
            UI.ResetAnchorPosition(decResourceButtonSizesUIDButton.gameObject, resoureButtonSizeHOffset + 20, navButtonsVOffset);

            UI.ResetAnchorPosition(resourceCountTextBox.gameObject, resourceCountHOffset, resourceCountVOffset);

            UI.ResetAnchorPosition(clearMultiSelectResourcesUIDButton.gameObject, navButtonsHOffset, navButtonsVOffset - 17f);
            UI.ResetAnchorPosition(selectAllResourcesUIDButton.gameObject,/* navButtonsHOffset + 53f*/0f, navButtonsVOffset - 17f);
            UI.ResetAnchorPosition(selectedResourceCountTextBox,-48f, navButtonsVOffset - 17f);
            SetMultiSelectClear();
        }

        public void SetMultiSelectClear()
        {
            if (browserFiltersUI.selectModeJSEnum.val == BrowserSelectMode.multi && parentBrowser.resourceSelectionsUI.SelectionCount(browserFiltersUI.currentResourceType, browserFiltersUI.currentAtomPresetResourceType) > 1)
            {
                clearMultiSelectResourcesUIDButton.gameObject.SetActive(true);
                selectedResourceCountTextBox.SetActive(true);
            }
            else
            {
                clearMultiSelectResourcesUIDButton.gameObject.SetActive(false);
                selectedResourceCountTextBox.SetActive(false);
            }

            if (browserFiltersUI.selectModeJSEnum.val == BrowserSelectMode.multi)
            {
                selectAllResourcesUIDButton.gameObject.SetActive(true);
            }
            else
            {
                selectAllResourcesUIDButton.gameObject.SetActive(false);
            }
        }
    }



}
